/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("bozos", ["mysql", "node-uuid", "emailjs"], function(mysql, uuid, emailjs) {
    'use strict';
    
    const exports = {};

    let ConnectionConfig = {
        hostname: "localhost",
        user: "Bozos",
        password: "jihqprewgop",
        database: "Bozos"
    };

    exports.Connect = function Connect(config) {
        config = Object.assign({}, ConnectionConfig, config);
        return mysql.createPool({
            connectionLimit: 100,
            host: config.hostname,
            user: config.user,
            password: config.password,
            database: config.database,
            charset: "utf8mb4_unicode_ci"
            /* IMPORTANT: before creating tables, you want mysql to be using full multibyte UTF8.

This goes in /etc/mysql/mysql.conf.d/mysqld.cnf:
----
init_connect='SET collation_connection = utf8mb4_unicode_ci'
init_connect='SET NAMES utf8mb4'
character-set-server=utf8mb4
collation-server=utf8mb4_unicode_ci
skip-character-set-client-handshake
----
If you already make the tables, you have to fix them:
----
ALTER DATABASE Models CHARACTER SET = utf8mb4 COLLATE = utf8mb4_unicode_ci;
USE Models;
ALTER TABLE Person CONVERT TO CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;
ALTER TABLE Model CONVERT TO CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;
ALTER TABLE LinkModel CONVERT TO CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;
ALTER TABLE Simulation CONVERT TO CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci;
-----
*/
        });
    };

    // should really configure with shared pool=modeldb.Connect()
    exports.Session = function(config) {
        this.config = Object.assign({}, ConnectionConfig, config);
        this.pool = this.config.pool || exports.Connect(this.config)
        this.signin = '';
        this.person = {};
    };
    exports.Session.prototype = {
        query: function(q, inserts) {
            return new Promise((resolve, reject) => {
                this.pool.getConnection((err,connection) => {
                    if (err) {
                        reject("No database connection: "+err);
                        return;
                    }
                    connection.query(q, inserts||[], (err, rows) => {
                        connection.release();
                        if ( err ) {
                            reject("Database error: "+err);
                        } else {
                            resolve(rows);
                        }
                    }).on('error', err => {
                        reject("Database error: "+err);
                        return;
                    });
                });
            });
        },

        isAllowed: function(ip, email) {
            var step1 = Promise.resolve(true);
            var subnet = '';
            if ( ip.startsWith('::ffff:') ) {
                subnet = ip.substring(7, ip.lastIndexOf('.'));
                step1 = this.query(`SELECT SubnetId FROM Subnet WHERE SubnetAddr=${mysql.escape(subnet)};`)
                    .then(rows => {
                        if ( ! rows.length ) {
                            return true;
                        } else {
                            this.query(`UPDATE Subnet SET SubnetEmail=${mysql.escape(email)}, SubnetWhen=NOW() WHERE SubnetId=${rows[0].SubnetId};`)
                                .catch(err => {
                                    console.log(err);
                                });
                            return false;
                        }
                    });
            }
            return step1.then(allowed => {
                if ( ! allowed ) {
                    return false;
                }
                return this.query(`SELECT IpId FROM Ip WHERE IpAddr=${mysql.escape(ip)};`)
                    .then(rows => {
                        if ( ! rows.length ) {
                            this.query(`INSERT INTO Ip (IpAddr, IpEmail, IpWhen) VALUES (${mysql.escape(ip)}, ${mysql.escape(email)}, NOW());`)
                                .catch(err => {
                                    console.log(err);
                                });
                            subnet && this.query(`INSERT INTO Subnet (SubnetAddr, SubnetEmail, SubnetWhen) VALUES (${mysql.escape(subnet)}, ${mysql.escape(email)}, NOW());`)
                                .catch(err => {
                                    console.log(err);
                                });
                            return true;
                        } else {
                            this.query(`UPDATE Ip SET IpEmail=${mysql.escape(email)}, IpWhen=NOW() WHERE IpId=${rows[0].IpId};`)
                                .catch(err => {
                                    console.log(err);
                                });
                        }
                    });
            });
        },
        remove: function(ip) {
            return this.query(`DELETE FROM Ip WHERE IpAddr=${mysql.escape(ip)};`)
                .then(() => {
                    if ( ip.startsWith('::ffff:') ) {
                        let subnet = ip.substring(7, ip.lastIndexOf('.'));
                        return this.query(`DELETE FROM Subnet WHERE SubnetAddr=${mysql.escape(subnet)};`);
                    }
                });
        }
    };
    
    
    return exports;
});
