CREATE DATABASE qubcloud;
GRANT SELECT, INSERT, DELETE, UPDATE ON qubcloud.* TO qubcloud@'localhost' IDENTIFIED BY 'jihqprewgop';
USE qubcloud;

CREATE TABLE Groups (
       groupId INT PRIMARY KEY AUTO_INCREMENT,
       groupActive BOOLEAN,
       groupName VARCHAR(128),
       groupAbout TEXT,
       groupMaxPerson INT,
       groupMaxAdmin INT,
       groupMaxSeats INT,
       groupComputeMinutes INT,
       groupFreeMB INT,
       groupDataKey VARCHAR(36),
       groupComputeKey VARCHAR(36)
);
CREATE TABLE Person (
       personId INT PRIMARY KEY AUTO_INCREMENT,
       personGroup INT,
       personActive BOOLEAN,
       personIsAdmin BOOLEAN,
       personIsContact BOOLEAN,
       personIsSuper BOOLEAN,
       personName VARCHAR(128),
       personAbout TEXT,
       personEmail VARCHAR(128)
);
CREATE TABLE Hopeful(
       hopefulId INT PRIMARY KEY AUTO_INCREMENT,
       hopefulGroup INT,
       hopefulGroupInfo VARCHAR(128),
       hopefulName VARCHAR(128),
       hopefulEmail VARCHAR(128),
       hopefulAbout TEXT
);
CREATE TABLE Seat (
       seatId INT PRIMARY KEY AUTO_INCREMENT,
       seatPerson INT,/*negative for Hopeful*/
       seatKey VARCHAR(36), INDEX(seatKey),
       seatWhen DATETIME,
       seatIP VARCHAR(48),
       seatAgent VARCHAR(255)
);
CREATE TABLE Invite (
       inviteId INT PRIMARY KEY AUTO_INCREMENT,
       invitePerson INT,
       inviteKey VARCHAR(36),
       inviteExpires DATETIME,
       inviteUsesRemain INT
);
CREATE TABLE Model (
       modelId INT PRIMARY KEY AUTO_INCREMENT,
       modelPerson INT,
       modelGroup INT,
       modelPublic BOOLEAN,
       modelName VARCHAR(128),
       modelQmj TEXT,
       modelAbout TEXT,
       modelWhenAdded DATETIME,
       modelWhenModified DATETIME,
       modelWhenPub DATETIME,
       modelAuthors TEXT,
       modelDOI VARCHAR(128),
       modelStates INT,
       modelClasses INT,
       modelDigest TEXT,
       modelFolder VARCHAR(128)
);
CREATE TABLE Dataset (
       datasetId INT PRIMARY KEY AUTO_INCREMENT,
       datasetPerson INT,
       datasetGroup INT,
       datasetPublic BOOLEAN,
       datasetName VARCHAR(128),
       datasetAbout TEXT,
       datasetWhenRec DATETIME,
       datasetWhenAdded DATETIME,
       datasetWhenIdl DATETIME,
       datasetSampling DOUBLE,
       datasetPoints INT,
       datasetEvents INT,
       datasetSegments INT,
       datasetStimuli TEXT,
       datasetConstants TEXT,
       datasetMolecule VARCHAR(128),
       datasetEtc TEXT,
       datasetDigest TEXT,
       datasetFolder VARCHAR(128)
);
CREATE TABLE DataServer (
       serverId INT PRIMARY KEY AUTO_INCREMENT,
       serverGroup INT,
       serverUUID VARCHAR(36),
       serverWhenFirst DATETIME,
       serverWhenLast DATETIME,
       serverCoverage FLOAT,
       serverFreeMB INT, /* ///## ? */
       serverBaseUrl VARCHAR(128)
);
CREATE TABLE DataInstance (
       instanceId INT PRIMARY KEY AUTO_INCREMENT,
       instanceDataset INT,
       instanceServer INT,
       instanceSum CHAR(32),
       instanceBytes INT,
       instancePath VARCHAR(255)
);
CREATE TABLE Project (
       projectId INT PRIMARY KEY AUTO_INCREMENT,
       projectPerson INT,
       projectGroup INT,
       projectPublic BOOLEAN,
       projectName VARCHAR(255),
       projectAbout TEXT,
       projectWhenCreated DATETIME,
       projectWhenModified DATETIME,
       projectWhenPub DATETIME,
       projectAuthors TEXT,
       projectDOI VARCHAR(128),
       projectType VARCHAR(32),
       projectConfig TEXT,
       projectDigest TEXT,
       projectFolder VARCHAR(128)
);
CREATE TABLE ProjectModel (
       pmId INT PRIMARY KEY AUTO_INCREMENT,
       pmProject INT,
       pmModel INT,
       pmActive BOOLEAN
);
CREATE TABLE ProjectDataset (
       pdId INT PRIMARY KEY AUTO_INCREMENT,
       pdProject INT,
       pdDataset INT,
       pdActive BOOLEAN
);
CREATE TABLE Result (
       resultId INT PRIMARY KEY AUTO_INCREMENT,
       resultProject INT,
       resultConfig TEXT,
       resultLL DOUBLE,
       resultWhen DATETIME,
       resultResults TEXT
);
