define("faqui", ["d3"], function(d3) {
    var exports = {};
    var ui = {};

    exports.start = function(config) {
        exports.node = config.node;
        ui.contents = exports.node.querySelector("#contents");
        ui.sections = [].map.call(exports.node.querySelectorAll(".faqSection"), function(node) {
            return {
                node: node,
                title: node.querySelector('.faqSectionTitle').textContent,
                entries: [].map.call(node.querySelectorAll('.faq'), function(enode) {
                    return {
                        node: enode,
                        name: enode.querySelector("a").name,
                        title: enode.querySelector('.title').textContent
                    };
                })
            };
        });

        ui.ask = {
            btn: exports.node.querySelector("#ask"),
            box: exports.node.querySelector("#askbox"),
            form: exports.node.querySelector("askform"),
            name: exports.node.querySelector(".askName"),
            email: exports.node.querySelector(".askEmail"),
            title: exports.node.querySelector(".askTitle"),
            question: exports.node.querySelector("#askQuestion"),
            warning: exports.node.querySelector("#askWarning"),
            send: exports.node.querySelector("#askSend"),
            conf: exports.node.querySelector("#askconf")
        };
        ui.ask.btn.addEventListener('click', function() {
            ui.ask.btn.classList.add("hidden");
            ui.ask.box.classList.remove("hidden");
            ui.ask.title.value = "";
            ui.ask.question.value = "";
            ui.ask.warning.innerHTML = "";
        });
        ui.ask.send.addEventListener('click', function() {
            if ( ! ui.ask.name.value ) {
                ui.ask.warning.innerHTML = "Please enter your name";
                ui.ask.name.focus();
                return;
            }
            if ( ui.ask.email.value.indexOf("@") <= 0 ) {
                ui.ask.warning.innerHTML = "Please enter your email address";
                ui.ask.email.focus();
                return;
            }
            if ( ! ui.ask.title.value ) {
                ui.ask.warning.innerHTML = "Please enter an informative title for this question";
                ui.ask.title.focus();
                return;
            }
            if ( ! ui.ask.question.value ) {
                ui.ask.warning.innerHTML = "The question appears to be blank";
                ui.ask.question.focus();
                return;
            }
            ui.ask.send.disabled = true;
            post("question", {
                name: ui.ask.name.value,
                email: ui.ask.email.value,
                title: ui.ask.title.value,
                question: ui.ask.question.value
            }).then(function(response) {
                if ( response.t === 'HOLD' ) {
                    setTimeout(function() {
                        alert('We can only accept one submission at a time, because too many people are trying to post advertisements on this page.\n\nPlease wait until your previous submission is approved.');
                    }, 10)
                } else {
                    ui.ask.send.disabled = false;
                    ui.ask.box.classList.add("hidden");
                    ui.ask.btn.classList.remove("hidden");
                    ui.ask.conf.classList.remove("hidden");
                    ui.ask.btn.innerHTML = "Ask another question...";
                }
            }).catch(function(err) {
                console.log(err);
            });
        });
        
        ui.twoCentsTemplate = exports.node.querySelector("#twoCents");
        ui.twoCents = ui.sections.map(function(section, i) {
            return section.entries.map(function(entry, j) {
                // first (mainly for t.o.c.) set up named anchors for each entry
                var anchor = entry.node.querySelector("a");
                if ( ! anchor.name ) {
                    anchor.name = 'q'+(i+1)+'_'+(j+1);
                }
                
                var node = ui.twoCentsTemplate.cloneNode(true);
                node.className = '';
                entry.node.appendChild(node);
                
                var tc = {
                    node: node,
                    offer: node.querySelector(".twoCentsOffer"),
                    show: node.querySelector(".twoCentsShow"),
                    msg: node.querySelector(".twoCentsMsg"),
                    info: node.querySelector(".twoCentsInfo"),
                    name: node.querySelector(".askName"),
                    email: node.querySelector(".askEmail"),
                    answer: node.querySelector(".twoCentsAnswer"),
                    warning: node.querySelector(".twoCentsWarning"),
                    send: node.querySelector(".twoCentsSend"),
                };
                
                tc.show.addEventListener('click', function() {
                    tc.offer.classList.add("hidden");
                    tc.info.classList.remove("hidden");
                    tc.answer.value = "";
                    tc.warning.innerHTML = "";
                });
                
                tc.send.addEventListener('click', function() {
                    if ( ! tc.name.value ) {
                        tc.warning.innerHTML = "Please enter your name";
                        tc.name.focus();
                        return;
                    }
                    if ( tc.email.value.indexOf("@") <= 0 ) {
                        tc.warning.innerHTML = "Please enter your email address";
                        tc.email.focus();
                        return;
                    }
                    if ( ! tc.answer.value ) {
                        tc.warning.innerHTML = "The answer appears to be blank";
                        tc.answer.focus();
                        return;
                    }
                    tc.warning.innerHTML = "Sending...";
                    tc.send.disabled = true;
                    post("answer", {
                        name: tc.name.value,
                        email: tc.email.value,
                        qname: anchor.name,
                        answer: tc.answer.value
                    }).then(function(response) {
                        if ( response.t === 'HOLD' ) {
                            setTimeout(function() {
                                alert('We can only accept one submission at a time, because too many people are trying to post advertisements on this page.\n\nPlease wait until your previous submission is approved.');
                            }, 10)
                        } else {
                            tc.send.disabled = false;
                            tc.info.classList.add("hidden")
                            tc.offer.classList.remove("hidden");
                            tc.msg.innerHTML = "Thanks for your input. We'll post it after review.";
                        }
                    });
                });
                return tc;
            });
        });
        
        var divSections = d3.select(ui.contents)
            .append("div");
        divSections.classed('faqTOCsections', true);
        var selSection = divSections.selectAll('.faqTOCsection').data(ui.sections);
        var divSection = selSection.enter().append('div');
        divSection.classed('faqTOCsection', true);
        divSection.style('display', function(d,i) { return d.entries.length ? '' : 'none'; });
        divSection.append('div')
            .text(function(d,i) { return d.title; })
            .classed('faqTOCsectionTitle', true);
        divSection.each(function(d,i) {
            let lst = d3.select(this).append('ol');
            lst.datum(function() { return null; }); // separate parent from child d3 selections
            lst.selectAll('li').data(d.entries)
                .enter().append('li')
                .append('a')
                .attr('href', function(dd, ii) { return '#'+dd.name; })
                .text(function(dd, ii) { return dd.title; });
        });
    };

    function post(method, data) {
        return new Promise(function(resolve, reject) {
            var request = new XMLHttpRequest();
            request.open('POST', method, true);
            request.onload = function() {
                var response = JSON.parse(request.responseText||'{"t": "Error", "msg": "Empty response"}');
                if ( response.t === 'Error' ) {
                    reject(response.msg);
                } else {
                    resolve(response);
                }
            };
            request.onerror = function(err) { reject(err) };
            request.setRequestHeader("Content-Type", "application/json;charset=UTF-8");
            request.send(JSON.stringify(data));
        });
    }

    return exports;
});
