"use strict";

define("server", ["express", "body-parser", "compression", "modeldb"], function(express, bodyParser, compression, modeldb) {
    let PORT = 1785;
    
    let exports = {};
    exports.app = express();
    exports.app.use(compression());
    exports.app.use(bodyParser.json());
    exports.app.use(bodyParser.urlencoded({ extended: true }));
    exports.app.use('/js', express.static('../static/js'));
    
    let db = new modeldb.Connection();

    exports.app.get('/', (req, res) => {
        res.sendFile('index.html', {root: '.'});
    });
    
    exports.app.post('/ping', (req, res) => {
        console.log(req.body);
        res.json(req.body);
    });
    exports.app.post("/signIn", (req, res) => {
        db.signIn(""+req.body.inviteSecret, req.ip, req.headers['user-agent'])
            .then(signinSecret => {
                res.json({signinSecret: signinSecret});
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/signOut", (req, res) => {
        db.signOut(""+req.body.signinSecret)
            .then(result => {
                res.json({t: "OK"});
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/aboutMe", (req, res) => {
        db.aboutMe(""+req.body.signinSecret, req.ip, req.headers['user-agent'])
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/setAbout", (req, res) => {
        db.setAbout(""+req.body.signinSecret, req.ip, req.headers['user-agent'], ""+req.body.about)
            .then(result => {
                res.json({t: "OK"});
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/setEmail", (req, res) => {
        db.setEmail(""+req.body.signinSecret, req.ip, req.headers['user-agent'], ""+req.body.email)
            .then(result => {
                res.json({t: "OK"});
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/selfInvite", (req, res) => {
        db.selfInvite(""+req.body.signinSecret, req.ip, req.headers['user-agent'], req.body.expiresHours|0, req.body.usesRemain|0)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/invite", (req, res) => {
        let b = req.body;
        db.invite(""+b.signinSecret, req.ip, req.headers['user-agent'],
                 {
                     personName: b.person.name ? (""+b.person.name) : "",
                     personAbout: b.person.about ? (""+b.person.about) : "",
                     personEmail: b.person.email ? (""+b.person.email) : "",
                     personCanSave: b.person.canSave|0,
                     personCanSelfInvite: b.person.canSelfInvite?1:0,
                     personCanInvite: b.person.canInvite?1:0,
                     personCanAdmin: b.person.canAdmin?1:0
                 },
                 b.expiresHours|0, b.usesRemain|0)
            .then(result => {
                res.json({inviteSecret: result});
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/listPeople", (req, res) => {
        db.listPeople(""+req.body.signinSecret, req.ip, req.headers['user-agent'])
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/reinvite", (req, res) => {
        db.reinvite(""+req.body.signinSecret, req.ip, req.headers['user-agent'],
                   req.body.personId|0, req.body.expiresHours|0, req.body.usesRemain|0)
            .then(result => {
                res.json({inviteSecret: result});
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/updatePerson", (req, res) => {
        var uu = req.body.updates || {};
        db.updatePerson(""+req.body.signinSecret, req.ip, req.headers['user-agent'],
                        uu.personId|0, ""+uu.personName, ""+uu.personAbout, ""+uu.personEmail, !! uu.personActive,
                        !! uu.personCanSave, !! uu.personCanSelfInvite, !! uu.personCanInvite, !! uu.personCanAdmin)
            .then(result => {
                res.json({t: "OK"});
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/listSignIns", (req, res) => {
        db.listSignIns(""+req.body.signinSecret, req.ip, req.headers['user-agent'])
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/unSignIn", (req, res) => {
        db.unSignIn(""+req.body.signinSecret, req.ip, req.headers['user-agent'], ""+req.body.otherSecret)
            .then(result => {
                res.json({t: 'OK'});
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/saveModel", (req, res) => {
        db.saveModel(""+req.body.signinSecret, req.ip, req.headers['user-agent'],
                     ""+req.body.name, req.body.qmj, ""+req.body.about, ""+req.body.doi,
                     req.body.whenPub ? new Date(""+req.body.whenPub) : "",
                     ""+req.body.authors)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/listModels", (req, res) => {
        db.listModels(req.body.query ? (""+req.body.query) : req.body.query)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/getModelQmj", (req, res) => {
        db.getModelQmj(req.body.modelId|0)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/getModelInfo", (req, res) => {
        db.getModelInfo(req.body.modelId|0)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/requestInvite", (req, res) => {
        var badchar = /"/g;
        db.requestInvite(req.ip, req.headers['user-agent'],
                         req.body.personName.replace(badchar, ""),
                         req.body.personAbout.replace(badchar, ""),
                         req.body.personEmail.replace(badchar, ""));
        res.json({t: 'OK'});
    });
    exports.app.post("/newPerson", (req, res) => {
        var badchar = /"/g;
        db.newPerson(req.ip, req.headers['user-agent'],
                     req.body.personName.replace(badchar, ""),
                     req.body.personAbout.replace(badchar, ""),
                     req.body.personEmail.replace(badchar, ""))
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/requestPermissions", (req, res) => {
        db.requestPermissions(req.ip, req.headers['user-agent'], ""+req.body.what);
        res.json({t: 'OK'});
    });
    exports.app.post("/createLink", (req, res) => {
        db.createLink(""+req.body.signinSecret, req.ip, req.headers['user-agent'],
                     ""+req.body.name, req.body.qmj, ""+req.body.about)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/loadLink", (req, res) => {
        db.loadLink(req.body.linkId|0, ""+req.body.linkKey)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/listLinks", (req, res) => {
        db.listLinks(""+req.body.signinSecret, req.ip, req.headers['user-agent'],
                     req.body.queryString ? (""+req.body.queryString) : req.body.queryString,
                     !!req.body.allUsers)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/saveSim", (req, res) => {
        db.saveSim(""+req.body.signinSecret, req.ip, req.headers['user-agent'],
                   ""+req.body.name, ""+req.body.about, req.body.config||{}, req.body.modelId|0)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });
    exports.app.post("/listSim", (req, res) => {
        db.listSim(('simId' in req.body) ? (req.body.simId|0) : -1)
            .then(result => {
                res.json(result);
            })
            .catch(err => { res.json({t: 'Error', msg: err}); });
    });

    // final (default) handler:
    exports.app.use((req, res, next) => {
        res.status(404).send('Not found (404)');
    });
    
    exports.run = function() {
        exports.app.listen(PORT);
    };
    
    return exports;
});
