/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qub-online", ["express", "body-parser", "compression", "cloud", "dataconv", "modeldb", "modelsrv"], function(express, bodyParser, compression, cloud, dataconv, modeldb, modelsrv) {
    'use strict';
    
    const exports = express.Router();
    //exports.use(bodyParser.json());
    //exports.use(bodyParser.urlencoded({ extended: true }));

    const pool = modeldb.Connect();
    const badchar = /"/g;

    exports.post("/reqErr", (req, res) => {
        let data='';
        req.setEncoding('utf8');
        req.on('data', chunk => { 
            data += chunk;
        });
        
        req.on('end', () => {
            try {
                data = JSON.parse(data);
            } catch(err) {}
            console.log('client error in require():');
            if ( data.stack ) {
                console.log(data.stack);
                console.log(data.userAgent);
                console.log(data.href);
                if ( typeof(data.other) !== 'undefined' ) {
                    console.log(data.other);
                }
            } else {
                console.log(data);
            }
        });
    });
        
    exports.setupWS = () => {
        exports.ws('/errs', (ws, req) => {
            ws.on('message', (data, flags) => {
                try {
                    data = JSON.parse(data);
                }
                catch (err)
                {}
                console.log('client error:');
                if ( data.stack ) {
                    console.log(data.stack);
                    console.log(data.userAgent);
                    console.log(data.href);
                    if ( typeof(data.other) !== 'undefined' ) {
                        console.log(data.other);
                    }
                } else {
                    console.log(data);
                }
            });
        });  
        
        exports.ws("/modeldb", (ws, req) => {
            const db = new modeldb.Session({pool: pool});
            let conv = {};
            //ws.binaryType = 'arraybuffer';
            ws.on('close', () => {
            });
            ws.on('message', (data, flags) => {
                let msg;
                try {
                    msg = JSON.parse(data);
                } catch(err) {
                    console.log('online:modeldb:parse error:',data);
                    return;
                }
                const reply = r => {
                    r.tkt = msg.tkt;
                    ws.send(JSON.stringify(r));
                    return r;
                };
                const replyEmpty = r => {
                    ws.send(JSON.stringify({
                        tkt: msg.tkt
                    }));
                    return r;
                };
                const replyInField = f => r => {
                    const rmsg = {tkt: msg.tkt};
                    rmsg[f] = r;
                    ws.send(JSON.stringify(rmsg));
                    return rmsg;
                };
                return (mType => {
                    if ( mType === 'ping' ) {
                        return Promise.resolve(reply(msg));
                    } else if ( mType === 'getModelQmj' ) {
                        return db.getModelQmj(msg.modelId|0)
                            .then(reply);
                    } else if ( mType === 'getModelInfo' ) {
                        return db.getModelInfo(msg.modelId|0)
                            .then(reply);
                    } else if ( mType === 'loadLink' ) {
                        return db.loadLink(msg.linkId|0, ''+msg.linkKey)
                            .then(reply);
                    }
                    
                    return Promise.reject('Unknown message type:'+mType+': '+JSON.stringify(msg));
                })(msg.t).catch(err => {
                    console.log('online:modeldb:err:',err);
                    ws.send(JSON.stringify({
                        tkt: msg.tkt,
                        tktErr: err
                    }));
                });
            });
        });

        cloud.setupWS();
        dataconv.setupWS();
    }

    exports.use('/cloud', cloud.router);
    exports.use('/data', dataconv.router);
    
    // public POST-based interface is still used by qub express to generate links
    // (versions <= 1.11.7 use the dedicated models server on port 1785)
    exports.use("/models", modelsrv);
    
    exports.use('/macsample', (req, res) => {
        res.redirect('/online/#link=7&k=212fdadc-1d58-42a2-bb8e-1e0127d288db&macqdf=sample&welcome=1');
    });
    exports.use('/piezo1', (req, res) => {
        res.redirect('/online/piezo1.html');
    });

    exports.use("/", express.static('static'));

    return exports;
});

    
