/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubdl", ["express", "body-parser", "fs", "mysql", "emailjs"], function(express, bodyParser, fs, mysql, emailjs) {
    'use strict';
    
    const db = {
        pool: mysql.createPool({
            connectionLimit: 100,
            host: 'localhost',
            user: 'dlposter',
            password: 'qubdl',
            database: 'downloads'
        }),
        query: function(q, inserts) {
            return new Promise(function(resolve, reject) {
                db.pool.getConnection((err,connection) => {
                    if (err) {
                        reject("No database connection: "+err);
                        return;
                    }
                    connection.query(q, inserts||[], (err, rows) => {
                        connection.release();
                        if ( err ) {
                            reject("Database error: "+err);
                        } else {
                            resolve(rows);
                        }
                    }).on('error', err => reject("Database error: "+err));
                });
            });
        },
        listApps: function(os, code, home) {
            const codeExpr = code ? ` WHERE Code = ${mysql.escape(code)}` : "";
            const apps = this.query(`SELECT Id, Name, Changelog, About, Description, Code FROM Apps${codeExpr};`);
            if ( os ) {
                return apps.then(rows => {
                    // copy them so we can add fields from other tables
                    const rowsCopy = rows.map(function(row) {
                        const rowOut = {};
                        for (var field in row) {
                            rowOut[field] = row[field];
                        }
                        return rowOut;
                    });
                    const appsForOs = [];
                    let work = Promise.resolve(appsForOs);
                    rowsCopy.forEach(function(row) {
                        work = work.then(function() {
                            return db.listVersions(os, row.Id, 1, home);
                        }).then(vrows => {
                            if ( vrows.length ) {
                                {let field; for (field in vrows[0]) {
                                    row[field] = vrows[0][field];
                                }}
                                appsForOs.push(row);
                            }
                            return appsForOs;
                        });
                    });
                    return work;
                });
            } else {
                return apps;
            }
        },
        listVersions: function(os, id, limit, home) {
            const limitExpr = limit ? ` ORDER BY Versions.Id DESC LIMIT ${limit}` : "";
            return this.query(`SELECT Filename, Upgrade, UpgradeSince, Number, Date FROM Files INNER JOIN Versions ON Version = Versions.Id WHERE (OS=${mysql.escape(os)}) AND (App=${id})${limitExpr}`)
		.then(function(rows) {
		    if ( home ) {
			rows.forEach(function(r) {
			    const paid = 'paid/'+r.Filename;
			    try {
				fs.statSync('/var/www/dl/'+paid);
				r.Filename = paid;
			    } catch (e) {}
			});
		    }
		    return rows;
		});
        },
        regPerson: function(name, email, date, ip) {
            return this.query("INSERT INTO Requests SET ?", {
                Name: name,
                Email: email,
                Date: date,
                Updates: 0,
                IP: ip
            });
        },
        regFile: function(name, email, date, filename, ip) {
            return this.query("INSERT INTO RequestFiles SET ?", {
                Date: date,
                Filename: filename,
                IP: ip
            });
        }
    };

    db.query(`SELECT Id, Filename FROM Files`)
        .then(rows => {
            rows.forEach(row => {
                try {
                    fs.statSync('/var/www/dl/'+row.Filename);
                } catch (err) {
                    console.log('deleted dl/'+row.Filename);
                    db.query(`DELETE FROM Files WHERE Id=${row.Id}`);
                }
            });
        }).catch(err => {
            console.log('cleaning Files table...',err);
        });

    const router = express.Router();
    router.use(bodyParser.json());
    router.use(bodyParser.urlencoded({ extended: true }));
    
    router.post("/listFiles", (req, res) => {// code name filename ext date version description upgrade upgradeSince aboutUrl changelogUrl size(MB)
        db.listApps(req.body.os, null, req.ip.indexOf('ffff:128.205'))
            .then(rows => {
                res.json({
                    files: rows.map(row => ({
                        code: row.Code,
                        name: row.Name,
                        changelogUrl: row.Changelog,
                        aboutUrl: row.About,
                        description: row.Description,
                        filename: row.Filename,
                        ext: row.Filename.slice(row.Filename.lastIndexOf(".")),
                        date: row.Date,
                        version: row.Number,
                        upgrade: row.Upgrade,
                        upgradeSince: row.UpgradeSince,
                        size: row.Filename ? (fs.statSync('/var/www/dl/'+row.Filename).size / (1024*1024)) : 0
                    }))
                });
            })
            .catch(err => res.json({t: "Error", msg: ""+err}));
    });
    router.post("/listPrev", (req, res) => {
        let name;
        db.listApps(req.body.os, req.body.code)
            .then(apps => {
                if ( apps.length ) {
                    name = apps[0].Name;
                    return db.listVersions(req.body.os, apps[0].Id);
                }
                throw new Error("No such app");
            }).then(versions => {
                res.json({
                    files: versions.map(vers => ({
                        name: name,
                        filename: vers.Filename,
                        date: vers.Date,
                        version: vers.Number
                    }))
                });
            })
            .catch(err => res.json({t: "Error", msg: ""+err}));
    });
    router.post("/reg", (req, res) => {
        console.log('reg', req.body);
        db.regPerson(req.body.name, req.body.email, req.body.date, req.ip)
            .then(result => {
                res.json({t: 'OK'});
            })
            .catch(err => res.json({t: "Error", msg: ""+err}));
    });
    router.post("/regFile", (req, res) => {
        console.log('regFile', req.body);
        db.regFile(req.body.name, req.body.email, req.body.date, req.body.filename, req.ip)
            .then(result => {
                res.json({t: 'OK'});
            })
            .catch(err => {
                console.log('in regFile:',err);
                res.json({t: "Error", msg: ""+err});
            });
                  
    });

    router.use('/', express.static('./download'));

    return router;
});
