/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubfaq", ["express", "body-parser", "emailjs", "bozos"], function(express, bodyParser, emailjs, bozos) {
    'use strict';

    const answerer = 'qub@www.qub.buffalo.edu';
    
    const router = express.Router();
    router.use(bodyParser.json());
    router.use(bodyParser.urlencoded({ extended: true }));

    const bozoCheck = new bozos.Session();
    
    router.post("/question", (req, res) => {
        bozoCheck.isAllowed(req.ip, req.body.email||'')
            .then(allowed => {
                if ( allowed ) {
                    emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                        .send({
                            from: `${req.body.name} <${req.body.email}>`,
                            to: answerer,
                            subject: "QUB Q: "+req.body.title,
                            text: req.body.question+'\n\n'+`IP: ${req.ip}`
                        }, (err, msg) => {
                            if ( err ) {
                                res.json({t: 'Error', msg: ""+err});
                            } else {
                                res.json({t: 'OK'});
                            }
                        });
                } else {
                    res.json({t: 'HOLD'});
                }
            }).catch(err => {
                res.json({t: 'Error', msg: ''+err});
            });
    });
    router.post("/answer", (req, res) => {
        bozoCheck.isAllowed(req.ip, req.body.email||'')
            .then(allowed => {
                if ( allowed ) {
                    emailjs.server.connect({host: '127.0.0.1', tls: false, domain: '127.0.0.1'})
                        .send({
                            from: `${req.body.name} <${req.body.email}>`,
                            to: answerer,
                            subject: "Re: QUB "+req.body.qname,
                            text: `Regarding https://www.qub.buffalo.edu/faq/#${req.body.qname}

${req.body.answer}

IP: ${req.ip}`
                        }, (err, msg) => {
                            if ( err ) {
                                res.json({t: 'Error', msg: ""+err});
                            } else {
                                res.json({t: 'OK'});
                            }
                        });
                } else {
                    res.json({t: 'HOLD'});
                }
            }).catch(err => {
                res.json({t: 'Error', msg: ''+err});
            });
    });

    return router;
});
