"use strict";

define("server", ["express", "body-parser", "compression", "optdb"], function(express, bodyParser, compression, optdb) {
    let PORT = 1784;
    
    let exports = {};
    exports.app = express();
    exports.app.use(compression());
    exports.app.use(bodyParser.json());
    exports.app.use(bodyParser.urlencoded({ extended: true }));
    exports.app.use('/js', express.static('../static/js'));
    
    let db = new optdb.Connection();

    exports.app.get('/', function(req, res) { res.sendFile('index.html', {root: '.'}) });
    
    exports.app.post('/ping', function(req, res) {
        console.log(req.body);
        res.json(req.body);
    });

    exports.app.post('/trial', function(req, res) {
        let b = req.body;
        let mPre = b.modelPre;
        if ( mPre && mPre.rates && b.modelPost && b.modelPost.rates ) {
            db.trial({algorithmName: ""+b.algorithmName,
                      modelPre: mPre,
                      modelPost: b.modelPost,
                      nautyNumber: b.nautyNumber|0,
                      stimLbl: ""+b.stimLbl,
                      dataCount: b.dataCount|0,
                      tDead: b.tDead||0.0,
                      sampling: b.sampling||0.0,
                      app: ""+b.app,
                      os: ""+b.os})
                .catch(function(err) {
                    console.log("/trial: "+err, b);
                });
        }
        res.json({t: 'OK'});
    });
    
    exports.app.post('/listAlgo', function(req, res) {
        db.listAlgo()
            .then(function(algos) { res.json(algos); })
            .catch(function(err) { res.json({t: 'Error', msg: err}); });
    });
    
    exports.app.post('/listNauty', function(req, res) {
        db.listNauty()
            .then(function(models) { res.json(models); })
            .catch(function(err) { res.json({t: 'Error', msg: err}) });
    });
    
    exports.app.post('/listStim', function(req, res) {
        db.listStim()
            .then(function(stims) { res.json(stims); })
            .catch(function(err) { res.json({t: 'Error', msg: err}); });
    });
    
    exports.app.post('/getNauty', function(req, res) {
        db.getNauty(req.body.nautyId|0)
            .then(function(model) { res.json(model); })
            .catch(function(err) { res.json({t: 'Error', msg: err}); });
    });
    
    exports.app.post('/getTrial', function(req, res) {
        db.getTrial(req.body.trialId|0)
            .then(function(t) { res.json(t); })
            .catch(function(err) { res.json({t: 'Error', msg: err}); });
    });
    
    exports.app.post('/getRates', function(req, res) {
        db.getRates(req.body)
            .then(function(rates) { res.json(rates); })
            .catch(function(err) { res.json({t: 'Error', msg: err}); });
    });
    
    
    // final (default) handler:
    exports.app.use(function(req, res, next) { res.status(404).send('Not found (404)'); });
    
    exports.run = function() {
        exports.app.listen(PORT);
    };
    
    return exports;
});
