"use strict";

define("server", ["express", "express-ws", "body-parser", "compression", "http", "spdy", "fs", "qub-online", "url", "morgan"], function(express, expressWS, bodyParser, compression, http, spdy, fs, online, url, morgan) {
    const exports = {};
    const PORT = (+process.env.HTTP_PORT) || 13806;
    const HTTPS_PORT = (+process.env.HTTPS_PORT) || 13807;
    const CERT_PATH = '/etc/letsencrypt/live/www.qub.buffalo.edu';
    let httpsOptions = null;
    
    function handleError(e) {
        console.error('ERROR', new Date(), e);
    }
    process.on('uncaughtException', handleError);
    
    try {
        httpsOptions = {
            key: fs.readFileSync(process.env.HTTPS_KEY_PATH || (CERT_PATH+'/privkey.pem')),
            cert: fs.readFileSync(process.env.HTTPS_CERT_PATH || (CERT_PATH+'/fullchain.pem'))
        };
    } catch (err) {
        console.log('https error:',err,'serving http only');
    }

    exports.app = express();
    exports.app.use(compression({
        filter: function(req, res) {
            return true;
        }
    }));

    exports.app.use(function(req, res, next) {
        res.header("Access-Control-Allow-Origin", "*");
        res.header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept");
        next();
    });

    const bigbots = ['Googlebot', 'Slurp', 'bingbot', 'Baiduspider']
    exports.app.use(morgan('combined', {
        skip: (req, res) => {
            const ua = req.headers['user-agent'];
            if ( ! ua ) {
                return false;
            }
            {let i, n=bigbots.length; for (i=0; i<n; ++i) {
                if ( ua.indexOf(bigbots[i]) >= 0 ) {
                    return true;
                }
            }}
            return false;
        }
    }));
    // can't use bodyParser yet because it breaks proxied form submission in qubsite.js
    //exports.app.use(bodyParser.json());
    //exports.app.use(bodyParser.urlencoded({ extended: true }));

    exports.run = function() {
        exports.app.use('/online', online);
        exports.app.use('/', express.static('/var/www'));
        
        // final (default) handler:
        exports.app.use(function(req, res, next) { res.status(404).send('Not found (404)'); });
    
        //exports.app.listen(PORT);
        const httpServer = http.createServer(exports.app)
            .listen(PORT);
        const httpsServer = httpsOptions && spdy.createServer(httpsOptions, exports.app)
            .listen(HTTPS_PORT);

        httpsServer && expressWS(exports.app, httpsServer);
        online.setupWS();
    };
    
    return exports;
});
