define("ajaxIframeClient", [], function() {
    var exports = {};

    exports.client = function(origin, iframeId, methodNames) {
        return new Promise(function(resolve, reject) {
            var iframe = document.createElement('iframe');
            iframe.style.display = 'none';
            iframe.sandbox = 'allow-scripts allow-same-origin';
            iframe.id = iframeId;
            iframe.src = origin;
            document.body.appendChild(iframe);
            var remote = iframe[0].contentWindow;
            var serial = 0;
            var handlers = {};

            function defineMethod(name) {
                return function(arg) {
                    return new Promise(function(resolve, reject) {
                        var data = {serial: serial, method: name, arg: arg};
                        handlers[serial] = {
                            resolve: resolve,
                            reject: reject
                        };
                        serial += 1;
                        remote.postMessage(data, origin);
                    });
                };
            }
            function defineMethods(names, target) {
                for (var i=0, n=names.length; i<n; ++i) {
                    target[names[i]] = defineMethod(names[i]);
                }
            }
            
            window.addEventListener('message', function(e) {
                if ( e.origin === origin && e.source === remote ) {
                    if ( e.data.t === 'Init' ) {
                        var client = {};
                        // all methods take one argument, a JSON-able object, and return a promise which resolves to an object. Ping echoes its arg.
                        defineMethods(methodNames, client);
                        resolve(client);
                    } else {
                        var s = e.data.serial;
                        if ( e.data.t === 'Error' ) {
                            handlers[s].reject(e.data.msg);
                        } else {
                            handlers[s].resolve(e.data.arg);
                        }
                        delete handlers[s];
                    }
                }
            });
        });
    };
    
    return exports;
});
