/* Copyright 2016-2107 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("fitness_ui", ["qubx", "qubx_ui", "qubx_chart", "fitness_core", "d3"],
    (qubx, qubx_ui, qubx_chart, fitness_core, d3) => {
        'use strict';

        const exports = {};

        const links = {
            rss: 'https://en.wikipedia.org/wiki/Residual_sum_of_squares',
            r2: 'https://en.wikipedia.org/wiki/Coefficient_of_determination',
            ww: 'https://en.wikipedia.org/wiki/Wald–Wolfowitz_runs_test'
        };
        const tips = {
            rss: 'Residual sum of squares, a.k.a. sum square difference',
            r2: 'R-squared, a.k.a coefficient of determination',
            ww: 'Probability the curve crossed the data often enough to be random (unbiased), after Wald and Wolfowitz'
        };

        exports.presets = [
            {
                name: 'Linear',
                expr: 'm*x + b'
            },
            {
                name: 'Gaussian',
                expr: 'Amp * (1/(Std * s2p)) * exp( -(x - Mean)*(x - Mean) / (2 * dVar) ) + Baseline # Amp: 1, Mean: Data.mean("x"), Std: .25*Data.std("x")||.1, Baseline: {val: 0, canFit: false}, common: {s2p: sqrt(2*PI), dVar: "2*Std*Std"}'
            },
            {
                name: 'Gaussian (2)',
                expr: 'Amp1*(1/(Std1*s2p))*exp(-(x-Mean1)*(x-Mean1)/dVar1) + Amp2*(1/(Std2*s2p))*exp(-(x-Mean2)*(x-Mean2)/dVar2) + Baseline # Amp1: .5, Amp2: .5, Mean1: Data.min("x")+.25*(Data.max("x")-Data.min("x")), Mean2: Data.min("x")+.75*(Data.max("x")-Data.min("x")), Std1: .1*Data.std("x"), Std2: .1*Data.std("x"), Baseline: {val: 0, canFit: false}, common: {s2p: sqrt(2*PI), dVar1: "2*Std1*Std1", dVar2: "2*Std2*Std2"}'
            },
            {
                name: 'Gaussian (3)',
                expr: 'Amp1*(1/(Std1*s2p))*exp(-(x-Mean1)*(x-Mean1)/dVar1) + Amp2*(1/(Std2*s2p))*exp(-(x-Mean2)*(x-Mean2)/dVar2)+ Amp3*(1/(Std3*s2p))*exp(-(x-Mean3)*(x-Mean3)/dVar3) + Baseline # Amp1: .3, Amp2: .3, Amp3: .3, Mean1: Data.min("x")+.25*(Data.max("x")-Data.min("x")), Mean2: Data.min("x")+.5*(Data.max("x")-Data.min("x")), Mean3: Data.min("x")+.75*(Data.max("x")-Data.min("x")), Std1: .1*Data.std("x"), Std2: .1*Data.std("x"), Std3: .1*Data.std("x"), Baseline: {val: 0, canFit: false}, common: {s2p: sqrt(2*PI), dVar1: "2*Std1*Std1", dVar2: "2*Std2*Std2", dVar3: "2*Std3*Std3"}'
            },
            {
                name: 'Gaussian (4)',
                expr: 'Amp1*(1/(Std1*s2p))*exp(-(x-Mean1)*(x-Mean1)/dVar1) + Amp2*(1/(Std2*s2p))*exp(-(x-Mean2)*(x-Mean2)/dVar2)+ Amp3*(1/(Std3*s2p))*exp(-(x-Mean3)*(x-Mean3)/dVar3) + Amp4*(1/(Std4*s2p))*exp(-(x-Mean4)*(x-Mean4)/dVar4) + Baseline # Amp1: .25, Amp2: .25, Amp3: .25, Amp4: .25, Mean1: Data.min("x")+.2*(Data.max("x")-Data.min("x")), Mean2: Data.min("x")+.4*(Data.max("x")-Data.min("x")), Mean3: Data.min("x")+.6*(Data.max("x")-Data.min("x")), Mean4: Data.min("x")+.8*(Data.max("x")-Data.min("x")), Std1: .1*Data.std("x"), Std2: .1*Data.std("x"), Std3: .1*Data.std("x"), Std4: .1*Data.std("x"), Baseline: {val: 0, canFit: false}, common: {s2p: sqrt(2*PI), dVar1: "2*Std1*Std1", dVar2: "2*Std2*Std2", dVar3: "2*Std3*Std3", dVar4: "2*Std4*Std4"}'
            },
            {
                name: 'Gaussian (5)',
                expr: 'Amp1*(1/(Std1*s2p))*exp(-(x-Mean1)*(x-Mean1)/dVar1) + Amp2*(1/(Std2*s2p))*exp(-(x-Mean2)*(x-Mean2)/dVar2)+ Amp3*(1/(Std3*s2p))*exp(-(x-Mean3)*(x-Mean3)/dVar3) + Amp4*(1/(Std4*s2p))*exp(-(x-Mean4)*(x-Mean4)/dVar4) + Amp5*(1/(Std5*s2p))*exp(-(x-Mean5)*(x-Mean5)/dVar5) + Baseline # Amp1: .2, Amp2: .2, Amp3: .2, Amp4: .2, Amp5: .2, Mean1: Data.min("x")+.1*(Data.max("x")-Data.min("x")), Mean2: Data.min("x")+.3*(Data.max("x")-Data.min("x")), Mean3: Data.min("x")+.5*(Data.max("x")-Data.min("x")), Mean4: Data.min("x")+.7*(Data.max("x")-Data.min("x")), Mean5: Data.min("x")+.9*(Data.max("x")-Data.min("x")), Std1: .1*Data.std("x"), Std2: .1*Data.std("x"), Std3: .1*Data.std("x"), Std4: .1*Data.std("x"), Std5: .1*Data.std("x"), Baseline: {val: 0, canFit: false}, common: {s2p: sqrt(2*PI), dVar1: "2*Std1*Std1", dVar2: "2*Std2*Std2", dVar3: "2*Std3*Std3", dVar4: "2*Std4*Std4", dVar5: "2*Std5*Std5"}'
            },
            {
                name: 'Declining Exponential',
                expr: 'a + (b-a)*(1 - exp(-(x-x0) / Tau)) + Slope*(x-x0) # Tau: .25*(Data.max("x")-Data.min("x")), a: +Data.y[0], b: +Data.y[Data.n-1], Slope: {val: 0, canFit: false}, x0: {val: 0, canFit: false}'
            },
            {
                name: 'Declining Exponential (2)',
                expr: 'a + (b-c)*(1 - exp(-(x-x0) / Tau1)) + c*(1-exp(-(x-x0) / Tau2)) + Slope*(x-x0) # Tau1: .25*(Data.max("x")-Data.min("x")), Tau2: (Data.max("x")-Data.min("x")), a: +Data.y[0], b: +Data.y[Data.n-1], c: -.5*(Data.y[Data.n-1]-Data.y[0]), Slope: {val: 0, canFit: false}, x0: {val: 0, canFit: false}'
            },
            {
                name: 'Exponential (log x)',
                expr: 'Area*(exp(-px/r/Tau) - exp(-px/Tau)) # Tau: pow(10, Data.mean("x")) # common: {r: "exp(log(logbase)*(Data.x[1]-Data.x[0]))"} # common: {px: "pow(logbase, x)"} # logbase: {val: 10, canFit: false}'
            },
            {
                name: 'Exponential (log x) (2)',
                expr: 'Area1*(exp(-px/r/Tau1) - exp(-px*/Tau1)) + Area2*(exp(-px/r/Tau2) - exp(-px/Tau2)) # Tau1: pow(10, Data.min("x")+.25*(Data.max("x")-Data.min("x"))), Tau2: pow(10, Data.min("x")+.75*(Data.max("x")-Data.min("x"))) # common: {r: "exp(log(logbase)*(Data.x[1]-Data.x[0]))"} # common: {px: "pow(logbase, x)"} # logbase: {val: 10, canFit: false}'
            },
            {
                name: 'Exponential (log x) (3)',
                expr: 'Area1*(exp(-px/r/Tau1) - exp(-px/Tau1)) + Area2*(exp(-px/r/Tau2) - exp(-px/Tau2)) + Area3*(exp(-px/r/Tau3) - exp(-px/Tau3)) # Tau1: pow(10, Data.min("x")+.25*(Data.max("x")-Data.min("x"))), Tau2: pow(10, Data.min("x")+.5*(Data.max("x")-Data.min("x"))), Tau3: pow(10, Data.min("x")+.75*(Data.max("x")-Data.min("x"))) # common: {r: "exp(log(logbase)*(Data.x[1]-Data.x[0]))"} # common: {px: "pow(logbase, x)"} # logbase: {val: 10, canFit: false}'
            },
            {
                name: 'Exponential (log x) (4)',
                expr: 'Area1*(exp(-px/r/Tau1) - exp(-px/Tau1)) + Area2*(exp(-px/r/Tau2) - exp(-px/Tau2)) + Area3*(exp(-px/r/Tau3) - exp(-px/Tau3)) + Area4*(exp(-px/r/Tau4) - exp(-px/Tau4)) # common: {r: "exp(log(logbase)*(Data.x[1]-Data.x[0]))"} # Tau1: pow(10, Data.min("x")+.2*(Data.max("x")-Data.min("x"))), Tau2: pow(10, Data.min("x")+.4*(Data.max("x")-Data.min("x"))), Tau3: pow(10, Data.min("x")+.6*(Data.max("x")-Data.min("x"))), Tau4: pow(10, Data.min("x")+.8*(Data.max("x")-Data.min("x"))) # common: {px: "pow(logbase, x)"} # logbase: {val: 10, canFit: false}'
            },
            {
                name: 'Exponential (log x) (5)',
                expr: 'Area1*(exp(-px/r/Tau1) - exp(-px/Tau1)) + Area2*(exp(-px/r/Tau2) - exp(-px/Tau2)) + Area3*(exp(-px/r/Tau3) - exp(-px/Tau3)) + Area4*(exp(-px/r/Tau4) - exp(-px/Tau4)) + Area5*(exp(-px/r/Tau5) - exp(-px/Tau5)) # common: {r: "exp(log(logbase)*(Data.x[1]-Data.x[0]))"} # Tau1: pow(10, Data.min("x")+.1*(Data.max("x")-Data.min("x"))), Tau2: pow(10, Data.min("x")+.3*(Data.max("x")-Data.min("x"))), Tau3: pow(10, Data.min("x")+.5*(Data.max("x")-Data.min("x"))), Tau4: pow(10, Data.min("x")+.7*(Data.max("x")-Data.min("x"))), Tau5: pow(10, Data.min("x")+.9*(Data.max("x")-Data.min("x"))) # common: {px: "pow(logbase, x)"} # logbase: {val: 10, canFit: false}'
            },
            {
                name: 'Hill',
                expr: 'pow(10, x*Hill) / (pow(10, x*Hill) + pow(EC50, Hill))'
            }
        ];
        exports.userPresets = [];
        try {
            exports.userPresets = JSON.parse(localStorage.getItem('fitness_presets')) || [];
        } catch (err)
        {}
        exports.savePresets = function savePresets() {
            localStorage.setItem('fitness_presets', JSON.stringify(exports.userPresets));
        };

        exports.FitFrame = class FitFrame {
            constructor(config) {
            config = config || {};
                this.config = config;
                this.ownCtx = ! config.context;
                this.context = config.context || new fitness_core.Context(config.fitness);
                this.config.fitness = this.context.config;
                this.node = config.node || document.createElement('div');
                this.node.classList.add('fitness_ui_frame');
                const controlConfig = Object.assign({}, config.controls, {
                    context: this.context,
                    oniteration: () => this.view.updateSoon()
                });
                const viewConfig = Object.assign({}, config.view, {
                    context: this.context,
                    onrss: this.onRSS.bind(this)
                });
                this.controls = new exports.FitControls(controlConfig);
                this.view = new exports.FitView(viewConfig);
                this.config.controls = this.controls.config;
                this.config.view = this.view.config;
                this.node.appendChild(this.view.node);
                this.node.appendChild(this.controls.node);
                this.fit = this.controls.fit.bind(this.controls);
            }
            destroy() {
                this.ownCtx && this.context.destroy();
            }
            getConfig() {
                const config = Object.assign({}, this.config);
                config.view = Object.assign({}, this.config.view);
                config.controls = Object.assign({}, this.config.controls);
                [config, config.view, config.controls].forEach(sub => {
                    delete sub.context;
                    delete sub.node;
                });
                return config;
            }
            ensureXY(colNames) {
                const allNames = colNames.concat(['i']);
                let xName = this.context.xName,
                    yName = this.context.yName;
                if ( allNames.indexOf(xName) < 0 ) {
                    xName = 'i';
                }
                if ( ((xName === 'i') && (yName === 'i')) || (allNames.indexOf(yName) < 0) ) {
                    yName = colNames.length ? colNames[0] : 'i';
                }
                if ( (xName !== this.context.xName) || (yName !== this.context.yName) ) {
                    this.context.setXY(xName, yName);
                }
            }
            onRSS(info) {
                this.controls.showRSS(info);
            }
        };
        
        const FitControlsDefs = {
            maxIter: 200
            // oniteration: function(fitter, params, score, r2, wwRuns, iteration)
        };
        exports.FitControls = class FitControls {
            constructor(config) {
                this.config = Object.assign({}, FitControlsDefs, config);
                this.ownCtx = ! config.context;
                this.ctx = config.context || new fitness_core.Context();
                this.node = config.node || document.createElement('div');
                this.node.classList.add('fitness_ui_ctrls');
                this.panEqs = d3.select(this.node).append('div');
                this.panEqs.classed('fitness_ui_eqs', true);
                let eq = this.panEqs.append('span');
                eq.append('span').text('y=');
                this.txtExpr = new qubx_ui.AcceptedInput(
                    eq.append('input').node(),
                    this.ctx.config.expr,
                    {
                        accept: this.__acceptExpr.bind(this),
                        format: this.__formatExpr.bind(this)
                    }
                );
                this.txtExpr.node.title = 'Enter f(x), involving\n'+
                    ' - constants: PI, E\n'+
                    ' - math fns: sin, pow, log, log10, sqrt, ...\n'+
                    ' - columns by name, and the row number i\n'+
                    'Multiply using *, e.g. m*x + b, not m x + b.\n'+
                    '\n'+
                    'To fit multiple columns, enter eqns separated by hash marks, e.g.\n'+
                    '  FOO=m*x+b # BAR=m*x+b\n'+
                    'where FOO and BAR are column names.';
                this.txtExpr.ev.change.sub(this.__onChangeExpr.bind(this));
                this.mnuPresets = new qubx_ui.Popup({
                    node: eq.append('select').node(),
                    getOptions: this.__getPresets.bind(this),
                    onChoose: this.__choosePreset.bind(this)
                });
                eq = this.panEqs.append('span')
                    .style('flex', '.25 1 auto');
                eq.append('span').text('Weight:');
                this.txtWeight = new qubx_ui.AcceptedInput(
                    eq.append('input').node(),
                    this.ctx.config.weightExpr,
                    {
                        accept: this.__acceptWeight.bind(this),
                        format: this.__formatWeight.bind(this),
                        widthChars: 4
                    }
                );
                this.txtWeight.ev.change.sub(this.__onChangeWeight.bind(this));
                this.txtWeight.node.title = 'Scales the residual, to give some points more importance or confidence.\n\n'+
                    'For even weights, enter 1.0.\n\n'+
                    'To weight by a column or its inverse, enter its name, e.g.\n'+
                    '  FOO\n'+
                    '  1/FOO\n'+
                    'Or enter any f(x, i, column names...)';
                this.panParams = d3.select(this.node).append('div');
                this.panParams.classed('fitness_ui_params', true);
                this.panParams.append('div')
                    .text('Parameters:')
                    .classed('fitness_ui_params_lbl', true);
                this.panParamsAll = this.panParams.append('div');
                this.panParamsAll.classed('fitness_ui_params_all', true);
                this.panParamsAdjust = this.panParams.append('div');
                this.panParamsAdjust.classed('fitness_ui_params_adjust', true);
                this.panParamsAdjust.style({display: 'none'});
                let row = this.panParamsAdjust.append('div');
                row.classed('fitness_ui_params_adjust1', true);
                this.lblParamAdjust = row.append('b'); // e.g. 'm:' 
                this.__adjusting = null; // name of adjusting param
                this.txtParamVal = new qubx_ui.AcceptedInput(
                    row.append('input')
                        .property('type', 'number')
                        .node(),
                    1.0,
                    {
                        accept: qubx.acceptFloat,
                        format: qubx.formatK0,
                        widthChars: 8
                    }
                );
                this.txtParamVal.ev.change.sub(this.__onChangeParamVal.bind(this));
                this.focusParamItem = this.txtParamVal;
                this.txtParamVal.node.addEventListener('focus', function() {
                    this.focusParamItem = this.txtParamVal;
                }.bind(this));
                this.btnParamLimits = row.append('button')
                    .text('Limits...')
                    .on('click', this.__onClickParamLimits.bind(this));
                row = this.panParamsAdjust.append('div');
                row.classed('fitness_ui_params_adjust1', true);
                this.sliderParam = new qubx_ui.SnapSlider({
                    node: this.panParamsAdjust.append('input').node()
                });
                this.sliderParam.ev.change.sub(this.__onSlideParam.bind(this));
                this.panBottom = d3.select(this.node).append('div');
                this.panBottom.classed('fitness_ui_bottom', true);
                this.panResults = this.panBottom.append('div');
                this.panResults.classed('fitness_ui_results', true);
                this.panRSS = this.panResults.append('div');
                this.panRSS.classed('fitness_ui_result', true);
                this.panRSS.append('a').html('RSS')
                    .property({
                        target: '_blank',
                        href: links.rss
                    });
                this.panRSS.property('title', tips.rss);
                this.panRSS.append('span').html(':&nbsp;');
                this.spanRSS = this.panRSS.append('span').text('0.0');
                this.spanRSS.classed('fitness_ui_result_val', true);
                this.panR2 = this.panResults.append('div');
                this.panR2.classed('fitness_ui_result', true);
                this.panR2.append('a').html('R<sup>2</sup>')
                    .property({
                        target: '_blank',
                        href: links.r2
                    });
                this.panR2.property('title', tips.r2);
                this.panR2.append('span').html(':&nbsp;');
                this.spanR2 = this.panR2.append('span').text('0.0');
                this.spanR2.classed('fitness_ui_result_val', true);
                this.panWW = this.panResults.append('div');
                this.panWW.classed('fitness_ui_result', true);
                this.panWW.append('a').html('P(runs)')
                    .property({
                        target: '_blank',
                        href: links.ww
                    });
                this.panWW.property('title', tips.ww);
                this.panWW.append('span').html(':&nbsp;');
                this.spanWW = this.panWW.append('span').text('0.0');
                this.spanWW.classed('fitness_ui_result_val', true);
                this.panControls = this.panBottom.append('div');
                this.panControls.classed('fitness_ui_controls', true);
                this.userControls = this.panControls.append('div');
                this.userControls.classed('fitness_ui_controls_l', true);
                this.btnSettings = this.panControls.append('button')
                    .on('click', this.__onClickSettings.bind(this));
                this.btnSettings.append('img')
                    .property({
                        src: "images/appbar.settings.svg",
                        alt: "[Settings...]",
                        title: "Settings"
                    });
                this.btnFit = this.panControls.append('button')
                    .on('click', this.__onClickFit.bind(this));
                this.btnFit.append('b')
                    .text('Fit');
                this.btnFit.style({
                    padding: '.55em'
                });
                
                this.dlgSettings = new qubx_ui.Modal({
                    title: 'QUB Fitness settings'
                });
                this.dlgSettings.dom.then(dom => {
                    const root = d3.select(dom.content);
                    root.classed('fitness_ui_settings', true);
                    const line = root.append('div');
                    line.append('span').text('Max iterations:');
                    this.txtMaxIter = new qubx_ui.AcceptedInput(
                        line.append('input').node(),
                        this.config.maxIter,
                        {
                            accept: qubx.acceptIntAtLeast(1),
                            format: qubx.formatStr,
                            widthChars: 5
                        }
                    );
                    root.append('div').html('<h4>About Fitness</h4><p>'+
                                            'Nonlinear least-squares curve fitting; '+
                                            'part of the <a href="https://qub.mandelics.com">QUB</a> project for analyzing single-molecule data.'+
                                            '</p><p>'+
                                            '<a target="_blank" href="fitness.html">Stand-alone curve fitting app</a><br>'+
                                            '<a target="_blank" href="qub-online/sources.html">Browse source code</a><br>'+
                                            '<a target="_blank" download="qub-online.zip" href="qub-online.zip">Download source zip</a><br>'+
                                            '</p>');
                });
                
                this.dlgLimits = new qubx_ui.Modal({
                    title: 'QUB Fitness - Parameter limits'
                });
                this.dlgLimits.dom.then(dom => {
                    const root = d3.select(dom.content);
                    root.classed('fitness_ui_limits', true);
                    let line = root.append('div');
                    line.append('span').text('Optional limits on parameter ');
                    this.spanLimitName = line.append('b');
                    line.append('span').html(':');
                    line = root.append('div');
                    line.append('span').html('Low:&nbsp;');
                    this.txtLimitLo = new qubx_ui.AcceptedInput(
                        line.append('input').node(),
                        qubx.UNSET,
                        {
                            accept: qubx.acceptFloatOrUnset,
                            format: qubx.formatOrUnset(qubx.formatK0),
                            widthChars: 10
                        }
                    );
                    line = root.append('div');
                    line.append('span').html('High:&nbsp;');
                    this.txtLimitHi = new qubx_ui.AcceptedInput(
                        line.append('input').node(),
                        qubx.UNSET,
                        {
                            accept: qubx.acceptFloatOrUnset,
                            format: qubx.formatOrUnset(qubx.formatK0),
                            widthChars: 10
                        }
                    );
                });
                
                this.dlgResults = new qubx_ui.Modal({
                    title: 'QUB Fitness - Results',
                    canCancel: false,
                    okLabel: 'Close'
                });
                this.dlgResults.dom.then(dom => {
                    const root = d3.select(dom.content);;
                    root.classed('fitness_ui_fitted', true);
                    const line = root.append('div');
                    line.append('span').text('Done fitting after ');
                    this.spanResultIter = line.append('b');
                    line.append('span').text(' iterations.');
                    this.tblResultScores = new qubx_ui.ColumnTable({
                        node: root.append('table').node(),
                        columns: [
                            qubx_ui.ColumnTable.htmlCol('', 'name', qubx.formatStr),
                            qubx_ui.ColumnTable.txtCol('', 'val', qubx.formatK0)
                        ]
                    });
                    root.append('div').append('b').text('Parameters:');
                    this.tblResultParams = new qubx_ui.ColumnTable({
                        node: root.append('table').node(),
                        columns: [
                            qubx_ui.ColumnTable.txtCol('Name', 'name', qubx.formatStr),
                            qubx_ui.ColumnTable.txtCol('Value', 'val', qubx.formatK0),
                            qubx_ui.ColumnTable.txtCol('\u00B1', 'stdErr', qubx.formatOrUnset(qubx.formatK0))
                        ]
                    });
                });
                
                this.ctx.ev.data.sub(this.__onData.bind(this));
                this.ctx.ev.xy.sub(this.__onXY.bind(this));
                this.ctx.ev.expr.sub(this.__onExpr.bind(this));
                this.ctx.ev.weightExpr.sub(this.__onWeightExpr.bind(this));
                this.ctx.ev.param.sub(this.__onParam.bind(this));
                this.ctx.ev.startFit.sub(this.__onStartFit.bind(this));
                this.ctx.ev.endFit.sub(this.__onEndFit.bind(this));
                this.__onExpr(this.ctx.config.expr, this.ctx.exprInfo);
            }
            destroy() {
                this.ownCtx && this.ctx.destroy();
            }
            showRSS(info) {
                this.spanRSS.text(qubx.formatK0u(info.rss));
                this.spanR2.text(qubx.formatK0u(info.r2));
                this.spanWW.text(qubx.formatK0u(info.wwRuns));
            }
            __onData() {
                if ( (! this.ctx.data[this.ctx.xName]) || (! this.ctx.data[this.ctx.yName]) ) {
                    this.ctx.setExpr(this.ctx.config.expr)
                        .catch(qubx.reportErr);
                    this.ctx.setWeightExpr(this.ctx.config.weightExpr)
                        .catch(qubx.reportErr);
                }
            }
            __onXY(xName, yName) {
                this.ctx.setExpr(this.ctx.config.expr)
                    .catch(qubx.reportErr);
                this.ctx.setWeightExpr(this.ctx.config.weightExpr)
                    .catch(qubx.reportErr);
            }
            __acceptExpr(expr) {
                if ( this.fitting ) {
                    throw new Error('busy');
                }
                return expr; /// ?
            }
            __formatExpr(x) {
                return x;
            }
            __onChangeExpr(x) {
                this.ctx.setExpr(x)
                    .then(() => this.txtExpr.node.classList.remove('fitness_expr_invalid'))
                    .catch(err => this.txtExpr.node.classList.add('fitness_expr_invalid'));
            }
            __acceptWeight(expr) {
                if ( this.fitting ) {
                    throw new Error('busy');
                }
                return expr; /// ?
            }
            __formatWeight(x) {
                return x;
            }
            __onChangeWeight(x) {
                this.ctx.setWeightExpr(x)
                    .then(() => this.txtWeight.node.classList.remove('fitness_expr_invalid'))
                    .catch(err => this.txtWeight.node.classList.add('fitness_expr_invalid'));
            }
            __onChangeParamVal(v) {
                if ( ! (this.fitting || this.ctx.setParam(this.__adjusting, v)) ) {
                    this.txtParamVal.update(this.ctx.params[this.__adjusting].val);
                } else {
                    this.sliderParam.value(v);
                }
            }
            __onClickParamLimits() {
                this.spanLimitName.text(this.__adjusting);
                const p = this.__paramStore[this.__adjusting];
                this.txtLimitLo.update(p.lo);
                this.txtLimitHi.update(p.hi);
                this.dlgLimits.run().then(() => {
                    this.ctx.setParam(this.__adjusting, {
                        lo: this.txtLimitLo.value,
                        hi: this.txtLimitHi.value
                    });
                }).catch(qubx.reportErr);
            }
            __onSlideParam(v) {
                if ( this.fitting ) {
                    return;
                }
                this.focusParamItem = this.sliderParam;
                this.ctx.setParam(this.__adjusting, v);
            }
            __getPresets() {
                this.__allPresets = exports.presets.concat(exports.userPresets);
                return ['Save new preset...'].concat(this.__allPresets.map(qubx.m('name')));
            }
            __choosePreset(label, i) {
                try {
                    if ( i ) {
                        const p = this.__allPresets[i-1];
                        this.txtExpr.update(p.expr);
                        this.ctx.setExpr(p.expr);
                        p.params && p.params.forEach(param => this.ctx.setParam(param.name, param));
                    } else {
                        const name = prompt('Title of new preset:', 'Untitled');
                        if ( name ) {
                            exports.userPresets.push({
                                name: name,
                                expr: this.txtExpr.value,
                                params: this.__paramList.map(p => ({
                                    name: p.name,
                                    val: p.val,
                                    lo: p.lo,
                                    hi: p.hi,
                                    canFit: p.canFit
                                }))
                            });
                        }
                    }
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
            __onExpr(expr, info) {
                try {
                    const self = this;
                    this.txtExpr.update(expr);
                    this.__paramList = info.params.map(name => {
                        this.__paramStore = this.__paramStore || {};
                        let p = this.__paramStore[name];
                        if ( ! p ) {
                            p = Object.create(this.ctx.params[name]);
                            this.__paramStore[name] = p;
                        }
                        return p;
                    });
                    const pdivs = this.panParamsAll.selectAll('div')
                          .data(this.__paramList, qubx.m('name'));
                    const pdiv = pdivs.enter().append('div')
                          .each(function(d,i) {
                              d.div = this;
                          });
                    pdiv.classed('fitness_ui_param', true);
                    pdiv.each(function(d,i) {
                        d.chkCanFit = d3.select(this).append('input')
                            .property({
                                type: 'checkbox'
                            }).on('change', function(d,i) {
                                self.ctx.setParam(d.name, {canFit: d.chkCanFit.node().checked});
                            });
                        d.lblName = d3.select(this).append('span')
                            .text(d.name+'=');
                        d.lblVal = d3.select(this).append('span');
                    }).on('click', d => {
                        this.__adjust(d.name);
                        pdivs.classed('fitness_ui_param_sel', qubx.m('adjusting'));
                    }).classed('fitness_ui_param_const', d => ! d.canFit);
                    pdivs.exit().remove();
                    pdivs.each(d => {
                        d.chkCanFit.node().checked = d.canFit;
                        d.lblVal.text(qubx.formatK0u(d.val));
                    });
                    if ( this.__adjusting ) {
                        if ( info.params.indexOf(this.__adjusting) < 0 ) {
                            this.__adjust(null);
                        }
                        this.__labelLimits();
                    }
                } catch (err) {
                    qubx.reportErr(err);
                }
            }                
            __adjust(name) {
                try {
                    if ( this.__adjusting ) {
                        this.__paramStore[this.__adjusting].adjusting = false;
                    }
                    this.__adjusting = name;
                    if ( name ) {
                        const p = this.__paramStore[name];
                        p.adjusting = true;
                        this.panParamsAdjust.style('display', '');
                        this.lblParamAdjust.html(name+':&nbsp;');
                        this.txtParamVal.update(p.val);
                        this.sliderParam.value(p.val);
                        p.canFit && (this.focusParamItem.select ? this.focusParamItem.select() : this.focusParamItem.focus());
                        this.__labelLimits(p);
                    } else {
                        this.panParamsAdjust.style('display', 'none');
                    }
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
            __labelLimits(param) {
                param = param || this.__paramStore[this.__adjusting];
                this.btnParamLimits.html('Limits: ['+((param.lo===qubx.UNSET) ? '\u2212&infin;' : qubx.formatK0u(param.lo))
                                         +', '+((param.hi===qubx.UNSET) ? '&infin;' : qubx.formatK0u(param.hi))
                                         +']');
            }
            __onWeightExpr(expr) {
                this.txtWeight.update(expr);
            }
            __onParam(p) {
                if ( p.name === this.__adjusting ) {
                    this.txtParamVal.update(p.val);
                    this.sliderParam.value(p.val, true);
                    this.__labelLimits(p);
                }
                this.__paramList.forEach(param => {
                    if ( param.name === p.name ) {
                        param.chkCanFit.node().checked = p.canFit;
                        param.lblVal.text(qubx.formatK0u(p.val));
                        d3.select(param.div).classed('fitness_ui_param_const', ! p.canFit);
                    }
                });
            }
            __onStartFit() {
                this.fitting = true;
                this.btnFit.textContent = 'Stop';
            }
            __onEndFit() {
                this.config.onfit && this.config.onfit();
                this.fitting = false;
                this.btnFit.textContent = 'Fit';
            }
            __onClickSettings() {
                this.txtMaxIter.update(this.config.maxIter);
                this.dlgSettings.run().then(() => {
                    this.config.maxIter = this.txtMaxIter.value;
                });
            }
            __onClickFit() {
                if ( this.fitting ) {
                    this.ctx.stop();
                    return;
                }
                this.fit()
                    .then(result => this.showFit(result));
            }
            fit() {
                if ( this.fitting) {
                    return Promise.reject('Busy');
                }
                return this.ctx.fit({
                    maxIter: this.config.maxIter,
                    oniteration: (fitter, params, score, r2, wwRuns, iteration) => {
                        this.__paramList.forEach(d => d.lblVal.text(qubx.formatK0u(d.val)));
                        if ( this.__adjusting ) {
                            const v = this.ctx.params[this.__adjusting].val;
                            this.txtParamVal.update(v);
                            this.sliderParam.value(v);
                        }
                        this.config.oniteration && this.config.oniteration(fitter, params, score, r2, wwRuns, iteration);
                        this.showRSS({
                            rss: score,
                            r2: r2,
                            wwRuns: wwRuns
                        });
                    }
                }).then(result => {
                    this.showRSS({
                        rss: result.score,
                        r2: result.r2,
                        wwRuns: result.wwRuns
                    });
                    return result;
                }).catch(qubx.reportErr);
            }
            showFit(result) {
                // {params, score, r2, wwRuns, iteration}
                return this.ctx.getErr(result).then(errLim => {
                    this.spanResultIter.text(''+result.iteration);
                    this.tblResultScores.tbl.html('');
                    this.tblResultScores.render([
                        {
                            name: 'RSS',
                            val: result.score
                        },
                        {
                            name: 'R<sup>2</sup>',
                            val: result.r2
                        },
                        {
                            name: 'P(runs)',
                            val: result.wwRuns
                        }
                    ], 'name');
                    this.tblResultParams.tbl.html('');
                    this.tblResultParams.render(this.__paramList.map(p => ({
                        name: p.name,
                        val: result.params[p.name],
                        stdErr: p.canFit ? errLim.stdErrEst[p.name] : qubx.UNSET
                    })), 'name');
                    
                    //// correlation? does it work?
                    
                    return this.dlgResults.run();
                }).catch(qubx.reportErr);
            }
        };
        
        const FitViewDefs = {
            onrss: function(info) {} // {rss, r2, wwRuns}
        };
        exports.FitView = class FitView {
            constructor(config) {
                config = config || {};
                this.config = Object.assign({}, FitViewDefs, config);
                this.ownCtx = ! config.context;
                this.ctx = config.context || new fitness_core.Context();
                this.node = config.node || document.createElement('div');
                this.node.classList.add('fitness_ui_plotsvars');
                this.panVars = d3.select(this.node).append('div');
                this.panVars.classed('fitness_ui_vars', true);
                this.panVars.append('span').text('x:');
                this.mnuX = new qubx_ui.Select({
                    node: this.panVars.append('select').node(),
                    onChange: this.pickX.bind(this)
                });
                this.panVars.append('span').text('y:');
                this.mnuY = new qubx_ui.Select({
                    node: this.panVars.append('select').node(),
                    onChange: this.pickY.bind(this)
                });
                this.boxPlots = d3.select(this.node).append('div');
                this.boxPlots.classed('fitness_ui_plotsbox', true);
                this.panPlots = new qubx_ui.AspectGrid3({
                    node: this.boxPlots.append('div').node()
                });
                this.panPlots.node.classList.add('fitness_ui_plots');
                
                this.plotsByName = {};
                this.updateSoon = qubx.batchDelayed(this.update.bind(this), 300, true);
                this.ctx.ev.data.sub(this.__onData.bind(this));
                this.ctx.ev.xy.sub(this.__onXY.bind(this));
                this.ctx.ev.expr.sub(this.__onExpr.bind(this));
                this.ctx.ev.param.sub(this.__onParam.bind(this));
                this.ctx.ev.startFit.sub(this.__onStartFit.bind(this));
                this.ctx.ev.endFit.sub(this.__onEndFit.bind(this));
                this.__onExpr(this.ctx.config.expr, this.ctx.exprInfo);
            }
            destroy() {
                this.ownCtx && this.ctx.destroy();
            }
            pickX(name, index) {
                name && this.ctx.setXY(name, this.ctx.yName);
                this.__staleX = true;
            }
            pickY(name, index) {
                name && this.ctx.setXY(this.ctx.xName, name);
            }
            __onData() {
                const names = Object.keys(this.ctx.data);
                if ( names.indexOf('i') < 0 ) {
                    names.push('i');
                }
                names.sort((a,b) => a.toLowerCase().localeCompare(b.toLowerCase()));
                this.mnuX.options(names);
                this.mnuY.options(names);
                this.__resetPlots();
                this.updateSoon();
            }
            __resetPlots() {
                this.panPlots.data([], qubx.m('name'), () => {}, () => {}, p => { // teardown
                                       p.fitted = p.chart = p.node = null;
                });
                this.plots = [];
                this.__staleX = false;
                this.__replot();
            }
            __onXY(xName, yName) {
                this.mnuX.selected(xName);
                this.mnuY.selected(yName);
                if ( this.__staleX ) {
                    this.__resetPlots();
                }
                this.updateSoon();
            }
            __onExpr(expr, info) {
                this.__replot(info);
                this.updateSoon();
            }
            __onParam(p) {
                this.updateSoon();
            }
            __onStartFit() {
                this.fitting = true;
            }
            __onEndFit() {
                this.fitting = false;
            }
            __replot(info) {
                try {
                    info = info || this.exprInfo;
                    if ( ! info ) {
                        return;
                    }
                    this.exprInfo = info;
                    const colsUsed = [...(new Set(info.series.map(ser => ser.name || this.ctx.yName).filter(qubx.i)))];
                    colsUsed.sort((a,b) => {
                        if ( a === this.ctx.yName ) {
                            return -1; // yName plotted first
                        }
                        return a.toLowerCase().localeCompare(b.toLowerCase());
                    });
                    const plots = colsUsed.map(name => {
                        let p = this.plotsByName[name];
                        if ( ! p ) {
                            p = {
                                name: name
                            };
                            this.plotsByName[name] = p;
                        }
                        return p;
                    });
                    plots.forEach(p => {
                        for (const series of info.series) {
                            if ( (series.name || this.ctx.yName) === p.name ) {
                                p.series = series;
                                break;
                            }
                        }
                        p.data = this.ctx.data[p.name];
                        if ( p.data && ((! p.fitted) || (p.fitted.length !== p.data.length)) ) {
                            p.fitted = new Float32Array(p.data.length);
                        }
                    });
                    const self = this,
                          iSeries = this.ctx.data.i && Array.prototype.slice.call(this.ctx.data.i);
                    this.plots = plots;
                    this.panPlots.data(plots, qubx.m('name'), function(p) { // init
                        p.node = this;
                        p.node.classList.add('fitness_ui_plot');
                        p.chart = new qubx_chart.Chart({
                            node: p.node,
                            legend: {
                                show: true
                            }
                        });
                        const chartSeries = {};
                        chartSeries[p.name] = p.data && Array.prototype.slice.call(p.data);
                        chartSeries.i = iSeries;
                        const tableColumns = ((self.ctx.xName==='i')?[]:['i']).concat([self.ctx.xName, p.name]);
                        const chartDef = {
                            series: chartSeries,
                            yvar: p.name,
                            color: "#000",
                            columns: tableColumns
                        };
                        chartDef.xvar = self.ctx.xName;
                        const typedArr = self.ctx.data[self.ctx.xName];
                        chartSeries[self.ctx.xName] = typedArr && Array.prototype.slice.call(typedArr);
                        p.chart.setScatterPlot(chartDef)
                            .trendCount(1, () => 'Fit', () => '#f00');
                    }, p => { // layout
                        p.chart.resize();
                    }, p => { // teardown
                        p.fitted = p.chart = p.node = null;
                    });
                    this.updateSoon();
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
            update() {
                if ( ! (this.ctx.xName && this.ctx.yName && this.plots.length) ) {
                    return;
                }
                if ( this.__updating ) {
                    this.__wantUpdate = true;
                    return;
                }
                this.__updating = true;
                this.__wantUpdate = false;
                const cleanup = () => {
                    this.__updating = false;
                    this.__wantUpdate && this.updateSoon();
                };
                (this.fitting ? Promise.resolve(false) : this.ctx.getRSS()).then(rssResult => {
                    rssResult && this.config.onrss(rssResult);
                    return this.ctx.getFit(this.plots.map(qubx.m('fitted'))).then(arrs => {
                        this.plots.forEach((p,i) => {
                            p.fitted = arrs[i];
                            p.chart.trends[0].setData(Array.prototype.slice.call(p.fitted));
                        });
                        cleanup();
                    });
                }).catch(qubx.reportErr)
                    .catch(cleanup);
            }
        };

        exports.SharedFitter = class SharedFitter {
            constructor(config) {
                this.ctx = new fitness_core.Context(config && config.fitness);
                this._sendCurveSoon = qubx.batchDelayed(this._sendCurve.bind(this), 250, true);
                this.ctx.ev.expr.sub(this._sendCurveSoon);
                this.ctx.ev.param.sub(this._sendCurveSoon);
                this.dlg = new qubx_ui.Modal({
                    title: 'Curve Fitting',
                    canCancel: false,
                    okLabel: 'Close'
                });
                this.dlg.dom.then(dom => {
                    dom.root.style.opacity = '0.9';
                    dom.content.classList.add('fitness_shared');
                    dom.content.appendChild(this.controls.node);
                });
                this.controls = new exports.FitControls({
                    context: this.ctx,
                    oniteration: (fitter, params, score, r2, wwRuns, iteration) => {
                        ///?
                    },
                    onfit: () => {
                        this._sendCurve()
                            .then(() => this._onfit && this._onfit());
                    }
                });
                this.controls.node.classList.add('qubx_fullsize');
            }
            _sendCurve() {
                if ( ! this._oncurve ) {
                    return Promise.resolve(true);
                }
                return this.ctx.getRSS()
                    .then(info => this.ctx.getFit(this.fitted || []).then(fitted => {
                        this.fitted = fitted;
                        this._oncurve(fitted[0], info);
                    })).catch(qubx.reportErr);
            }
            run(opts) {
                this._oncurve = opts.oncurve;
                this._onfit = opts.onfit;
                let saved = localStorage.getItem('fitness_shared_'+opts.category);
                try {
                    saved = JSON.parse(saved);
                } catch (err) {
                    saved = null;
                }
                if ( saved && saved.config ) {
                    this.ctx.restoreState(saved);
                } else {
                    (opts.xName && opts.yName && this.ctx.setXY(opts.xName, opts.yName) || Promise.resolve(true)).then(() => {
                        opts.weightExpr && this.ctx.setWeightExpr(opts.weightExpr);
                        opts.expr && this.ctx.setExpr(opts.expr);
                    });
                }
                this.ctx.setColumns(opts.columns);
                return this.dlg.run().then(() => {
                    this._onfit = undefined;
                    const state = this.ctx.getState();
                    localStorage.setItem('fitness_shared_'+opts.category, JSON.stringify(state));
                    this.ctx.setColumns({});
                    return {
                        expr: this.ctx.config.expr,
                        weightExpr: this.ctx.config.weightExpr,
                        params: state.params
                    };
                }).catch(qubx.reportErr);
            }
        };

        let sharedFitDlg = null;
        exports.runFitControls = function runFitControls(opts) {
            sharedFitDlg = sharedFitDlg || new exports.SharedFitter();
            return sharedFitDlg.run(opts);
        };
        
        return exports;
    });
