/* Copyright 2016 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("hjcfit", ["asmffi"], function(asmffi) {
    'use strict';

    return asmffi.load('qubxhjc', {
        'qub_hjc_bursts_alloc': ['void*', []],
        'qub_hjc_bursts_free': ['void', ['void*']],
        'qub_hjc_bursts_append': ['void', ['void*', 'float*', 'int']],
        'qub_hjc_likelihood': ['double', ['void*', 'double*', 'int', 'int', 'double', 'double', 'int']]
    }, {}, {
        TOTAL_MEMORY: 16*1024*1024
    }).then(function(lib) {
        lib.Bursts = function Bursts() {
            this.obj = lib.qub_hjc_bursts_alloc();
            this.dwellTransfer = lib.fvec();
            this.qTransfer = lib.dvec();
        };
        lib.Bursts.prototype.destroy = function() {
            lib.qub_hjc_bursts_free(this.obj);
            this.dwellTransfer.destroy();
            this.qTransfer.destroy();
        };
        lib.Bursts.prototype.append = function(dwells) {
            if ( ! (dwells.length % 2) ) {
                throw Error('Burst length must be odd (starting with open event)');
            }
            this.dwellTransfer.set(dwells);
            lib.qub_hjc_bursts_append(this.obj, this.dwellTransfer.p, dwells.length);
        };

        lib.likelihood = function likelihood(bursts, Q, numOpen, tDead_ms, tCrit_ms, useCHSvectors) {
            numOpen = numOpen|0;
            tDead_ms = +tDead_ms;
            tCrit_ms = (+tCrit_ms) || 1000.0;
            useCHSvectors = useCHSvectors|0;
            var numState = Q.length;
            if ( numState && Q[0].length ) { // is 2d?
                numState = Q[0].length;
                bursts.qTransfer.size(numState*numState);
                for (var i=0; i<numState; ++i) {
                    for (var j=0; j<numState; ++j) {
                        bursts.qTransfer.set(i*numState+j, Q[i][j]);
                    }
                }
            } else {
                numState = Math.sqrt(numState)|0;
                bursts.qTransfer.set(Q);
            }
            return lib.qub_hjc_likelihood(bursts.obj, bursts.qTransfer.p, numState, numOpen, tDead_ms, tCrit_ms, useCHSvectors);
        };

        return lib;
    });
});
