define("quboptstatview", ["jquery", "jqplot", "quboptstat"], function($, jqplot, quboptstat) {
    "use strict";
    var exports = {};
    
    var ViewConfig = {};
    exports.View = function(config) {
        this.config = $.extend({}, ViewConfig, config);
        this.node = this.config.node;
        this.plot = this.node.find("#qosPlot");
        this.plot.bind('jqplotDataClick', function (ev, seriesIndex, pointIndex, data) {
            console.log('click', ev, seriesIndex, pointIndex, data);
        });
        this.node.find("#qosUpdate").on("click", this._onClickUpdate.bind(this));
        this.client = null;
        quboptstat.client.then(function(client) {
            this.client = client;
            this.update();
        }.bind(this));
    };
    exports.View.prototype._onClickUpdate = function(ev) {
        this.client && this.update();
    };
    exports.View.prototype.update = function() {
        var q = {};
        var input = this.node.find("#qosAlgorithm");
        var val;
        if ( input.val() !== '-Any-' && input[0].selectedOptions.length ) {
            q.algorithmId = parseInt(input[0].selectedOptions[0].getAttribute('data-algorithmId'));
        }
        input = this.node.find("#qosNauty");
        if ( input.val() !== '-Any-' && input[0].selectedOptions.length ) {
            q.nautyId = parseInt(input[0].selectedOptions[0].getAttribute('data-nautyId'));
        }
        input = this.node.find("#qosStim");
        if ( input.val() !== '-Any-' && input[0].selectedOptions.length ) {
            q.stimId = parseInt(input[0].selectedOptions[0].getAttribute('data-stimId'));
        }
        input = this.node.find("#qosDataCount");
        if ( input.val() ) {
            val = parseInt(input.val());
            if ( val !== NaN ) {
                q.dataCount = val;
            }
        }
        input = this.node.find("#qosChannelCount");
        if ( input.val() ) {
            val = parseInt(input.val());
            if ( val !== NaN ) {
                q.channelCount = val;
            }
        }
        input = this.node.find("#qosSampling");
        if ( input.val() ) {
            val = parseFloat(input.val());
            if ( val && (val !== NaN) ) {
                q.sampling = 1e-3/val; // kHz -> seconds
            }
        }
        this.client.getRates(q).then(function(res) {
            function iseries(rateIndex, rateType) { return 3*rateIndex + rateType; }
            var src_series = [];
            res.rates.forEach(function(r) {
                var i = iseries(r.rateIndex, r.rateType);
                if ( typeof(src_series[i]) === 'undefined' ) {
                    src_series[i] = [];
                }
                src_series[i].push(r);
            });
            var series = src_series.map(function(s) {
                return s.map(function(r) {
                    var sgn = r.rateFactor < 0 ? -1 : 1;
                    return [r.rateInit, sgn*Math.log10(sgn*r.rateFactor)];
                });
            });
            var seriesCfg = src_series.map(function(s, i) {
                var rateType = i % 3;
                var rateIndex = Math.floor(i/3) | 0;
                return {
                    label: "k"+rateType+"_"+rateIndex,
                    showLine: false,
                    showMarker: true,
                    markerOptions: {
                        style: ['filledCircle', 'filledDiamond', 'filledSquare'][rateType]
                        // size: 9
                    }
                };
            });
            for (var i=series.length-1; i>=0; --i) {
                if ( ! series[i] ) {
                    series.splice(i, 1);
                    seriesCfg.splice(i, 1);
                }
            }
            /// replot doesn't clear axes first, looks bad
            /// error no data specified when there are no points
            this.plot.jqplot(series, {
                title: "Title",
                series: seriesCfg
            });
        }.bind(this))
            .catch(function(err) {
                console.log(err);
            });
        // update selects
        this.client.listAlgo().then(function(result) {
            var sel = this.node.find("#qosAlgorithm");
            var v = sel.val();
            sel.find(".qosOptions").remove();
            var opts = $('<optgroup class="qosOptions">');
            result.algorithms.forEach(function(a) {
                var opt = $('<option value="'+a.algorithmName+'">'+a.algorithmName+'</option>');
                opt.attr('data-algorithmId', a.algorithmId);
                if ( v === a.algorithmName ) {
                    opt.attr("selected", "selected");
                }
                opt.appendTo(opts);
            });
            opts.appendTo(sel);
        }.bind(this));
        this.client.listNauty().then(function(result) {
            var sel = this.node.find("#qosNauty");
            var v = sel.val();
            sel.find(".qosOptions").remove();
            var opts = $('<optgroup class="qosOptions">');
            result.nauty.forEach(function(n) {
                var opt = $('<option value="'+n.nautyNumber+'">'+n.nautyNumber+'</option>');
                opt.attr('data-nautyId', n.nautyId);
                opt.attr('data-nautyQmj', JSON.stringify(n.nautyQmj));
                if ( v == n.nautyNumber ) {
                    opt.attr("selected", "selected");
                }
                opt.appendTo(opts);
            });
            opts.appendTo(sel);
        }.bind(this));
        this.client.listStim().then(function(result) {
            var sel = this.node.find("#qosStim");
            var v = sel.val();
            sel.find(".qosOptions").remove();
            var opts = $('<optgroup class="qosOptions">');
            result.stimulus.forEach(function(s) {
                var opt = $('<option value="'+s.stimLbl+'">'+s.stimLbl+'</option>');
                opt.attr('data-stimId', s.stimId);
                if ( v === s.stimLbl ) {
                    opt.attr("selected", "selected");
                }
                opt.appendTo(opts);
            });
            opts.appendTo(sel);
        }.bind(this));
    };
    
    return exports;
});
