/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_app", ["qubx", "qubx_ui", "screenfull"],
    (qubx, qubx_ui, screenfull) => {
        "use strict";
        const exports = {};

        const MainDefs = {
            fullscreen: false,
            topframe: "#qubx_topframe"
            // readHash()
        };
        exports.Main = class Main {
            constructor(config) {
                this.config = Object.assign({}, MainDefs, config);
                this._mode = null;
                this.panels = qubx_ui.loadTemplates([this.config.topframe, "#qubx_banner"])
                    .then(items => {
                        const [top, banner] = items;
                        const p = {
                            top: top,
                            main: top.querySelector("#qubx_main"),
                            banner: banner
                        };
                        top.appendChild(banner);
                        
                        p.controls = p.main.querySelector("#qubx_controlbar");
                        
                        p.fullscreen = p.banner.querySelector(".qubx_btn_fullscreen");
                        p.fullscreen.addEventListener("click", this._onClickFullscreen.bind(this));
                        if (screenfull.enabled) {
                            document.addEventListener(screenfull.raw.fullscreenchange, this._updateFullscreen.bind(this), false);
                        } else {
                            p.fullscreen.style.display = 'none';
                        }
                        
                        config.node && config.node.appendChild(top);
                        
                        this.welcome = new qubx_ui.Modal({
                            title: "QUB Online",
                            template: "#qubx_welcome",
                            canCancel: false,
                            okLabel: 'Start',
                            onOK: () => {
                                this.config.fullscreen = this.welcome.fullscreen.checked;
                                if ( screenfull.enabled && this.config.fullscreen ) {
                                    screenfull.request();
                                }
                            }
                        });
                        this.welcome.dom.then(dom => {
                            this.welcome.fullscreen = dom.content.querySelector("#qubx_welcome_fullscreen");
                            this.welcome.msg = dom.content.querySelector("#qubx_welcome_msg");
                            if ( ! (screenfull.enabled && this.config.fullscreen) ) {
                                this.welcome.fullscreen.checked = false;
                            }
                        }).catch(qubx.reportErr);
                        
                        if ( this.config.readHash ) {
                            setTimeout(() => {
                                this.config.readHash();
                                document.location.hash = "";
                            }, 10);
                        }
                        
                        return p;
                    }).catch(qubx.reportErr);
            }
            dispose() {
            }
            showWelcome(msg) {
                if ( typeof(msg) !== 'undefined' ) {
                    this.welcome.msg.innerHTML = msg;
                }
                return this.welcome.run();
            }
            mode(m) {
                if ( typeof(m) === 'undefined' ) {
                    return this._mode;
                }
                m = m || null;
                if ( m === this._mode ) {
                    return Promise.resolve(true);
                }
                return (this._mode && this._mode.deactivate && this._mode.deactivate() || Promise.resolve(true))
                    .then(() => {
                        this._mode = m;
                        m && m.activate();
                    }).catch(qubx.reportErr);
            }
            _onClickFullscreen(event) {
                screenfull.toggle();
            }
            _updateFullscreen() {
                this.config.fullscreen = screenfull.isFullscreen;
                this.panels.then(function(panels) {
                    const arrows = panels.fullscreen.querySelectorAll("polygon");
                    var bounds;
                    if ( screenfull.isFullscreen ) {
                        let i = 0;
                        for (; i<4; ++i) {
                            bounds = arrows[i].getBBox();
                            arrows[i].setAttributeNS(null, "transform", "rotate(180 "+(bounds.x + .5*bounds.width)+" "+(bounds.y + .5*bounds.height)+")");
                        }
                    } else {
                        let i = 0;
                        for (; i<4; ++i) {
                            arrows[i].setAttributeNS(null, "transform", "");
                        }
                    }
                }).catch(qubx.reportErr);
            }
        };

        return exports;
    });

