/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_chart", ["d3", "c3", "qubx", "qubx_ui", "qubx_fit"], (d3, c3, qubx, qubx_ui, qubx_fit) => {
    "use strict";
    const exports = {};

    const templates = qubx_ui.loadTemplates([".qubx_chart_flip_table", ".qubx_chart_flip_chart", ".qubx_icon_download"])
          .then(items => ({
              flipTable: items[0],
              flipChart: items[1],
              download: items[2]
          })).catch(qubx.reportErr);
    
    const saver = new qubx_ui.Saver();
    let chartStyle = "";
    qubx.get("css/c3-embed.css")
        .then(styleText => {
            chartStyle = styleText;
        }).catch(qubx.reportErr);
    
    exports.trendColors = ['#000', '#f00', '#00f', '#0f0', '#00ffff', '#ff0', '#f0f', '#800', '#865203', '#728c07', '#26930b', '#0f9946', '#149f9f', '#1951a6', '#3b1eac', '#9624b3', '#b92a80', '#bf3030', '#c68c36', '#afcc3d'];

    const ChartDefaults = {
        legend: {
            show: false
        },
        /*zoom: {
            enabled: true,
            rescale: true
        },*/
        onselected: null,
    };
    exports.Chart = class Chart {
        constructor(config) {
            this.trends = [];
            this.config = Object.assign({}, ChartDefaults, config);
            this.node = config.node || document.createElement("div");
            this._setupHistSoon = qubx.batchDelayed(this._setupHist.bind(this), 200, true);
            this._setupScatterSoon = qubx.batchDelayed(this._setupScatter.bind(this), 200, true);
            
            this.cbox = d3.select(this.node).append("div");
            this.cbox.classed("qubx_fullsize", true);
            this.cdiv = this.cbox.append("div");
            this.cdiv.classed("qubx_fullsize", true);
            this.buttons = this.cbox.append('div');
            this.buttons.classed('qubx_chart_buttons', true);
            this.tbox = d3.select(this.node).append("div");
            this.tbox.classed("qubx_fullsize", true);
            this.tbox.style({
                overflow: 'auto',
                display: 'none'
            });
            this.tbltop = this.tbox.append('div');
            this.fit = this.tbox.append("div")
                .classed("qubx_chart_fit", true);
            this.fitInfo = this.fit.append("div")
                .classed("qubx_chart_fitInfo", true);
            this.fitParams = this.fit.append("div")
                .classed("qubx_chart_fitParams", true);
            this.table = this.tbox.append("table");
            this.table.classed('qubx_chart_tbl', true);
            this.thead = this.table.append("thead");
            this.tbody = this.table.append("tbody");
            this.tbuttons = this.tbox.append('div');
            this.tbuttons.classed('qubx_chart_buttons', true);
            this.laidOut = templates.then(t => {
                this.flipTable = d3.select(t.flipTable.cloneNode(true));
                this.buttons.node().appendChild(this.flipTable.node());
                this.dlChart = d3.select(t.download.cloneNode(true));
                this.dlChart.on('click', this._onClickDownloadChart.bind(this));
                this.buttons.node().appendChild(this.dlChart.node());
                this.flipChart = d3.select(t.flipChart.cloneNode(true));
                this.tbuttons.node().appendChild(this.flipChart.node());
                this.dlTable = d3.select(t.download.cloneNode(true));
                this.dlTable.on('click', this._onClickDownloadTable.bind(this));
                this.tbuttons.node().appendChild(this.dlTable.node());
                [this.dlTable, this.dlChart].forEach(dl => {
                    dl.style({
                        width: '1.4em',
                        height: '2em',
                        background: 'rgba(255,255,255,0.8)'
                    });
                });
                [this.flipChart, this.flipTable].forEach((el, i) => {
                    el.style({
                        width: '2em',
                        height: '2em'
                    });
                    el.on('click', d => {
                        try {
                            this.showTable(!!i);
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    });
                });
            }).catch(qubx.reportErr);
        }
        resize(w, h) {
            const cNode = this.cbox.node();
            this.chart && this.chart.resize({
                width: w || cNode.clientWidth,
                height: h || cNode.clientHeight
            });
        }
        showTable(show) {
            this.tbox.style({display: show ? '' : 'none'});
            this.cbox.style({display: show ? 'none' : ''});
            show || this.resize();
        }
        setTable(columns, columnData, clear) {
            this.columnNames = columns;
            this.columnData = columnData;
            const tblRows = [],
                  n = columnData.length&&columnData[0].length;
            let i = 0;
            for (; i<n; ++i) {
                tblRows.push(i);
            }
            const ths = this.thead
                .selectAll("th")
                .data(columns);
            ths.html(qubx.i);
            ths.enter()
                .append("th")
                .html(qubx.i);
            ths.exit().remove();
            if ( clear ) {
                this.tbody.html('');
            }
            const trs = this.tbody
                .selectAll("tr")
                .data(tblRows);
            trs.selectAll("td")
                .data(r => columnData.map(col => qubx.formatK0(col[r])))
                .text(qubx.i);
            const tds = trs.enter()
                  .append("tr")
                  .selectAll("td")
                  .data(r => columnData.map(col => qubx.formatK0(col[r])))
                  .text(qubx.i);
            tds.enter()
                .append("td")
                .html(qubx.i);
            tds.exit().remove();
            trs.exit().remove();
        }
        select(i) {
            this.chart && this.chart.select(this.spec.yvar, (i>=0) ? [i] : [], true);
            //this.config.onselected && this.config.onselected(i);
        }
        _onSelected(d, element) {
            try {
                this.config.onselected && this.config.onselected(d.index);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onClickDownloadChart() {
            try {
                saver.prompt({name: "Chart.svg"}).then(fi => saver.saveText(fi.fullname, this.asSVG().outerHTML, "image/svg+xml"))
                    .catch(qubx.reportErr);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        _onClickDownloadTable() {
            try {
                saver.prompt({name: "Table.txt"}).then(fi => saver.saveText(fi.fullname, this.asTSV(), "application/octet-stream;charset=UTF-8"))
                    .catch(qubx.reportErr);
            } catch (err) {
                qubx.reportErr(err);
            }
        }
        asSVG() {
            const svg = this.cdiv.select("svg").node().cloneNode(true);
            svg.setAttribute("xmlns", "http://www.w3.org/2000/svg");
            svg.setAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink");
            svg.setAttribute("version", "1.1");
            const style = document.createElement("style");
            style.innerHTML = chartStyle;
            svg.insertBefore(style, svg.firstChild);
            return svg;
        }
        asTSV() {
            const m = this.columnNames.length;
            if ( m === 0 ) {
                return "";
            }
            const n = this.columnData[0].length,
                  cc = this.columnData;
            let result = this.columnNames.join('\t'),
                i = 0;
            for (; i<n; ++i) {
                result += '\n'+cc[0][i];
                let j = 1;
                for (; j<m; ++j) {
                    result += '\t'+cc[j][i];
                }
            }
            result += '\n\n';
            return result;
        }
        setScatter(spec) {
            // {series: {name: arr}, xvar, yvar, columns, fitCurve, fitMarks}
            this.spec = spec;
            this.trends = [];
            const data = {
                xs: {},
                unload: true,
                columns: [[spec.yvar].concat(spec.series[spec.yvar])],
                //type: 'scatter',
                selection: {
                    enabled: true,
                    multiple: false
                },
                onselected: this._onSelected.bind(this)
            };
            const axes = {
                x: {label: spec.xvar||""},
                y: {label: spec.yvar||""}
            };
            const grid = {
                x: {lines: []},
                y: {lines: []}
            };
            if ( spec.xvar in spec.series ) {
                data.x = spec.xvar;
                data.xs[spec.yvar] = spec.xvar;
                data.columns.push([spec.xvar].concat(spec.series[spec.xvar]));
            }
            var ff, fitPars, fitted;
            if ( spec.fitCurve ) {
                const xx = spec.series[spec.xvar];
                const yy = spec.series[spec.yvar];
                ff = xx.map(() => 0.0);
                fitPars = qubx_fit.makeParams(spec.fitCurve);
                spec.fitCurve.curveInit(fitPars, xx, yy);
                fitted = qubx_fit.simplexFit(spec.fitCurve, fitPars, xx, yy, ff);
                data.xs.Fit = spec.xvar;
                data.columns.push(['Fit'].concat(ff));
                if ( spec.fitMarks && spec.fitMarks.x ) {
                    spec.fitMarks.x.forEach(name => grid.x.lines.push({
                        value: fitPars[name].val,
                        text: ''+name+'='+qubx.formatK0(fitPars[name].val)
                    }));
                } //// actually it's supposed to be log10(ligand) (but non-log Voltage and Pressure)???
            }
            if ( ! this.chart ) {
                this.chart = c3.generate({
                    data: data,
                    axis: axes,
                    bindto: this.cdiv,
                    legend: this.config.legend,
                    zoom: this.config.zoom,
                    grid: grid
                });
            } else {
                this.chart.load(data);
                this.chart.data.axes(axes);
            }
            
            let columns = spec.columns || (spec.xvar ? [spec.xvar, spec.yvar] : [spec.yvar]);
            if ( fitted ) {
                columns = columns.slice();
                columns.push('Fit');
                this.fitInfo.text('Fit: '+spec.fitCurve.curveInfo);
                const pOrder = [];
                spec.fitCurve.curveParams.forEach(p => p.constant || pOrder.push({
                    name: p.name,
                    val: qubx.formatK0(fitPars[p.name].val)
                }));
                const ps = this.fitParams
                    .selectAll("span")
                    .data(pOrder);
                function pFormat(p) {
                    return ''+p.name+'='+p.val+'; ';
                }
                ps.text(pFormat);
                ps.enter()
                    .append("span")
                    .text(pFormat);
                ps.exit().remove();
            }
            const tblCols = columns.map(name => (name === 'Fit') ? ff : spec.series[name]);
            this.setTable(columns, tblCols);
            return this;
        }
        _setupScatter() {
            // {series: {name: arr}, xvar, yvar, columns, fitCurve, fitMarks}
            const spec = this.spec;
            if ( typeof(spec.unload) === 'undefined' ) {
                spec.unload = true;
            }
            this.trends = this.trends || [];
            const trends = this.trends,
                  colors = {},
                  types = {};
            if ( spec.color ) {
                colors[spec.yvar] = spec.color;
            }
            types[spec.yvar] = 'scatter';
            trends.forEach(t => {
                colors[t.name] = t.color;
                types[t.name] = 'line';
            });
            const data = {
                //x: spec.xvar,
                unload: true,
                columns: [[spec.yvar].concat(spec.series[spec.yvar])]
                    .concat(trends.map(t => [t.name].concat(t.data))),
                colors: colors,
                types: types,
                selection: {
                    enabled: true,
                    multiple: false
                },
                onselected: this._onSelected.bind(this)
            };
            const axes = {
                x: {
                    label: spec.xlabel || spec.xvar || "",
                    tick: {
                        format: spec.xformat || qubx.formatK0,
                        culling: {
                            max: spec.xticks || 5
                        }
                    }
                },
                y: {
                    label: spec.ylabel || spec.yvar || "",
                    tick: {
                        format: spec.yformat || qubx.formatK0,
                        culling: {
                            max: spec.yticks || 5
                        }
                    }
                }
            };
            const grid = {
                x: {lines: []},
                y: {lines: []}
            };
            if ( spec.xvar in spec.series ) {
                data.x = spec.xvar;
                data.columns.push([spec.xvar].concat(spec.series[spec.xvar]));
            }
            if ( ! this.chart ) {
                this.chart = c3.generate({
                    data: data,
                    axis: axes,
                    bindto: this.cdiv,
                    legend: this.config.legend,
                    zoom: this.config.zoom,
                    grid: grid,
                    size: {
                        width: this.cbox.node().clientWidth,
                        height: this.cbox.node().clientHeight
                    },
                    transition: {
                        duration: 10
                    }
                });
            } else {
                this.chart.load(data);
                this.chart.data.axes(axes);
                setTimeout(() => this.resize(), 500);
            }
            spec.unload = false;
            
            let columns = spec.columns || (spec.xvar ? [spec.xvar, spec.yvar] : [spec.yvar]);
            columns = columns.slice();
            const tblCols = columns.map(name => spec.series[name]);
            if ( spec.xlabel && (columns[0] === spec.xvar) ) {
                columns[0] = spec.xlabel;
            }
            if ( spec.ylabel && (columns[columns.length-1] === spec.yvar) ) {
                columns[columns.length-1] = spec.ylabel;
            }
            trends.forEach(t => {
                columns.push(t.name);
                tblCols.push(t.data);
            });
            // trends: t.name, t.color, t.data
            this.setTable(columns, tblCols);
            return this;
        }
        _setupHist() {
            const spec = this.spec;
            if ( typeof(spec.unload) === 'undefined' ) {
                spec.unload = true;
            }
            this.trends = this.trends || [];
            const trends = this.trends,
                  colors = {},
                  types = {};
            if ( spec.color ) {
                colors[spec.yvar] = spec.color;
            }
            trends.forEach(t => {
                colors[t.name] = t.color;
                types[t.name] = 'line';
            });
            const data = {
                x: spec.xvar,
                unload: spec.unload && this.chart && this.chart.columns,
                columns: [
                    [spec.xvar].concat(spec.series[spec.xvar]),
                    [spec.yvar].concat(spec.series[spec.yvar])
                ].concat(trends.map(t => [t.name].concat(t.data))),
                colors: colors,
                type: 'bar',
                types: types
            };
            const axes = {
                x: {
                    label: spec.xlabel || spec.xvar || "",
                    tick: {
                        format: spec.xformat || qubx.formatK0,
                        culling: {
                            max: spec.xticks || 5
                        }
                    }
                },
                y: {
                    label: spec.ylabel || spec.yvar || "",
                    tick: {
                        format: spec.yformat || qubx.formatK0,
                        culling: {
                            max: spec.yticks || 5
                        }
                    }
                }
            };
            if ( ! this.chart ) {
                this.chart = c3.generate({
                    data: data,
                    axis: axes,
                    bindto: this.cdiv,
                    legend: this.config.legend,
                    zoom: this.config.zoom,
                    size: {
                        width: this.cbox.node().clientWidth,
                        height: this.cbox.node().clientHeight
                    },
                    transition: {
                        duration: 10
                    }
                });
            } else {
                this.chart.load(data);
                this.chart.data.axes(axes);
                setTimeout(() => this.resize(), 500);
            }
            spec.unload = false;
            
            const columns = spec.columns || (spec.xvar ? [spec.xvar, spec.yvar] : [spec.yvar]);
            const tblCols = columns.map(name => spec.series[name]);;
            if ( spec.xlabel && (columns[0] === spec.xvar) ) {
                columns[0] = spec.xlabel;
            }
            if ( spec.ylabel && (columns[columns.length-1] === spec.yvar) ) {
                columns[columns.length-1] = spec.ylabel;
            }
            trends.forEach(t => {
                columns.push(t.name);
                tblCols.push(t.data);
            });
            // trends: t.name, t.color, t.data
            this.setTable(columns, tblCols, true);
        }
        setHist(spec) {
            this.spec = spec;
            this.trends = null;
            this.isHist = true;
            this._setupHistSoon();
            return this;
        }
        setScatterPlot(spec) {
            this.spec = spec;
            this.trends = null;
            this.isHist = false;
            this._setupScatterSoon();
            return this;
        }
        trendCount(n, getName, getColor, getData) {
            const self = this,
                  spec = this.spec;
            this.trends = this.trends || [];
            if ( typeof(n) === 'undefined' ) {
                return this.trends.length;
            }
            n = n|0;
            if ( n < this.trends.length ) {
                this.spec.unload = true;
                this.trends.splice(n, this.trends.length-n);
            } else if ( n > this.trends.length ) {
                const zeros = getData ? null : spec.series[spec.yvar].map(() => 0.0);
                while ( n > this.trends.length ) {
                    const i = this.trends.length,
                          self = this;
                    this.trends.push({
                        name: getName ? getName(i) : (''+i),
                        color: getColor ? getColor(i) : exports.trendColors[i],
                        data: getData ? getData(i) : zeros,
                        setData: function(d) {
                            this.data = d;
                            if ( self.isHist ) {
                                self._setupHistSoon();
                            } else {
                                self._setupScatterSoon();
                            }
                            return this;
                        }
                    });
                }
            } else {
                // no change
                return this;
            }
            if ( this.isHist ) {
                this._setupHistSoon();
            } else {
                this._setupScatterSoon();
            }
            return this;
        }
    };

    return exports;
});
