/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_data_prep", ["d3", "qubx", "qubx_ui", "qubtrace", "qubx_lib"], (d3, qubx, qubx_ui, qubtrace, qubx_lib) => {
    'use strict';
    
    const exports = {};
    
    exports.openDlg = new qubx_ui.Modal({
        title: "Open data file",
        template: "#qubx_mac_opendata"
    });
    exports.openDlg.dom.then(dom => qubx_lib.then(lib => {
        const dotdot = qubx_ui.querydotdot(dom.content);
        const mnuDepvars = dotdot("#qmod_depvar"),
              visible = [], // visible[i] if signal i is marked L, V, or P
              PREVIEW_MAX_SECONDS = 1.0;
        let sigData = [], // signalDB entries corresponding to sigNames
            sigInd = []; // sigData - depvar
        var signalDB, // passed in via .runSignals()
            rdr, // passed in
            worker, // assigned while running
            previewSeg = 0,
            previewOffset = 0,
            previewLength;
        
        const showPrevSeg = function showPrevSeg() {
            if ( previewSeg ) {
                --previewSeg;
                segnum.textContent = ""+(previewSeg+1);
                loadPreviews();
            }
        };
        const showNextSeg = function showNextSeg() {
            if ( (previewSeg+1) < rdr.counts.length ) {
                ++previewSeg;
                segnum.textContent = ""+(previewSeg+1);
                loadPreviews();
            }
        };
        const showPrevScreen = function showPrevScreen() {
            if ( previewOffset ) {
                previewOffset -= previewLength/2;
                loadPreviews();
            }
        };
        const showNextScreen = function showNextScreen() {
            if ( (previewOffset + (previewLength*3/2)) <= rdr.maxN ) {
                previewOffset += previewLength/2;
                loadPreviews();
            }
        };
        
        const depPreview = new qubtrace.Plot({
            node: dotdot("#qmod_depvar_preview"),
            scalebarLores: false,
            raster_samples: lib.raster_samples
        });
        depPreview.node.width = document.body.clientWidth;
        depPreview.node.height = document.body.clientHeight/4;
        const preview = new qubtrace.Plot({
            node: dotdot("#qmod_preview_lo"),
            hires: dotdot("#qmod_preview_hi"),
            fitStyle: "rgba(32,255,64,0.7)",
            scalebarLores: false,
            scalebarHires: false,
            raster_samples: lib.raster_samples
        });
        dotdot.listen("#qmod_preview_prev", 'click', showPrevSeg);
        dotdot.listen("#qmod_preview_next", 'click', showNextSeg);
        dotdot.listen("#qmod_preview_bwd", 'click', showPrevScreen);
        dotdot.listen("#qmod_preview_fwd", 'click', showNextScreen);
        const segnum = dotdot("#qmod_preview_segnum");
        const spinner = dotdot(".spinBox");
        const showSpin = function showSpin() {
            const spin = dotdot('.spinner');
            spinner.style.display = '';
            spinner.style.top = 0.5*(spinner.parentNode.clientHeight - spin.clientHeight);
            spinner.style.left = 0.5*(spinner.parentNode.clientWidth - spin.clientWidth);
        };
        const hideSpin = function hideSpin() {
            spinner.style.display = 'none';
        };
        
        const maxDepPreviewLength = 65536;
        let depPreviewLength = 0;
        const updateDepvar = function updateDepvar(names) {
            if ( names ) {
                qubx_ui.setSelectOptions(mnuDepvars, names);
                mnuDepvars.selectedIndex = 0;
                depPreviewLength = Math.min(maxDepPreviewLength, rdr.counts.length && rdr.counts[0] || 0);
                depPreview.segmentCount(depPreviewLength ? 1 : 0);
                depPreview.segmentSize(0, depPreviewLength, 0);
                depPreview.doneSegmentSize();
            }
            depPreview.clear();
            depPreviewLength && rdr.readSegment(0, (rdr, seg) => {
                depPreview.set(0, 0, seg.buffers[mnuDepvars.selectedIndex], depPreviewLength);
            });
            setTimeout(() => depPreview.reshape(), 10);
        };
        
        const idealizePreview = function idealizePreview(iSig) {
            const sig = sigData[iSig],
                  chunkSize = 256;
            return worker.idealizeMDL(preview.segments[iSig].points, 1/*sampling, so duration==count*/, sig.delta, false, chunkSize, true)
                .then(idl => {
                    const classify = new lib.IdlClassifier(sig.delta);
                    if ( rdr ) { // else the dialog is long gone
                        idl.classes = classify.add(rdr.sampling, idl.classes, idl.durations, idl.amps, idl.stds);
                        const dist = classify.read();
                        idl.amps = dist.amps;
                        idl.stds = dist.stds;
                    }
                    classify.destroy();
                    if ( ! rdr ) {
                        return idl;
                    }
                    //    return idl;
                    //})
                    //.then(function(idl) {
                    const fit = new Float32Array(previewLength),
                          ndwt = idl.classes.length;
                    let at = 0,
                        i = 0;
                    for (; i<ndwt; ++i) {
                        const n = idl.durations[i]|0,
                              a = idl.amps[idl.classes[i]];
                        let j = 0;
                        for (; j<n; ++j, ++at) {
                            fit[at] = a;
                        }
                    }
                    preview.setFit(iSig, 0, fit);
                    return idl;
                });
        };
        const loadPreview = function loadPreview(iSig) {
            showSpin();
            try {
                rdr.readSegment(previewSeg, (rdr, seg) => {
                    preview.set(iSig, 0, seg.buffers[iSig], previewLength, previewOffset, qubx.extrema(seg.buffers[iSig]));
                    idealizePreview(iSig)
                        .then(hideSpin);
                });
            } catch (err) {}
            visible[iSig] = true;
            return true;
        };
        const loadPreviews = function loadPreviews() {
            showSpin();
            preview.clear();
            try {
                rdr.readSegment(previewSeg, (rdr, seg) => visible.forEach((v, iSig) => {
                    if ( v ) {
                        preview.set(iSig, 0, seg.buffers[iSig], previewLength, previewOffset, qubx.extrema(seg.buffers[iSig]));
                        if ( (iSig+1) < seg.buffers.length ) {
                            idealizePreview(iSig);
                        } else {
                            idealizePreview(iSig)
                                .then(hideSpin);
                        }
                    }
                }));
            } catch (err) {}
        };
        const clearPreview = function clearPreview(iSig) {
            if ( typeof(iSig) === 'undefined' ) {
                preview.segmentCount(0);
            } else {
                preview.clear(iSig);
                visible[iSig] = false;
            }
        };
        const setupPreview = function setupPreview() {
            if ( (rdr.signalNames.length < 2) || (! rdr.maxN) ) {
                clearPreview();
                return;
            }
            previewSeg = previewOffset = 0;
            previewLength = Math.floor(Math.min(rdr.maxN, PREVIEW_MAX_SECONDS/rdr.sampling))|0;
            segnum.textContent = "1";
            
            preview.segmentCount(rdr.signalNames.length);
            preview.segments.forEach((seg, i) => preview.segmentSize(i, previewLength, 0));
            preview.doneSegmentSize();
            
            visible.forEach((v,i) => v && loadPreview(i));
            setTimeout(() => preview.reshape(), 10);
        };
            
        const checkSignalVisible = function checkSignalVisible(row, col, val, infoOnly) {
            const vis = row.isLigand || row.isVoltage || row.isPressure,
                  i = rdr.signalNames.indexOf(row.name);
            if ( vis === visible[i] ) {
                return;
            }
            if ( infoOnly ) {
                visible[i] = vis;
            } else {
                vis && loadPreview(i) || clearPreview(i);
            }
            return vis;
        };
        const checkIdealize = function checkIdealize(row, col, val) {
            const i = rdr.signalNames.indexOf(row.name);
            if ( visible[i] ) {
                showSpin();
                idealizePreview(i).then(hideSpin);
            }
        };
        
        const indvar = new qubx_ui.ColumnTable({
            node: dotdot("#qmod_indvar"),
            columns: [
                qubx_ui.ColumnTable.inputCol('Name', 'rename', 'text', qubx.acceptStrNonempty),
                qubx_ui.ColumnTable.chkCol('Ligand', 'isLigand', checkSignalVisible),
                qubx_ui.ColumnTable.chkCol('Voltage', 'isVoltage', checkSignalVisible),
                qubx_ui.ColumnTable.chkCol('Pressure', 'isPressure', checkSignalVisible),
                qubx_ui.ColumnTable.numCol('Delta', 'delta', qubx.acceptFloatGreaterThan(0.0), qubx.formatK0, checkIdealize)
            ]
        });
        const updateIndvars = function updateIndvars(skipPreviews) {
            sigInd = [];
            sigData.forEach((sig, i) => {
                if ( i === mnuDepvars.selectedIndex ) {
                    visible[i] = false;
                } else {
                    sigInd.push(sig);
                }
            });
            indvar.render(sigInd, 'name');
            sigInd.forEach(row => checkSignalVisible(row, 0, 0, true));
            if ( ! skipPreviews ) {
                loadPreviews();
            }
        };
        d3.select(mnuDepvars)
            .on('change', () => {
                try {
                    updateDepvar();
                    updateIndvars();
                } catch (err) {
                    qubx.reportErr(err);
                }
            });
        
        exports.openDlg.runSignals = function runSignals(_worker, filename, _rdr, _signalDB) {
            return new Promise((resolve, reject) => {
                worker = _worker;
                rdr = _rdr;
                signalDB = _signalDB;
                dom.content.querySelector("#qmod_filename").textContent = filename;
                dom.content.querySelector("#qmod_indbox").className = (rdr.signalNames.length<=1) ? 'nodisplay' : '';
                updateDepvar(rdr.signalNames.slice());
                sigData = rdr.signalNames.map(signalDB.get.bind(signalDB));
                updateIndvars(true);
                setupPreview();
                const resizeHandle = qubx_ui.onResize(dom.content, info => preview.reshape());
                exports.openDlg.run().then(() => {
                    qubx_ui.clearResize(resizeHandle);
                    clearPreview();
                    worker = null;
                    signalDB = null;
                    rdr = null;
                    resolve({
                        depVar: mnuDepvars.selectedIndex
                    });
                });
            });
        };
    })).catch(qubx.reportErr);
    
    exports.prepData = function prepData(worker, rdr, iCurrent, signalDB, onProgress) {
        // reorders signals so dependent var is first, and only signals chosen as independent variables are retained;
        // idealizes independent variables;
        return new Promise((resolve, reject) => qubx_lib.then(lib => {
            // set signals in order
            let haveV = false,
                haveP = false;
            const signals = {
                index: [iCurrent],
                name: [rdr.signalNames[iCurrent]],
                delta: [0.0],
                idealized: []
            };
            rdr.signalNames.forEach((nm, iSig) => {
                signals.idealized[iSig] = rdr.idealized && rdr.idealized[iSig] || [];
                if ( iSig === iCurrent ) {
                    return;
                }
                const sig = signalDB.get(nm);
                if ( sig.isLigand ) {
                    signals.index.push(iSig);
                    signals.name.push(sig.rename);
                    signals.delta.push(sig.delta);
                }
                if ( sig.isVoltage && ! haveV ) {
                    signals.index.push(iSig);
                    signals.name.push('Voltage');
                    signals.delta.push(sig.delta);
                    haveV = true;
                }
                if ( sig.isPressure && ! haveP ) {
                    signals.index.push(iSig);
                    signals.name.push('Pressure');
                    signals.delta.push(sig.delta);
                    haveP = true;
                }
            });
            
            const classify = signals.delta.map((delta, iSig) => iSig ? (new lib.IdlClassifier(delta)) : null);
            function cleanReject(err) {
                classify.forEach(cl => cl && cl.destroy());
                return reject(err);
            }
            const Nseg = rdr.counts.length;
            let Ndone = 0;
            const segprog = new qubx.Kalman1D();
            segprog.Q = 1e-5;
            segprog.R = 1e-3;
            const chunkSize = Math.max(512, (rdr.maxN / 256)|0),
                  segPrep = []; // independent promises for stimulus idl
            // going to copy all of the stimulus arrays at once up front, not the best way
            if ( rdr.isSampled && rdr.isSampled.length && rdr.isSampled[0] ) { /// assuming all sampled or none--could be a problem with other file types?
                let iSeg = 0;
                for (; iSeg<Nseg; ++iSeg) {
                    rdr.readSegment(iSeg, (rdr, seg) => {
                        signals.index.forEach((iSig, iOut) => {
                            if ( iOut && ! signals.idealized[iSig].length ) {
                                segPrep.push(
                                    worker.idealizeMDL(new Float32Array(seg.buffers[iSig]), 1, signals.delta[iOut], false, chunkSize, true, prog => {
                                        segprog.next((Ndone + prog)/segPrep.length);
                                        onProgress && onProgress(segprog.X);
                                    }).then(events => {
                                        signals.idealized[iOut].push({
                                            classes: classify[iOut].add(1, events.classes, events.durations, events.amps, events.stds),
                                            counts: new Int32Array(events.durations)
                                        });
                                        ++Ndone;
                                        return events;
                                    })
                                );
                            }
                        });
                    });
                }
            }
            Promise.all(segPrep).then(() => {
                onProgress && onProgress(1.0);
                classify.forEach((clsf, iOut) => {
                    if ( iOut ) {
                        const dist = clsf.read();
                        signals.idealized[iOut].forEach(events => {
                            events.amps = dist.amps;
                            events.stds = dist.stds;
                        });
                        clsf.destroy();
                    }
                });
                resolve(signals);
            });
        }));
    };

    exports.reScan = function reScan(worker, src, onProgress) {
        // scans N of resampled depvar for various delta, to calibrate the fast/accurate slider
        // resample to save time and transfer memory
        const rsChunk = 4096,
              chunk = new Float32Array(rsChunk),
              scale = src.length / rsChunk;
        let i = 0;
        for (; i<rsChunk; ++i) {
            chunk[i] = src[Math.round(scale*i)|0];
        }
        return worker.resampleScan(chunk, onProgress);
    };

    exports.idlStim = function idlStim(worker, segSamples, delta, onProgress) {
        return qubx_lib.then(lib => {
            const classify = new lib.IdlClassifier(delta);
            function cleanReject(err) {
                classify.destroy();
                return Promise.reject(err);
            }
            const Nseg = segSamples.length;
            let maxN = 0;
            segSamples.forEach(arr => {
                maxN = Math.max(maxN, arr.length);
            });
            let Ndone = 0;
            const segprog = new qubx.Kalman1D();
            segprog.Q = 1e-5;
            segprog.R = 1e-3;
            const chunkSize = Math.max(512, (maxN / 256)|0);
            // going to copy all of the stimulus arrays at once up front, not the best way
            return Promise.all(segSamples.map(arr => worker.idealizeMDL(arr, 1, delta, false, chunkSize, true, prog => {
                segprog.next((Ndone + prog)/segSamples.length);
                onProgress && onProgress(segprog.X);
            }).then(events => {
                ++Ndone;
                return {
                    classes: classify.add(1, events.classes, events.durations, events.amps, events.stds),
                    counts: new Int32Array(events.durations)
                };;
            }))).then(segEvents => {
                onProgress && onProgress(1.0);
                const dist = classify.read();
                segEvents.forEach(events => {
                    events.amps = dist.amps;
                    events.stds = dist.stds;
                });
                classify.destroy();
                return segEvents;
            }).catch(cleanReject);
        });
    };
    
    return exports;
});
