/* Copyright 2016-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_demo_dataentry", ["qubx", "qubx_ui", "qubx_data", "qubx_data_prep", "qubtrace", "cloud", "qubx_cloud_ui", "d3"],
    (qubx, qubx_ui, qubx_data, qubx_data_prep, qubtrace, cloud, qubx_cloud_ui, d3) => {
        "use strict";
        const exports = {};

        const DataEntryModeDefs = {
            constants: {}
        };
        exports.DataEntryMode = class DataEntryMode {
            constructor(main, config) {
                this.main = main;
                let storedPrefs = localStorage.getItem("qubx_de_prefs");
                storedPrefs = storedPrefs ? JSON.parse(storedPrefs) : {};
                this.config = Object.assign({}, DataEntryModeDefs, config, storedPrefs);
                this.active = false;
                this.main.panels.then(p => {
                    this.modelView = p.modelView;
                    this.data = p.data;
                    this.controls = p.controls.querySelector("#qubx_demo_data_ctrls");
                    this._makePanels(p);
                    this._makeDialogs(p);
                    this._makeMenus(p);
                }).catch(qubx.reportErr);
            }
            activate() {
                this.main.panels.then(p => {
                    this.active = true;
                    this.menus.forEach(node => this.controls.appendChild(node));
                    const modelParent = this.modelView.node.parentNode;
                    this.hOnSelect = this.data.plot.ev.select.sub(this._onSelect.bind(this));
                    this.hOnChoose = this.data.plot.ev.choose.sub(this._onChoose.bind(this));
                    this.hOnHilite = this.data.plot.ev.hilite.sub(this._onHilite.bind(this));
                    modelParent.removeChild(this.modelView.node);
                    modelParent.appendChild(this.tools.node);
                    this.data.plot.reSelect(this._onSelect.bind(this));
                    this.data.plot.reHiSel();
                    if ( this.main.topbot.config.skewType === this.main.topbot.config.skews.length ) {
                        this.main.topbot.moreB();
                    }
                    this.main.vis.pick('Data info');
                }).catch(qubx.reportErr);
            }
            deactivate() {
                return this.tools.pick(null).then(() => {
                    this.active = false;
                    this.menus.forEach(node => this.controls.removeChild(node));
                    const modelParent = this.tools.node.parentNode;
                    modelParent.removeChild(this.tools.node);
                    modelParent.appendChild(this.modelView.node);
                    this.savePrefs();
                    this.data.plot.ev.select.unsub(this.hOnSelect);
                    this.data.plot.ev.choose.unsub(this.hOnChoose);
                    this.data.plot.ev.hilite.unsub(this.hOnHilite);
                }).catch(qubx.reportErr);
            }
            savePrefs() {
                try {
                    const prefs = {};
                    for (const k of Object.keys(DataEntryModeDefs)) { // to avoid e.g. config.node
                        prefs[k] = this.config[k];
                    }
                    localStorage.setItem("qubx_de_prefs", JSON.stringify(prefs));
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
            _makePanels(panels) {
                qubx_ui.loadTemplate('#qubx_dataentry_components').then(cmps => {
                    this._linkModes(cmps);
                    this.tools = new qubx_ui.Tabs({
                        position: 'left',
                        defaultContent: cmps.querySelector('#qubx_de_welcome')
                    });
                    this.tools.node.classList.add('qubx_fullsize');
                    this.tools.ev.pick.sub(this._onSetTool.bind(this));
                    this.toolScaling = new ScalingTool(this, cmps.querySelector('#qubx_de_scaling'));
                    this.toolBaseline = new BaselineTool(this, cmps.querySelector('#qubx_de_baseline'));
                    this.toolStimBuild = new StimBuildTool(this, cmps.querySelector('#qubx_de_stimbuild'));
                    this.toolSelect = new SelectTool(this, cmps.querySelector('#qubx_de_select'));
                    this.toolAnalyze = new AnalyzeTool(this, cmps.querySelector('#qubx_de_analyze'));
                    [this.toolScaling, this.toolBaseline, this.toolStimBuild, this.toolSelect, this.toolAnalyze]
                        .forEach(t => this.tools.add(t));
                });
            }
            _linkModes(components) {
                const main = this.main;
                [
                    {
                        selector: '.qubx_de_go_mac',
                        mode: main.macMode,
                        title: 'Start Macroscopic Analysis...'
                    },
                    {
                        selector: '.qubx_de_go_single',
                        mode: main.singleMode,
                        title: 'Start Single Molecule Analysis...'
                    }
                ].forEach(mode => d3.select(components)
                          .selectAll(mode.selector)
                          .property({
                              href: '#',
                              title: mode.title
                          }).on('click', () => {
                              main.mode(mode.mode);
                              d3.event.preventDefault();
                              d3.event.stopPropagation();
                          }));
            }
            _makeDialogs(p) {
            }
            _makeMenus(p) {
                this.mnuTools = new qubx_ui.Menu({
                    caption: 'Data',
                    groups: [[
                        {
                            caption: 'Scaling',
                            action: () => this.tools.pick(this.toolScaling)
                        },
                        {
                            caption: 'Baseline',
                            action: () => this.tools.pick(this.toolBaseline)
                        },
                        {
                            caption: 'Stimulus builder',
                            action: () => this.tools.pick(this.toolStimBuild)
                        },
                        {
                            caption: 'Select and Upload',
                            action: () => this.tools.pick(this.toolSelect)
                        },
                        {
                            caption: 'Analyze...',
                            action: () => this.tools.pick(this.toolAnalyze)
                        }
                    ]]
                });
                this.mnuTools.config.node.style.width = '5em';
                
                this.menus = [this.mnuTools.config.node];
            }
            _onSetTool(tool) {
                this.tool = tool;
            }
            _reSelectSeg(iSeg) {
                if ( this._selSegs ) {
                    if ( iSeg < 0 ) {
                        if ( this._selSegs.length > 1 ) {
                            this._sel = null;
                        } else {
                            this._sel = this._selSegs[0];
                        }
                    } else {
                        this._sel = this._selSegs[iSeg];
                    }
                } else {
                    this._sel = null;
                }
                if ( this._hiliteSegs ) {
                    if ( iSeg < 0 ) {
                        if ( this._hiliteSegs.length > 1 ) {
                            this._hilite = null;
                        } else {
                            this._hilite = this._hiliteSegs[0];
                        }
                    } else {
                        this._hilite = this._hiliteSegs[iSeg];
                    }
                } else {
                    this._hilite = null;
                }
            }
            _onSelect(offset, count, segments) {
                this._selSegs = segments;
                this._reSelectSeg(this.data.plot.chosen());
                this.tool && this.tool.onSelect && this.tool.onSelect(offset, count, segments);
            }
            _onChoose(iSeg) {
                this._reSelectSeg(iSeg);
                this.tool && this.tool.onChoose && this.tool.onChoose(iSeg);
            }
            _onHilite(offset, count, segments) {
                this._hiliteSegs = segments;
                this._reSelectSeg(this.data.plot.chosen());
                this.tool && this.tool.onHilite && this.tool.onHilite(offset, count, segments);
            }
            bestSel() {
                const hiSel = this._hiSel || (this._hiliteSegs && this._hiliteSegs.length && this._hiliteSegs[0]) || null,
                      sel = this._sel || (this._selSegs && this._selSegs.length && this._selSegs[0]) || null;
                if ( hiSel && hiSel.to ) {
                    return Object.assign({}, hiSel, {
                        sources: sel && sel.sources || []
                    });
                }
                if ( sel ) {
                    return Object.assign({}, sel, {
                        from: sel.selFrom,
                        to: sel.selTo
                    });
                }
                return null;
            }
        };
        
        class ScalingTool {
            constructor(deMode, panScaling) {
                this.deMode = deMode;
                this.node = panScaling;
                this.tabTitle = 'Scaling';
                this.tabIcon = '<img src="images/scaling.svg">';
                const dotdot = qubx_ui.querydotdot(panScaling);
                this.tbl = new qubx_ui.ColumnTable({
                    node: dotdot('table'),
                    columns: [
                        qubx_ui.ColumnTable.txtCol('Signal', 'name'),
                        qubx_ui.ColumnTable.inputCol('sel. mean', 'mean', 'number', qubx.acceptFloat, qubx.formatK0, this._onChangeMean.bind(this)),
                        qubx_ui.ColumnTable.txtCol('= x*', '_u'),
                        qubx_ui.ColumnTable.inputCol('Scaling', 'scaling', 'number', qubx.acceptFloat, qubx.formatK0, this._onChangeScaling.bind(this)),
                        qubx_ui.ColumnTable.inputCol('+ Offset', 'offset', 'number', qubx.acceptFloat, qubx.formatK0, this._onChangeOffset.bind(this)),
                        qubx_ui.ColumnTable.inputCol('Delta', 'delta', 'number', qubx.acceptFloatGreaterThan(0.0), qubx.formatK0, this._onChangeDelta.bind(this))
                    ]
                });
                this.txtSamplingKHZ = new qubx_ui.AcceptedInput(
                    dotdot('#qubx_de_sampling_khz'),
                    1.0,
                    {
                        accept: qubx.acceptFloatGreaterThan(0),
                        format: qubx.formatK0,
                        widthChars: 8
                    }
                );
                this.txtSamplingMS = new qubx_ui.AcceptedInput(
                    dotdot('#qubx_de_sampling_ms'),
                    1.0,
                    {
                        accept: qubx.acceptFloatGreaterThan(0),
                        format: qubx.formatK0,
                        widthChars: 8
                    }
                );
                const txtsSampling = [this.txtSamplingKHZ, this.txtSamplingMS];
                txtsSampling.forEach((txt, which) => {
                    txt.ev.change.sub(v => {
                        txtsSampling[(which+1)%2].update(1/v);
                        this.panEdited.style.display = '';
                    });
                });
                this.panEdited = dotdot('#qubx_de_scaling_edited', pan => {
                    pan.style.display = 'none';
                    pan.querySelector('#qubx_de_scaling_ok').addEventListener('click', this._onClickOK.bind(this));
                    pan.querySelector('#qubx_de_scaling_cancel').addEventListener('click', this._onClickCancel.bind(this));
                });
                this.remeasureSoon = qubx.batchDelayed(this._remeasure.bind(this), 50);
            }
            onShow(active) {
                active = !! active;
                this.deMode.data.plot.rawStim(active);
                if ( active ) {
                    this.hOnChangeSignals = this.deMode.data.data.ev.changeSignals.sub(this._onChangeSignals.bind(this));
                    this._onChangeSignals(this.deMode.data.data.signals);
                    const samplingMS = 1e3 * this.deMode.data.data.sampling();
                    this.txtSamplingMS.update(samplingMS);
                    this.txtSamplingKHZ.update(1/samplingMS);
                } else if ( this.edited ) {
                    return qubx_ui.confirm({
                        hdr: 'Scaling: unsaved changes',
                        q: "Changes in the Scaling panel haven't been saved. Switch tools anyway?"
                    }).then(() => this._rowsReset())
                        .catch(err => {
                            this.deMode.data.plot.rawStim(true);
                            return Promise.reject(err);
                        });
                } else {
                    this.deMode.data.data.ev.changeSignals.unsub(this.hOnChangeSignals);
                    this.signals = [];
                }
                return Promise.resolve(true);
            }
            _onChangeSignals(signals) {
                this.signals = this.deMode.data.data.signals.map((sig, iSig) => ({
                    iSig: iSig,
                    name: sig.name,
                    scaling: 1.0,
                    offset: 0.0,
                    edited: false
                })).filter(sig => this.deMode.data.data.segments[0] && this.deMode.data.data.segments[0].sampled[sig.iSig])
                    .map(sig => {
                        sig.mainSrc = this.deMode.main.signals.get(sig.name);
                        sig.delta = sig.mainSrc.delta;
                        return sig;
                    });
                this.remeasureSoon();
            }
            onChoose(iSeg) {
                this.remeasureSoon();
            }
            onSelect(offset, count, segments) {
                this.deMode._hilite || this.remeasureSoon();
            }
            onHilite(offset, count, segments) {
                this.remeasureSoon();
            }
            _remeasure() {
                const sel = this.deMode.bestSel();
                if ( ! sel ) {
                    return;
                }
                this.signals.forEach(sig => {
                    sig.meanRaw = qubx.meanstd(sel.sources[sig.iSig], sel.from, (sel.to - sel.from + 1)).mean;
                    sig.mean = sig.meanRaw * sig.scaling + sig.offset;
                });
                this._rerender();
            }
            _rerender() {
                this.tbl.render(this.signals, 'name');
                this.tbl.tbody.selectAll("tr")
                    .classed('qubx_de_scaling_rowEdited', qubx.m('edited'));
            }
            _rowEdited(r) {
                const sig = this.signals[r];
                this.edited = sig.edited = true;
                sig.mean = sig.meanRaw * sig.scaling + sig.offset;
                this.panEdited.style.display = '';
                this._rerender();
            }
            _rowsReset() {
                this.signals.forEach(sig => {
                    sig.scaling = 1.0;
                    sig.offset = 0.0;
                    sig.mean = sig.meanRaw;
                    sig.delta = sig.mainSrc.delta;
                    sig.edited = false;
                });
                this.edited = false;
                this.panEdited.style.display = 'none';
                this._rerender();
            }
            _onChangeMean(row, col, val) {
                const sig = this.signals[row.iSig];
                sig.scaling = (val - sig.offset) / (sig.meanRaw || 1.0);
                this._rowEdited(row.iSig);
            }
            _onChangeScaling(row, col, val) {
                this._rowEdited(row.iSig);
            }
            _onChangeOffset(row, col, val) {
                this._rowEdited(row.iSig);
            }
            _onChangeDelta(row, col, val) {
                this._rowEdited(row.iSig);
            }
            _onClickOK() {
                this.deMode.data.data.sampling(1e-3*this.txtSamplingMS.value);
                this.signals.forEach(sig => {
                    if ( sig.edited ) {
                        if ( ! sig.iSig ) {
                            this.deMode.main.getDeltaForFrac = null;
                        }
                        const data = this.deMode.data.data;
                        sig.mainSrc.delta = sig.delta;
                        const segSamples = qubx.range(data.segments.length).map(iSeg =>data.editSampled(iSeg, sig.iSig, (arr, n) => {
                            const m = sig.scaling,
                                  b = sig.offset;
                            let i = 0;
                            for (; i<n; ++i) {
                                arr[i] = arr[i] * m + b;
                            }
                            sig.iSig || data.segments[iSeg].sampled[0].set(arr);
                            return arr;
                        }));
                        if ( sig.iSig ) {
                            qubx_data_prep.idlStim(this.deMode.main.dataWorker, segSamples, sig.delta) /// , onProgress
                                .then(segEvents => data.segments.forEach((seg, iSeg) => {
                                    const events = segEvents[iSeg];
                                    data.setIdeal(iSeg, sig.iSig, events.counts, events.classes, 0, events.amps, events.stds);
                                })).catch(qubx.reportErr);
                        }
                    }
                });
                this._rowsReset();
            }
            _onClickCancel() {
                const samplingMS = 1e3 * this.deMode.data.data.sampling();
                this.txtSamplingMS.update(samplingMS);
                this.txtSamplingKHZ.update(1/samplingMS);
                this._rowsReset();
            }
        };
        
        class BaselineTool {
            constructor(deMode, panBaseline) {
                this.deMode = deMode;
                this.node = panBaseline;
                this.tabTitle = 'Baseline';
                this.tabIcon = '<img src="images/baseline.png">';
                const dotdot = qubx_ui.querydotdot(panBaseline);
                this.welcome = dotdot('#qubx_de_bp_welcome');
                this.canvas = dotdot('canvas', canvas => {
                    qubx_ui.onResize(canvas.parentNode, this._onResizeNodePreview.bind(this));
                });
                this.btnNodePrev = dotdot.listen('#qubx_de_bn_prev', 'click', this._onClickNodePrev.bind(this));
                this.btnNodeNext = dotdot.listen('#qubx_de_bn_next', 'click', this._onClickNodeNext.bind(this));
                this.txtNodeSample = dotdot.listen('#qubx_de_bn_sample', 'input', this._onInputNodeSample.bind(this));
                this.txtNodeTime = dotdot.listen('#qubx_de_bn_time', 'input', this._onInputNodeTime.bind(this));
                this.txtNodeLevel = dotdot.listen('#qubx_de_bn_level', 'input', this._onInputNodeLevel.bind(this));
                dotdot.listen('#qubx_de_bn_del', 'click', this._onClickDelNode.bind(this));
                this.panNode = dotdot('#qubx_de_baseline_node', pan => {
                    pan.style.visibility = 'hidden';
                });
                this.btnSubtract = dotdot.listen('#qubx_de_baseline_subtract', 'click', this._onClickSubtract.bind(this));
                this.btnSubtract.disabled = true;
                this.updateNodePreviewSoon = qubx.batchDelayed(this._updateNodePreview.bind(this), 100, true);
            }
            onShow(active) {
                active = !! active;
                this.deMode.data.plot.baselining = active;
                if ( active ) {
                    this.hOnPickNode = this.deMode.data.plot.ev.pickNode.sub(this._onPickNode.bind(this));
                    this.hOnChangeBaseline = this.deMode.data.data.ev.changeBaseline.sub(this._onChangeBaseline.bind(this));
                    this.welcome.style.width = this.deMode.data.plot.hiresCanvas.clientWidth+'px';
                    this.updateNodePreviewSoon();
                    if ( this.deMode.data.plot.chosen() < 0 ) {
                        this.deMode.data.plot.chosen(0);
                    }
                } else if ( ! this.btnSubtract.disabled ) {
                    return qubx_ui.confirm({
                        hdr: 'Baseline: unsaved changes',
                        q: "Baseline is not yet subtracted from the data. Switch tools anyway?"
                    }).catch(err => {
                        this.deMode.data.plot.baselining = true;
                        return Promise.reject(err);
                    });
                } else {
                    this.deMode.data.plot.ev.pickNode.unsub(this.hOnPickNode);
                    this.deMode.data.data.ev.changeBaseline.unsub(this.hOnChangeBaseline);
                    this._onPickNode(null);
                }
                return Promise.resolve(true);
            }
            onChoose(iSeg) {
                this.updateNodePreviewSoon();
                this._onPickNode(null);
            }
            onSelect(offset, count, segments) {
                this.updateNodePreviewSoon();
            }
            _onPickNode(baselineSel) {
                this.bnode = baselineSel;
                if ( baselineSel ) {
                    this.panNode.style.visibility = '';
                    this.txtNodeSample.value = baselineSel.ix;
                    this.txtNodeTime.value = baselineSel.t;
                    this.txtNodeTime.step = this.deMode.data.data.sampling();
                    this.txtNodeLevel.value = qubx.formatK0(baselineSel.level);
                    this.txtNodeLevel.step = qubx.msd(this.deMode.data.plot.bounds[0].dy/100);
                    this._nodeIxs = this.deMode.data.data.segments[this.deMode.data.plot.chosen()].baseline.getNodes().indices;
                    this._nodeIx = this._nodeIxs.indexOf(baselineSel.ix);
                    this.btnNodePrev.disabled = (this._nodeIx === 0);
                    this.btnNodeNext.disabled = (this._nodeIx === (this._nodeIxs.length-1));
                } else {
                    this.panNode.style.visibility = 'hidden';
                }
            }
            _onClickNodePrev() {
                this.deMode.data.plot.pickNode(this._nodeIxs[this._nodeIx-1]);
            }
            _onClickNodeNext() {
                this.deMode.data.plot.pickNode(this._nodeIxs[this._nodeIx+1]);
            }
            _onChangeBaseline(iSeg) {
                this.updateNodePreviewSoon();
            }
            _onResizeNodePreview(info) {
                this.canvas.width = (this.deMode.data.plot.dpRatio * this.canvas.clientWidth)|0;
                this.canvas.height = (this.deMode.data.plot.dpRatio * this.canvas.clientHeight)|0;
                this.welcome.style.width = this.deMode.data.plot.hiresCanvas.clientWidth+'px';
                this.updateNodePreviewSoon();
            }
            _updateNodePreview() {
                const ctx = this.canvas.getContext('2d');
                ctx.clearRect(0, 0, this.canvas.width, this.canvas.height);
                const iSeg = this.deMode.data.plot.chosen();
                if ( (iSeg < 0) || ! this.deMode._sel ) {
                    this.welcome.style.display = '';
                    this.btnSubtract.disabled = true;
                    return;
                }
                const seg = this.deMode.data.data.segments[iSeg];
                if ( ! seg.baseline ) {
                    this.welcome.style.display = '';
                    this.btnSubtract.disabled = true;
                    return;
                }
                this.welcome.style.display = 'none';
                this.btnSubtract.disabled = false;
                const src = seg.sampled[0],
                      bl = seg.baseline.toSampled(this.deMode._sel.selFrom, this.deMode._sel.selTo),
                      i0 = this.deMode._sel.selFrom;
                let i = 0;
                for (; i<bl.length; ++i) {
                    bl[i] = src[i0+i] - bl[i];
                }
                const ex = qubx.extrema(bl),
                      c = .5*(ex.hi + ex.lo);
                let dy = ex.hi - ex.lo;
                dy *= 1.1;
                const lo = c - .5*dy,
                      hi = c + .5*dy,
                      pxw = this.deMode.data.plot.wHi,
                      pxh = this.canvas.height,
                      spp = bl.length / pxw;
                if ( spp >= this.deMode.data.plot.config.minSppRaster ) {
                    ctx.lineWidth = 1.0;
                    this.deMode.data.plot._drawRaster(ctx, 0, pxw, pxh, bl, 0, bl.length, this.deMode.data.plot.config.dataStyle, this.deMode.data.plot.config.normStyle, spp, lo, dy/pxh);
                } else {
                    ctx.lineWidth = 0.5*Math.max(1, Math.round(pxh*this.deMode.data.plot.config.lineHeight)|0);
                    qubtrace.drawLines(ctx, 0, pxw, pxh, bl, 0, bl.length, this.deMode.data.plot.config.lineStyle, spp, lo, dy/pxh);
                }
            }
            _onInputNodeSample() {
                try {
                    this._doInputNodeSample(true, qubx.acceptIntAtLeast(0)(this.txtNodeSample.value));
                } catch (err) {}
            }
            _onInputNodeTime() {
                try {
                    this._doInputNodeSample(false, Math.round(qubx.acceptFloatAtLeast(0.0)(this.txtNodeTime.value) / this.deMode.data.data.sampling()));
                } catch (err) {}
            }
            _doInputNodeSample(enteredSample, ix) {
                const from = this.bnode.ix;
                this.bnode.ix = ix;
                this.bnode.t = ix*this.deMode.data.data.sampling();
                if ( enteredSample ) {
                    this.txtNodeTime.value = this.bnode.t;
                } else {
                    this.txtNodeSample.value = ix;
                }
                const lvl = this.bnode.level;
                this.deMode.data.data.setBaseline(this.deMode.data.plot.chosen(), b => {
                    b.clearNodes(from, from);
                    b.setNodes([ix], [lvl]);
                    this._nodeIxs = b.getNodes().indices;
                    this._nodeIx = this._nodeIxs.indexOf(ix);
                    this.btnNodePrev.disabled = (this._nodeIx === 0);
                    this.btnNodeNext.disabled = (this._nodeIx === (this._nodeIxs.length-1));
                });
            }
            _onInputNodeLevel() {
                var lvl;
                try {
                    lvl = qubx.acceptFloat(this.txtNodeLevel.value);
                } catch (err) {
                    return;
                }
                this.bnode.level = lvl;
                const ix = this.bnode.ix;
                this.deMode.data.data.setBaseline(this.deMode.data.plot.chosen(), b => b.updateNode(ix, lvl));
            }
            _onClickDelNode() {
                const ix = this.bnode.ix;
                this.deMode.data.data.setBaseline(this.deMode.data.plot.chosen(), b => b.clearNodes(ix, ix));
                this._onPickNode(null);
            }
            _onClickSubtract() {
                const iSeg = this.deMode.data.plot.chosen(),
                      seg = this.deMode.data.data.segments[iSeg],
                      src = seg.raw,
                      proc = seg.sampled[0],
                      bl = seg.baseline.toSampled(),
                      n = src.length;
                let i = 0;
                for (; i<n; ++i) {
                    src[i] -= bl[i];
                    proc[i] = src[i];
                }
                if ( this.deMode.data.data.ev.changeRaw.subscribed() ) {
                    this.deMode.data.data.ev.changeRaw.trigger(iSeg, 0, src.length);
                } else {
                    this.deMode.data.data.ev.changeSamples.trigger(iSeg, 0, 0, src.length);
                }
                this.deMode.data.data.reboundSoon(0);
                this.deMode.data.data.clearBaseline(iSeg);
                this.btnSubtract.disabled = true;
                this._onPickNode(null);
            }
        };
        
        class StimBuildTool {
            constructor(deMode, panSB) {
                this.deMode = deMode;
                this.node = panSB;
                this.tabTitle = 'Stimulus Builder';
                this.tabIcon = '<img src="images/stimbuild.svg">';
                const dotdot = qubx_ui.querydotdot(panSB);
                this.welcome = dotdot('#qubx_de_sb_welcome');
                this.canvas = dotdot('canvas', canvas => {
                    qubx_ui.onResize(canvas.parentNode, this._onResizePreview.bind(this));
                });
                this.mnuType = dotdot('#qubx_de_sb_type', mnu => {
                    qubx_ui.setSelectOptions(mnu, ['Ligand', 'Voltage', 'Pressure']);
                    mnu.selectedIndex = 1;
                    mnu.addEventListener('change', this._onChangeSignalType.bind(this));
                });
                this.txtName = dotdot('#qubx_de_sb_name', txt => {
                    txt.style.display = 'none';
                });
                dotdot.listen('#qubx_de_sb_add', 'click', this._onClickAdd.bind(this));
                this.txtDelta = dotdot('#qubx_de_sb_delta');
                this.btnFinish = dotdot('#qubx_de_sb_finish', btn => {
                    btn.disabled = true;
                    btn.addEventListener('click', this._onClickFinish.bind(this));
                });
                this.steps = [];
                this.tbl = new qubx_ui.ColumnTable({
                    node: dotdot('table'),
                    columns: [
                        qubx_ui.ColumnTable.selectCol('Type', 't', ['Step', 'Ramp'], this._onChangeType.bind(this)),
                        qubx_ui.ColumnTable.inputCol('to Level', 'level', 'text', qubx.acceptList(qubx.acceptFloat), qubx.formatList(qubx.formatK0), this._onChangeLevel.bind(this)),
                        qubx_ui.ColumnTable.inputCol('after Duration [ms]', 'dur', 'text', qubx.acceptList(qubx.acceptFloatAtLeast(0.0)), qubx.formatList(qubx.formatK0), this._onChangeDur.bind(this)),
                        qubx_ui.ColumnTable.btnCol('', '<span style="color:red">&#x2716;</span>', this._onClickDel.bind(this))
                    ]
                });
                this.updatePreviewSoon = qubx.batchDelayed(this._updatePreview.bind(this), 100, true);
            }
            onShow(active) {
                active = !! active;
                if ( active ) {
                    this.welcome.style.width = this.deMode.data.plot.hiresCanvas.clientWidth+'px';
                    this.updatePreviewSoon();
                } else if ( this.edited ) {
                    return qubx_ui.confirm({
                        hdr: 'Stimulus builder: unsaved changes',
                        q: "The new stimulus hasn't been created yet. Switch tools anyway?"
                    });
                } else {
                }
                return Promise.resolve(true);
            }
            onChoose(iSeg) {
                this.updatePreviewSoon();
            }
            onSelect(offset, count, segments) {
                this.updatePreviewSoon();
            }
            _onChangeSignalType() {
                this.txtName.style.display = this.mnuType.selectedIndex ? 'none' : '';
            }
            _onResizePreview(info) {
                this.canvas.width = (this.deMode.data.plot.dpRatio * this.canvas.clientWidth)|0;
                this.canvas.height = (this.deMode.data.plot.dpRatio * this.canvas.clientHeight)|0;
                this.welcome.style.width = this.deMode.data.plot.hiresCanvas.clientWidth+'px';
                this.updatePreviewSoon();
            }
            _updatePreview() {
                const ctx = this.canvas.getContext('2d');
                ctx.clearRect(0, 0, this.canvas.width, this.canvas.height);
                if ( this.steps.length === 0 ) {
                    this.welcome.style.display = '';
                    return;
                }
                this.welcome.style.display = 'none';
                const nseg = this.deMode.data.data.segments.length,
                      steps = fullSteps(this.steps, nseg),
                      ex = qubx.extremas(steps.map(qubx.m('level'))),
                      c = .5*(ex.hi + ex.lo);
                let dy = ex.hi - ex.lo;
                dy *= 1.1;
                const lo = c - .5*dy,
                      hi = c + .5*dy,
                      pxw = this.deMode.data.plot.wHi,
                      pxh = this.canvas.height,
                      chosenSeg = this.deMode.data.plot.chosen();
                let i0 = 0;
                var n;
                const sel = this.deMode._selSegs && this.deMode._selSegs.length && this.deMode._selSegs[Math.max(0, chosenSeg)];
                if ( sel ) {
                    i0 = sel.selFrom;
                    n = sel.selTo - i0;
                } else {
                    n = Math.max.apply(null, this.deMode.data.data.segments.map(qubx.m('n')));
                }
                const mx = pxw / n,
                      my = pxh / dy,
                      samp_per_ms = 1e-3 / this.deMode.data.data.sampling();
                ctx.lineWidth = Math.max(1, Math.round(pxh*this.deMode.data.plot.config.lineHeight)|0);
                ctx.strokeStyle = 'black';
                let iSeg = 0;
                for (; iSeg<=nseg; ++iSeg) {
                    let iDraw = iSeg;
                    if ( iSeg === nseg ) {
                        if ( chosenSeg >= 0 ) {
                            ctx.strokeStyle = 'red';
                            iDraw = chosenSeg;
                        } else {
                            break;
                        }
                    } else if ( chosenSeg === iSeg ) {
                        continue;
                    }
                    let px = 0,
                        py = my * (hi - steps[0].level[iDraw]),
                        t = 0,
                        i = 1;
                    ctx.beginPath();
                    ctx.moveTo(px, py);
                    for (; (i<steps.length) && (px<pxw); ++i) {
                        const step = steps[i];
                        t += step.dur[iDraw];
                        px = mx * (samp_per_ms*t - i0);
                        const py1 = my * (hi - step.level[iDraw]);
                        if ( step.t === 'Step' ) {
                            ctx.lineTo(px, py);
                        }
                        ctx.lineTo(px, py1);
                        py = py1;
                    }
                    if ( px < pxw ) {
                        ctx.lineTo(pxw, py);
                    }
                    ctx.stroke();
                }
            }
            _reTable() {
                this.tbl.render(this.steps);
                this.btnFinish.disabled = (this.steps.length === 0);
                this.updatePreviewSoon();
                this.edited = !! this.steps.length;
            }
            _onClickAdd() {
                this.steps.push({
                    t: 'Step',
                    level: [0.0],
                    dur: [10.0]
                });
                this._reTable();
            }
            _onChangeType(row, col, val) {
                this.updatePreviewSoon();
            }
            _onChangeLevel(row, col, val) {
                this.updatePreviewSoon();
            }
            _onChangeDur(row, col, val) {
                this.updatePreviewSoon();
            }
            _onClickDel(d,i) {
                this.steps.splice(i, 1);
                this._reTable();
            }
            _onClickFinish() {
                const data = this.deMode.data.data,
                      isV = this.mnuType.selectedIndex === 1,
                      isP = this.mnuType.selectedIndex === 2,
                      name = (isV ? 'Voltage' : (isP ? 'Pressure' : this.txtName.value));
                if ( ! name ) {
                    alert('Please enter a name for the ligand.');
                    return;
                }
                let step1 = Promise.resolve(true),
                    iSig = data.signals.length;
                data.signals.forEach((sig, i) => {
                    if ( sig.name === name ) {
                        step1 = qubx_ui.confirm({
                            hdr: 'Replace entire '+name+' signal?',
                            q: 'There is already a signal named "'+name+'."  Replace it?'
                        });
                        iSig = i;
                    }
                });
                step1.then(() => {
                    if ( iSig === data.signals.length ) {
                        data.signalCount(data.signalCount()+1, (sig, j) => {
                            if ( j === iSig ) {
                                sig.name = name;
                            }
                        });
                    }
                    const nseg = data.segments.length,
                          steps = fullSteps(this.steps, nseg),
                          mx = 1e-3/data.sampling();
                    data.segments.forEach((seg, iSeg) => {
                        const buf = new Float32Array(data.segments[iSeg].n);
                        let t = 0,
                            y0 = 0,
                            iOut = 0;
                        for (const step of steps) {
                            t += step.dur[iSeg];
                            const iPast = Math.min(buf.length, Math.floor(t*mx)+1);
                            if ( step.t === 'Step' ) {
                                for (; iOut<iPast; ++iOut) {
                                    buf[iOut] = y0;
                                }
                            } else {
                                const my = (step.level[iSeg] - y0) / (t*mx - iOut);
                                let iHere = 0;
                                for (; iOut<iPast; ++iOut, ++iHere) {
                                    buf[iOut] = y0 + my*iHere;
                                }
                            }
                            y0 = step.level[iSeg];
                        }
                        for (; iOut<buf.length; ++iOut) {
                            buf[iOut] = y0;
                        }
                        data.setSampled(iSeg, iSig, 0, buf, buf.length);
                        data.reboundSoon(iSig);
                        this.edited = false;
                        const delta = (+this.txtDelta.value) || 0.01;
                        this.deMode.main.signals.get(name).delta = delta;
                        qubx_data_prep.idlStim(this.deMode.main.dataWorker, [buf], delta) /// , onProgress
                            .then(segEvents => {
                                const events = segEvents[0];
                                data.setIdeal(iSeg, iSig, events.counts, events.classes, 0, events.amps, events.stds);
                            }).catch(qubx.reportErr);
                    });
                }).catch(qubx.reportErr);
            }
        };

        function fullArray(a, nseg) {
            const z = a.slice(),
                  x = a.length ? a[a.length-1] : 0.0;
            while ( z.length < nseg ) {
                z.push(x);
            }
            return z;
        }
        function fullSteps(steps, nseg) {
            return [{
                t: 'Step',
                level: [0.0],
                dur: [0.0]
            }].concat(steps).map(step => ({
                t: step.t,
                level: fullArray(step.level, nseg),
                dur: fullArray(step.dur, nseg)
            }));
        }

        class SelectTool {
            constructor(deMode, panSelect) {
                this.deMode = deMode;
                this.node = panSelect;
                this.tabTitle = 'Select and Upload';
                this.tabIcon = '<img src="images/lists.svg">';
                const dotdot = qubx_ui.querydotdot(panSelect);
                dotdot.listen("#qubx_de_sel_add", 'click', this._onClickAdd.bind(this));
                dotdot.listen("#qubx_de_sel_none", 'click', this._onClickNone.bind(this));
                dotdot.listen("#qubx_de_sel_all", 'click', this._onClickAll.bind(this));
                dotdot.listen("#qubx_de_sel_chop", 'click', this._onClickChop.bind(this));
                dotdot.listen("#qubx_de_sel_chopidl", 'click', this._onClickChopIdl.bind(this));
                dotdot.listen("#qubx_de_sel_bystim", 'click', this._onClickByStim.bind(this));
                this.chkJoin = dotdot.listen("#qubx_de_sel_join");
                this.tbl = new qubx_ui.ColumnTable({
                    node: dotdot.listen('#qubx_de_sel_tbl'),
                    columns: [
                        qubx_ui.ColumnTable.htmlCol('', 'html'),
                        qubx_ui.ColumnTable.btnCol('', '<span style="color:red">&#x2716;</span>', this._onClickDel.bind(this))
                    ]
                });
                this.btnCrop = dotdot.listen("#qubx_de_sel_crop", 'click', this._onClickCrop.bind(this));
                this.btnCrop.disabled = true;
                this.btnUpload = dotdot.listen("#qubx_de_sel_upload", 'click', this._onClickUpload.bind(this));
                this.btnUpload.disabled = true;
                cloud.ev.signin.sub(() => {
                    this.btnUpload.disabled = true; //// this.sels && this.sels.length;
                });
                cloud.ev.signout.sub(() => {
                    this.btnUpload.disabled = true;
                });
                this.sels = [];
                
                this.dlgChop = new qubx_ui.Modal({
                    template: '#qubx_de_chop'
                });
                this.dlgChop.dom.then(dom => {
                    const dlgdot = qubx_ui.querydotdot(dom.content);
                    this.txtChopSkipPre = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chop_skipPre'),
                        0.0,
                        {
                            accept: qubx.acceptFloatAtLeast(0.0),
                            format: qubx.formatK0,
                            localKey: 'qubx_de_chop_skipPre',
                            widthChars: 8
                        }
                    );
                    this.txtChopGet = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chop_get'),
                        100.0,
                        {
                            accept: qubx.acceptFloatGreaterThan(0.0),
                            format: qubx.formatK0,
                            localKey: 'qubx_de_chop_get',
                            widthChars: 8
                        }
                    );
                    this.txtChopSkipPost = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chop_skipPost'),
                        0.0,
                        {
                            accept: qubx.acceptFloatAtLeast(0.0),
                            format: qubx.formatK0,
                            localKey: 'qubx_de_chop_skipPost',
                            widthChars: 8
                        }
                    );
                    this.chkChopOnce = dlgdot('#qubx_de_chop_once');
                    this.txtChopSkipSegFirst = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chop_skipSegFirst'),
                        0.0,
                        {
                            accept: qubx.acceptFloatAtLeast(0.0),
                            format: qubx.formatK0,
                            localKey: 'qubx_de_chop_skipSegFirst',
                            widthChars: 8
                        }
                    );
                    this.txtChopSkipSegLast = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chop_skipSegLast'),
                        0.0,
                        {
                            accept: qubx.acceptFloatAtLeast(0.0),
                            format: qubx.formatK0,
                            localKey: 'qubx_de_chop_skipSegLast',
                            widthChars: 8
                        }
                    );
                    this.chkChopDiscardPartial = dlgdot('#qubx_de_chop_discardPartial');
                });
                
                this.dlgChopIdlFail = new qubx_ui.Modal({
                    template: '#qubx_de_chopidl_fail',
                    okLabel: 'Switch now'
                });
                
                this.dlgChopIdl = new qubx_ui.Modal({
                    template: '#qubx_de_chopidl'
                });
                this.dlgChopIdl.dom.then(dom => {
                    const dlgdot = qubx_ui.querydotdot(dom.content);
                    this.chkChopIdlBrkCls = dlgdot('#qubx_de_chopidl_brk_bycls');
                    this.txtChopIdlBrkCls = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chopidl_brk_cls'),
                        0,
                        {
                            accept: qubx.acceptIntAtLeast(0),
                            format: qubx.formatStr,
                            localKey: 'qubx_de_chopidl_brk_cls',
                            widthChars: 3
                        }
                    );
                    this.txtChopIdlBrkTm = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chopidl_brk_tm'),
                        1000.0,
                        {
                            accept: qubx.acceptFloatGreaterThan(0.0),
                            format: qubx.formatK0,
                            localKey: 'qubx_de_chopidl_brk_tm',
                            widthChars: 8
                        }
                    );
                    this.txtChopIdlBrkN = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chopidl_brk_n'),
                        100,
                        {
                            accept: qubx.acceptIntAtLeast(1),
                            format: qubx.formatStr,
                            localKey: 'qubx_de_chopidl_brk_n',
                            widthChars: 5
                        }
                    );
                    this.chkChopIdlDiscardByCls = dlgdot('#qubx_de_chopidl_discard_bycls');
                    this.txtChopIdlDiscardCls = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chopidl_discard_cls'),
                        2,
                        {
                            accept: qubx.acceptIntAtLeast(1),
                            format: qubx.formatStr,
                            localKey: 'qubx_de_chopidl_discard_cls',
                            widthChars: 3
                        }
                    );
                    this.chkChopIdlDiscardByDur = dlgdot('#qubx_de_chopidl_discard_bydur');
                    this.txtChopIdlDiscardDur = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chopidl_discard_dur'),
                        1.0,
                        {
                            accept: qubx.acceptFloatGreaterThan(0.0),
                            format: qubx.formatK0,
                            localKey: 'qubx_de_chopidl_discard_dur',
                            widthChars: 8
                        }
                    );
                    this.chkChopIdlDiscardByCount = dlgdot('#qubx_de_chopidl_discard_bycount');
                    this.txtChopIdlDiscardCount = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_chopidl_discard_count'),
                        2,
                        {
                            accept: qubx.acceptIntAtLeast(1),
                            format: qubx.formatStr,
                            localKey: 'qubx_de_chopidl_discard_count',
                            widthChars: 5
                        }
                    );
                    this.chkChopIdlDropFinal = dlgdot('#qubx_de_chopidl_dropfinal');
                });
                
                this.dlgByStim = new qubx_ui.Modal({
                    template: '#qubx_de_bystim'
                });
                this.dlgByStim.dom.then(dom => {
                    const dlgdot = qubx_ui.querydotdot(dom.content);
                    this.mnuByStimSignal = dlgdot('#qubx_de_bystim_signal');
                    this.chkByStimAbove = dlgdot('#qubx_de_bystim_above');
                    this.txtByStimThresh = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_bystim_thresh'),
                        0.0,
                        {
                            accept: qubx.acceptFloat,
                            format: qubx.formatK0,
                            localKey: 'qubx_de_bystim_thresh',
                            widthChars: 8
                        }
                    );
                    this.txtByStimAddPre = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_bystim_addpre'),
                        0.0,
                        {
                            accept: qubx.acceptFloatAtLeast(0.0),
                            format: qubx.formatK0,
                            localKey: 'qubx_de_bystim_addpre',
                            widthChars: 8
                        }
                    );
                    this.txtByStimAddPost = new qubx_ui.AcceptedInput(
                        dlgdot('#qubx_de_bystim_addpost'),
                        0.0,
                        {
                            accept: qubx.acceptFloatAtLeast(0.0),
                            format: qubx.formatK0,
                            localKey: 'qubx_de_bystim_addpost',
                            widthChars: 8
                        }
                    );
                });
                
                this.dlgUpload = new qubx_ui.Modal({
                    template: '#qubx_de_upload'
                });
                this.dlgUpload.dom.then(dom => {
                    cloud.groupNames = ['Admin', 'Examples']; ///
                    
                    const dlgdot = qubx_ui.querydotdot(dom.content);
                    this.dlgUpload.outline = dlgdot('#qubx_upload_outline', node => new qubx_cloud_ui.FileListing({
                        canSelectFolder: true,
                        canSelectFile: true,
                        onSelect: item => console.log('onSelect',item),
                        node: node,
                        autoExpand: false
                    })); /// focus and arrows; padding; pair with metadata preview and/or edit (toggle fields editable? or confirm edits?)
                    this.dlgUpload.outline.data(
                        'Datasets', [
                            {datasetName: 'Data2015893282392', datasetPoints: 1000008, datasetSegments: 1, datasetFolder: '201589', datasetGroup: 0},
                            {datasetName: 'Data2015893282395', datasetPoints: 2000008, datasetSegments: 2, datasetFolder: '201589', datasetGroup: 0},
                            {datasetName: 'Data2015893282391', datasetPoints: 3000008, datasetSegments: 3, datasetFolder: '201589', datasetGroup: 1},
                            {datasetName: 'Data2015893282393', datasetPoints: 4000008, datasetSegments: 4, datasetFolder: '201589', datasetGroup: 1},
                            {datasetName: 'Data2015893282399', datasetPoints: 5000008, datasetSegments: 5, datasetFolder: '201589', datasetGroup: 1},
                        ],
                        f => f.datasetName + ' ('+f.datasetPoints+' points in '+f.datasetSegments+' segments)',
                        f => f.datasetFolder,
                        f => cloud.groupNames[f.datasetGroup]
                    );
                });
            /// move to qubx_cloud_ui
            ///  remember last folder and all metadata:
            ///   public, name, about, whenRec, molecule, free-entry key/value (etc)
            ///  folder dropdown with '+' button (display folder==='' as '[none]')
            ///   '...' expands to folder browser with gray collapsed files, selectable folder (display folder==='' at root level below folders)

            /// also need file picker for model/dataset/project
            ///  same folder browser, only files pickable
            ///  preview metadata
            ///   (also most-recent project result doc?)
            /// include 'Unsorted' folder even if it's empty

            /// Outline restore highlight on expand

            /// both sides spinner while listing
            
                qubx_ui.loadTemplate('.qubx_de_sel_common').then(common => {
                    [this.dlgChop, this.dlgChopIdl, this.dlgByStim].forEach(dlg => {
                        dlg.dom.then(dom => {
                            const box = dom.content.querySelector('.qubx_de_sel_commonbox');
                            box.appendChild(common.cloneNode(true));
                            dlg.chkUseSel = box.querySelector('.qubx_de_sel_sel');
                            dlg.chkUseFile = box.querySelector('.qubx_de_sel_file');
                            dlg.chkReplace = box.querySelector('.qubx_de_sel_replace');
                            dlg.chkUseFile.checked = true;
                            dlg.chkReplace.checked = true;
                        });
                        dlg.readCommon = () => ({
                            useSel: dlg.chkUseSel.checked,
                            useFile: dlg.chkUseFile.checked,
                            replace: dlg.chkReplace.checked
                        });
                    });
                });
                this.deMode.main.data.data.ev.doneSegSize.sub(this._onDoneSegSize.bind(this));
            }
            onShow(active) {
                return Promise.resolve(true);
            }
            _onDoneSegSize() {
                this.sels = [];
                this._reTable();
            }
            _onClickAdd() {
                if ( this.deMode._hiliteSegs && this.deMode._hiliteSegs.length && this.deMode._hiliteSegs[0].to ) {
                    this.deMode._hiliteSegs.forEach(seg => this.sels.push({
                        s: seg.index,
                        f: seg.from,
                        l: seg.to
                    }));
                } else if ( this.deMode._sel ) {
                    this.sels.push({
                        s: this.deMode.main.data.plot.chosen(),
                        f: this.deMode._sel.selFrom,
                        l: this.deMode._sel.selTo
                    });
                } else if ( this.deMode._selSegs ) {
                    this.deMode._selSegs.forEach((seg, s) => this.sels.push({
                        s: s,
                        f: seg.selFrom,
                        l: seg.selTo
                    }));
                }
                this._reTable();
            }
            _onClickAll() {
                this.sels = this.deMode.data.data.segments.map((seg, s) => ({
                    s: s,
                    f: 0,
                    l: seg.n
                }));
                this._reTable();
            }
            _onClickNone() {
                this.sels = []; //// undo?
                this._reTable();
            }
            _onClickChop() {
                const sampling = this.deMode.data.data.sampling();
                [this.txtChopSkipPre, this.txtChopGet, this.txtChopSkipPost, this.txtChopSkipSegFirst, this.txtChopSkipSegLast]
                    .forEach(txt => txt.update(undefined, 1e3*sampling));
                this.dlgChop.run().then(() => {
                    const common = this.dlgChop.readCommon(),
                          useSel = common.useSel,
                          replace = common.replace,
                          msToSamples = 1e-3/sampling,
                          skipPre = Math.round(this.txtChopSkipPre.value*msToSamples)|0,
                          get = Math.round(this.txtChopGet.value*msToSamples)|0,
                          skipPost = Math.round(this.txtChopSkipPost.value*msToSamples)|0,
                          once = this.chkChopOnce.checked,
                          skipSegFirst = Math.round(this.txtChopSkipSegFirst.value*msToSamples)|0,
                          skipSegLast = Math.round(this.txtChopSkipSegLast.value*msToSamples)|0,
                          discardPartial = this.chkChopDiscardPartial.checked;
                    if ( replace ) {
                        this.sels = [];
                    }
                    getChopSource(this.deMode, useSel, skipSegFirst, skipSegLast).forEach(seg => {
                        let at = seg.f;
                        while ( at <= seg.l ) {
                            const f = at + skipPre;
                            let l = f + get - 1;
                            if ( (l > seg.l) && discardPartial ) {
                                break;
                            }
                            l = Math.min(l, seg.l);
                            this.sels.push({
                                s: seg.index,
                                f: f,
                                l: l
                            });
                            at = l + skipPost + 1;
                            if ( once ) {
                                break;
                            }
                        }
                    });
                    this._reTable();
                }).catch(qubx.reportErr);
            }
            _onClickChopIdl() {
                const data = this.deMode.data.data;
                if ( ! (data.segments.length && data.segments[0].ideal[0] && (0 <= data.segments[0].ideal[0].classAt(0))) ) {
                    this.dlgChopIdlFail.run()
                        .then(() => this.deMode.main.mode(this.deMode.main.singleMode));
                    return;
                }
                const sampling = data.sampling();
                [this.txtChopIdlBrkTm, this.txtChopIdlDiscardDur]
                    .forEach(txt => txt.update(undefined, 1e3*sampling));
                this.dlgChopIdl.run().then(() => {
                    const common = this.dlgChopIdl.readCommon(),
                          msToSamples = 1e-3/sampling,
                          useSel = common.useSel,
                          replace = common.replace,
                          brkByCls = this.chkChopIdlBrkCls.checked,
                          brkCls = this.txtChopIdlBrkCls.value,
                          brkSamples = Math.round(msToSamples*this.txtChopIdlBrkTm.value)|0,
                          brkN = this.txtChopIdlBrkN.value,
                          discardByCls = this.chkChopIdlDiscardByCls.checked,
                          discardCls = this.txtChopIdlDiscardCls.value,
                          discardByDur = this.chkChopIdlDiscardByDur.checked,
                          discardSamples = Math.round(msToSamples*this.txtChopIdlDiscardDur.value)|0,
                          discardByCount = this.chkChopIdlDiscardByCount.checked,
                          discardCount = this.txtChopIdlDiscardCount.value,
                          dropFinal = this.chkChopIdlDropFinal.checked;
                    if ( replace ) {
                        this.sels = [];
                    }
                    const checkBreak = brkByCls ? breakOnTcrit(brkCls, brkSamples) : breakOnCount(brkN);
                    if ( brkByCls ) {
                        this.deMode.data.data.tCrit(+this.txtChopIdlBrkTm.value);
                    }
                    const discards = [];
                    if ( discardByCls ) {
                        discards.push(discardOnClass(discardCls));
                    }
                    if ( discardByDur ) {
                        discards.push(discardOnLength(discardSamples));
                    }
                    if ( discardByCount ) {
                        discards.push(discardOnNevent(discardCount));
                    }
                    chopIdl(data, getChopSource(this.deMode, useSel), checkBreak, brkByCls, discards, dropFinal)
                        .forEach(seg => this.sels.push({
                            s: seg.index,
                            f: seg.f,
                            l: seg.l
                        }));
                    this._reTable();
                }).catch(qubx.reportErr);
            }
            _onClickByStim() {
                const data = this.deMode.data.data,
                      sampling = data.sampling();
                [this.txtByStimAddPre, this.txtByStimAddPost]
                    .forEach(txt => txt.update(undefined, 1e3*sampling));
                qubx_ui.setSelectOptions(this.mnuByStimSignal, data.signals.slice(1).map(qubx.m('name')));
                this.dlgByStim.run().then(() => {
                    const common = this.dlgChopIdl.readCommon(),
                          msToSamples = 1e-3/sampling,
                          useSel = common.useSel,
                          replace = common.replace,
                          iSig = this.mnuByStimSignal.selectedIndex + 1,
                          isAbove = this.chkByStimAbove.checked,
                          thresh = this.txtByStimThresh.value,
                          addPre = Math.round(msToSamples*this.txtByStimAddPre.value)|0,                        
                          addPost = Math.round(msToSamples*this.txtByStimAddPost.value)|0;
                    if ( replace ) {
                        this.sels = [];
                    }
                    chopByStim(data, getChopSource(this.deMode, useSel), iSig, isAbove, thresh, addPre, addPost)
                        .forEach(seg => this.sels.push({
                            s: seg.index,
                            f: seg.f,
                            l: seg.l
                        }));
                    this._reTable();
                }).catch(qubx.reportErr);
            }
            _onClickDel(d, i) {
                this.sels.splice(i, 1);
                this._reTable();
            }
            _onClickSel(d, i) {
                const maxNpoint = this.deMode.data.plot.maxNpoint;
                this.deMode.data.plot.chosen(d.s);
                this.deMode.data.plot.zoom(d.f/maxNpoint, d.l/(maxNpoint-1));
            }
            _reTable() {
                this.sels.sort((a,b) =>(a.s - b.s) || (a.f - b.f) || (a.l - b.l));
                this.sels.forEach((sel, i) => {
                    sel.html = `${i+1}. trace ${sel.s+1}:<a data-sel-list-index="${i}">${sel.f}&ndash;${sel.l}</a>`;
                });
                this.tbl.render(this.sels);
                const sels = this.sels,
                      handler = this._onClickSel.bind(this);
                d3.select(this.tbl.config.node)
                    .selectAll('a')
                    .property('href', '#')
                    .each(function() {
                        const i = this.dataset.selListIndex|0;
                        this.onclick = function(event) {
                            event.preventDefault();
                            event.stopPropagation();
                            handler(sels[i], i);
                        }
                    });
                this.btnCrop.disabled = ! this.sels.length;
                this.btnUpload.disabled = true; //// ! (this.sels.length && cloud.personName);
            }
            _onClickCrop() {
                if ( ! this.sels.length ) {
                    return;
                }
                const sels = this.sels,
                      main = this.deMode.main,
                      srcdata = main.data.data,
                      join = this.chkJoin.checked,
                      counts = [],
                      signalNames = srcdata.signals.map(qubx.m('name')),
                      constants = {};
                srcdata.constants.entries.forEach(c => {
                    constants[c.Name] = c.Value;
                });
                constants['Tdead (ms)'] = srcdata.tDead();
                constants['Tcrit (ms)'] = srcdata.tCrit();
                const idealized = srcdata.signals.map((sig, iSig) => sels.map(sel => {
                    let idl = srcdata.segments[sel.s].ideal[iSig];
                    const events = idl ? idl.getDwellsCD(sel.f, sel.l, null, null, 1) : {
                        classes: [],
                        durations: []
                    };
                    idl = idl || {
                        amp: [],
                        std: []
                    };
                    const nd = events.durations.length,
                          counts = new Int32Array(nd);
                    let i = 0;
                    for (; i<nd; ++i) {
                        counts[i] = events.durations[i];
                    }
                    return {
                        classes: events.classes,
                        counts: counts,
                        amps: idl.amp,
                        stds: idl.std
                    };
                }));
                const cropped = { // implements data reader interface
                    buffers: sels.map((sel, iSel) => {
                        const i0 = sel.f,
                              n = sel.l - i0 + 1,
                              seg = srcdata.segments[sel.s];
                        counts[iSel] = n;
                        return seg.sampled.map((src, iSrc) => {
                            if ( ! src ) {
                                return src;
                            }
                            if ( ! iSrc ) {
                                src = seg.raw;
                            }
                            const dest = new Float32Array(n);
                            let i = 0;
                            for (; i<n; ++i) {
                                dest[i] = src[i+i0];
                            }
                            return dest;
                        });
                    }),
                    counts: counts,
                    isSampled: srcdata.signals.map(sig => true),
                    idealized: idealized,
                    sampling: srcdata.sampling(),
                    signalNames: signalNames,
                    signalDeltas: srcdata.signals.map(qubx.m('delta')),
                    sigData: signalNames.map(main.signals.get.bind(main.signals)),
                    constants: constants,
                    readSegment: (iSeg, onread) => onread(cropped, {
                        n: cropped.counts[iSeg],
                        buffers: cropped.buffers[iSeg]
                    }, iSeg),
                    readSegments: onread => cropped.buffers.map((segbufs, iSeg) => cropped.readSegment(iSeg, onread))
                }
                const joined = join && {
                    isSampled: cropped.isSampled,
                    sampling: cropped.sampling,
                    signalNames: cropped.signalNames,
                    signalDeltas: cropped.signalDeltas,
                    sigData: cropped.sigData,
                    constants: cropped.constants,
                    readSegment: function(iSeg, onread) {
                        return onread(joined, {
                            n: this.counts[0],
                            buffers: this.buffers
                        }, 0);
                    },
                    readSegments: function(onread) {
                        return [this.readSegment(0, onread)];
                    },
                    preread: function() {
                        let n = 0;
                        cropped.counts.forEach(c => {
                            n += c;
                        });
                        this.counts = [n];
                        if ( ! n ) {
                            return;
                        }
                        this.buffers = cropped.buffers[0].map(() => new Float32Array(n));
                        let p = 0;
                        cropped.readSegments((cr, seg, iSeg) => {
                            this.buffers.forEach((buf, s) => buf.set(seg.buffers[s], p));
                            p += seg.n;
                        });
                        let nEvent = 0;
                        cropped.idealized[0].forEach(idlseg => {
                            nEvent += idlseg.counts.length;
                        });
                        const idlOut = {
                            classes: new Int32Array(nEvent),
                            counts: new Int32Array(nEvent),
                            amps: [],
                            stds: []
                        };
                        this.idealized = [ [idlOut] ];
                        p = 0;
                        cropped.idealized[0].forEach(idlseg => {
                            idlOut.classes.set(idlseg.classes, p);
                            idlOut.counts.set(idlseg.counts, p);
                            p += idlseg.counts.length;
                            // "ideally" we'd disambiguate classes within delta I and return the union
                            // instead we preserve class index, keeping the last (amp,std) encountered for each class
                            idlseg.amps.forEach((a, i) => {
                                idlOut.amps[i] = a;
                                idlOut.stds[i] = idlseg.stds[i];
                            });
                        });
                        return this;
                    }
                }.preread();
                const reader = joined || cropped;
                
                // show progress
                main.progressDlg.show('Cropping data...');
                qubx_data_prep.prepData(main.dataWorker, reader, 0, main.signals, prog => {
                    main.progressDlg.progress(prog);
                    return ! main.progressDlg.canceled;
                }).then(info => {
                        main._readData(reader, reader.signalNames.map(qubx.di), reader.signalNames, reader.signalDeltas,
                                       reader.idealized);
                        main.data.plot.zoom(0, 1);
                        main.progressDlg.hide();
                }).catch(qubx.reportErr)
                    .catch(err => {
                        main.progressDlg.hide();
                        reject(err);
                    });
            }
            _onClickUpload() {
                /// this.dlgUpload.prepare(this.sels); ///
                this.dlgUpload.run().then(meta => {
                    ///   folder, public, name, about, whenRec, molecule, free-entry key/value (etc)
                    /// we also know: sampling, points, events, segments, stimuli, constants
                    
                    /// find out what we need to satisfy *Reader
                    ///  put all but samples and idl into index.json
                    ///  get upload auth
                    ///  post index.json
                    ///  cloud.uploading()...
                    ///  post idealized.jsotan
                    ///  one-by-one post I${i}.jsotan and S${i].jsotan
                    
                    /// ask if they want to start a new project with it now
                });
            }
        };

        function getChopSource(deMode, useSel, skipSegFirst, skipSegLast) {
            skipSegFirst = skipSegFirst|0;
            skipSegLast = skipSegLast|0;
            const segments = (useSel && deMode._sel && [deMode._sel]) || deMode.data.data.segments.map((seg, iSeg) => ({
                selFrom: 0,
                selTo: seg.n-1,
                index: iSeg,
                n: seg.n
            }));
            const offset = useSel ? deMode._selSegs[0].selFrom : 0;
            return segments.map((seg, iSeg) => {
                const hilite = deMode._hiliteSegs && deMode._hiliteSegs[iSeg];
                if ( useSel && hilite && hilite.to ) {
                    const isect = {
                        f: Math.max(hilite.from, seg.selFrom),
                        l: Math.min(hilite.to, seg.selTo)
                    };
                    if ( isect.f <= isect.l ) {
                        return {
                            index: seg.index,
                            f: isect.f + skipSegFirst,
                            l: isect.l - skipSegLast
                        };
                    }
                }
                return {
                    index: seg.index,
                    f: offset + skipSegFirst,
                    l: useSel ? Math.min(seg.n-1, seg.selTo-skipSegLast) : (seg.n - 1)
                };
            }).filter(seg => seg.f <= seg.l);
        }

        function chopIdl(data, segments, checkBreak, consumeBreak, discards, dropFinal) {
            let maxNevent = 0;
            var firsts, lasts, classes;
            const segmentsOut = [];
            segments.forEach(seg => {
                const idl = data.segments[seg.index].ideal[0],
                      Nevent = idl.countDwells(seg.f, seg.l+1);
                if ( Nevent > maxNevent ) {
                    maxNevent = Nevent;
                    firsts = new Int32Array(Nevent);
                    lasts = new Int32Array(Nevent);
                    classes = new Int32Array(Nevent);
                }
                idl.getDwellsFLC(seg.f, seg.l, firsts, lasts, classes);
                let ie = 0;
                while ( ie < Nevent ) {
                    const ie0 = ie;
                    while ( (ie < Nevent) && ! checkBreak(firsts[ie], lasts[ie], classes[ie]) ) {
                        ++ie;
                    } // now ie is on a break event, which will be consumed if consumeBreak, or re-checked next iter
                    if ( ie0 < ie ) {
                        let keep = true,
                            d = 0;
                        for (; d<discards.length; ++d) {
                            if ( discards[d](firsts, lasts, classes, ie0, ie) ) {
                                keep = false;
                                break;
                            }
                        }
                        if ( keep ) {
                            const ieAdj = (dropFinal && (ie > (ie0+1))) ? (ie-1) : ie;
                            segmentsOut.push({
                                index: seg.index,
                                f: firsts[ie0],
                                l: lasts[ieAdj-1]
                            });
                        }
                    }
                    if ( consumeBreak ) {
                        ++ie;
                    }
                }
            });
            return segmentsOut;
        }
        
        function breakOnTcrit(cls, count) { // with consumeBreak=true
            return (f, l, c) => (c === cls) && ((l-f+1) > count);
        }
        function breakOnCount(n) { // with consumeBreak=false
            let i = -1;
            return (f, l, c) => {
                ++i;
                if ( i === n ) {
                    i = -1;
                    return true;
                }
                return false;
            };
        }
        function discardOnClass(cls) {
            return (firsts, lasts, classes, start, end) => {
                let i = start;
                for (; i<end; ++i) {
                    if ( classes[i] >= cls ) {
                        return true;
                    }
                }
                return false;
            };
        }
        function discardOnLength(count) {
            return (firsts, lasts, classes, start, end) => {
                const n = lasts[end-1] - firsts[start] + 1;
                return (n < count);
            };
        }
        function discardOnNevent(count) {
            return (firsts, lasts, classes, start, end) => {
                const n = end - start;
                return (n < count);
            };
        }

        function chopByStim(data, segments, iSig, isAbove, threshold, countPre, countPost) {
            const segmentsOut = [],
                  passes = isAbove ? (v => v > threshold) : (v => v < threshold);
            segments.forEach(seg => {
                const stim = data.segments[seg.index].sampled[iSig];
                let at = seg.f,
                    i0 = at,
                    passing = passes(stim[at]);
                for (;at<=seg.l; ++at) {
                    const nextPassing = passes(stim[at]);
                    if ( nextPassing && ! passing ) {
                        i0 = Math.max(seg.f, at-countPre);
                    } else if ( passing && ! nextPassing ) {
                        segmentsOut.push({
                            index: seg.index,
                            f: i0,
                            l: Math.min(seg.l, at+countPost-1)
                        });
                    }
                    passing = nextPassing;
                }
                if ( passing ) {
                    segmentsOut.push({
                        index: seg.index,
                        f: i0,
                        l: seg.l
                    });
                }
            });
            return segmentsOut;
        }
        
        class AnalyzeTool {
            constructor(deMode, panAnalyze) {
                this.deMode = deMode;
                this.node = panAnalyze;
                this.tabTitle = 'Analyze';
                this.tabIcon = '...';
            }
        }
        
        return exports;
    });

//// should update datasetEvents when replacing idl
        
/// group names are public for file listing purposes
///   unprivileged listGroups with less info

