/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_demo_mac", ["qubx", "qubx_ui", "qubx_ui_data", "qubx_lib", "qubx_mac", "qubx_data_prep", "d3"],
    (qubx, qubx_ui, qubx_ui_data, qubx_lib, qubx_mac, qubx_data_prep, d3) => {
        "use strict";
        const exports = {};

        const MacModeDefs = {
            minimize: {
                lm: false,
                dfp: true,
                simplex: true,
                maxIter: 60,
                maxStep: 0.1,
                dx: 1e-7,
                optNchannel: false
            },
            workers: 3
        };
        const MacOptDefs = {
            rates: true,
            cc: false,
            cond: false,
            variance: false
        };

        class WaitModel {
            constructor(rec) {
                this.model = rec.model;
                this.isChanged = false;
                this.rates = new Map();
                return this;
            }
            changed() {
                this.rates = null;
                this.isChanged = true;
            }
            rateChanged(info) {
                if ( this.rates ) {
                    this.rates.set([info.row, info.column], info);
                }
            }
            ccChanged(channelCount) {
                this.channelCount = channelCount;
            }
            changedRates() {
                return this.rates ? this.rates.values() : [];
            }
        };
        
        exports.MacMode = class MacMode {
            constructor(main, config) {
                config = config || {};
                this.main = main;
                let storedPrefs = localStorage.getItem("qubx_mac_prefs");
                storedPrefs = storedPrefs ? JSON.parse(storedPrefs) : {};
                this.config = Object.assign({}, MacModeDefs, config, storedPrefs);
                this.config.minimize = Object.assign({}, MacModeDefs.minimize, config.minimize, storedPrefs.minimize);
                this.prepDataSoon = qubx.batchDelayed(this.prepData.bind(this), 1000);
                this.sendModelsSoon = qubx.batchDelayed(this.sendModels.bind(this), 10);
                this.main.panels.then(p => {
                    this.modelView = p.modelView;
                    this.data = p.data;
                    this.controls = p.controls.querySelector("#qubx_demo_data_ctrls");
                    this._makeMenus(p);
                    this._makeDialogs(p);
                }).catch(qubx.reportErr);
                this.sentConst = {};
                this.resultStyle = qubx.get("css/qubx_resultDoc.css");
            }
            activate() {
                this.mac = new qubx_mac.MacWorkgroup({count: this.config.workers});
                this.main.panels.then(p => {
                    this._modelInsertModel = p.modelView.models.ev.insert.sub(this._onInsertModel.bind(this));
                    this._modelRemoveModel = p.modelView.models.ev.remove.sub(this._onRemoveModel.bind(this));
                    this._modelActivateModel = p.modelView.models.ev.activate.sub(this._onActivateModel.bind(this));
                    this._modelSetChannelCount = p.modelView.models.ev.setChannelCount.sub(this._onSetChannelCount.bind(this));
                    this._modelSetIeqFv = p.modelView.models.ev.setIeqFv.sub(this._onSetIeqFv.bind(this));
                    this._modelSetVRev = p.modelView.models.ev.setVRev.sub(this._onSetVRev.bind(this));
                    this._modelSetState = p.modelView.models.ev.setState.sub(this._onSetState.bind(this));
                    this._modelSetClass = p.modelView.models.ev.setClass.sub(this._onSetClass.bind(this));
                    this._modelInsertedRate = p.modelView.models.ev.insertRate.sub(this._onInsertedRate.bind(this));
                    this._modelRemovedRate = p.modelView.models.ev.removeRate.sub(this._onRemovedRate.bind(this));
                    this._modelSetRate = p.modelView.models.ev.setRate.sub(this._onSetRate.bind(this));
                    this._dataSetConst = this.data.data.constants.ev.set.sub(this._onSetConst.bind(this));
                    this._dataChangeSamples = this.data.data.ev.changeSamples.sub(this._onChangeSamples.bind(this));
                    this.menus.forEach(node => this.controls.appendChild(node));
                    this.sendModels();
                    p.modelView.models.models.forEach(m => m.model.rates.entries.forEach(r => {
                        r.Ligand && this.ensureConst(r.Ligand);
                        r.Voltage && this.ensureConst(r.Voltage);
                        r.Pressure && this.ensureConst(r.Pressure);
                    }));
                    this.prepDataSoon();
                }).catch(qubx.reportErr);
            }
            deactivate() {
                return ((this.btnRun.innerHTML === 'Stop') ? qubx_ui.confirm({
                    hdr: 'Optimizer is still running',
                    q: 'Switch tools anyway, interrupting the optimizer?'
                }).then(() => (this.btnRun.innerHTML === 'Stop') && this._onClickRun()) : Promise.resolve(true)).then(() => {
                    this.menus.forEach(node => this.controls.removeChild(node));
                    this.savePrefs();
                    this.data.data.ev.changeSamples.unsub(this._dataChangeSamples);
                    this.data.data.constants.ev.set.unsub(this._dataSetConst);
                    this.modelView.models.ev.setChannelCount.unsub(this._modelSetChannelCount);
                    this.modelView.models.ev.setIeqFv.unsub(this._modelSetIeqFv);
                    this.modelView.models.ev.setVRev.unsub(this._modelSetVRev);
                    this.modelView.models.ev.setState.unsub(this._modelSetState);
                    this.modelView.models.ev.setClass.unsub(this._modelSetClass);
                    this.modelView.models.ev.insertRate.unsub(this._modelInsertedRate);
                    this.modelView.models.ev.removeRate.unsub(this._modelRemovedRate);
                    this.modelView.models.ev.setRate.unsub(this._modelSetRate);
                    this.modelView.models.ev.insert.unsub(this._modelInsertModel);
                    this.modelView.models.ev.remove.unsub(this._modelRemoveModel);
                    this.modelView.models.ev.activate.unsub(this._modelActivateModel);
                    this.mac.destroy();
                });
            }
            savePrefs() {
                const prefs = {};
                for (const k of Object.keys(MacModeDefs)) { // to avoid e.g. config.node
                    prefs[k] = this.config[k];
                }
                localStorage.setItem("qubx_mac_prefs", JSON.stringify(prefs));
                qubx_ui.resultDlg.savePrefs();
            }
            _makeMenus(p) {
                qubx_ui.loadTemplate("#qubx_mac_data_ctrls").then(m => {
                    const dotdot = qubx_ui.querydotdot(m);
                    this.btnRun = dotdot.listen("#qubx_mac_run", 'click', this._onClickRun.bind(this));
                    this.llOut = dotdot("#qubx_mac_ll_out");
                    this.btnResults = dotdot.listen("#qubx_mac_showResults", 'click', () => qubx_ui.resultDlg.run());
                    this.resampleBox = dotdot("#qubx_mac_resamplebox");
                    this.slideResample = dotdot.listen("#qubx_mac_resample", 'input', this.prepData.bind(this));
                    this.resampleN = dotdot("#qubx_mac_resample_n");
                    this.resampleNbox = dotdot("#qubx_mac_resampleNbox");
                    
                    this.menus = [this.resampleBox, this.resampleNbox, this.btnRun, this.llOut, this.btnResults];
                }).catch(qubx.reportErr);
            }
            _makeDialogs(p) {
                this.optDlg = new qubx_ui.Modal({
                    title: "Optimize",
                    template: "#qubx_mac_optstart"
                });
                Promise.all([this.optDlg.dom, qubx_ui.loadTemplate(".qmos_model")]).then(htmlItems => {
                    const [dom, modelTemplate] = htmlItems;
                    const dotdot = qubx_ui.querydotdot(dom.content),
                          constBox = dom.content.querySelector("#qmos_constants_box"),
                          constTable = new qubx_ui_data.ConstantsTable({
                              node: dotdot("#qmos_constants"),
                              plot: this.data.plot
                          }),
                          chks = {
                              lm: {host: this.config.minimize, node: dotdot("#qmos_lm")},
                              simplex: {host: this.config.minimize, node: dotdot("#qmos_simplex")},
                              dfp: {host: this.config.minimize, node: dotdot("#qmos_dfp")},
                              optNchannel: {host: this.config.minimize, node: dotdot("#qmos_opt_cc")}
                          };
                    for (const nm of Object.keys(chks)) {
                        (function (nm,host,node) {
                            node.checked = host[nm];
                            node.addEventListener('change', () => {
                                host[nm] = node.checked;
                            });
                        })(nm, chks[nm].host, chks[nm].node);
                    }
                    const divModels = dotdot("#qmos_models");
                    dotdot.sub('#qmos_maxIter', input => new qubx_ui.AcceptedInput(
                        input,
                        this.config.minimize.maxIter,
                        {
                            accept: qubx.acceptIntAtLeast(1),
                            format: qubx.formatStr,
                            widthChars: 5
                        }
                    ), 'change', val => {
                        this.config.minimize.maxIter = val;
                    });
                    dotdot.sub('#qmos_maxStep', input => new qubx_ui.AcceptedInput(
                        input,
                        this.config.minimize.maxStep,
                        {
                            accept: qubx.acceptFloatGreaterThan(0),
                            format: qubx.formatK0,
                            widthChars: 8
                        }
                    ), 'change', val => {
                        this.config.minimize.maxStep = val;
                    });
                    dotdot.sub('#qmos_dx', input => new qubx_ui.AcceptedInput(
                        input,
                        this.config.minimize.dx,
                        {
                            accept: qubx.acceptFloatGreaterThan(0),
                            format: qubx.formatK0,
                            widthChars: 8
                        }
                    ), 'change', val => {
                        this.config.minimize.dx = val;
                    });
                    const advanced = dotdot("#qmos_advanced", adv => {
                        adv.className = 'nodisplay';
                    });
                    const showadv = dotdot.listen("#qmos_showadv", 'click', () => {
                        if ( advanced.classList.contains('nodisplay') ) {
                            advanced.classList.remove('nodisplay');
                            showadv.querySelector("span").textContent = 'Hide';
                        } else {
                            advanced.classList.add('nodisplay');
                            showadv.querySelector("span").textContent = 'Show';
                        }
                    });
                    dotdot.sub('#qmos_workers', input => new qubx_ui.AcceptedInput(
                        input,
                        this.config.workers,
                        {
                            accept: qubx.acceptIntAtLeast(1),
                            format: qubx.formatStr,
                            widthChars: 5
                        }
                    ), 'change', val => {
                        this.config.workers = val;
                        this.mac.count(val);
                    });
                    const coreRun = this.optDlg.run.bind(this.optDlg);
                    this.optDlg.run = () => {
                        const modelStim = this.modelView.getStimulus();
                        constTable.layout(modelStim);
                        constBox.style.display = constTable.node.querySelector('tr') ? '' : 'none';

                        const boundModel = d3.select(divModels)
                            .selectAll(".qmos_modelframe")
                            .data(this.modelView.active, qubx.m('seq'));
                        boundModel.each(d => {
                            d.mac.spanName.textContent = d.model.modelName || 'Untitled';
                            for (const nm of Object.keys(d.mac.chks)) {
                                const c = d.mac.chks[nm];
                                c.node.checked = c.host[nm];
                            }
                            d.mac.chkBal.checked = d.model.balanceLoops();
                            d.mac.constraints.render(d.model.constraints);
                        });
                        const enterModel = boundModel.enter().append("div");
                        enterModel.classed("qmos_modelframe", true);
                        enterModel.each(function(d, i) {
                            d.mac = d.mac || {};
                            const node = modelTemplate.cloneNode(true);
                            d.mac.opt = d.mac.opt || Object.assign({}, MacOptDefs);
                            d.mac.spanName = node.querySelector(".qmos_model_name");
                            d.mac.spanName.textContent = d.model.modelName || 'Untitled';
                            d.mac.chks = {
                                rates: {host: d.mac.opt, node: node.querySelector(".qmos_opt_rates")},
                                cond: {host: d.mac.opt, node: node.querySelector(".qmos_opt_cond")},
                                variance: {host: d.mac.opt, node: node.querySelector(".qmos_opt_var")},
                            };
                            for (const nm of Object.keys(d.mac.chks)) {
                                (function (nm,host,node) {
                                    node.checked = host[nm];
                                    node.addEventListener('change', function() {
                                        const yes = node.checked;
                                        host[nm] = yes;
                                    });
                                })(nm, d.mac.chks[nm].host, d.mac.chks[nm].node);
                            }
                            d.mac.chkBal = node.querySelector(".qmos_balance_loops");
                            d.mac.chkBal.checked = d.model.balanceLoops();
                            d.mac.chkBal.addEventListener('change', () => d.model.balanceLoops(chkBal.checked));
                            d.mac.constraints = new qubx_ui.TableView({
                                node: node.querySelector(".qmos_constraints"),
                                columns: ['Type', 'States', 'delete'],
                                table: d.model.constraints,
                                autohide: true
                            });
                            node.querySelector(".qmos_add_constraint").addEventListener('click', () => d.model.constraints.append({}));
                            d3.select(node).datum(() => null); // prevent d3 data binding propagation to children
                            this.appendChild(node);
                        });
                        boundModel.exit()
                            .each(d => {
                                d.mac.chks = d.mac.chkBal = d.mac.constraints = null;
                            }).remove();
                        return coreRun().then(result => {
                            for (const cname of Object.keys(modelStim)) {
                                this.mac.setConstant(cname, this.data.data.constants.getByName(cname));
                            }
                            return result;
                        });
                    };
                }).catch(qubx.reportErr);
            }
            ensureConst(name) {
                if ( ! this.data ) {
                    setTimeout(() => this.ensureConst(name), 500);
                    return;
                }
                if ( this.waitRun ) {
                    return;
                }
                const sig = this.main.signals.get(name),
                      val = this.data.data.constants.getByName(name);
                if ( this.sentConst[name] !== val ) {
                    this.sentConst[name] = val;
                    this.mac.setConstant(name, val);
                }
            }
            sendModels() {
                this.mac.setModels(this.modelView.models.models.map(m => m.model.active ? m.model.toJson() : ''));
                this.tryLive();
            }
            _onInsertModel(m, i) {
                if ( this.waitRun ) {
                    this.waitModels = true;
                } else {
                    this.sendModels();
                }
            }
            _onRemoveModel(m, i) {
                if ( this.waitRun ) {
                    this.waitModels = true;
                } else {
                    this.sendModelsSoon();
                }
            }
            _onActivateModel(m, i, active) {
                if ( this.waitRun ) {
                    this.waitModels = true;
                } else {
                    this.sendModels(); /// incl. tryLive()
                }
            }
            _onInsertedRate(m, row) {
                row.Ligand && this.ensureConst(row.Ligand);
                row.Voltage && this.ensureConst(row.Voltage);
                row.Pressure && this.ensureConst(row.Pressure);
                if ( row.Index % 2 ) {
                    if ( this.waitRun && this.waitRun[m.index] ) {
                        this.waitRun[m.index].changed();
                    } else {
                        this.mac.setModel(m.index, m.model.toJson());
                    }
                    this.tryLive();
                }
            }
            _onRemovedRate(m, row) {
                if ( ! (row.Index % 2) ) {
                    if ( this.waitRun && this.waitRun[m.index] ) {
                        this.waitRun[m.index].changed();
                    } else {
                        this.mac.setModel(m.index, m.model.toJson());
                    }
                    this.tryLive();
                }
            }
            _onSetRate(m, info) {
                if ( (-1 < ['Ligand', 'Voltage', 'Pressure'].indexOf(info.column)) &&
                     (typeof(this.sentConst[info.value]) === 'undefined') ) {
                    this.ensureConst(info.value);
                }
                if ( this.waitRun && this.waitRun[m.index] ) {
                    this.waitRun[m.index].rateChanged(info);
                } else {
                    this.mac.setRate(m.index, info.row, info.column, info.value);
                }
                this.tryLive();
            }
            _onSetClass(m, info) {
                if ( this.waitRun && this.waitRun[m.index] ) {
                    this.waitRun[m.index].changed();
                } else {
                    this.mac.setModel(m.index, m.model.toJson());
                }
                this.tryLive();
            }
            _onSetState(m, info) {
                if ( this.waitRun && this.waitRun[m.index] ) {
                    this.waitRun[m.index].changed();
                } else {
                    this.mac.setModel(m.index, m.model.toJson());
                }
                this.tryLive();
            }
            _onSetChannelCount(m, cc) {
                if ( this.waitRun && this.waitRun[m.index] ) {
                    this.waitRun[m.index].ccChanged(cc);
                } else {
                    this.mac.setChannelCount(m.index, cc);
                }
                this.tryLive();
            }
            _onSetIeqFv(m, b) {
                this.ensureConst("Voltage");
                if ( this.waitRun && this.waitRun[m.index] ) {
                    this.waitRun[m.index].changed();
                } else {
                    this.mac.setModel(m.index, m.model.toJson());
                    this.tryLive();
                }
            }
            _onSetVRev(m, vrev) {
                if ( this.waitRun && this.waitRun[m.index] ) {
                    this.waitRun[m.index].changed();
                } else {
                    this.mac.setModel(m.index, m.model.toJson());
                    this.tryLive();
                }
            }
            _onSetConst(info) {
                if ( info.column === 'Value' ) {
                    const name = this.data.data.constants.get(info.row, 'Name'),
                          sig = this.main.signals.get(name);
                    if ( ! this.waitRun ) {
                        this.mac.setConstant(name, info.value);
                        this.sentConst[name] = info.value;
                    }
                    this.tryLive();
                }
            }
            _sendDelayed() {
                let anyChange = (this.waitModels || this.waitRun.filter(mwait => {
                    return mwait && (mwait.isChanged || mwait.channelCount || mwait.rates.size);
                }).length);
                for (const name of Object.keys(this.modelView.getStimulus())) {
                    const val = this.data.data.constants.getByName(name);
                    if ( this.sentConst[name] !== val ) {
                        this.mac.setConstant(name, val);
                        this.sentConst[name] = val;
                        anyChange = true;
                    }
                }
                if ( this.waitModels ) {
                    this.sendModels();
                } else if ( this.waitRun.length ) {
                    this.waitRun.forEach((mwait, i) => {
                        if ( ! mwait ) {
                            
                        } else if ( mwait.isChanged ) {
                            this.mac.setModel(i, mwait.model.toJson());
                        } else {
                            if ( mwait.channelCount ) {
                                this.mac.setChannelCount(i, mwait.channelCount);
                            }
                            mwait.changedRates().forEach(info => this.mac.setRate(i, info.row, info.column, info.value));
                        }
                    });
                }
                this.waitRun = null;
                this.waitModels = false;
                if ( anyChange ) {
                    this.tryLive();
                }
            }
            _onChangeSamples() {
                this.prepDataSoon();
            }
            fileOpened() {
            }
            
            prepData() {
                const data = this.data && this.data.data;
                if ( ! (data && data.segmentCount() && data.segments[0].sampled[0] ) ) {
                    return;
                }
                const step1 = this.main.getDeltaForFrac ? qubx_lib : new Promise((resolve, reject) => {
                    let iExtreme = -1;
                    var mostExtreme;
                    data.segments.forEach((seg, i) => {
                        const ex = qubx.extrema(seg.sampled[0]);
                        ex.dy = ex.hi - ex.lo;
                        if ( (! mostExtreme) || (ex.dy > mostExtreme.dy) ) {
                            mostExtreme = ex;
                            iExtreme = i;
                        }
                    });
                    qubx_data_prep.reScan(this.main.dataWorker, data.segments[iExtreme].sampled[0])
                        .then(getDelta => {
                            this.main.getDeltaForFrac = getDelta;
                            this.slideResample.value = getDelta.cpuStrength - 1.0;
                        }).catch(qubx.reportErr)
                            .then(() => resolve(qubx_lib));
                });
                step1.then(lib => {
                    const dt = data.sampling(),
                          delays = {},
                          stims = this.modelView.getStimulus();
                    data.signals.forEach(sig => {
                        const nm = sig.name;
                        delays[nm] = this.main.signals.get(nm).delay;
                        delete stims[nm];
                    });
                    const defaults = {};
                    for (const nm of Object.keys(stims)) {
                        let v = data.constants.getByName(nm);
                        if ( typeof(v) === 'undefined' ) {
                            v = this.main.signals.get(nm).defval;
                            data.constants.setByName(nm, v);
                        }
                        defaults[nm] = v;
                    }
                    this.sentConst = defaults;
                    let Ntot = 0;
                    const segments = data.segments.map((seg, i) => {
                        const stimuli = {};
                        data.signals.slice(1).forEach((sig,j) => {
                            stimuli[sig.name] = seg.ideal[j+1].getDwellsCD(0, seg.n-1, null, null, dt*1e3); // in msec
                            stimuli[sig.name].amps = seg.ideal[j+1].amp;
                        });
                        const segOut = {
                            n: seg.n,
                            chunks: [{
                                n: seg.n,
                                dur: dt*seg.n,
                                t0: 0.0,
                                stimuli: stimuli
                            }]
                        };
                        Ntot += seg.n;
                        return segOut;
                    });
                    const rsRaw = +this.slideResample.value,
                          isResampled = !! rsRaw;
                    if ( isResampled ) {
                        const rsFrac = 1 + rsRaw,
                              delta = this.main.getDeltaForFrac(rsFrac),
                              rs = new lib.AdaptiveResampler(dt);
                        Ntot = 0;
                        data.segments.forEach((seg, i) => {
                            seg.resampled = rs.resample(seg.sampled[0], delta);
                            segments[i].n = seg.resampled.n;
                            segments[i].chunks[0].n = seg.resampled.n;
                            Ntot += seg.resampled.n;
                        });
                        rs.destroy();
                    } else {
                        data.segments.forEach((seg, i) => {
                            seg.resampled = null;
                        });
                    }
                    this.data.plot.drawHiresSoon(); // update resampled display
                    this.resampleN.innerHTML = ''+Ntot;
                    
                    return this.mac.setLayout({
                        sampling: dt,
                        signalNames: data.signals.map(qubx.m('name')),
                        delays: delays,
                        defaults: defaults,
                        resampled: isResampled,
                        segments: segments
                    }, (arrays, iSeg, iChunk) => {
                        const seg = data.segments[iSeg];
                        if ( isResampled ) {
                            arrays[0].set(seg.resampled.means);
                            arrays[1].set(seg.resampled.T);
                        } else {
                            arrays[0].set(seg.sampled[0]);
                        }
                    }).then(result => {
                        this.tryLive();
                        return result;
                    });
                }).catch(qubx.reportErr);
            }
            tryLive(ifAgain) {
                if ( ifAgain && ! this.againLive ) {
                    return;
                }
                if ( this.waitLive ) {
                    this.againLive = true;
                    return;
                }
                this.waitLive = true;
                this.againLive = false;
                const stepOne = this.waitRun ? Promise.resolve({}) : this.mac.getLL();
                stepOne.then(result => {
                    if (result.ll ) {
                        this.llOut.innerHTML = 'LL: '+result.ll.toPrecision(4);
                    }
                }).then(() => this.mac.getOutput((iSeg, iChunk, arr, chunk) => {
                    const seg = this.data.data.segments[iSeg],
                          pointLL = new Float32Array(seg.n);
                    if ( seg.resampled ) {
                        const fit = new Float32Array(2*seg.n),
                              lasts = seg.resampled.lasts,
                              ln = lasts.length;
                        let i = 0,
                            j = 0;
                        for (; i<ln; ++i) {
                            const l = lasts[i],
                                  y = arr[4*i],
                                  s = arr[4*i+1],
                                  pll = arr[4*i+2];
                            while ( j < l ) {
                                fit[2*j] = y;
                                fit[2*j+1] = s;
                                pointLL[j] = pll;
                                ++j;
                            }
                        }
                        this.data.data.setFitStd(iSeg, 0, fit, seg.n, 0, 2, true);
                    } else {
                        let i = 0;
                        for (; i<seg.n; ++i) {
                            pointLL[i] = arr[4*i+2];
                        }
                        this.data.data.setFitStd(iSeg, 0, arr, chunk.n, 0, 4, true);
                    }
                    this.data.data.setPointLL(iSeg, pointLL);
                })).then(() => {
                    this.waitLive = false;
                    this.tryLive(true);
                }).catch(qubx.reportErr)
                    .catch(err => {
                        this.llOut.innerHTML = ''+err;
                        this.waitLive = false;
                        // tryLive? not on err?
                    });
            }
            _beginRun() {
                try {
                    this.btnRun.innerHTML = 'Stop';
                    this.llOut.innerHTML = 'Running...';
                    this.waitRun = this.modelView.models.models.map(rec => {
                        if ( rec.model.active ) {
                            rec.model.fromJson(rec.model.toJson());
                            rec.model.undo.disable();
                            return new WaitModel(rec);
                        } else {
                            return null;
                        }
                    });
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
            _endRun(msg) {
                try {
                    this.btnRun.innerHTML = 'Run';
                    this.llOut.innerHTML = msg;
                    this.waitRun.forEach(mwait => {
                        if ( mwait ) {
                            mwait.model.undo.enable();
                            mwait.model.fromJson(mwait.model.toJson());
                            mwait.model.undo.sealUndo("Optimize");
                        }
                    });
                    this.mac.optPars([], false, this.config.opt); // unconstrained model editing -> no surprises
                    this._sendDelayed();
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
            _onClickRun() {
                try {
                    if ( this.waitRun ) {
                        this.llOut.innerHTML = "Stopping...";
                        this.stopRun = true;
                        return;
                    }
                    this.stopRun = false;
                    const updateModels = models => models.forEach((model, i) => {
                        if ( model ) {
                            model.constraints = this.waitRun[i].model.toJson().constraints;
                            this.waitRun[i].model.fromJson(model);
                        }
                    });
                    if ( ! this.modelView.active.length ) {
                        return;
                    }
                    this._beginRun();
                    this.optDlg.run().then(() => this.mac.optimize(
                        this.modelView.models.models.map((rec, i) => {
                            if ( ! rec.model.active ) {
                                return null;
                            }
                            return {
                                index: i,
                                opt: rec.mac.opt,
                                balance: rec.model.balanceLoops(),
                                constraints: rec.model.constraints.toJson()
                            };
                        }),
                        Object.assign({}, this.config.minimize, {
                            onIteration: (models, params, ll, iteration, grad) => {
                                console.log('mac iter',iteration,ll,params,grad,models);
                                this.llOut.innerHTML = 'LL: '+ll.toPrecision(4)+' (iter '+iteration+')';
                                updateModels(models);
                                return ! this.stopRun;
                            }
                        }))).then(result => {
                            console.log('mac done',result);
                            updateModels(result.models);
                            const used = this.waitRun;
                            this._endRun('LL: '+result.ll.toPrecision(4));
                            
                            this.modelView.context.unselect(); // for clean svg capture /// or is it each modelRec.unselected()?
                            Promise.all(used.map((u,i) => u ? this.modelView.models.models[i].getSVG() : Promise.resolve(false))
                                        .concat([this.resultStyle]))
                                .then(items => {
                                    const css = items[items.length-1],
                                          formatNumbers = qubx.formatList(qubx.formatK0),
                                          fragment = document.createDocumentFragment(),
                                          outer = d3.select(fragment).append("div");
                                    outer.append("p").html('Results of <a href="https://qub.mandelics.com/online/mac.html">QUB Online: Macroscopic</a>, '+new Date());
                                    outer.append("p").text('Finished optimizing after '+result.iteration+' iterations.');
                                    outer.append("p").text('Data points: '+result.ndata+', Parameters: '+result.params.length);
                                    outer.append("p").text('LL: '+qubx.formatK0(result.ll));
                                    outer.append("p").text('AIC: '+qubx.formatK0(result.aic));
                                    outer.append("p").text('BIC: '+qubx.formatK0(result.bic));
                                    outer.append("p").text('Gradient: '+formatNumbers(result.grad));
                                    //// per used model:
                                    result.models.forEach((model, iModel) => {
                                        if ( ! used[iModel] ) {
                                            return;
                                        }
                                        outer.append("p").text('Model: '+(used[iModel].model.modelName || 'Untitled'));
                                        if ( this.config.minimize.optNchannel ) {
                                            outer.append("p").text('Channel count: '+model.channelCount+' +/- '+qubx.formatK0(model.channelCount_std));
                                        } else {
                                            outer.append("p").text('Channel count: '+model.channelCount);
                                        }
                                        const condFields = [];
                                        if ( this.modelView.models.models[iModel].mac.opt.cond ) {
                                            if ( model.IeqFv ) {
                                                condFields.push(qubx_ui.ColumnTable.txtCol('Cond (pS)', 'Cond', qubx.formatK0));
                                            } else {
                                                condFields.push(qubx_ui.ColumnTable.txtCol('Amp (pA)', 'Amp', qubx.formatK0));
                                            }
                                        }
                                        if ( this.modelView.models.models[iModel].mac.opt.variance ) {
                                            if ( model.IeqFv ) {
                                                condFields.push(qubx_ui.ColumnTable.txtCol('Std', 'CondStd', qubx.formatK0));
                                            } else {
                                                condFields.push(qubx_ui.ColumnTable.txtCol('Std', 'Std', qubx.formatK0));
                                            }
                                        }
                                        if ( condFields.length ) {
                                            outer.append("p").text("Conductance classes:");
                                            condFields.splice(0, 0, qubx_ui.ColumnTable.txtCol('Class', 'Index'));
                                            new qubx_ui.ColumnTable({
                                                node: outer.append("table").node(),
                                                columns: condFields
                                            }).render(model.classes);
                                        }
                                        if ( this.modelView.models.models[iModel].mac.opt.rates ) {
                                            const rateFields = [
                                                qubx_ui.ColumnTable.txtCol('From', 'From'),
                                                qubx_ui.ColumnTable.txtCol('To', 'To'),
                                                qubx_ui.ColumnTable.txtCol('k0', 'k0', qubx.formatK0),
                                                qubx_ui.ColumnTable.txtCol('k0_std', 'k0_std', qubx.formatK0)
                                            ];
                                            for (const r of model.rates) {
                                                if ( r.k1 ) {
                                                    rateFields.push(qubx_ui.ColumnTable.txtCol('k1', 'k1', qubx.formatK0));
                                                    rateFields.push(qubx_ui.ColumnTable.txtCol('k1_std', 'k1_std', qubx.formatK0));
                                                    break;
                                                }
                                            }
                                            for (const r of model.rates) {
                                                if ( r.k2 ) {
                                                    rateFields.push(qubx_ui.ColumnTable.txtCol('k2', 'k2', qubx.formatK0));
                                                    rateFields.push(qubx_ui.ColumnTable.txtCol('k2_std', 'k2_std', qubx.formatK0));
                                                    break;
                                                }
                                            }
                                            outer.append("p").text("Rate constants:");
                                            new qubx_ui.ColumnTable({
                                                node: outer.append("table").node(),
                                                columns: rateFields
                                            }).render(model.rates);
                                        }
                                        const svg = items[iModel];
                                        svg.classList.add('resultModel');
                                        outer.node().appendChild(svg);
                                    });
                                    qubx_ui.resultDlg.add('<html><head><style type="text/css">'+css+'</style></head><body>'+outer.node().innerHTML+'</body></html>',
                                                          "Macroscopic");
                                });
                        }).catch(qubx.reportErr)
                            .catch(err => {
                                this._endRun(err);
                                this.tryLive(true);
                            });
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
        };
        
        return exports;
    });

