/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_demo_main", ["qubx", "qubx_app", "qubx_ui", "qubx_ui_model", "qubx_ui_data", "qubx_hist", "qubx_demo_simu", "qubx_demo_mac", "qubx_demo_single", "qubx_demo_dataentry", "qubx_demo_vid", "qubx_data", "qubx_data_axg", "qubx_data_prep", "qubx_data_work", "models", "jsotan", "cloud", "qubx_cloud_ui"],
    (qubx, qubx_app, qubx_ui, qubx_ui_model, qubx_ui_data, qubx_hist, qubx_demo_simu, qubx_demo_mac, qubx_demo_single, qubx_demo_dataentry, qubx_demo_vid, qubx_data, qubx_data_axg, qubx_data_prep, qubx_data_work, models, jsotan, cloud, qubx_cloud_ui) => {
        "use strict";
        const exports = {};

        class MacSignals {
            constructor() {
                this.storage = {};
            }
            get(name) {
                if ( name in this.storage ) {
                    const sig = this.storage[name];
                    sig.defval = sig.defval || 0.0;
                    sig.delay = sig.delay || 0.0;
                    return sig;
                } else {
                    const sig = {
                        name: name,
                        defval: 0.0,
                        delta: 0.05,
                        delay: 0.0,
                        rename: name,
                        isLigand: false,
                        isVoltage: (name[0] in ['V', 'v']),
                        isPressure: (name[0] in ['P', 'p'])
                    };
                    this.storage[name] = sig;
                    return sig;
                }
            }
            savePrefs() {
                localStorage.setItem("qubx_mac_signals", JSON.stringify(this.storage));
            }
            readPrefs() {
                var saved = localStorage.getItem("qubx_mac_signals");
                if ( saved ) {
                    this.storage = JSON.parse(saved);
                }
            }
        }

        const DemoMainDefs = {
            fullscreen: true,
            dataWorkers: 3
        };
        exports.start = function start(config) {
            try {
                var main;
                let storedPrefs = localStorage.getItem("qubx_demo_prefs");
                storedPrefs = storedPrefs ? JSON.parse(storedPrefs) : {};
                config = Object.assign({}, DemoMainDefs, config, storedPrefs, {
                    readHash: () => {
                        const h = qubx.unhash(document.location.hash);
                        main.panels.then(panels => {
                            // now gui is setup, and main.modelLink is ready
                            let oldNmodel = panels.modelView.models.models.length;
                            function closeOld() {
                                while ( oldNmodel ) {
                                    --oldNmodel;
                                    panels.modelView.models.models[oldNmodel].close();
                                }
                            };
                            if ( h.model ) {
                                models.getModelQmj(h.model|0)
                                    .then(qmj => models.getModelInfo(h.model|0)
                                          .then(info => {
                                              info.modelQmj = qmj;
                                              main.loadDBModel(info);
                                          })).then(closeOld);
                            }
                            if ( h.link ) {
                                models.loadLink(h.link, h.k)
                                    .then(main.loadDBModel)
                                    .then(closeOld);
                            }
                            h.macqdf && main.loadMacQDF(h.macqdf);
                            h.welcome || main.showWelcome();
                        }).catch(qubx.reportErr);
                    }
                });
                main = new qubx_app.Main(config);
                const _dispose = main.dispose.bind(main);
                
                Object.assign(main, {
                    dataWorker: new qubx_data_work.DataWorkgroup({count: main.config.dataWorkers}),
                    signals: new MacSignals(),
                    getDeltaForFrac: null,
                    saver: new qubx_ui.Saver(),
                    dispose: function() {
                        main.mode(null);
                        main.savePrefs();
                        main.panels.then(panels => {
                            panels.data.dispose();
                            panels.modelView.dispose();
                            _dispose();
                        }).catch(qubx.reportErr);
                        main.dataWorker.destroy();
                    },
                    savePrefs: function() {
                        main.signals.savePrefs();
                        main.simMode.savePrefs();
                        main.macMode.savePrefs();
                        main.singleMode.savePrefs();
                        localStorage.setItem("qubx_demo_model", JSON.stringify(main.modelView.toJson()));
                        const prefs = {};
                        for (const k of Object.keys(DemoMainDefs)) { // to avoid e.g. config.node
                            prefs[k] = main.config[k];
                        }
                        prefs.topbot = {skewType: main.topbot.config.skewType};
                        prefs.toplr = {skewType: main.toplr.config.skewType};
                        localStorage.setItem("qubx_demo_prefs", JSON.stringify(prefs));
                        qubx_ui.SaveInputs();
                    },
                    _makeMenus: function(p) {
                        main.opener = new qubx_ui.Opener();
                        main.dataconv = new qubx_ui_data.Converter();
                        main.vidMaker = new qubx_demo_vid.VidMaker({
                            modelView: main.modelView,
                            dataView: main.data,
                            saver: main.saver,
                            progress: main.progressDlg
                        });
                        p.mnuQUB = new qubx_ui.Menu({
                            node: p.controls.querySelector("#qubx_mnu_qub"),
                            caption: 'QUB...',
                            groups: [[
                                {
                                    caption: 'About...',
                                    action: main.showAbout
                                }
                            ], [
                                {
                                    caption: 'Simulator',
                                    action: () => {
                                        main.mode(main.simMode);
                                        main.constTable.node.style.display = 'none';
                                    }
                                },
                                {
                                    caption: 'Macroscopic Solver',
                                    action: () => {
                                        main.mode(main.macMode);
                                        main.constTable.node.style.display = (main.mode() === main.simMode) ? 'none' : '';
                                    }
                                },
                                {
                                    caption: 'Single Molecule',
                                    action: () => {
                                        main.mode(main.singleMode);
                                        main.constTable.node.style.display = (main.mode() === main.simMode) ? 'none' : '';
                                    }
                                },
                                {
                                    caption: 'Data Entry',
                                    action: () => {
                                        main.mode(main.deMode);
                                        main.constTable.node.style.display = (main.mode() === main.simMode) ? 'none' : '';
                                    }
                                }
                            ], [
                                {
                                    caption: 'Not signed in',
                                    action: () => qubx_cloud_ui.editMember(),
                                    needPerson: true,
                                    recaptionName: true
                                },
                                {
                                    caption: 'Request account...',
                                    action: () => qubx_cloud_ui.requestAccount(),
                                    noPerson: true
                                },
                                {
                                    caption: 'Administer...',
                                    action: () => qubx_cloud_ui.administer(),
                                    admin: true
                                },
                                {
                                    caption: 'Sign out',
                                    action: () => qubx_ui.confirm({
                                        hdr: 'Sign out',
                                        q: "Sign out now?"
                                    }).then(() => cloud.signOut()),
                                    needPerson: true
                                }
                            ]],
                            onfocus: function(d, i) {
                                if ( d.admin ) {
                                    d3.select(p.mnuQUB.config.node.options[i])
                                        .style({display: cloud.personIsAdmin ? '' : 'none'});
                                } else if ( d.needPerson ) {
                                    d3.select(p.mnuQUB.config.node.options[i])
                                        .style({display: cloud.personName ? '' : 'none'});
                                } else if ( d.noPerson ) {
                                    d3.select(p.mnuQUB.config.node.options[i])
                                        .style({display: cloud.personName ? 'none' : ''});
                                }
                                if ( d.recaptionName ) {
                                    d3.select(p.mnuQUB.config.node.options[i])
                                        .text(cloud.personName ? ('Signed in as '+cloud.personName+'...') : 'Not signed in');
                                }
                            }
                        });

                        p.mnuProject = new qubx_ui.Menu({
                            node: p.controls.querySelector("#qubx_mnu_project"),
                            caption: 'Project',
                            groups: [[
                            ]],
                            onfocus: (d, i) => {
                                if ( d.admin ) {
                                    d3.select(p.mnuProject.config.node.options[i])
                                        .style({display: cloud.personIsAdmin ? '' : 'none'});
                                } else if ( d.needPerson ) {
                                    d3.select(p.mnuProject.config.node.options[i])
                                        .style({display: cloud.personName ? '' : 'none'});
                                }
                            }
                        });
                        p.mnuProject.config.node.style.display = 'none';
                        
                        p.mnuLoad = new qubx_ui.Menu({
                            node: p.controls.querySelector("#qubx_mnu_file"),
                            caption: 'File',
                            groups: [[
                                {
                                    caption: 'Load data file...',
                                    action: () => {
                                        const wasSim = (main.mode() === main.simMode);
                                        if ( wasSim ) {
                                            main.mode(main.deMode);
                                        }
                                        return main.opener.run(undefined, p.dataFormatInfo, true)
                                            .then(main.openFiles)
                                            .then(() => main.opener.reset())
                                            .catch(err => {
                                                main.opener.reset();
                                                wasSim && main.mode(main.simMode);
                                                return Promise.reject(err);
                                            }).catch(qubx.reportErr);
                                    }
                                },
                                {
                                    caption: 'Save data as...',
                                    action: () => main.saver.prompt({
                                        info: "Time, Current[, and stimulus] columns separated by tabs (.txt) or commas (.csv), with a blank line between segments.",
                                        name: 'Data',
                                        ext: '.txt',
                                        exts: ['.txt', '.csv']
                                    }).then(fi => main.saver.saveText(fi.fullname, p.data.data.getTXT({
                                        sep: (fi.ext === '.csv') ? ',' : '\t'
                                    }), "application/octet-stream;charset=UTF-8"))
                                },
                                {
                                    caption: 'Save idealized traces as...',
                                    action: () => main.saver.prompt({
                                        info: "Time, Current[, and stimulus] columns separated by tabs (.txt) or commas (.csv), with a blank line between segments.",
                                        name: 'Idealized',
                                        ext: '.txt',
                                        exts: ['.txt', '.csv']
                                    }).then(fi => main.saver.saveText(fi.fullname, p.data.data.getTXT({
                                        sep: (fi.ext === '.csv') ? ',' : '\t',
                                        idealTraces: true
                                    }), "application/octet-stream;charset=UTF-8"))
                                },
                                {
                                    caption: 'Save idealized events (DWT) as...',
                                    action: () => main.saver.prompt({
                                        info: "<b>DWT</b> is a text-based format with columns for event class and duration (ms)",
                                        name: 'Events',
                                        ext: '.dwt',
                                        exts: ['.dwt']
                                    }).then(fi => main.saver.saveText(fi.fullname, p.data.data.getDWT(), "application/octet-stream;charset=UTF-8"))
                                },
                                {
                                    caption: 'Save event statistics...',
                                    action: () => main.saver.prompt({
                                        info: "text-based <b>DWT</b> file with additional columns: each event's mean and std. dev.",
                                        name: 'Events',
                                        ext: '.txt',
                                        exts: ['.txt', '.dwt']
                                    }).then(fi => main.saver.saveText(fi.fullname, p.data.data.getDWT(true), "application/octet-stream;charset=UTF-8"))
                                },
                                {
                                    caption: 'Save data image as PNG...',
                                    action: () => {
                                        // canvas.toBlob() not standard yet
                                        const url = p.data.plot.getComposite().toDataURL("image/png"),
                                              data = atob(url.replace(/^.*?base64,/,'')),
                                              asArray = new Uint8Array(data.length),
                                              n = data.length;
                                        let i = 0;
                                        for (; i<n; ++i) {
                                            asArray[i] = data.charCodeAt(i);
                                        }
                                        main.saver.prompt({name: 'Data.png'})
                                            .then(fi => main.saver.save(fi.fullname, new Blob([asArray.buffer], {type: 'image/png'})));
                                    }
                                },
                                {
                                    caption: 'Save video of data...',
                                    action: () => main.vidMaker.run()
                                }
                            ], [
                                {
                                    caption: 'New model',
                                    action: () => p.modelView.models.insert()
                                },
                                {
                                    caption: 'Load model file...',
                                    action: () => main.opener.run()
                                        .then(files => p.modelView.openFiles(files))
                                        .then(() => main.opener.reset())
                                        .catch(err => {
                                            main.opener.reset();
                                            return Promise.reject(err);
                                        }).catch(qubx.reportErr)
                                },
                                {
                                    caption: 'Save model as QMJ...',
                                    action: () => main.saver.prompt({
                                        info: "<b>QMJ</b> model files are plain text, and can also be read by QUB Express.",
                                        name: p.modelView.model.modelName || 'Model',
                                        ext: '.qmj',
                                        exts: ['.qmj']
                                    }).then(fi => main.saver.saveJSON(fi.fullname, (p.modelView.model || p.modelView.models.models[0].model).toJson()))
                                        .catch(qubx.reportErr)
                                },
                                {
                                    caption: 'Save model image as SVG...',
                                    action: () => p.modelView.getSVG()
                                        .then(svg => main.saver.prompt({
                                            name: ((main.modelView.model || main.modelView.models.models[0].model).modelName || 'Model')+'.svg'
                                        }).then(fi => main.saver.saveText(fi.fullname, svg.outerHTML, "image/svg+xml")))
                                        .catch(qubx.reportErr)
                                }
                            ]]
                        });
                    },
                    _makeDialogs: function(p) {
                        main.aboutDlg = new qubx_ui.Modal({
                            title: 'QUB Online - About',
                            template: '#qubx_demo_about',
                            canCancel: false,
                            okLabel: 'Close'
                        });
                        
                        main.filterDlg = new qubx_ui.Modal({
                            title: 'Filter options',
                            template: "#qubx_filter_options",
                            canCancel: false,
                            okLabel: 'Close'
                        });
                        main.filterDlg.dom.then(dom => {
                            main.filterDlg.chkFilter = dom.content.querySelector('#qfo_filter');
                            main.filterDlg.chkFilter.addEventListener('change', main._onCheckFilter);
                            main.filterDlg.detail = dom.content.querySelector('#qfo_detail');
                            main.filterDlg.txtFreq = new qubx_ui.AcceptedInput(
                                main.filterDlg.detail.querySelector("#qfo_freq"),
                                1.0,
                                {
                                    accept: qubx.acceptFloatGreaterThan(0.0),
                                    format: qubx.formatK0,
                                    widthChars: 6
                                }
                            );
                            main._onCheckFilter();
                        }).catch(qubx.reportErr);
                    },
                    showAbout: function() {
                        return main.aboutDlg.run();
                    },
                    
                    _setupModel: function(p) {
                        main.updateModelStimSoon = qubx.batchDelayed(main.updateModelStim, 50); // mainly useful to strip the arguments from loosely related triggering callback
                        const modelNode = document.createElement('div');
                        modelNode.className = 'qubx_fullsize';
                        main.topbot.second.appendChild(modelNode);
                        p.modelView = new qubx_ui_model.ModelView({
                            node: modelNode,
                            listStimuli: () => {
                                const names = main.data.data.signals.slice(1).map(qubx.m('name'));
                                return names.length ? names : ['Ligand'];
                            },
                            grabAmp: () => {
                                const iSeg = Math.max(0, main.data.plot.chosen()),
                                      sel = main.data.plot.hilite();
                                if ( ! (sel.from || sel.to) ) {
                                    return Promise.reject('No data selected');
                                }
                                return Promise.resolve(qubx.meanstd(main.data.data.segments[iSeg].sampled[0], sel.offset, sel.count));
                            },
                            grabAmps: model => {
                                const iSeg = Math.max(0, main.data.plot.chosen()),
                                      sel = main.data.plot.hilite();
                                if ( ! (sel.from || sel.to) ) {
                                    return Promise.reject('No data selected');
                                }
                                const data = new Float32Array(sel.count),
                                      src = main.data.data.segments[iSeg].sampled[0];
                                let i = 0;
                                for (; i<sel.count; ++i) {
                                    data[i] = src[sel.offset+i];
                                }
                                const extrema = qubx.extrema(data),
                                      modelJ = model.toJson();
                                modelJ.channelCount = 1;
                                return main.dataWorker.AMP(data, main.data.data.sampling(), modelJ,
                                                           true, Math.abs(extrema.lo) > Math.abs(extrema.hi));
                            }
                        });
                        main.modelView = p.modelView;
                        p.modelView.models.ev.setIeqFv.sub(main._onSetIeqFv);
                        p.modelView.models.ev.insertRate.sub(main._onInsertedRate);
                        p.modelView.models.ev.setRate.sub(main._onSetRate);
                        
                        var mj = localStorage.getItem("qubx_demo_model");
                        if ( mj ) {
                            p.modelView.fromJson(JSON.parse(mj));
                        }
                        if ( p.modelView.models.models.length === 0 ) {
                            p.modelView.models.insert();
                            p.modelView.model.states.append({x: 25, y: 50});
                            p.modelView.model.states.append({Class: 1, x: 75, y: 50});
                            p.modelView.model.rates.append({From: 0, To: 1, k0: 10});
                            p.modelView.model.rates.append({From: 1, To: 0, k0: 100});
                            p.modelView.model.undo.clear();
                        }
                        p.modelView.context.select(p.modelView.models.models[0]);
                    },

                    loadDBModel: function(m) {
                        return main.panels.then(p => {
                            if ( main.mode() === main.simMode ) {
                                main.simMode.sim.paused = true;
                            }
                            p.modelView.models.insert();
                            p.modelView.model.fromJson(m.modelQmj);
                            p.modelView.model.undo.sealUndo("Load model");
                            p.modelView.model.modelName = m.modelName;
                            p.modelView.model.modelAbout = m.modelAbout;
                            p.modelView.model.modelId = m.modelId;
                            p.modelView.model.changed = false;
                            if ( main.mode() === main.simMode ) {
                                main.simMode.sim.paused = false;
                                main.simMode.sim.update();
                            }
                        });
                    },
                    
                    _setupData: function(p, components) {
                        const ctrls = components.querySelector("#qubx_demo_data_ctrls");
                        p.controls.appendChild(ctrls);
                        main.btnFilter = d3.select(p.controls)
                            .append('button')
                            .text('Filter... (off)')
                            .on('click', main.promptFilter)
                            .node();
                        p.data = main.data = new qubx_ui_data.PlotFrame();
                        p.data.plot.workers = main.dataWorker;
                        p.data.node.classList.add('qubx_demo_data');
                        p.data.node.classList.add('qubx_fullsize');
                        p.data.node.addEventListener('dragenter', qubx_ui.dragEnter(p.data.node));
                        p.data.node.addEventListener('dragleave', qubx_ui.dragLeave(p.data.node));
                        p.data.node.addEventListener('dragover', qubx.stopAndPrevent);
                        p.data.node.addEventListener('drop', main._onDrop);
                        main.toplr.first.appendChild(p.data.node);
                        main.lastData = null;
                        p.data.plot.ev.changeData.sub(d => {
                            if ( main.lastDataFile ) {
                                main.lastDataFile.ev.changeTDead.unsub(main._onChangeTDead);
                            }
                            main.lastDataFile = d;
                            if ( d ) {
                                d.ev.changeTDead.sub(main._onChangeTDead);
                                main._onChangeTDead(d.tDead());
                            }
                        });
                        p.data.plot.ev.changeFilter.sub(main._onChangeFilter);
                        main.divTdead = document.createElement('div');
                        main.divTdead.className = 'qubx_opt_line';
                        d3.select(main.divTdead).append('span')
                            .html('t<sub>dead</sub> (ms):')
                            .classed('qubx_opt_name', true);
                        var span = d3.select(main.divTdead).append('span');
                        span.classed('qubx_opt_data', true);
                        var txt = span.append('input').property({
                            type: 'number',
                            min: 0
                        });
                        main.txtTdead = new qubx_ui.AcceptedInput(txt.node(),
                                                                  0.0,
                                                                  {
                                                                      accept: qubx.acceptFloatAtLeast(0.0),
                                                                      format: qubx.formatK0,
                                                                      widthChars: 6
                                                                  });
                        main.txtTdead.ev.change.sub(main._onEnterTDead);
                        
                        main.constTable = new qubx_ui_data.ConstantsTable({
                            plot: p.data.plot
                        });
                        
                        setTimeout(() => {
                            main.modelView.context.node.appendChild(main.divTdead);
                            main.modelView.context.node.appendChild(main.constTable.node);
                        }, 500);
                        
                        p.data.setData(new qubx_data.Data());
                        main.constTable.layout(main.modelView.getStimulus());
                        p.data.data.constants.ev.inserted.sub(main.updateModelStimSoon);
                        p.data.data.constants.ev.set.sub(main.updateModelStimSoon);
                    },
                    
                    promptFilter: function() {
                        var freq = main.data.plot.filterKHz();
                        main.filterDlg.chkFilter.checked = !! freq;
                        if ( freq ) {
                            main.filterDlg.txtFreq.value = freq;
                        }
                        main._onCheckFilter();
                        return main.filterDlg.run()
                            .then(() => main.toggleFilter(main.filterDlg.chkFilter.checked))
                            .catch(qubx.reportErr);
                    },
                    toggleFilter: function(state) {
                        if ( typeof(state) === 'undefined' ) {
                            state = ! main.data.plot.filterKHz();
                        }
                        state = !! state;
                        main.data.plot.filterKHz(state ? main.filterDlg.txtFreq.value : 0);
                    },
                    _onChangeFilter: function(freq) {
                        if ( freq ) {
                            main.filterDlg.chkFilter.checked = true;
                            main.filterDlg.txtFreq.value = freq;
                        } else {
                            main.filterDlg.chkFilter.checked = false;
                        }
                        main.btnFilter.textContent = freq ? ('Filter: '+freq.toPrecision(2)+' kHz') : 'Filter... (off)';
                        main._onCheckFilter();
                    },
                    _onCheckFilter: function() {
                        var filter = main.filterDlg.chkFilter.checked;
                        main.filterDlg.detail.style.display = filter ? '' : 'none';
                    },
                    
                    _onChangeTDead: function(td) {
                        main.txtTdead.update(td, .5*1e3*main.data.data.sampling());
                        main.modelView.tDead(td);
                    },
                    _onEnterTDead: function(td) {
                        main.data.plot.data().tDead(td);
                    },
                    
                    _setupVis: function() {
                        main.cursorInfo = new qubx_ui_data.CursorInfo({
                            plot: main.data.plot
                        });
                        main.cursorInfo.tabTitle = 'Data info';
                        main.cursorInfo.tabIcon = '\u2316';
                        //main.cursorInfo.tabOffscreen = true;
                        main.vis.add(main.cursorInfo);
                        main.vis.pick(main.cursorInfo);
                        
                        main.ampHist = new qubx_hist.AmpHist({
                            dataView: main.data.plot,
                            modelView: main.modelView
                        });
                        main.ampHist.tabTitle = 'Amplitude histogram';
                        main.ampHist.tabIcon = '<img src="images/appbar.graph.bar.svg"><i style="font-size:30%">A</i>';
                        main.ampHist.tabOffscreen = true;
                        main.vis.add(main.ampHist);
                        
                        main.durHist = new qubx_hist.DurHists({
                            dataView: main.data.plot,
                            modelView: main.modelView
                        });
                        main.durHist.tabTitle = 'Duration histogram';
                        main.durHist.tabIcon = '<img src="images/appbar.graph.bar.svg"><i style="font-size:30%">D</i>'
                        main.durHist.tabOffscreen = true;
                        main.vis.add(main.durHist)

                        main.energyLandscape = new qubx_ui_model.EnergyLandscapeVis({
                            modelView: main.modelView
                        });
                        main.energyLandscape.tabTitle = 'Energy landscape';
                        main.energyLandscape.tabIcon = '<img src="images/landscape.svg">';
                        main.energyLandscape.tabOffscreen = true;
                        main.vis.add(main.energyLandscape);
                    },
                    
                    _setupModes: function() {
                        main.simMode = new qubx_demo_simu.SimMode(main);
                        main.macMode = new qubx_demo_mac.MacMode(main);
                        main.singleMode = new qubx_demo_single.SingleMode(main);
                        main.deMode = new qubx_demo_dataentry.DataEntryMode(main);
                        main.mode(main.simMode);
                        main.constTable.node.style.display = (main.mode() === main.simMode) ? 'none' : '';
                    },
                    
                    _readData: function(rdr, index, name, delta, idealized) {
                        main.resetModelStim();
                        const data = main.data.data;
                        data.constants.clear();
                        for (const cname of Object.keys(rdr.constants)) {
                            const cvalue = rdr.constants[cname];
                            if ( cname === 'Tdead (ms)' ) {
                                data.tDead(cvalue);
                                main.modelView.tDead(cvalue);
                            } else if ( cname === 'Tcrit (ms)' ) {
                                data.tCrit(cvalue);
                            } else {
                                data.constants.append({
                                    Name: cname,
                                    Value: cvalue
                                });
                            }
                        }
                        data.segmentCount(0);
                        data.clear();
                        data.sampling(rdr.sampling);
                        data.signalCount(index.length, (sig, iSig) => {
                            sig.name = name[iSig]; //?
                        });
                        //index.forEach(function(iSrc, i) {
                        //    data.signals[i].name = name[i]; // = rdr.signalNames[iSrc];
                        //});
                        data.segmentCount(rdr.counts.length);
                        rdr.counts.forEach((count, iSeg) => data.segmentSize(iSeg, count));
                        data.doneSegmentSize();
                        
                        idealized.forEach((idlSegs, iSig) => idlSegs.forEach((events, iSeg) => {
                            data.setIdeal(iSeg, iSig, events.counts, events.classes, 0, events.amps, events.stds);
                                //events.durations = new Float32Array(events.counts.length);
                                //for (var i=0, n=events.counts.length; i<n; ++i) {
                                //    events.durations[i] = events.counts[i] * rdr.sampling*1e3;
                                //}
                        }));
                        rdr.readSegments((rdr, seg, iSeg) => index.forEach((iSrc, iSig) => {
                            data.setSampled(iSeg, iSig, 0, seg.buffers[iSrc], rdr.counts[iSeg]);
                                /// what if there's no buffer (e.g. DWT, only idealized)?
                                /*
                                  var idlSegs = idealized[iSig]; // TODO make rdr interface consistent iSig or iSrc
                                  if ( idlSegs && idlSegs.length ) {
                                  var events = idlSegs[iSeg];
                                  data.setIdeal(iSeg, iSig, events.counts, events.classes, 0, events.amps, events.stds);
                                  events.durations = new Float32Array(events.counts.length);
                                  for (var i=0, n=events.counts.length; i<n; ++i) {
                                  events.durations[i] = events.counts[i] * rdr.sampling*1e3;
                                  }
                                  return events;
                                  }
                                */
                        }));
                        
                        main.txtTdead.update(data.tDead(), .5*1e3*data.sampling());
                        main.getDeltaForFrac = null;
                        main.mode().fileOpened && main.mode().fileOpened();
                    },
                    
                    progressDlg: new qubx_ui.ModalProgress(),
                    
                    _onDrop: function(event) {
                        event.stopPropagation();
                        event.preventDefault();
                        var files = Array.prototype.map.call(event.dataTransfer.files, file => ({
                            name: file.name,
                            get: function() { return Promise.resolve(file); }
                        }));
                        if ( main.mode() === main.simMode ) {
                            main.mode(main.deMode);
                        }
                        main.openFiles(files);
                        main.data.node.classList.remove('qubx_dragtarget');
                    },
                    openFiles: function(files) {
                        //// handle multi-file esp. *+qsf, pul+dat
                        if ( files.length ) {
                            const lowName = files[0].name.toLowerCase();
                            if ( lowName.endsWith('.qmf') || lowName.endsWith('.qmj') ) {
                                main.modelView.openFiles([files[0]]);
                            } else {
                                main.dataconv.receiveFile(files[0])
                                    .then(f => main.openData(f, f.name));
                            }
                        }
                    },
                    openData: function(f, name, assignSignals) {
                        name = name || '';
                        var rdrType;
                        const reader = new FileReader();
                        reader.onload = e => {
                            const rdr = new rdrType(e.target.result);
                            while ( isNaN(1/rdr.sampling) ) {
                                rdr.sampling = rdr.sampling || 1e-3/(+prompt("Sampling rate [kHz]:", "10.0"));
                            }
                            (assignSignals ? assignSignals(rdr, main.signals)
                             : qubx_data_prep.openDlg.runSignals(main.dataWorker, name, rdr, main.signals))
                                .then(info => {
                                    main.progressDlg.show('Preparing data...');
                                    return qubx_data_prep.prepData(main.dataWorker, rdr, info.depVar, main.signals, prog => {
                                        main.progressDlg.progress(prog);
                                        return ! main.progressDlg.canceled;
                                    });
                                }).then(info => {
                                    main.progressDlg.hide();
                                    main._readData(rdr, info.index, info.name, info.delta, info.idealized);
                                }).catch(qubx.reportErr)
                                .catch(err => main.progressDlg.hide());
                        };
                        let ext = '',
                            iDot = name.lastIndexOf('.');
                        if ( iDot > 0 ) {
                            ext = name.substring(iDot).toLowerCase();
                        }
                        if ( ext === '.qdf' ) { // really ought to open/multi with .qsf if available
                            rdrType = qubx_data.QDFReader;
                            reader.readAsArrayBuffer(f);
                        } else if ( ext === '.dwt' ) {
                            rdrType = qubx_data.DWTReader;
                            reader.readAsText(f);
                        } else if ( ext === '.atf' ) {
                            rdrType = qubx_data.ATFReader;
                            reader.readAsText(f);
                        } else if ( (ext === '.axgx') || (ext === '.axgr') ) {
                            rdrType = qubx_data_axg.AXGReader;
                            reader.readAsArrayBuffer(f);
                        } else/*if ( ext === '.txt' )*/{
                            rdrType = qubx_data.TextReader;
                            reader.readAsText(f);
                        }
                    },
                    
                    loadMacQDF: function(name) {
                        main.mode(main.macMode);
                        main.constTable.node.style.display = '';
                        name = name.replace(/\//g, '');
                        main.progressDlg.show('Downloading...');
                        return qubx.get('macqdf/'+name+'.qdf', 'blob', progress => {
                            main.progressDlg.progress(progress);
                            return main.progressDlg.canceled;
                        }).then(qdfBlob => {
                            main.progressDlg.hide();
                            main.openData(qdfBlob, name+'.qdf', (rdr, signalDB) => {
                                // signals.get(name): delta, delay, rename, isLigand, isVoltage, isPressure
                                let depVar = 0;
                                const signals = rdr.signalNames.map(signalDB.get.bind(signalDB));
                                signals.forEach((sig, i) => {
                                    sig.isLigand = sig.isVoltage = sig.isPressure = false;
                                    sig.delay = 0.0;
                                    const firstChar = sig.name[0].toLowerCase();
                                    if ( (firstChar === 'i') || (firstChar === 'c') ) {
                                        depVar = i;
                                    } else if ( sig.name === 'pro' ) {
                                        sig.isLigand = true;
                                        sig.rename = sig.name;
                                        sig.delta = 0.005;
                                    } else if ( firstChar === 'p' ) {
                                        sig.isPressure = true;
                                        sig.delta = 0.1;
                                    } else if ( firstChar === 'v' ) {
                                        sig.isVoltage = true;
                                        sig.delta = 1.0;
                                    }
                                });
                                return Promise.resolve({
                                    depVar: depVar
                                });
                            });
                        }).catch(qubx.reportErr)
                            .catch(() => main.progressDlg.hide());
                    },
                    
                    ensureConst: function(name) {
                        if ( name ) {
                            main.signals.get(name);
                        }
                        main.modelView && main.constTable && main.constTable.layout(main.modelView.getStimulus());
                    },
                    _onInsertedRate: function(m, row) {
                        row.Ligand && main.ensureConst(row.Ligand);
                        row.Voltage && main.ensureConst(row.Voltage);
                        row.Pressure && main.ensureConst(row.Pressure);
                    },
                    _onSetRate: function(m, info) {
                        if ( -1 < ['Ligand', 'Voltage', 'Pressure'].indexOf(info.column) ) {
                            main.ensureConst(info.value);
                        }
                    },
                    _onSetIeqFv: function(m, b) {
                        b && main.ensureConst("Voltage");
                    },
                    
                    resetModelStim: function() {
                        main._stimSelOffset = 0;
                        main._stimSelCount = 0;
                        main._stimSelSegs = [];
                    },
                    updateModelStim: function(offset, count, segments) {
                        try {
                            if ( typeof(offset) === 'undefined' ) {
                                offset = main._stimSelOffset|0;
                            }
                            if ( typeof(count) === 'undefined' ) {
                                count = main._stimSelCount|0;
                            }
                            if ( typeof(segments) === 'undefined' ) {
                                segments = main._stimSelSegs||[];
                            }
                            main._stimSelOffset = offset;
                            main._stimSelCount = count;
                            main._stimSelSegs = segments;
                            
                            var stim = {};
                            if ( segments.length ) {
                                var hiSeg = segments[segments.length-1];
                                var seg = main.data.data.segments[hiSeg.index];
                                var mid = hiSeg.from + ((hiSeg.count/2)|0);
                                main.data.data.signals.forEach((sig, iSig) => {
                                    if ( iSig ) {
                                        const idl = seg.ideal[iSig];
                                        let cls = idl.classAt(mid);
                                        if ( cls < 0 ) {
                                            cls = idl.classAt(hiSeg.from);
                                        }
                                        if ( cls < 0 ) {
                                            cls = idl.classAt(hiSeg.to);
                                        }
                                        if ( cls >= 0 ) {
                                            stim[sig.name] = idl.amp[cls];
                                        } else if ( seg.sampled[iSig] ) {
                                            stim[sig.name] =  seg.sampled[iSig][mid];
                                        }
                                    }
                                });
                            }
                            main.modelView.setStimulus(stim, name => main.data.data.constants.getByName(name) || 0.0);
                            main.energyLandscape.showing && main.energyLandscape.renderSoon();
                        }
                        catch (err) {
                            qubx.reportErr(err);
                        }
                    },
                    checkUpdates: function() {
                        try {
                            navigator.serviceWorker.getRegistration('/common/mandelic.ico')
                                .then(reg => {
                                    reg && reg.update();
                                    setTimeout(main.checkVersion, main.version && 60000 || 0);
                                });
                        } catch (err) {
                            main.checkVersion();
                        }
                    },
                    checkVersion: function() {
                        qubx.get('version').then(resp => {
                            if ( ! main.version ) {
                                main.version = resp;
                                main.aboutDlg.dom.then(dom => {
                                    dom.content.querySelector("#qubx_demo_buildno").textContent = main.version;
                                });
                            } else if ( (! main.versionNotified) && (main.version !== resp) ) {
                                main.versionNotified = true;
                                qubx_ui.confirm({
                                    hdr: 'QUB - Update available',
                                    q: '<p>An update is available for this page. Reload and start using the update?</p><p>If you choose Cancel, you can continue using QUB as-is until next time.</p>'
                                }).then(() => window.location.reload());
                                navigator.serviceWorker.controller.postMessage('update');
                            }
                        });
                    }
                });
                
                setInterval(main.checkUpdates, 30*60000);
                main.checkUpdates();
                
                main.resetModelStim();
                
                qubx_ui.loadTemplate("#qubx_demo_components").then(components => {
                    main.panels.then(p => {
                        window.addEventListener("pagehide", main.savePrefs.bind(main));
                        
                        main.topbot = new qubx_ui.TwoPane({
                            node: p.main.querySelector("#qubx_topbot"),
                            vertical: true,
                            skewType: qubx.fieldOrDef(main.config.topbot, 'skewType', 2),
                            firstLabel: 'Show Data',
                            secondLabel: 'Show Model'
                        });
                        const toplr = document.createElement('div');
                        toplr.classList.add('qubx_fullsize');
                        main.topbot.first.appendChild(toplr);
                        main.toplr = new qubx_ui.TwoPane({
                            node: toplr,
                            skewType: qubx.fieldOrDef(main.config.toplr, 'skewType', 3),
                            firstLabel: 'Show Data',
                            secondLabel: 'More...'
                        });
                        
                        main.vis = new qubx_ui.Tabs({position: 'top'});
                        main.vis.node.classList.add('qubx_fullsize');
                        main.toplr.second.appendChild(main.vis.node);

                        p.dataFormatInfo = components.querySelector('#qubx_dataformatinfo');
                        
                        main._setupModel(p);
                        main._setupData(p, components);
                        main._makeMenus(p);
                        main._makeDialogs(p);
                        main._setupVis();
                        main._setupModes();
                        
                        main.data.plot.ev.hilite.sub(main.updateModelStim);

                        // for safari:
                        setTimeout(() => {
                            main.modelView.onResize();
                            main.modelView.grid.__layout();
                        }, 2000);
                    }).catch(qubx.reportErr);
                });

                return main;
            } catch (err) {
                qubx.reportErr(err);
            }
        };
        
        return exports;
    });


// error catching (incl. asm out of memory etc) and possible auto-reload
// slow smooth durhist?
// touch events: twopane, ...?
// ? use optimal nbin via fred's 05?/2016 email
// landowne expo fit
// more native data formats

// cloud projects
// choose (default?) to load events not samples (esp. default for stim)
// idl stats, select, etc...?
// multiple models, auto-mmerge (auto-sum in macroscopic, heuristic choice in simu)
// better results browsing
// idl wizard: pick sel, tell Nlevel, init classes by {hilite, click 'Initialize from sel' or 'Initialize all levels from sel'}
//             tell mean lifetime, see chain model, see idl of preview sel
// group compute

// curve fittings?
// macror with ll-rescaling
// chain, star, classify, perturb
// model search


// https://localhost:13807/online/index.html#link=1&k=20c7f6b7-d93b-4a79-950c-a8f5a000a89f&macqdf=sample&welcome=1
// https://www.qub.buffalo.edu/online/#link=7&k=212fdadc-1d58-42a2-bb8e-1e0127d288db&macqdf=sample&welcome=1


/// option to simulate merged
/// merged multi-model idealize
/// merged multi-model MIL and HJC

/// view of merged?
/// 2-way rate binding

///// uncle meg


/// safari:
///   idealize
///   load macroscopic
///   results layout


/// complain if no localStorage

/// mfp dialog +name of model for clarity?
/// edit model name?

/// scalebars:
//   adjust if py < 1em, but make sure py < plotH
//   units; smart pick sec/ms/us/ps
//   option to hide it?
//   separate x and y components?

// encapsulate mac session within worker to allow multiples e.g. 1 local 1 remote
// websocket client.main.mac <-> server

// optimize stimulus delay

// We're going to approximate the stimulus as a staircase. How big should the smallest step be?
//  initialize proportional to (hi - lo)

/// say: adaptive resampling greatly speeds up fitting at the expense of accurac
/// say: samples frequently in rapidly changing signals, and infrequently in steady signals


// results:
//   param std ~== value => tell user it's bad (highlight/mark)
//   pretty-print date (time only if it's today; "yesterday", weekday name within past week; else just date
//   view result list as sortable table (ll, npar, ndata, aic, bic)


// ask moffatt about error analysis and confidence limits
// ask moffatt for feedback on mac online


// spinner while waiting mac/live
/// server clear temps on startup
/// multi-file open especially *+qdf and pul+dat?

/// copy images to clipboard


///// DB:

// Hopeful: name, about, group
// Group: name, about, active, limits:{people, admins, ...? and counts}, dataKey[to id and connect data servers]
// Person: name, about, email, active, group, isAdmin, isContact, isSuper
// Signin: person, secret, when, ip, agent
// Invite: person, secret, expires, usesRemain
// Model: public, name, qmj AS JSON, about, person, group, whenAdded, whenPub, authors, doi, nstate, nclass, digest
// Dataset: public, uuid, name, about, whenRecorded, whenAdded, person, group, isSampled, isIdealized, npoint, nevent, nsegment, stimuli, constants, sampling, moleculename, etc as JSON
// DataServer: uuid, lastContact, coverage[fraction,self-reported], baseUrl, (nondb/runtime: putPath[unique at server startup, to limit spam])
// DataInstance: dataset, dataserver [after first for dataset, verify byte count or checksum]
// Project: public, name, about, whenCreated, whenModified, whenPub, authors, doi, type, config
// ProjectModel: project, model
// ProjectData: project, dataset
// Result: project, ll, config, results as JSON

///// ROLES:

// dataserver:
//   store separate jsotan: meta+idl, each segment's current signal, each segment's other sampled signals
//     {uuid}/index.jsotan, {uuid}/I/%d.jsotan, {uuid}/X/%d.jsotan
//   also look for unwanted files on startup and delete them?

// super:
//   groups: list, new, del, edit {name, active, limits}
//     -> group-admin functions
//   person: group, isSuper
//   out-of-group new/del/edit: person, model, dataset, project
//   out-of-group invites

// contact:
//  receive (re-)invite requests, approve or deny
//       remember seat id of request so they can be linked for next page load

// admin (in group):
//   group: about, dataKey; list counts
//   person: list, new, del, edit except {group, super}
//   (re-)invite
//   del and edit: model, dataset, project (except person, group, whenAdded/Created, isSampled, isIdealized, npoint, nevent, nsegment, stimuli, sampling, projType)
//   del project*
//   list dataservers, forget one

// person:
//   (re-)sign in
//   edit own about, email
//   self-invite
//   (single sign-in, older ones signed out)
//   add model, dataset, project*, result
//     put data to dataserver
//   edit own model {name, about, whenPub, authors, doi}
//   edit own dataset {name, about, whenRecorded, constants, moleculename, etc}
//   edit own project {about, whenPub, authors, doi}
//   del own project(+*), dataset, model
//   del result
//   get link to project


/// "guest" group for randoms with its own data servers
/// for new signins:
///   if already signed in elsewhere: tell them how to self-invite
///   pick your group
///   send msg to admin, get id token
///   find out about approval via email, or at next page-load, or by ws if still loaded

/// bug report page



/// option to not allow multi-model-sim
/// ask when they first have 2 models open?

///// mac: limit n points between stop flag checks


/// landscape about: maybe some instructions for making mmm dwt and graphics


/// use reader.signals esp. units; deprecate signalNames

/// load traces in columns from plain text?

/// redirect obviously mis-dropped data from model to dataview
/// "RangeError: Array buffer allocation failed" -- handle or tell user

