/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_demo_single", ["qubx", "qubx_ui", "qubx_ui_data", "qubx_single", "d3"],
    (qubx, qubx_ui, qubx_ui_data, qubx_single, d3) => {
        "use strict";
        const exports = {};

        const SingleModeDefs = {
            idealize: {
                reestimate: true,
                maxIter: 10,
                reestRates: true,
                fixAmp: {},
                fixStd: {}
            },
            minimize: {
                dfp: true,
                simplex: true,
                maxIter: 100,
                maxStep: 1.0,
                dx: 1e-7,
                hjc: false,
                chs: true,
                msl: false
            },
            workers: 3
        };

        exports.SingleMode = class SingleMode {
            constructor(main, config) {
                config = config || {};
                this.main = main;
                let storedPrefs = localStorage.getItem("qubx_single_prefs");
                storedPrefs = storedPrefs ? JSON.parse(storedPrefs) : {};
                this.config = Object.assign({}, SingleModeDefs, config, storedPrefs);
                this.config.idealize = Object.assign({}, SingleModeDefs.idealize, config.idealize, storedPrefs.idealize);
                this.config.minimize = Object.assign({}, SingleModeDefs.minimize, config.minimize, storedPrefs.minimize);
                this.prepDataSoon = qubx.batchDelayed(this.prepData.bind(this), 1000);
                this.main.panels.then(p => {
                    this.modelView = p.modelView;
                    this.data = p.data;
                    this.controls = p.controls.querySelector("#qubx_demo_data_ctrls");
                    this._makeMenus(p);
                    this._makeDialogs(p);
                    this._boundOnChangeTCrit = this._onChangeTCrit.bind(this);
                    this._lastDataFile = null;
                    this.data.plot.ev.changeData.sub(d => {
                        if ( this._lastDataFile ) {
                            this._lastDataFile.ev.changeTCrit.unsub(this._boundOnChangeTCrit);
                        }
                        this._lastDataFile = d;
                        if ( d ) {
                            d.ev.changeTCrit.sub(this._boundOnChangeTCrit);
                            this._onChangeTCrit(d.tCrit());
                        }
                    });
                }).catch(qubx.reportErr);
                this.resultStyle = qubx.get("css/qubx_resultDoc.css");
            }
            activate() {
                try {
                    this.menus.forEach(node => this.controls.appendChild(node));
                    this.workgroup = new qubx_single.SingleWorkgroup({count: this.config.workers});
                    this._dataChangeSamples = this.data.data.ev.changeSamples.sub(this._onChangeSamples.bind(this));
                    this._onChangeTCrit(this.data.data.tCrit());
                    this.prepDataSoon();
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
            deactivate() {
                return (this.waitRun ? qubx_ui.confirm({
                    hdr: 'Optimizer is still running',
                    q: 'Switch tools anyway, interrupting the optimizer?'
                }).then(() => this.waitRun && this._onClickRun()) : Promise.resolve(true)).then(() => {
                    this.data.data.ev.changeSamples.unsub(this._dataChangeSamples);
                    this.menus.forEach(node => this.controls.removeChild(node));
                    this.savePrefs();
                    this.workgroup.destroy();
                });
            }
            savePrefs() {
                try {
                    const prefs = {};
                    for (const k of Object.keys(SingleModeDefs)) { // to avoid e.g. config.node
                        prefs[k] = this.config[k];
                    }
                    localStorage.setItem("qubx_single_prefs", JSON.stringify(prefs));
                    qubx_ui.resultDlg.savePrefs();
                } catch (err) {
                    qubx.reportErr(err);
                }
            }
            _makeMenus(p) {
                qubx_ui.loadTemplate("#qubx_single_data_ctrls").then(m => {
                    const dotdot = qubx_ui.querydotdot(m);
                    this.btnIdealize = dotdot.listen("#qubx_idl_run", 'click', this._onClickIdealize.bind(this));
                    this.btnRun = dotdot.listen("#qubx_single_opt", 'click', this._onClickRun.bind(this));
                    this.llOut = dotdot("#qubx_single_ll_out");
                    this.btnResults = dotdot.listen("#qubx_single_showResults", 'click', () => qubx_ui.resultDlg.run());
                    this.menus = [this.btnIdealize, this.btnRun, this.llOut, this.btnResults];
                }).catch(qubx.reportErr);
            }
            _makeDialogs(p) {
                this.idlDlg = new qubx_ui.Modal({
                    title: "Idealize",
                    template: "#qubx_idealize_options",
                    okLabel: 'Idealize'
                })
                this.idlDlg.dom.then(dom => {
                    const dotdot = qubx_ui.querydotdot(dom.content);
                    this.idlDlg.chkReest = dotdot.listen("#qubx_idl_reest", 'change', () => {
                        try {
                            const reest = this.idlDlg.chkReest.checked;
                            this.config.idealize.reestimate = reest;
                            this.idlDlg.panelReest.style.display = reest ? '' : 'none';
                            this.idlDlg.panelMaxIter.style.display = reest ? '' : 'none';
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    });
                    this.idlDlg.chkReest.checked = this.config.idealize.reestimate;
                    this.idlDlg.numMaxIter = dotdot.sub("#qubx_idl_maxiter", input => new qubx_ui.AcceptedInput(input, this.config.idealize.maxIter, {
                        accept: qubx.acceptIntAtLeast(1),
                        format: qubx.formatStr,
                        widthChars: 5
                    }), 'change', val => {
                        this.config.idealize.maxIter = val;
                    });
                    this.idlDlg.chkReestRates = dotdot.listen("#qubx_idl_reest_rates", 'change', () => {
                        this.config.idealize.reestRates = this.idlDlg.chkReestRates.checked;
                    });
                    this.idlDlg.chkReestRates.checked = this.config.idealize.reestRates;
                    this.idlDlg.panelReest = dotdot("#qubx_idl_reest_options", p => {
                        p.style.display = this.config.idealize.reestimate ? '' : 'none';
                    });
                    this.idlDlg.panelMaxIter = dotdot("#qubx_idl_maxiterbox", p => {
                        p.style.display = this.config.idealize.reestimate ? '' : 'none';
                    });
                    this.idlDlg.layoutCond = model => {
                        const classOccupied = [];
                        model.states.entries.forEach(s => {
                            classOccupied[s.Class] = 1;
                        });
                        const chks = [];
                        classOccupied.forEach((occupied, c) => {
                            chks.push({
                                field: 'fixAmp',
                                index: c
                            });
                            chks.push({
                                field: 'fixStd',
                                index: c
                            });
                        });
                        const bound = d3.select(dotdot("#qubx_idl_reest_cond"))
                            .selectAll("div")
                            .data(chks);
                        const self = this;
                        const lbl = bound.enter().append("div")
                              .append("label");
                        lbl.append("span").text(d => d.field.slice(3)+' '+d.index);
                        lbl.append("input")
                            .attr("type", "checkbox")
                            .on("change", function(d, i) {
                                try {
                                    self.config.idealize[d.field][''+d.index] = ! this.checked;
                                } catch (err) {
                                    qubx.reportErr(err);
                                }
                            });
                        bound.exit().remove();
                        bound.select("label").select("input")
                            .property("checked", d => ! this.config.idealize[d.field][''+d.index]);
                    };
                }).catch(qubx.reportErr);

                this.idlResultDlg = new qubx_ui.Modal({
                    title: "Idealization Result",
                    template: "#qubx_idl_result",
                    canCancel: false,
                    okLabel: 'Close'
                });
                this.idlResultDlg.dom.then(dom => {
                    this.idlResultDlg.tblSegments = new qubx_ui.ColumnTable({
                        node: dom.content.querySelector("#idl_result_segments")
                    });
                }).catch(qubx.reportErr);
                
                this.optDlg = new qubx_ui.Modal({
                    title: "Optimize",
                    template: "#qubx_single_optstart"
                });
                this.optDlg.dom.then(dom => {
                    const dotdot = qubx_ui.querydotdot(dom.content),
                          constBox = dotdot("#qsos_constants_box"),
                          constTable = new qubx_ui_data.ConstantsTable({
                              node: dotdot("#qsos_constants"),
                              plot: this.data.plot
                          }),
                          chks = {
                              simplex: {host: this.config.minimize, node: dotdot("#qsos_simplex")},
                              dfp: {host: this.config.minimize, node: dotdot("#qsos_dfp")},
                              hjc: {host: this.config.minimize, node: dotdot("#qsos_hjc")},
                              chs: {host: this.config.minimize, node: dotdot("#qsos_chs")},
                              msl: {host: this.config.minimize, node: dotdot("#qsos_msl")}
                          },
                          boxHJC = dotdot("#qsos_hjcbox"),
                          boxCHS = dotdot("#qsos_chsbox");
                    boxHJC.style.display = this.config.minimize.msl ? 'none' : '';
                    boxCHS.style.display = (this.config.minimize.hjc && ! this.config.minimize.msl) ? '' : 'none';
                    for (const nm of Object.keys(chks)) {
                        (function (nm,host,node) {
                            node.checked = host[nm];
                            node.addEventListener('change', () => {
                                try {
                                    const yes = node.checked;
                                    host[nm] = yes;
                                } catch (err) {
                                    qubx.reportErr(err);
                                }
                            });
                        })(nm, chks[nm].host, chks[nm].node);
                    }
                    chks.msl.node.addEventListener('change', () => {
                        try {
                            boxHJC.style.display = this.config.minimize.msl ? 'none' : '';
                            boxCHS.style.display = (this.config.minimize.hjc && ! this.config.minimize.msl) ? '' : 'none';
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    });
                    chks.hjc.node.addEventListener('change', () => {
                        try {
                            boxCHS.style.display = (this.config.minimize.hjc && ! this.config.minimize.msl) ? '' : 'none';
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    });
                    this.optDlg.txtTcrit = dotdot.sub("#qsos_tcrit", input => new qubx_ui.AcceptedInput(input, 0.0, {
                        accept: qubx.acceptFloatAtLeast(0.0),
                        format: qubx.formatK0,
                        widthChars: 6
                    }), 'change', this._onEnterTCrit.bind(this));
                    dotdot.sub('#qsos_maxIter', input => new qubx_ui.AcceptedInput(input, this.config.minimize.maxIter, {
                        accept: qubx.acceptIntAtLeast(1),
                        format: qubx.formatStr,
                        widthChars: 5
                    }), 'change', val => {
                        this.config.minimize.maxIter = val;
                    });
                    dotdot.sub('#qsos_maxStep', input => new qubx_ui.AcceptedInput(input, this.config.minimize.maxStep, {
                        accept: qubx.acceptFloatGreaterThan(0),
                        format: qubx.formatK0,
                        widthChars: 8
                    }), 'change', val => {
                        this.config.minimize.maxStep = val;
                    });
                    dotdot.sub('#qsos_dx', input => new qubx_ui.AcceptedInput(input, this.config.minimize.dx, {
                        accept: qubx.acceptFloatGreaterThan(0),
                        format: qubx.formatK0,
                        widthChars: 8
                    }), 'change', val => {
                        this.config.minimize.dx = val;
                    });
                    const chkBal = dotdot.listen('#qsos_balance_loops', 'change', () => {
                        try {
                            this.optModel.balanceLoops(chkBal.checked);
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    });
                    const constraints = dotdot('#qsos_constraints', node => new qubx_ui.TableView({
                        node: node,
                        columns: ['Type', 'States', 'delete'],
                        table: this.modelView.model.constraints,
                        autohide: true
                    }));
                    dotdot.listen("#qsos_add_constraint", 'click', () => {
                        try {
                            this.optModel.constraints.append({});
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    });
                    const advanced = dotdot('#qsos_advanced', node => {
                        node.className = 'nodisplay';
                    });
                    const showadv = dotdot.listen("#qsos_showadv", 'click', () => {
                        try {
                            if ( advanced.classList.contains('nodisplay') ) {
                                advanced.classList.remove('nodisplay');
                                showadv.querySelector("span").textContent = 'Hide';
                            } else {
                                advanced.classList.add('nodisplay');
                                showadv.querySelector("span").textContent = 'Show';
                            }
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                    });
                    dotdot.sub('#qsos_workers', input => new qubx_ui.AcceptedInput(input, this.config.workers, {
                        accept: qubx.acceptIntAtLeast(1),
                        format: qubx.formatStr,
                        widthChars: 5
                    }), 'change', val => {
                        this.config.workers = val;
                        this.workgroup.count(val);
                    });
                    const coreRun = this.optDlg.run.bind(this.optDlg);
                    this.optDlg.run = () => {
                        this.optModelInfo = this.modelView.info;
                        this.optModel = this.modelView.model;
                        const modelStim = this.optModel.getStimulus();
                        try {
                            constTable.layout(modelStim);
                            constBox.style.display = constTable.node.querySelector('tr') ? '' : 'none';
                            chkBal.checked = this.optModel.balanceLoops();
                            constraints.render(this.optModel.constraints);
                        } catch (err) {
                            qubx.reportErr(err);
                        }
                        return coreRun().then(result => {
                            for (const cname of Object.keys(modelStim)) {
                                this.workgroup.setConstant(cname, this.data.data.constants.getByName(cname));
                            }
                            return result;
                        });
                    };
                }).catch(qubx.reportErr);
            }
            _onChangeTCrit(tc) {
                this.optDlg.txtTcrit.update(tc, 1e3*this.data.data.sampling());
            }
            _onEnterTCrit(tc) {
                this.data.data.tCrit(tc);
            }
            _onClickIdealize() {
                const model = this.modelView.model;
                try {
                    this.idlModel = model;
                    this.idlDlg.layoutCond(model);
                    this.btnIdealize.disabled = true;
                } catch (err) {
                    return qubx.reportErr(err);
                }
                return this.idlDlg.run().then(() => {
                    const data = this.data && this.data.data,
                          segments = data.segments.map((iseg, i) => ({
                              index: i,
                              n: iseg.n,
                              sampled: iseg.sampled[0]
                          })),
                          fixAmp = model.classes.entries.map((c,ic) => !! this.config.idealize.fixAmp[''+ic]),
                          fixStd = model.classes.entries.map((c,ic) => !! this.config.idealize.fixStd[''+ic]);
                    return this.workgroup.SKM(
                        model.toJson(), data.sampling(), segments, (seg, offset, n) => {
                            const copy = new Float32Array(n);
                            let i = 0;
                            for (; i<n; ++i, ++offset) {
                                copy[i] = seg.sampled[offset];
                            }
                            return copy;
                        }, this.config.idealize.reestimate ? this.config.idealize.maxIter : 0,
                           ! this.config.idealize.reestRates, fixAmp, fixStd)
                        .then(idlSegments => {
                            idlSegments.forEach(idlseg => data.setIdeal(idlseg.iSeg, 0, idlseg.counts, idlseg.classes, idlseg.offset, idlseg.amp, idlseg.std));
                            this.prepData();
                            this.btnIdealize.disabled = false;
                            this.showIdlResult(idlSegments);
                        });
                }).catch(qubx.reportErr)
                    .catch(err => {
                        this.btnIdealize.disabled = false;
                        return Promise.reject(err);
                    });
            }
            showIdlResult(segments) {
                if ( ! (segments && segments.length && segments[0].counts && segments[0].counts.length) ) {
                    return Promise.resolve(true);
                }
                const dtMS = 1e3 * this.data.data.sampling();
                let maxCls = -1;
                segments.forEach(seg => {
                    const n = seg.classes.length;
                    let i = 0;
                    for (; i<n; ++i) {
                        if ( maxCls < seg.classes[i] ) {
                            maxCls = seg.classes[i];
                        }
                    }
                });
                segments.forEach(seg => {
                    const sortCounts = [];
                    let c = 0;
                    for (; c<=maxCls; ++c) {
                        sortCounts.push([]);
                    }
                    const totalCounts = sortCounts.map(() => 0),
                          n = seg.classes.length;
                    let totalCount = 0,
                        i = 0;
                    for (; i<n; ++i) {
                        const c = seg.classes[i],
                              count = seg.counts[i];
                        sortCounts[c].push(count);
                        totalCounts[c] += count;
                        totalCount += count;
                    }
                    for (c=0; c<=maxCls; ++c) {
                        const nd = sortCounts[c].length;
                        seg['Index'] = c;
                        seg['ndwell'] = seg.classes.length;
                        seg['nevent'+c] = nd;
                        seg['amp'+c] = seg.amp[c];
                        seg['std'+c] = seg.std[c];
                        seg['occupancy'+c] = totalCounts[c] / totalCount;
                        const ltstats = qubx.meanstd(sortCounts[c]);
                        seg['lifetimeMean'+c] = dtMS * ltstats.mean;
                        seg['lifetimeStd'+c] = dtMS * ltstats.std;
                        sortCounts[c].sort();
                        if ( nd % 2 ) {
                            seg['lifetimeMedian'+c] = dtMS * sortCounts[c][(nd/2)|0];
                        } else {
                            seg['lifetimeMedian'+c] = dtMS * 0.5 * (sortCounts[c][nd/2] + sortCounts[c][nd/2 - 1]);
                        }
                    }
                });
                
                const columns = [
                    qubx_ui.ColumnTable.txtCol('Index', 'Index'),
                    qubx_ui.ColumnTable.txtCol('Events', 'ndwell')
                ];
                let c = 0;
                for (; c<=maxCls; ++c) {
                    columns.push(qubx_ui.ColumnTable.txtCol('nevent '+c, 'nevent'+c));
                }
                for (c=0; c<=maxCls; ++c) {
                    columns.push(qubx_ui.ColumnTable.txtCol('amp '+c, 'amp'+c, qubx.formatK0));
                    columns.push(qubx_ui.ColumnTable.txtCol('std '+c, 'std'+c, qubx.formatK0));
                }
                for (c=0; c<=maxCls; ++c) {
                    columns.push(qubx_ui.ColumnTable.txtCol('median lifetime '+c+' (ms)', 'lifetimeMedian'+c, qubx.formatK0));
                    columns.push(qubx_ui.ColumnTable.txtCol('mean lifetime '+c, 'lifetimeMean'+c, qubx.formatK0));
                    columns.push(qubx_ui.ColumnTable.txtCol('std lifetime '+c, 'lifetimeStd'+c, qubx.formatK0));
                }
                for (c=0; c<=maxCls; ++c) {
                    columns.push(qubx_ui.ColumnTable.txtCol('occupancy '+c, 'occupancy'+c, qubx.formatK0));
                }
                
                this.idlResultDlg.tblSegments.config.node.innerHTML = '';
                this.idlResultDlg.tblSegments.config.columns = columns;
                this.idlResultDlg.tblSegments.render(segments);
                
                return this.idlResultDlg.run();
            }
            _onChangeSamples() {
                this.prepDataSoon();
            }
            prepData() {
                const data = this.data && this.data.data;
                if ( ! (data && data.segmentCount()) ) {
                    return;
                }
                const dt = data.sampling(),
                      delays = {},
                      stims = this.modelView.model.getStimulus(); ////
                data.signals.forEach(sig => {
                    const nm = sig.name;
                    delays[nm] = this.main.signals.get(nm).delay;
                    delete stims[nm];
                });
                const defaults = {};
                for (const nm of Object.keys(stims)) {
                    let v = data.constants.getByName(nm);
                    if ( typeof(v) === 'undefined' ) {
                        v = this.main.signals.get(nm).defval;
                        data.constants.setByName(nm, v);
                    }
                    defaults[nm] = v;
                }
                this.btnIdealize.disabled = ! (data.segments.length && data.segments[0].sampled[0]);
                let haveIdl = true;
                const segments = data.segments.map((seg, i) => {
                    const idl = seg.ideal[0].getDwellsCD(0, seg.n-1, null, null, dt*1e3);
                    if ( idl.classes.length === 0 ) {
                        haveIdl = false;
                    }
                    const cn = idl.classes.length;
                    let ci = 0;
                    for (; ci<cn; ++ci) {
                        if ( idl.classes[ci] < 0 ) {
                            haveIdl = false;
                        }
                    }
                    const stimuli = {};
                    data.signals.slice(1).forEach((sig,j) => {
                        stimuli[sig.name] = seg.ideal[j+1].getDwellsCD(0, seg.n-1, null, null, dt*1e3); // in msec
                        stimuli[sig.name].amps = seg.ideal[j+1].amp;
                    });
                    return {
                        idl: idl,
                        stimuli: stimuli,
                        dur: dt*seg.n
                    };
                });
                this.btnRun.disabled = ! haveIdl;
                return this.workgroup.setLayout({
                    sampling: dt,
                    signalNames: data.signals.map(qubx.m('name')),
                    delays, defaults, segments
                });
            }
            _beginRun() {
                this.modelView.model.fromJson(this.modelView.model.toJson());
                this.modelView.model.undo.disable();
                this.btnRun.innerHTML = 'Stop';
                this.llOut.innerHTML = 'Running...';
                this.waitRun = true;
            }
            _endRun(msg) {
                this.optModel.undo.enable();
                this.optModel.fromJson(this.optModel.toJson());
                this.optModel.undo.sealUndo("Optimize");
                this.btnRun.innerHTML = 'Optimize...';
                this.llOut.innerHTML = msg;
                this.waitRun = false;
            }
            _onClickRun() {
                if ( this.waitRun ) {
                    this.llOut.innerHTML = "Stopping...";
                    this.stopRun = true;
                    return;
                }
                this.stopRun = false;
                const updateModel = model => {
                    if ( model ) {
                        this.optModel.fromJson(model);
                    }
                };
                const tdMS = this.data.data.tDead();
                this.optDlg.txtTcrit.update(this.data.data.tCrit(), 1e3*this.data.data.sampling());
                this._beginRun();
                this.optDlg.run().then(() => this.workgroup.optimize(
                    this.optModel.toJson(),
                    1e-3*this.data.data.tDead(), 1e-3*this.data.data.tCrit(),
                    Object.assign({}, this.config.minimize, {
                        onIteration: (model, params, ll, iteration, grad) => {
                            console.log('single iter',iteration,ll,params,grad,model);
                            this.llOut.innerHTML = 'LL: '+ll.toPrecision(4)+' (iter '+iteration+')';
                            updateModel(model);
                            return ! this.stopRun;
                        }
                    }))).then(result => {
                        console.log('single opt done',result);
                        this._endRun('LL: '+result.ll.toPrecision(4));
                        updateModel(result.model);
                        this.btnResults.disabled = true;
                        this.optModelInfo.unselected(); // for clean svg capture
                        Promise.all([this.optModelInfo.getSVG(), this.main.durHist.getSVG(), this.resultStyle]).then(items => {
                            const [svgModel, svgHists, css] = items;
                            const algoName = result.msl ? "MSL" : (result.hjc ? "HJC" : "MIL"),
                                  fragment = document.createDocumentFragment(),
                                  outer = d3.select(fragment).append("div");
                            outer.append("p").html('Results of <a href="https://qub.madelics.com/online">QUB Online: '+algoName+'</a>, '+new Date());
                            outer.append("p").text('Finished optimizing after '+result.iteration+' iterations.');
                            outer.append("p").html('Dwells: '+result.ndwell+', Parameters: '+result.params.length+', t<sub>dead</sub>: '+tdMS+' ms');
                            outer.append("p").text('LL: '+qubx.formatK0(result.ll));
                            outer.append("p").text('AIC: '+qubx.formatK0(result.aic));
                            outer.append("p").text('BIC: '+qubx.formatK0(result.bic));
                            const rateFields = [
                                qubx_ui.ColumnTable.txtCol('From', 'From'),
                                qubx_ui.ColumnTable.txtCol('To', 'To'),
                                qubx_ui.ColumnTable.txtCol('k0', 'k0', qubx.formatK0),
                                qubx_ui.ColumnTable.txtCol('k0_std', 'k0_std', qubx.formatK0)
                            ];
                            for (const r of result.model.rates) {
                                if ( r.Voltage ) {
                                    rateFields.push(qubx_ui.ColumnTable.txtCol('k1', 'k1', qubx.formatK0));
                                    rateFields.push(qubx_ui.ColumnTable.txtCol('k1_std', 'k1_std', qubx.formatK0));
                                    break;
                                }
                            }
                            for (const r of result.model.rates) {
                                if ( r.Pressure ) {
                                    rateFields.push(qubx_ui.ColumnTable.txtCol('k2', 'k2', qubx.formatK0));
                                    rateFields.push(qubx_ui.ColumnTable.txtCol('k2_std', 'k2_std', qubx.formatK0));
                                    break;
                                }
                            }
                            outer.append("p").text("Rate constants:");
                            new qubx_ui.ColumnTable({
                                node: outer.append("table").node(),
                                columns: rateFields
                            }).render(result.model.rates);
                            
                            const hists = outer.append("div");
                            hists.classed('resultHists', true);
                            svgHists.forEach(svg => hists.node().appendChild(svg));
                            svgModel.classList.add('resultModel');
                            outer.node().appendChild(svgModel);
                            qubx_ui.resultDlg.add('<html><head><style type="text/css">'+css+'</style></head><body>'+outer.node().innerHTML+'</body></html>',
                                                  algoName);
                            this.btnResults.disabled = false;
                        });
                    }).catch(qubx.reportErr)
                        .catch(err => {
                            console.log('single opt err',err);
                            this.btnResults.disabled = false;
                            this._endRun(err.errMsg || err);
                        });
            }
        };
        
        return exports;
    });

