/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_demo_vid", ["qubx", "qubx_lib", "qubx_ui", "qubtrace", "RecordRTC"],
    (qubx, qubx_lib, qubx_ui, qubtrace, RecordRTC) => {
        "use strict";
        const exports = {};

        const VidMakerDefs = {

        };
        exports.VidMaker = class VidMaker {
            constructor(config) {
                this.config = Object.assign({}, VidMakerDefs, config);
                this.modelView = this.config.modelView;
                this.dataView = this.config.dataView;
                this.saver = this.config.saver || new qubx_ui.Saver();
                this.progress = new qubx_ui.ModalProgress();
                qubx_lib.then(lib => {
                    this.raster_samples = lib.raster_samples;
                });
                this.setupDlg = new qubx_ui.Modal({
                    title: 'Make a video from data',
                    template: '#qubx_demo_vid_setup'
                });
                this.setupDlg.dom.then(dom => {
                    const dotdot = qubx_ui.querydotdot(dom.content);
                    this.txtWidthPx = new qubx_ui.AcceptedInput(dotdot("#qdvs_width_px"), 800, {
                        accept: qubx.acceptIntAtLeast(1),
                        format: qubx.formatStr,
                        localKey: 'qdvs_width_px'
                    });
                    this.txtWidthSec = new qubx_ui.AcceptedInput(dotdot("#qdvs_width_sec"), 1.0, {
                        accept: qubx.acceptFloatGreaterThan(0.0),
                        format: qubx.formatK0,
                        localKey: 'qdvs_width_sec'
                    });
                    this.txtHeightPx = new qubx_ui.AcceptedInput(dotdot("#qdvs_height_px"), 450, {
                        accept: qubx.acceptIntAtLeast(1),
                        format: qubx.formatStr,
                        localKey: 'qdvs_height_px'
                    });
                    this.chkWithPics = dotdot("#qdvs_withPics");
                    this.chkWithSound = dotdot("#qdvs_withSound");
                    this.chkWithSound.addEventListener('change', this._initSound.bind(this));
                    this.mnuSoundType = dotdot("#qdvs_soundType");
                    this.txtRate = new qubx_ui.AcceptedInput(dotdot("#qdvs_rate"), 1.0, {
                        accept: qubx.acceptFloatGreaterThan(0),
                        format: qubx.formatK0,
                        localKey: 'qdvs_rate'
                    });
                    this.classes = new qubx_ui.ColumnTable({
                        node: dotdot("#qdvs_classes"),
                        columns: [
                            qubx_ui.ColumnTable.txtCol('Index', 'index', qubx.formatStr),
                            qubx_ui.ColumnTable.inputCol('Freq (Hz)', 'freq', 'number', qubx.acceptFloatGreaterThan(0), qubx.formatK0),
                            qubx_ui.ColumnTable.inputCol('Volume', 'gain', 'number', qubx.acceptFloatBetween(0, 1), qubx.formatK0, undefined, 0.0, 1.0, 0.1),
                            qubx_ui.ColumnTable.fileCol('Image', (input, d, i) => {
                                if ( ! input.files.length ) {
                                    return;
                                }
                                const fr = new FileReader();
                                fr.onload = () => {
                                    d.row.image = new Image();
                                    d.row.image.src = fr.result;
                                };
                                fr.readAsDataURL(input.files[0]);
                            })
                        ]
                    });
                    this.progress.dom.then(progdom => {
                        const progbody = progdom.root.querySelector(".qubx_modal_prog_body");
                        progbody.insertBefore(dotdot("#qdvs_preview"), progbody.firstChild);
                    });
                    this.canvas = dotdot("#qdvs_canvas");
                    this.scaleCanvas = document.createElement('canvas');
                });
            }
            _initSound() {
                if ( this.chkWithSound.checked && ! this.audioCtx ) {
                    this.audioCtx = new (window.AudioContext || window.webkitAudioContext)();
                }
            }
            _updateClassSetup() {
                this.clsData = this.clsData || this.modelView.model.classes.entries.map((s,i) => ({
                    index: i,
                    freq: 220*Math.pow(2, i),
                    gain: 1.0
                }));
                const ncls = this.modelView.model.classes.size;
                while ( this.clsData.length < ncls ) {
                    this.clsData.push({
                        index: this.clsData.length,
                        freq: 220*Math.pow(2, this.clsData.length),
                        gain: 1.0
                    });
                }
                this.clsData.splice(ncls, this.clsData.length-ncls);
                this.classes.render(this.clsData);
            }
            _layout() {
                let w = this.canvas.width = this.txtWidthPx.value,
                    h = this.canvas.height = this.txtHeightPx.value;
                const dur = this.txtWidthSec.value,
                      withPics = this.chkWithPics.checked,
                      sampling = this.dataView.data.sampling(),
                      border = 10,
                      x = border,
                      y = border;
                w -= 2*border,
                h -= 2*border;
                const h1 = (.62*h)|0,
                      dh = withPics ? (h-h1) : h,
                      samples_screen = Math.round(dur / sampling)|0,
                      spp = samples_screen / w, // samples per pixel
                      halfscreenSamples = samples_screen >> 1,
                      bounds = this.dataView.plot.bounds[0],
                      cy = (bounds.hi + bounds.lo) / 2,
                      yLo = cy - .6*bounds.dy,
                      yHi = cy + .6*bounds.dy,
                      ypp = (yHi - yLo) / dh,
                      withSound = this.chkWithSound.checked,
                      soundType = this.mnuSoundType.value,
                      rate = this.txtRate.value;
                return {
                    border, w, h,
                    d: {
                        y: withPics ? h1 : 0,
                        h: dh,
                    },
                    p: {
                        h: h1
                    },
                    dur, withPics, samples_screen, spp, halfscreenSamples,
                    yLo, ypp, withSound, soundType, rate, sampling
                };
            }
            _drawData(ctx, l, iseg, i, points, n) {
                var px0, pxw, ndraw, offset;
                if ( i < 0 ) {
                    ndraw = Math.ceil(Math.min(i+l.samples_screen, n))|0;
                    px0 = Math.round(-i/l.spp)|0;
                    offset = 0;
                }
                else {
                    px0 = 0;
                    offset = Math.max(0,Math.round(i)|0);
                    ndraw = Math.max(0, Math.ceil(Math.min(l.samples_screen, n-offset))|0);
                }
                if ( ndraw ) {
                    pxw = Math.ceil(ndraw/l.spp)|0;
                    if ( l.spp >= this.dataView.plot.config.minSppRaster ) {
                        const raster = this.raster_samples || qubtrace.raster_samples;
                        const rastered = raster(points.buffer, offset, ndraw, l.w, l.spp, l.ypp);
                        qubtrace.drawRastered(ctx, px0, pxw, l.d.h, rastered, "rgb(100,100,100)", "rgb(0,0,0)", l.spp, l.yLo, l.ypp);
                    } else {
                        qubtrace.drawLines(ctx, px0, pxw, l.d.h, points, offset, ndraw, "rgb(0,0,0)", l.spp, l.yLo, l.ypp);
                    }
                }
                ctx.lineWidth = 3;
                ctx.beginPath();
                ctx.moveTo(l.w/2, .15*l.d.h);
                ctx.lineTo(l.w/2, .85*l.d.h);
                ctx.strokeStyle = 'rgba(255,55,55,.6)';
                ctx.stroke();
                const emsize = this.dataView.plot.emsize;
                qubtrace.drawScalebar(this.scaleCanvas, emsize, l.w, l.d.h, l.sampling*l.spp, [{
                    ypp: l.ypp,
                    units: 'pA',
                    color: "black"
                }], l.withPics ? 0.25 : .15, "sans-serif", "rgba(255,255,255,0.6)", "black");
                ctx.drawImage(this.scaleCanvas, l.w - this.scaleCanvas.width - emsize, emsize);
            }
            _animate(ctx, l, reject, onFinish) {
                if ( ! this.animState ) {
                    this.animState = {
                        ctx: ctx,
                        l: l,
                        reject: reject,
                        onFinish: onFinish,
                        iseg: -1
                    };
                } else {
                    ctx = ctx || this.animState.ctx;
                    l = l || this.animState.l;
                    reject = reject || this.animState.reject;
                    onFinish = onFinish || this.animState.onFinish;
                }
                const a = this.animState,
                      segments = this.dataView.data.segments,
                      nseg = segments.length;
                this.progress.progress(a.iseg/nseg);
                if ( this.progress.canceled ) {
                    reject('Canceled');
                    return;
                }
                if ( (a.iseg < 0) || (a.i >= a.n) ) {
                    ++a.iseg;
                    if ( a.iseg === nseg ) {
                        onFinish();
                        return;
                    }
                    a.i0 = 0;
                    a.n = segments[a.iseg].seen;
                    a.data = segments[a.iseg].sampled[0];
                    a.i = - l.halfscreenSamples;
                    a.t = window.performance ? performance.now() : new Date();
                } else {
                    const t = window.performance ? performance.now() : new Date();
                    a.i += l.rate * 1e-3*(t - a.t) / l.sampling;
                    a.t = t;
                }
                ctx.fillStyle = "rgb(255,255,255)";
                ctx.fillRect(0, 0, l.w, l.h);
                // sample idl at i+halfscreenSamples
                if ( a.i+l.halfscreenSamples < a.n ) {
                    const cls = segments[a.iseg].ideal[0].classAt(a.i+l.halfscreenSamples);
                    if ( cls >= 0 ) {
                        const cd = this.clsData[cls];
                        if ( l.withSound ) {
                            this.oscillator.frequency.value = cd.freq;
                            this.gain.gain.value = cd.gain;
                        }
                        if ( l.withPics && cd.image ) {
                            const scale = l.p.h / cd.image.height,
                                  w1 = scale * cd.image.width;
                            // draw cd.image scaled to l.h1; centered in l.w
                            ctx.drawImage(cd.image, (l.w - w1)/2, 0, w1, l.p.h);
                        }
                    } else {
                        if ( l.withSound ) {
                            this.gain.gain.value = 0.0;
                        }
                    }
                }
                ctx.save();
                ctx.translate(0, l.d.y);
                this._drawData(ctx, l, a.iseg, a.i, a.data, a.n);
                ctx.restore();
                requestAnimationFrame(() => this._animate());
            }
            run() {
                this._updateClassSetup();
                return this.setupDlg.run().then(() => new Promise((resolve, reject) => {
                    this.progress.show('Recording...');
                    const l = this._layout();
                    var recorder;
                    if ( l.withSound && this.audioCtx ) {
                        this.oscillator = this.audioCtx.createOscillator();
                        this.oscillator.type = l.soundType;
                        this.gain = this.audioCtx.createGain();
                        this.oscillator.connect(this.gain);
                        this.gain.connect(this.audioCtx.destination);
                        /// find something that can record audio, even if we have to disable video...
                        //this.audioStream = this.audioCtx.createMediaStreamDestination();
                        //
                        //this.gain.connect(this.audioStream);
                        //recorder = RecordRTC(this.audioStream.stream);
                    }
                    //} else {
                    recorder = new RecordRTC.CanvasRecorder(this.canvas, {
                        disableLogs: false
                    });
                    //}
                    const ctx = this.canvas.getContext('2d');
                    ctx.clearRect(0, 0, this.canvas.width, this.canvas.height);
                    ctx.translate(l.border+0.5, l.border); // for even raster pixels
                    recorder.record();
                    this.animState = null;
                    this._animate(ctx, l, reject, () => {
                        l.withSound && this.oscillator.stop();
                        recorder.stop(blob => {
                            this.progress.hide();
                            this.saver.prompt({name: 'data.webm'}).then(info => {
                                this.saver.save(info.fullname, blob);
                                resolve(true);
                            }).catch(reject);
                        });
                    });
                    l.withSound && this.oscillator.start();
                })).catch(err => {
                    this.progress.hide();
                    this.oscillator && this.oscillator.stop();
                    return Promise.reject(err);
                });
            }
        };

        return exports;
    });

// classes not states
//    because that's the kind of idealized data we have

/// record sound with video

/// preview layout
