/* Copyright 2016-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_fitness_main", ["qubx", "qubx_app", "qubx_coldata", "qubx_ui", "fitness_ui"],
    (qubx, qubx_app, qubx_coldata, qubx_ui, fitness_ui) => {
        "use strict";
        const exports = {};

        const DemoMainDefs = {
            fullscreen: false
        };
        exports.start = function start(config) {
            config = Object.assign({}, DemoMainDefs, config);
            const main = new qubx_app.Main(config),
                  _dispose = main.dispose.bind(main);
            
            Object.assign(main, {
                opener: new qubx_ui.Opener(),
                reader: new qubx_coldata.TableReader(),
                rows: [],
                columns: [],
                noData: true,

                dispose: function() {
                    main.savePrefs();
                    _dispose();
                },
                savePrefs: function() {
                    localStorage.setItem('fitness_config', JSON.stringify(main.fitting.getConfig()));
                    fitness_ui.savePresets();
                },
                _makeDialogs: function(p) {
                },
                _setupFitting: function(p) {
                    let config = localStorage.getItem('fitness_config');
                    try {
                        config = JSON.parse(config);
                    } catch (err) {
                        config = undefined;
                    }
                    main.fitting = new fitness_ui.FitFrame(config);
                    p.main.appendChild(main.fitting.node);
                    let btn = main.fitting.controls.userControls
                        .append('button')
                        .on('click', () => {
                            main.opener.run()
                                .then(main.reader.open.bind(main.reader))
                                .then(main.setData)
                                .then(() => main.opener.reset())
                                .catch(qubx.reportErr);
                        });
                    btn.append('img')
                        .property({
                            src: "images/appbar.folder.open.svg",
                            alt: "[Load...]",
                            title: "Load"
                        });
                    btn = main.fitting.controls.userControls
                        .append('button')
                        .on('click', () => main.reader.edit(main.rows, main.columns)
                            .then(main.setData)
                            .catch(qubx.reportErr));
                    btn.append('img')
                        .property({
                            src: "images/appbar.page.corner.cells.svg",
                            alt: "[Edit...]",
                            title: "Edit"
                        });
                },
                setData: function(data) {
                    if ( main.noData ) {
                        main.noData = false;
                    }
                    main.rows = data.rows;
                    main.columns = data.columns;
                    const columnNames = main.columns.map(col => {
                        let name = col.name;
                        if ( ! name.substring(0, 1).match(/[A-Za-z_]/) ) {
                            name = '_'+name;
                        }
                        name = name.replace(/[^A-Za-z0-9_]/g, '_');
                        if ( name !== col.name ) {
                            col.rename = name;
                        }
                        return name;
                    });
                    main.fitting.context.clear();
                    main.fitting.context.setColumns(qubx_coldata.readColumns(main.rows, main.columns))
                        .then(() => main.fitting.ensureXY(columnNames))
                        .catch(qubx.reportErr);
                },
                _onDrop: function(event) {
                    event.stopPropagation();
                    event.preventDefault();
                    main.reader.open(event.dataTransfer.files)
                        .then(main.setData)
                        .catch(qubx.reportErr);
                    main.panels.then(p => p.main.classList.remove('qubx_dragtarget'));
                }
            });
            main.panels.then(p => {
                window.addEventListener("pagehide", main.savePrefs.bind(main));
                
                p.main.removeChild(p.main.querySelector("#qubx_controlbar"));
                p.main.removeChild(p.main.querySelector("#qubx_topbot"));
                var bannerLabel = p.banner.querySelector("#qubx_banner_lbl");
                bannerLabel.innerHTML = bannerLabel.innerHTML.replace(' Online', '/Fitness');
                main._makeDialogs(p);
                main._setupFitting(p);

                main.setData({
                    rows: qubx.range(-5, 5, .1).map(x => ({
                        X: x,
                        Y: 0
                    })),
                    columns: [{name: 'X'}, {name: 'Y'}]
                });
                setTimeout(() => main.fitting.context.setXY('X', 'Y'), 500); ///
                
                p.main.addEventListener('dragenter', qubx_ui.dragEnter(p.main));
                p.main.addEventListener('dragleave', qubx_ui.dragLeave(p.main));
                p.main.addEventListener('dragover', qubx.stopAndPrevent);
                p.main.addEventListener('drop', main._onDrop);

                return p;
            }).catch(qubx.reportErr);
        
            return main;
        };
        
        return exports;
    });

/// hash urlencode(expr, weight, xName, yName, url or text)
///  try iframe embedding a fitness figure
/// similar hash file,mode startup for qubx_demo?

/// toggle display as histogram?
/// some way to clear out mistaken or obsolete user presets
/// show correlation? is it working?

/// persistent data? up to N rows?
/// also persist xmeans data?

/// CVODE
/// Levenburg-Marquardt

