/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

/*
try {
    define;
} catch (e) {
    define = function(name, dep, run) {
        GLOBAL[name] = run();

        i = new qubx_idl.Idealization(.1, 10000, 1024);
        i.setDwells([100, 1024, 3000, 5000], [1023, 2999, 4999, 7999], [0, 1, 0, 1], 0);
        console.log(i);
        i.eachChunk(function(chunk) {
            console.log(chunk.data);
        });
        console.log(i.getDwellsFLC());
        console.log(i.getDwellsCD());
        i.eachDwellFLC(function(f,l,c) {
            console.log(f,l,c);
        });
        i.eachDwellFLC(function(f,l,c) {
            console.log(f,l,c);
        }, 6000, 9000, true);
    };
}
*/
define("qubx_idl", [], () => {
    'use strict';
    
    const exports = {};

    const CHUNKSIZE = 1<<14;

    function indexAt(offset, data, iMin, iMax) {
        iMax = iMax || (((data.length/2)|0) - 1);
        iMin = iMin | 0;
        while ( iMin <= iMax ) {
            const iMid = ((iMin+iMax)/2)|0,
                  v = data[2*iMid];
            if ( iMin === iMax ) {
                return (v <= offset) ? iMin : -1;
            }
            if ( v > offset ) {
                iMax = iMid - 1;
            } else if ( offset >= data[2*(iMid+1)] ) {
                iMin = iMid + 1;
            } else {
                return iMid;
            }
        }
        return -1;
    }

    // data alternating relativeOffset, classIndex
    class Chunk {
        constructor(offset, chunkSize) {
            this.offset = offset;
            this.n = chunkSize;
            this.data = [0, -1];
        }
        clear() {
            this.data = [0, -1];
        }
        _setData(insertion, offset, offsetPost) {
            if ( (! offset) && (offsetPost === this.n) ) {
                this.data = insertion;
            } else if ( ! this.data.length ) {
                if ( (! insertion.length) || insertion[0] ) {
                    insertion.splice(0, 0, 0, -1);
                }
                if ( offsetPost < this.n ) {
                    insertion.push(offsetPost);
                    insertion.push(-1);
                }
                this.data = insertion;
            } else {
                let iLeft = indexAt(offset, this.data),
                    iRight = indexAt(offsetPost, this.data);
                if ( offsetPost < this.n ) {
                    if ( iLeft === iRight ) {
                        insertion.push(offsetPost);
                        insertion.push(this.data[2*iRight+1]);
                        ++iRight;
                    } else {
                        this.data[2*iRight] = offsetPost;
                    }
                } else {
                    iRight = (this.data.length/2)|0;
                }
                if ( this.data[2*iLeft] < offset ) {
                    ++iLeft;
                }
                Array.prototype.splice.apply(this.data, [2*iLeft, 2*(iRight-iLeft)].concat(insertion));
            }
        }
        setDwells(firsts, lasts, classes, deltaOffset, iStart, iEnd) {
            // returns index after last completely consumed dwell
            // modifies firsts[rtn] if partially consumed
            deltaOffset = deltaOffset | 0;
            iStart = iStart | 0;
            iEnd = iEnd || counts.length;
            const insertion = [];
            var insertAt;
            let rtn = iEnd,
                i = iStart;
            for (; i<iEnd; ++i) {
                if ( (i > iStart) && (firsts[i] > (lasts[i-1]+1)) ) {
                    insertion.push(lasts[i-1]+1);
                    insertion.push(-1);
                }
                insertAt = Math.max(0, firsts[i] + deltaOffset);
                const insertPost = lasts[i] + deltaOffset + 1;
                insertion.push(insertAt);
                insertion.push(classes[i]);
                const n = insertPost - insertAt;
                if ( ! n ) {
                    console.log('109!');
                }
                const extras = insertAt + n - this.n;
                insertAt += n - Math.max(0, extras);
                if ( extras > 0 ) {
                    rtn = i;
                    break;
                } else if ( extras === 0 ) {
                    rtn = i+1;
                    break;
                }
            }
            this._setData(insertion, insertion[0], insertAt);
            return rtn;
        }
        setDwellsCounts(counts, classes, offset, iStart, iEnd) {
            offset = offset | 0;
            iStart = iStart | 0;
            iEnd = iEnd || counts.length;
            const insertion = [];
            let insertAt = offset,
                rtn = iEnd,
                i = iStart;
            for (; i<iEnd; ++i) {
                insertion.push(insertAt);
                insertion.push(classes[i]);
                let n = counts[i];
                const dn = insertAt + n - this.n;
                insertAt += n;
                if ( dn > 0 ) {
                    n -= dn;
                    counts[i] -= n;
                    rtn = i;
                    break;
                }
                if ( n === 0 ) {
                    rtn = i;
                    break;
                }
            }
            this._setData(insertion, offset, insertAt);
            return rtn;
        }
        countDwells(offset, offsetPost, withBlanks, lastCls) {
            offset = offset | 0;
            offsetPost = offsetPost || this.n;
            let i = indexAt(offset, this.data),
                i1 = indexAt(offsetPost-1, this.data),
                count = 0;
            if ( typeof(lastCls) === 'undefined' ) {
                lastCls = withBlanks ? -2 : -1;
            }
            while ( i <= i1 ) {
                const cls = this.data[2*i+1];
                if ( cls !== lastCls ) {
                    lastCls = cls;
                    if ( withBlanks || (cls >= 0) ) {
                        ++count;
                    }
                }
                ++i;
            }
            this.lastCls = lastCls;
            return count;
        }
        getDwellsFLC(firsts, lasts, classes, i0, offset, offsetPost, lastCls, deltaOffset, withBlanks) {
            i0 = i0 | 0;
            offset = offset | 0;
            offsetPost = offsetPost || this.n;
            if ( typeof(lastCls) === 'undefined' ) {
                lastCls = withBlanks ? -2 : -1;
            }
            deltaOffset = deltaOffset | 0;
            const iRight = indexAt(offsetPost-1, this.data),
                  atEnd = this.offset + offsetPost + deltaOffset;
            let iLeft = indexAt(offset, this.data),
                at = this.offset + offset + deltaOffset;
            while ( iLeft <= iRight ) {
                const cls = this.data[2*iLeft+1];
                if ( i0 && (withBlanks || (lastCls >= 0)) && (cls === lastCls) ) {
                    lasts[i0-1] = atEnd;
                } else {
                    if ( i0 && (withBlanks || (lastCls >= 0)) ) {
                        lasts[i0-1] = at-1;
                    }
                    if ( withBlanks || (cls >= 0) ) {
                        firsts[i0] = at;
                        lasts[i0] = atEnd;
                        classes[i0] = cls;
                        ++i0;
                    }
                    lastCls = cls;
                }
                ++iLeft;
                at = this.offset + this.data[2*iLeft] + deltaOffset;
            }
            this.lastCls = lastCls;
            return i0;
        }
    };
    
    exports.Idealization = class Idealization {
        constructor(sampling, count, chunkSize) {
            this._sampling = sampling || 1.0;
            this.count = count | 0;
            this.chunkSize = chunkSize || CHUNKSIZE;
            this.chunks = [];
            let i = 0;
            for (; i<this.count; i+=this.chunkSize) {
                this.chunks.push(new Chunk(i, Math.min(this.count-i, this.chunkSize)));
            }
        }
        sampling(x) {
            if ( typeof(x) === 'undefined' ) {
                return this._sampling;
            }
            this._sampling = +x;
            return this;
        }
        eachChunk(callback, first, last) {
            first = first | 0;
            if ( typeof(last) === 'undefined' ) {
                last = this.count-1;
            }
            let iFirst = (first / this.chunkSize)|0;
            const iLast = (last / this.chunkSize)|0;
            while ( iFirst <= iLast ) {
                const chunk = this.chunks[iFirst],
                      chunkOffset = Math.max(0, first-chunk.offset),
                      chunkOffsetPost = Math.min(last-chunk.offset, chunk.offset+chunk.n);
                callback(chunk, chunkOffset, chunkOffsetPost);
                ++iFirst;
            }
        }
        clear(first, last) {
            first = first | 0;
            if ( typeof(last) === 'undefined' ) {
                last = this.count-1;
            }
            this.eachChunk(chunk => {
                if ( (first <= chunk.offset) && ((chunk.offset + chunk.n) <= (last+1)) ) {
                    chunk.clear();
                } else {
                    const f = Math.max(first, chunk.offset),
                          l = Math.min(last, chunk.offset+chunk.n-1);
                    chunk.setDwellsCounts([l-f+1], [-1], f-chunk.offset);
                }
            }, first, last);
        }
        setDwells(firsts, lasts, classes, offset) {
            offset = offset | 0;
            const iEnd = firsts.length;
            let iStart = 0,
                iChunk = (offset/this.chunkSize)|0;
            while ( iStart < iEnd ) {
                const chunk = this.chunks[iChunk],
                      f = Math.max(0, firsts[iStart] + offset - chunk.offset);
                iStart = chunk.setDwells(firsts, lasts, classes, -chunk.offset, iStart, iEnd);
                ++iChunk;
            }
        }
        setDwellsCounts(counts, classes, offset) {
            offset = offset | 0;
            const iEnd = counts.length;
            let iStart = 0,
                iChunk = (offset/this.chunkSize)|0;
            offset -= this.chunks[iChunk].offset;
            if ( counts && counts.length && ! counts[0] ) {
                debugger;
            }
            while ( iStart < iEnd ) {
                iStart = this.chunks[iChunk].setDwellsCounts(counts, classes, offset, iStart, iEnd);
                ++iChunk;
                offset = 0;
            }
        }
        classAt(offset) {
            const chunk = this.chunks[(offset/this.chunkSize)|0];
            return chunk.data[2*indexAt(offset - chunk.offset, chunk.data)+1];
        }
        countDwells(first, last, withBlanks) {
            first = first | 0;
            if ( typeof(last) === 'undefined' ) {
                last = this.count - 1;
            }
            let n = 0,
                lastCls = withBlanks ? -2 : -1;
            this.eachChunk((chunk, chunkOffset, chunkOffsetPost) => {
                n += chunk.countDwells(chunkOffset, chunkOffsetPost, withBlanks, lastCls);
                lastCls = chunk.lastCls;
            }, first, last);
            return n;
        }
        getDwellsFLC(first, last, firsts, lasts, classes, offset, withBlanks) {
            if ( ! firsts ) {
                const n = this.countDwells(first, last, withBlanks);
                firsts = new Int32Array(n);
                lasts = new Int32Array(n);
                classes = new Int32Array(n);
            }
            offset = offset | 0;
            let i0 = 0,
                lastCls = withBlanks ? -2 : -1;
            this.eachChunk((chunk, chunkOffset, chunkOffsetPost) => {
                i0 = chunk.getDwellsFLC(firsts, lasts, classes, i0, chunkOffset, chunkOffsetPost, lastCls, offset, withBlanks);
                lastCls = chunk.lastCls;
            }, first, last);
            return {
                firsts: firsts,
                lasts: lasts,
                classes: classes
            };
        }
        getDwellsCD(first, last, classes, durations, sampling) {
            const n = this.countDwells(first, last, true);
            if ( ! classes ) {
                classes = new Int32Array(n);
                durations = new Float32Array(n);
            }
            const firsts = new Int32Array(classes.length),
                  lasts = new Int32Array(classes.length);
            this.getDwellsFLC(first, last, firsts, lasts, classes, 0, true);
            sampling = sampling || this._sampling;
            let i = 0;
            for (; i<n; ++i) {
                durations[i] = sampling * (lasts[i] - firsts[i] + 1);
            }
            return {
                classes: classes,
                durations: durations
            };
        }
        eachDwellFLC(callback, first, last, withBlanks) {
            const n = this.countDwells(first, last, withBlanks),
                  firsts = new Int32Array(n),
                  lasts = new Int32Array(n),
                  classes = new Int32Array(n);
            this.getDwellsFLC(first, last, firsts, lasts, classes, 0, withBlanks);
            let i = 0;
            for (; i<n; ++i) {
                callback(firsts[i], lasts[i], classes[i]);
            }
        }
    };

    exports.sampleDwells = function sampleDwells(idl, amp, std, first, last, y_res, los, his, UNSET) {
        if ( typeof(UNSET) === 'undefined' ) {
            UNSET = (function(a) {
                const buf = new ArrayBuffer(4),
                      v = new DataView(buf);
                v.setInt32(0, a);
                return v.getFloat32(0);
            })(0x2aaaaaaa);
        }
        let Nsample = los && los.length;
        if ( ! Nsample ) {
            Nsample = los;
            los = new Float32Array(Nsample);
        }
        if ( ! his ) {
            his = new Float32Array(Nsample);
        }
        let i = 0;
        for (; i<Nsample; ++i) {
            los[i] = his[i] = UNSET;
        }
        const Ndwell = idl.classes.length;
        if ( ! Ndwell ) {
            return {lo: los, hi: his};
        }
        const Np = (last - first + 1),
              spp = Nsample / Np;
        let sam = ((idl.firsts[0] - first) * spp)|0;
        if ( sam < 0 ) {
            sam = 0;
        }
        let p_next = (sam/spp + first)|0,
            lastAlo = UNSET,
            lastAhi = UNSET,
            iDwell = 0,
            ef = idl.firsts[iDwell],
            el = idl.lasts[iDwell],
            ec = idl.classes[iDwell];
        for (; sam<Nsample; ++sam) {
            const p = p_next;
            let lo = lastAlo,
                hi = lastAhi;
            p_next = ((sam+1)/spp + first)|0;
            while ( iDwell < Ndwell ) {
                if ( (ef > p) && (ef >= p_next) ) {
                    lastAlo = lastAhi = UNSET;
                    break; // starts after this sample
                } else if ( ec >= 0 ) {
                    const a = amp[ec],
                          s = std[ec];
                    lastAlo = a - s;
                    lastAhi = a + s;
                    if ( lo == UNSET ) {
                        lo = lastAlo;
                        hi = lastAhi;
                    } else {
                        if ( a < lo ) {
                            lo = lastAlo;
                        }
                        if ( a > hi ) {
                            hi = lastAhi;
                        }
                    }
                } else {
                    lastAlo = lastAhi = UNSET;
                }
                if ( el <= (p_next-1) ) {
                    ++iDwell;
                    if ( iDwell < Ndwell ) {
                        ef = idl.firsts[iDwell];
                        el = idl.lasts[iDwell];
                        ec = idl.classes[iDwell];
                    }
                }
                if ( el >= (p_next-1) ) {
                    break;
                }
            }
            if ( lo != UNSET ) {
                const d = hi - lo;
                if ( d < y_res ) {
                    lo = 0.5*(hi + lo - y_res)
                    hi = lo + y_res;
                }
                los[sam] = lo;
                his[sam] = hi;
            }
        }
        return {
            lo: los,
            hi: his
        };
    };
    
    exports.dwellsToTrace = function dwellsToTrace(idl, amp, first, last) {
        first = first|0;
        const ndwell = idl.classes.length;
        if ( typeof(last) === 'undefined' ) {
            last = idl.lasts[ndwell-1];
        }
        const n = last - first + 1,
              trace = new Float32Array(n);
        let a = 0,
            p = 0,
            v = -1,
            l = idl.firsts[0] - 1;
        while ( p < n ) {
            while ( (p < n) && ((p + first) <= l) ) {
                trace[p] = a;
                ++p;
            }
            ++v;
            if ( v <= ndwell ) {
                a = amp[idl.classes[v]];
                l = idl.lasts[v];
            } else {
                a = 0;
                l = last;
            }
        }
        return trace;
    };

    exports.BaselineNodes = class BaselineNodes {
        constructor(sampling, count, chunkSize) {
            this.idl = new exports.Idealization(sampling, count, chunkSize);
            this.levels = [];
        }
        sampling(x) {
            if ( typeof(x) === 'undefined' ) {
                return this.idl.sampling();
            }
            this.idl.sampling(x);
            return this;
        }
        clear(first, last) {
            this.idl.clear(first, last);
            this.levels = [];
        }
        setNodes(indices, levels, offset) {
            const c = this.levels.length,
                  classes = indices.map((f,i) => c+i);
            this.levels = this.levels.concat(levels);
            this.idl.setDwells(indices, indices, classes, offset);
        }
        updateNode(ix, lvl) {
            const c = this.idl.classAt(ix);
            if ( (c >= 0) && (c < this.levels.length) ) {
                this.levels[c] = lvl;
            }
        }
        clearNodes(first, last) {
            this.idl.clear(first, last);
        }
        countNodes(first, last) {
            return this.idl.countDwells(first, last);
        }
        getNodes(first, last, indices, levels, offset) {
            var n;
            if ( ! indices ) {
                n = this.countNodes(first, last);
                indices = new Int32Array(n);
                levels = new Float32Array(n);
            } else {
                n = indices.length;
            }
            const lasts = new Int32Array(n),
                  classes = new Int32Array(n),
                  dwells = this.idl.getDwellsFLC(first, last, indices, lasts, classes, offset, false),
                  allLevels = this.levels;
            let i = 0;
            for (; i<n; ++i) {
                levels[i] = allLevels[classes[i]];
            }
            return {
                indices: indices,
                levels: levels
            };
        }
        toSampled(first, last, samples) {
            first = first | 0;
            if ( typeof(last) === 'undefined' ) {
                last = this.idl.count - 1;
            }
            samples = samples || new Float32Array(last-first+1);
            const nodes = this.getNodes();
            if ( ! nodes.levels.length ) {
                return samples;
            }
            let i1 = 0;
            while ( (i1 < nodes.indices.length) && (nodes.indices[i1] <= first) ) {
                ++i1;
            }
            if ( i1 ) {
                --i1;
            }
            let i2 = i1;
            while ( (i2 < nodes.indices.length) && (nodes.indices[i2] <= last) ) {
                ++i2;
            }
            if ( i2 === nodes.indices.length ) {
                --i2;
            }
            var ind, ii;
            let lvl = nodes.levels[i1];
            if ( nodes.indices[i1] > first ) {
                ind = first;
                ii = i1;
            } else {
                ind = nodes.indices[i1];
                ii = i1 + 1;
            }
            let iOut = 0,
                iInSeg = first;
            for (; ii<=i2; ++ii) {
                const ind2 = nodes.indices[ii],
                      lvl2 = nodes.levels[ii],
                      m = (lvl2 - lvl) / (ind2 - ind),
                      nextI = iInSeg + Math.min(ind2, last) - Math.max(ind, iInSeg) + 1;
                for (; iInSeg<nextI; ++iInSeg, ++iOut) {
                    samples[iOut] = m*(iInSeg - ind)+lvl;
                }
                ind = ind2;
                lvl = lvl2;
            }
            for (; iOut<samples.length; ++iOut) {
                samples[iOut] = lvl;
            }
            return samples;
        }
    };
    
    return exports;
});
