/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_lib", ["asmffi", "random"], (asmffi, Random) => {
    'use strict';

    let totalMemory = 8*16*1024*1024;
    try {
        totalMemory = QUBX_LIB_MEMORY;
    } catch (err)
    {}
    
    return asmffi.load('qubx_lib_asm', {
        /* --- max_mac_ll.h -------------------- (mac) */
        
        "qubx_mac_data_create": ['void*', ['int', 'int']],
        "qubx_mac_data_free": ['void', ['void*']],
        "qubx_mac_data_reset": ['void', ['void*', 'int', 'int']],
        "qubx_mac_data_append_segment": ['int', ['void*']],
        "qubx_mac_data_append_chunk": ['void', ['void*', 'int', 'double', 'double', 'int',
                                                "float*", "float*", "int*",
                                                'void*'/*"int**"*/, 'void*'/*"float**"*/, 'void*'/*"double**"*/]],
        "qubx_mac_data_get_stimclasses": ["double*", ['void*', "int*"]],
        "qubx_mac_data_get_stimclass_frac": ["double*", ['void*']],
        "qubx_mac_data_get_buffer": ["float*", ['void*']],
        "qubx_mac_param_create": ['void*', ['int']],
        "qubx_mac_param_free": ['void', ['void*']],
        "qubx_mac_param_add_model": ['void', ['void*', 'void*', 'void*', 'void*']],
        "qubx_mac_setup_constraints": ['int', ['void*']],
        "qubx_mac_setup_nchannel": ['void', ['void*']],
        "qubx_mac_do_fPars_to_model": ['void', ['void*']],
        "qubx_mac_ll": ['int', ['void*']],
        "qubx_mac_calc_std": ['void', ['void*', "double*"]],

        /* --- max_subi_ll.h ----------------------------- (single) */

        'msl_multiplex_bounds': ['void', ['int', 'int', 'int**', 'int***', 'float***', 'double***', 'int*', 'int*', 'int*']],
        'msl_multiplex': ['void', ['int', 'int', 'int**', 'int***', 'float***', 'double***', 'double', 'int*', 'int**', 'float**', 'int*', 'int*', 'int*', 'double*']],
        'msl_accel_context_init': ['void*', ['int*']],
        'msl_accel_context_free': ['void', ['void*']],
        'msl_accel_data_init': ['void*', ['void*', 'int', 'int*', 'int**', 'float**', 'int', 'int', 'int*', 'int', 'double*']],
        'msl_accel_data_free': ['void', ['void*']],
        'msl_accel_data_get_ll': ['double*', ['void*']],
        'msl_accel_models_init': ['void*', ['void*', 'int', 'int', 'int']],
        'msl_accel_models_free': ['void', ['void*']],
        'msl_accel_models_reset': ['void', ['void*']],
        'msl_accel_models_setup_model': ['int', ['void*', 'void*', 'int', 'int*', 'double*', 'double**', 'double**', 'double**', 'int**', 'int**', 'int**', 'double']],
        'msl_accel_models_get_ll': ['double*', ['void*']],
        'msl_accel_ll': ['int', ['void*', 'void*', 'void*', 'int']],
        'mil_accel_ll': ['int', ['void*', 'void*', 'void*', 'int']],
        
        /* --- qub_adaptive_resample.h -------------------- (main, data) */
        
        "qub_adaptive_resample": ['int', ["float*", 'int', 'double', "float*", "float*", "int*", "int*", "int*"]],
        
        /* --- qub_doseresponse.h -------------------- (main) */
        
        "qub_doseresponse_equil_init": ['void*', ['int']],
        "qub_doseresponse_equil_free": ['void', ['void*']],
        "qub_doseresponse_equil_add_data": ['void', ['void*', 'int', "float*", "float*"]],
        "qub_doseresponse_equil_result": ['int', ['void*', 'int', "float*", "float*", "float*"]],
        "qub_doseresponse_peak_init": ['void*', ['double', 'double', 'double']],
        "qub_doseresponse_peak_free": ['void', ['void*']],
        "qub_doseresponse_peak_add_data": ['void', ['void*', 'int', "float*", "float*"]],
        "qub_doseresponse_peak_result": ['int', ['void*', 'int', "float*", "float*"]],
        
        /* --- qub_filter.h -------------------- (sim, ...?) */
        
        "qub_filterdataf": ['int', ["float*", "float*", 'int', 'float', 'float']],
        
        /* --- qub_raster_samples.h -------------------- (main) */
        
        "qub_raster_samples": ['int', ['int', 'int', "float*", 'double', 'double', "float*", "float*", "float*", "float*"]],
        
        /* --- qubx_amp.h -------------------- (data) */
        
        'qubx_amp': ['int', ['double', 'int', "float*", 'void*', 'int', 'int', "int*", "int*"]],
        
        /* --- qubx_amphist.h -------------------- (main) */
        
        'qubx_amphist_init': ['void*', ['int']],
        'qubx_amphist_free': ['void', ['void*']],
        'qubx_amphist_reset': ['void', ['void*', 'int']],
        'qubx_amphist_setNbin': ['void', ['void*', 'int']],
        'qubx_amphist_scan': ['void', ['void*', "float*", 'int']],
        'qubx_amphist_bin': ['void', ['void*', "float*"]],
        'qubx_amphist_add': ['void', ['void*', "float*", 'int']],
        'qubx_amphist_hist': ['void', ['void*', "float*"]],
        
        /* --- qubx_durhist.h -------------------- (main) */
        
        'qubx_durhist_binning': ['void', ['float', 'float', 'float', "float*", 'float', "double*"]],
        'qubx_durhist_calhist': ['int', ['float', 'int', 'int', "int*", "float*", 'int', "float*", "float*", 'float']],
        'qubx_durhist_calhist_smooth': ['int', ['int', 'int', "int*", "float*", 'int', "float*", "float*", 'float', 'float', 'float']],
        
        /* --- qubx_idealize_mdl.h -------------------- (data) */
        
        "qubx_idealize_mdl_context_create": ['void*', []],
        "qubx_idealize_mdl_context_free": ['void', ['void*']],
        "qubx_idealizer_mdl_create": ['void*', ['void*']],
        "qubx_idealizer_mdl_free": ['void', ['void*']],
        "qubx_idealizer_mdl_idealize": ['int', ['void*', "float*", 'int', 'int', 'int', 'int', 'float']],
        "qubx_idealizer_mdl_interrupt": ['void', ['void*']],
        "qubx_idealizer_mdl_get_dwells": ['int', ['void*', "int*", "float*", "float*", 'int']],
        "qubx_idealizer_mdl_get_dwells_fl": ['int', ['void*', 'int', "int*", "int*", "float*", "float*", 'int']],
        
        /* --- qubx_idealize_mdl_single.h -------------------- (data) */
        
        "qubx_idealize_mdl32_context_create": ['void*', []],
        "qubx_idealize_mdl32_context_free": ['void', ['void*']],
        "qubx_idealizer_mdl32_create": ['void*', ['void*']],
        "qubx_idealizer_mdl32_free": ['void', ['void*']],
        "qubx_idealizer_mdl32_idealize": ['int', ['void*', "float*", 'int', 'int', 'int', 'int', 'float']],
        "qubx_idealizer_mdl32_interrupt": ['void', ['void*']],
        "qubx_idealizer_mdl32_get_dwells": ['int', ['void*', "int*", "float*", "float*", 'int']],
        "qubx_idealizer_mdl32_get_dwells_fl": ['int', ['void*', 'int', "int*", "int*", "float*", "float*", 'int']],
        
        /* --- qubx_mac_dataprep.h -------------------- */

        // (mac)
        "qubx_mac_data_append_chunk_signals": ['void', ['void*', 'int', 'double', 'double', 'double',
                                                        'int', 'int', 'void*'/*"float**"*/,
                                                        'int', "float*", "float*"]],
        // (main)
        "qubx_mac_idlcls_create": ['void*', ['double']],
        "qubx_mac_idlcls_free": ['void', ['void*']],
        "qubx_mac_idlcls_add": ['int', ['void*', 'double', 'int', "float*", "double*", "double*", "int*"]],
        "qubx_mac_idlcls_read": ['int', ['void*', 'int', "double*", "double*"]],
        
        /* --- qubx_model.h -------------------- (all) */
        
        "qubx_model_create": ['void*', ['int', 'int', "int*", 'void*', 'void*']],
        "qubx_model_free": ['void', ['void*']],
        "qubx_model_to_console": ['void', ['void*']],
        "qubx_multi_model_create": ['void*', ['void*']],
        "qubx_multi_model_free": ['void', ['void*']],
        "qubx_multi_model_update": ['void', ['void*']],
        "qubx_constrained_create": ['void*', ['int']],
        "qubx_constrained_free": ['void', ['void*']],
        "qubx_constrained_setup": ['int', ['void*']],
        "qubx_constrained_free_to_pars": ['void', ['void*']],
        "qubx_constrained_pars_to_free": ['void', ['void*']],
        "qubx_constrained_calc_std": ['void', ['void*', "double*", "double*"]],
        "qubx_stim_amps_create": ['void*', ['void*', 'int', 'int', "double*", "double*"]],
        "qubx_stim_amps_free": ['void', ['void*']],
        "qubx_stim_amps_setup_scs": ['void', ['void*']],
        "qubx_stim_amps_pars_to_model": ['void', ['void*']],
        "qubx_stim_amps_model_to_pars": ['void', ['void*']],
        "qubx_stim_amps_fix_amp0": ['void', ['void*']],
        "qubx_stim_amps_fix_var0": ['void', ['void*']],
        "qubx_stim_amps_fix_amp": ['void', ['void*', 'int']],
        "qubx_stim_amps_fix_var": ['void', ['void*', 'int']],
        "qubx_stim_amps_calc_std": ['void', ['void*', "double*"]],
        "qubx_stim_rates_create": ['void*', ['void*', 'int', 'int', "double*", "double*", 'double', 'void*']],
        "qubx_stim_rates_free": ['void', ['void*']],
        "qubx_stim_rates_setup_QQ": ['void', ['void*']],
        "qubx_stim_rates_setup_AA": ['void', ['void*']],
        "qubx_stim_rates_setup_A": ['void', ['void*', 'int', 'double']],
        "qubx_stim_rates_fix_rate": ['void', ['void*', 'int', 'int']],
        "qubx_stim_rates_fix_exp": ['void', ['void*', 'int', 'int']],
        "qubx_stim_rates_fix_pres": ['void', ['void*', 'int', 'int']],
        "qubx_stim_rates_scale_rate": ['void', ['void*', 'int', 'int', 'int', 'int']],
        "qubx_stim_rates_scale_exp": ['void', ['void*', 'int', 'int', 'int', 'int']],
        "qubx_stim_rates_scale_pres": ['void', ['void*', 'int', 'int', 'int', 'int']],
        "qubx_stim_rates_balance_loop": ['void', ['void*', 'int', "int*"]],
        "qubx_stim_rates_balance_loops": ['void', ['void*']],
        "qubx_stim_rates_fix_all": ['void', ['void*']],
        "qubx_stim_rates_setup_constraints": ['int', ['void*']],
        "qubx_stim_rates_pars_to_model": ['void', ['void*']],
        "qubx_stim_rates_model_to_pars": ['void', ['void*']],
        "qubx_stim_rates_calc_std": ['void', ['void*', "double*"]],
        
        /* --- qubx_simulate.h -------------------- (sim) */
        
        "qubx_simulate1": ['int', ['void*', "int*", 'int', 'int', 'int', 'float',
                                   'double', 'int', "int*", 'void*'/*"float**"*/,
                                   'int', "float*", "int*", "int*", "double*", "double*",
                                   'int', 'int']],
        "qubx_simulate_stimulus_steps": ['void', ['int', "float*", 'double', 'int', "float*", "float*"]],
        "qubx_simulate_stimulus_ramps": ['void', ['int', "float*", 'double', 'int', "float*", "float*"]],
        
        /* --- qubx_skm.h -------------------- (single) */
        
        "qubx_skm_session_create": ['void*', ['double', 'int', 'void*']],
        "qubx_skm_session_free": ['void', ['void*']],
        "qubx_skm_session_reset": ['void', ['void*']],
        "qubx_skm_viterbi": ['void', ['void*', 'int', "float*", "int*", "double*"]],
        "qubx_skm_iterate": ['int', ['void*', 'int', "float*", "int*", 'int', 'int', "int*", "int*", "double*"]],
        "qubx_skm_Nstate": ['int', ['void*']],
        "qubx_skm_read_A": ['void', ['void*', "double*"]],
        "qubx_skm_Nclass": ['int', ['void*']],
        "qubx_skm_read_amp": ['void', ['void*', "double*", "double*"]],
        "qubx_skm_Nevent": ['int', ['void*']],
        "qubx_skm_read_events": ['void', ['void*', "int*", "int*"]],
        
        /* --- qubx_util.h -------------------- (all -- move em out if you don't want replicated) */
        
        "qubx_QtoPe": ['void', ["double*", "double*", 'int']],
        "qubx_QtoQe": ['void', ["double*", "int", "int*", "double", "double*"]],
        "qubx_expm": ['void', ['int', "double*", "double*", 'double']],
        "qubx_QtoQflux": ['void', ['int', "double*", "double*"]],
        "qubx_MeanFirstPassage": ['void', ['int', "double*", "double*", 'double']],
        "qubx_FirstLatencies": ['void', ["double*", "double*", 'double', 'int', 'double*']],
        "qubx_mil_pdf": ['int', ['int', "double*", "int*", 'int', 'double', 'int', "float*", "float*", "float*", "float*"]],

        /* --- ublas_matrixutil.h ------------------- */

        "sv_invm": ['int', ['double**', 'int', 'double**']]
        
    }, {
        // rename:

    }, { // Module:
        TOTAL_MEMORY: totalMemory
    }).then(lib => {
        
        /* --- max_mac_ll.h -------------------- */
        
        lib.MacData = class MacData{
            constructor(Nstimsig, dataCap, chunkCap) {
                this.stimsigCap = Math.max(2, Nstimsig+1);
                this.Nstimsig = Nstimsig;
                this.dataCap = dataCap;
                this.chunkCap = chunkCap;
                this.pSignals = lib.falloc2(this.stimsigCap, chunkCap);
                this.ptr = lib.qubx_mac_data_create(Nstimsig, dataCap);
                this.insertOffset = 0;
            }
            destroy() {
                lib.qubx_mac_data_free(this.ptr);
                lib.ffree2(this.pSignals);
            }
            reset(Nstimsig, dataCap, chunkCap) {
                if ( typeof(Nstimsig) === 'undefined' ) {
                    lib.qubx_mac_data_reset(this.ptr, this.Nstimsig, this.dataCap);
                    return;
                }
                lib.qubx_mac_data_reset(this.ptr, Nstimsig, dataCap);
                this.Nstimsig = Nstimsig;
                this.dataCap = dataCap;
                this.insertOffset = 0;
                let stale = false;
                if ( this.stimsigCap < (Nstimsig+1) ) {
                    this.stimsigCap = Nstimsig+1;
                    stale = true;
                }
                if ( this.chunkCap < chunkCap ) {
                    this.chunkCap = chunkCap;
                    stale = true;
                }
                if ( stale ) {
                    lib.ffree2(this.pSignals);
                    this.pSignals = lib.falloc2(this.stimsigCap, this.chunkCap);
                }
            }
            appendSegment() {
                return lib.qubx_mac_data_append_segment(this.ptr);
            }
            appendChunk(iSeg, dur_sec, t0, Npoint, T, I, idl) {
                const offset = this.insertOffset,
                      pT = lib.frow2(this.pSignals, 0),
                      pI = lib.frow2(this.pSignals, 1);
                T && lib.memcpy_in_float(pT, T, 0, Npoint);
                I && lib.memcpy_in_float(pI, I, 0, Npoint);
                lib.qubx_mac_data_append_chunk(this.ptr, iSeg, dur_sec, t0, Npoint, T?pT:0, I?pI:0, idl.dwellCounts.p, idl.classeses.p, idl.durationses.p, idl.ampses.p);
                this.insertOffset += Npoint;
                return offset;
            }
            appendChunkSignals(iSeg, dur_sec, t0, dt, signals, defaultValues, combineDeltas) {
                // signals: array of Float32Array
                // defaultValues: array[0..max(Ligand, Voltage, Pressure index)] of holding value
                // combineDeltas: array[0..signals.length-1] of delta y for stimulus signal idealization (using Max. Desc. Len. idl)
                // if typedarrays make sure they're all Float32Array
                // returns start offset in output buffer
                const offset = this.insertOffset,
                      Nsignal = signals.length,
                      Npoint = Nsignal ? signals[0].length : 0,
                      Nstimulus = defaultValues.length - 1,
                      pDefaults = lib.falloc(defaultValues.length),
                      pCombineDeltas = lib.falloc(combineDeltas.length);
                lib.memcpy_in_float(pDefaults, defaultValues, 0, defaultValues.length);
                lib.memcpy_in_float(pCombineDeltas, combineDeltas, 0, combineDeltas.length);
                signals.forEach((arr, i) => lib.memcpy_in_float(lib.frow2(this.pSignals,i), arr, 0, arr.length));
                lib.qubx_mac_data_append_chunk_signals(this.ptr, iSeg, dur_sec, t0, dt, Nsignal, Npoint, this.pSignals, Nstimulus, pDefaults, pCombineDeltas);
                lib.ffree(pCombineDeltas);
                lib.ffree(pDefaults);
                this.insertOffset += Npoint;
                return offset;
            }
            getStimclasses() {
                // the pointers defining levels in all stimclasses need to be available for mac_param model setup
                // sets this.Nstimclass, this.stimclasses, this.sc_frac
                // return raw stimclasses (ptr)
                const pNstimclass = lib.ialloc(1);
                this.stimclasses = lib.qubx_mac_data_get_stimclasses(this.ptr, pNstimclass);
                this.sc_frac = lib.qubx_mac_data_get_stimclass_frac(this.ptr);
                this.Nstimclass = lib.iget(pNstimclass, 0);
                lib.ifree(pNstimclass);
                return this.stimclasses;
            }
            readOutput(arr, offset, count) {
                // arr[3*count] interleaved mean, std, ll
                // now 4*count with added scaled_err = abs(data - mean)/std
                lib.memcpy_out_float(arr, 0, lib.qubx_mac_data_get_buffer(this.ptr)+offset*(4*4), 4*count);
            }
        };
        
        lib.MacLLCalc = class MacLLCalc {
            constructor(data, Nmodel) {
                this.ptr = lib.qubx_mac_param_create(Nmodel);
                lib.addStructAccess(this, this.ptr,
                                    [{name: 'storage', typ: '*'},
                                     {name: 'data', typ: 'i32'}, // settable pointer
                                     {name: 'useVar', typ: 'i32'},
                                     {name: 'optNchannel', typ: 'i32'},
                                     {name: 'accel', typ: 'i32'},
                                     {name: 'Npar', typ: 'i32'},
                                     {name: 'NfPar', typ: 'i32'},
                                     {name: 'pars', typ: 'double*'},
                                     {name: 'fPars', typ: 'double*'},
                                     {name: 'gpu_fail', typ: 'i32'},
                                     {name: 'll', typ: 'double'},
                                     {name: 'float_Nchannel', typ: 'double'}
                                    ]);
                this.data(data.ptr);
                this.accel(0); // no openmp or opencl available so pick the simpler code path
                this.models = [];
                this.pInvHessian = lib.dvec();
            }
            destroy(preserveModel) {
                lib.qubx_mac_param_free(this.ptr);
                this.models.forEach(m => {
                    if ( ! preserveModel ) {
                        m.model.destroy();
                    }
                    m.ampm.destroy();
                    m.ratesm.destroy();
                });
            }
            addModel(model, ampm, ratesm) {
                // takes ownership of all
                this.models.push({
                    model: model,
                    ampm: ampm,
                    ratesm: ratesm
                });
                lib.qubx_mac_param_add_model(this.ptr, model.ptr, ampm.ptr, ratesm.ptr);
                //this.float_Nchannel(model.Nchannel());
            }
            setupConstraints() {
                return lib.qubx_mac_setup_constraints(this.ptr);
            }
            setupNchannel() {
                lib.qubx_mac_setup_nchannel(this.ptr);
            }
            do_fParsToModel() {
                lib.qubx_mac_do_fPars_to_model(this.ptr);
            }
            run() {
                return lib.qubx_mac_ll(this.ptr);
            }
            calcStd(invHessian) {
                this.pInvHessian.set(invHessian);
                lib.qubx_mac_calc_std(this.ptr, this.pInvHessian.p);
            }
        };
        
        const IdlArrSignal = class IdlArrSignal {
            constructor() {
                this.n = 0;
                this.classes = lib.ivec();
                this.durations = lib.fvec();
                this.amps = lib.dvec();
                this.duration0 = 0;
            }
            destroy() {
                this.classes.destroy();
                this.durations.destroy();
                this.amps.destroy();
            }
            size(n) {
                if ( typeof(n) === 'undefined' ) {
                    return this.n;
                }
                this.classes.size(n);
                this.durations.size(n);
                this.n = n;
                return this;
            }
            ampCount(n) {
                return this.amps.size(n);
            }
            set(idl, delay) {
                idl.classes && this.classes.set(idl.classes);
                idl.durations && this.durations.set(idl.durations);
                this.n = this.classes.size();
                idl.amps && this.amps.set(idl.amps);
                this.duration0 = (idl.durations && idl.durations.length) ? idl.durations[0] : 0;
                delay && this.delay(delay);
            }
            delay(d) {
                this.durations.size() && this.durations.set(0, Math.max(0, this.duration0 + d));
            }
        };
        
        lib.IdlArrays = class IdlArrays {
            constructor(signals) {
                this.signals = [];
                this.classeses = lib.ivec();
                this.durationses = lib.ivec();
                this.ampses = lib.ivec();
                this.dwellCounts = lib.ivec();
                signals && this.set(signals);
            }
            destroy() {
                this.dwellCounts.destroy();
                this.ampses.destroy();
                this.durationses.destroy();
                this.classeses.destroy();
                this.signals.forEach(sig => sig.destroy());
            }
            size(n) {
                if ( typeof(n) === 'undefined' ) {
                    return this.signals.length;
                }
                var nDwells;
                if ( n.length ) {
                    nDwells = n;
                    n = nDwells.length;
                }
                this.classeses.size(n);
                this.durationses.size(n);
                this.ampses.size(n);
                this.dwellCounts.size(n);
                while ( this.signals.length < n ) {
                    this.signals.push(new IdlArrSignal());
                }
                while ( this.signals.length > n ) {
                    this.signals.pop().destroy();
                }
                if ( nDwells ) {
                    this.signals.forEach((sig, i) => {
                        sig.size(nDwells[i]);
                        this.classeses.set(i, sig.classes.p);
                        this.durationses.set(i, sig.durations.p);
                        this.ampses.set(i, sig.amps.p);
                        this.dwellCounts.set(i, sig.classes.size());
                    });
                }
            }
            set(signals) {
                this.size(signals.length);
                this.signals.forEach((sig, i) => {
                    sig.set(signals[i]);
                    this.classeses.set(i, sig.classes.p);
                    this.durationses.set(i, sig.durations.p);
                    this.ampses.set(i, sig.amps.p);
                    this.dwellCounts.set(i, sig.classes.size());
                });
            }
        };

        lib.IdlArraySegments = class IdlArraySegments {
            constructor(segments) {
                this.segments = [];
                this.classeseses = lib.pvec();
                this.durationseses = lib.pvec();
                this.ampseses = lib.pvec();
                this.dwellCountses = lib.pvec();
                segments && this.set(segments);
            }
            destroy() {
                this.classeseses.destroy();
                this.durationseses.destroy();
                this.ampseses.destroy();
                this.dwellCountses.destroy();
                this.segments.forEach(seg => seg.destroy());
            }
            size(nSeg, nSignal) {
                if ( typeof(nSeg) === 'undefined' ) {
                    return this.segments.length;
                }
                nSignal = nSignal || 1;
                this.classeseses.size(nSeg);
                this.durationseses.size(nSeg);
                this.ampseses.size(nSeg);
                this.dwellCountses.size(nSeg);
                while ( this.segments.length < nSeg ) {
                    this.segments.push(new lib.IdlArrays(nSignal));
                }
                while ( this.segments.length > nSeg ) {
                    this.segments.pop().destroy();
                }
            }
            set(segments) {
                this.size(segments.length, segments.length ? segments[0].signals.length : 0);
                this.segments.forEach((seg, i) => {
                    seg.set(segments[i].signals);
                    this.classeseses.set(i, seg.classeses.p);
                    this.durationseses.set(i, seg.durationses.p);
                    this.ampseses.set(i, seg.ampses.p);
                    this.dwellCountses.set(i, seg.dwellCounts.p);
                });
            }
        };

        
        
        /* --- max_subi_ll.h -------------------- */

        lib.MSL = class MSL{
            constructor(accel) {
                this.iParam0 = lib.ivec().size(1);
                this.iParam1 = lib.ivec().size(1);
                this.iParam0.set(0, accel);
                this.ctx = lib.msl_accel_context_init(this.iParam0.p);
                this.accel = this.iParam0.get(0);
            }
            destroy() {
                lib.msl_accel_context_free(this.ctx);
                this.iParam0.destroy();
                this.iParam1.destroy();
            }
            newData(idl, tdead_sec) {
                // idl: IdlArraySegments
                tdead_sec = tdead_sec || 0.0;
                const tdMS = tdead_sec * 1e3,
                      Nseg = idl.segments.length,
                      Nsignal = Nseg ? idl.segments[0].signals.length : 0,
                      pDwellCounts = lib.ivec().size(Nseg);
                lib.msl_multiplex_bounds(Nseg, Nsignal, idl.dwellCountses.p, idl.classeseses.p, idl.durationseses.p, idl.ampseses.p,
                                         pDwellCounts.p, this.iParam0.p, this.iParam1.p);
                let Nplex = this.iParam0.get(0),
                    Nstim = this.iParam1.get(0);
                const multiplexed = new lib.IdlArrays(); // note its "signals" are actually segments
                multiplexed.size(pDwellCounts.contents());
                
                const plexiclasses = lib.ivec().size(2*Nplex),
                      stimclasses = lib.dvec().size(Nstim*Nsignal);
                lib.msl_multiplex(Nseg, Nsignal, idl.dwellCountses.p, idl.classeseses.p, idl.durationseses.p, idl.ampseses.p,
                                  tdMS, pDwellCounts.p, multiplexed.classeses.p, multiplexed.durationses.p,
                                  this.iParam0.p, plexiclasses.p, this.iParam1.p, stimclasses.p);
                Nplex = this.iParam0.get(0);
                Nstim = this.iParam1.get(0);
                const obj = lib.msl_accel_data_init(this.ctx, Nseg, multiplexed.dwellCounts.p, multiplexed.classeses.p, multiplexed.durationses.p,
                                                  Nsignal, Nplex, plexiclasses.p, Nstim, stimclasses.p);
                let Ndwell = 0;
                pDwellCounts.contents().forEach(nd => {
                    Ndwell += nd;
                });
                multiplexed.destroy();
                pDwellCounts.destroy();
                return {
                    obj,
                    tdead_sec,
                    Nseg,
                    Nsignal,
                    Nplex,
                    Nstim,
                    Ndwell,
                    plexiclasses,
                    stimclasses,
                    
                    destroy: function() {
                        // could destroy these two at end of newData?
                        this.stimclasses.destroy();
                        this.plexiclasses.destroy();
                        
                        lib.msl_accel_data_free(this.obj);
                    },
                    getLL: function() {
                        const n = Nseg * this.Nmodel,
                              ll = [],
                              pLL = lib.msl_accel_data_get_ll(this.obj);
                        let i = 0;
                        for (; i<n; ++i) {
                            ll.push(lib.dget(pLL, i));
                        }
                        return ll;
                    }
                };
            }
            newModels(dim, Nmodel, Nsc) {
                return {
                    obj: lib.msl_accel_models_init(this.ctx, dim, Nmodel, Nsc),
                    dim: dim,
                    Nmodel: Nmodel,
                    Nsc: Nsc,
                    destroy: function() {
                        lib.msl_accel_models_free(this.obj);
                    },
                    reset: function() {
                        lib.msl_accel_models_reset(this.obj);
                    },
                    setupModel: function(model, data) {
                        data.Nmodel = Nmodel;
                        return lib.msl_accel_models_setup_model(this.obj, data.obj, model.Nstate(),
                                                                lib.pget(model.clazz.ptr), 0, lib.pget(model.K0.ptr), lib.pget(model.K1.ptr), lib.pget(model.K2.ptr),
                                                                lib.pget(model.L.ptr), lib.pget(model.V.ptr), lib.pget(model.P.ptr), data.tdead_sec);
                    },
                    getLL: function() {
                        const pLL = lib.msl_accel_models_get_ll(this.obj),
                              ll = [];
                        let i = 0;
                        for (; i<this.Nmodel; ++i) {
                            ll.push(lib.dget(pLL, i));
                        }
                        return ll;
                    }
                };
            }
            mil(data, models, accel) {
                accel = (typeof(accel) === 'undefined') ? lib.MSL.ACCEL_OPENMP : (accel|0);
                return lib.mil_accel_ll(this.ctx, data.obj, models.obj, accel);
            }
            msl(data, models, accel) {
                accel = (typeof(accel) === 'undefined') ? lib.MSL.ACCEL_OPENMP : (accel|0);
                return lib.msl_accel_ll(this.ctx, data.obj, models.obj, accel);
            }
        };
        lib.MSL.ACCEL_NONE = 0;
        lib.MSL.ACCEL_OPENMP = 1;
        lib.MSL.ACCEL_OPENCL = 2;
        
        /* --- qub_adaptive_resample.h -------------------- */
        
        lib.AdaptiveResampler = class AdaptiveResampler {
            constructor(sampling) {
                this.sampling = sampling;
                this.pData = lib.fvec();
                this.pMean = lib.fvec();
                this.pStd = lib.fvec();
                this.pFirst = lib.ivec();
                this.pLast = lib.ivec();
                this.pClosest = lib.ivec();
            }
            destroy() {
                this.pData.destroy();
                this.pMean.destroy();
                this.pStd.destroy();
                this.pFirst.destroy();
                this.pLast.destroy();
                this.pClosest.destroy();
            }
            resample(data, maxStd) {
                this.pData.set(data);
                [this.pMean, this.pStd, this.pFirst, this.pLast, this.pClosest].forEach(buf => buf.size(data.length));
                const n = lib.qub_adaptive_resample(this.pData.p, data.length, maxStd,
                                                    this.pMean.p, this.pStd.p, this.pFirst.p, this.pLast.p, this.pClosest.p);
                const closests = this.pClosest.contents(n);
                let T = [];
                if ( this.sampling ) {
                    T = new Float32Array(n);
                    let i = 0;
                    for (; i<n; ++i) {
                        T[i] = closests[i]*this.sampling;
                    }
                }
                return {
                    n: n,
                    means: this.pMean.contents(n),
                    stds: this.pStd.contents(n),
                    firsts: this.pFirst.contents(n),
                    lasts: this.pLast.contents(n),
                    closests: closests,
                    T: T
                };
            }
        };
        
        /* --- qub_doseresponse.h -------------------- */
        
        lib.DoseResponseEquil = class DoseResponseEquil {
            constructor(Ntail, dy, maxN, maxResults) {
                if ( maxN < maxResults ) {
                    console.log("maxN must be >= maxResults; adjusting...");
                    maxN = maxResults;
                }
                this.Ntail = Ntail;
                this.dy = dy;
                this.maxN = maxN;
                this.maxResults = maxResults;
                this.bufDose = lib.fvec();
                this.bufResponse = lib.fvec();
                this.bufDose.size(this.maxN);
                this.bufResponse.size(this.maxN);
                this.bufStd = lib.fvec();
                this.bufStd.size(this.maxResults);
                this.obj = lib.qub_doseresponse_equil_init(Ntail, dy);
            }
            destroy() {
                lib.qub_doseresponse_equil_free(this.obj);
                this.bufDose.destroy();
                this.bufResponse.destroy();
                this.bufStd.destroy();
            }
            add(dose, response, offset, n) {
                offset = offset || 0;
                n = n || dose.length;
                if ( n > this.maxN ) {
                    console.log("N > maxN in DoseResponseEquil");
                    n = this.maxN;
                }
                lib.memcpy_in_float(this.bufDose.p, dose, offset, n);
                lib.memcpy_in_float(this.bufResponse.p, response, offset, n);
                lib.qub_doseresponse_equil_add_data(this.obj, n, this.bufDose.p, this.bufResponse.p);
            }
            getResult() {
                const count = lib.qub_doseresponse_equil_result(this.obj, this.maxResults, this.bufDose.p, this.bufResponse.p, this.bufStd.p);
                return {
                    dose: this.bufDose.contents(count),
                    mean: this.bufResponse.contents(count),
                    std: this.bufStd.contents(count)
                };
            }
        };
        
        lib.DoseResponsePeak = class DoseResponsePeak {
            constructor(sampling, filter_kHz, sign, maxN, maxResults) {
                if ( maxN < maxResults ) {
                    console.log("maxN must be >= maxResults; adjusting...");
                    maxN = maxResults;
                }
                this.sampling = sampling;
                this.filter_kHz = filter_kHz;
                this.sign = sign;
                this.maxN = maxN;
                this.maxResults = maxResults;
                this.bufDose = lib.fvec();
                this.bufDose.size(this.maxN);
                this.bufResponse = lib.fvec();
                this.bufResponse.size(this.maxN);
                this.obj = lib.qub_doseresponse_peak_init(sampling, filter_kHz, sign);
            }
            destroy() {
                lib.qub_doseresponse_peak_free(this.obj);
                this.bufResponse.destroy();
                this.bufDose.destroy();
            }
            add(dose, response, offset, n) {
                offset = offset || 0;
                n = n || dose.length;
                if ( n > this.maxN ) {
                    console.log("N > maxN in DoseResponsePeak");
                    n = this.maxN;
                }
                lib.memcpy_in_float(this.bufDose.p, dose, offset, n);
                lib.memcpy_in_float(this.bufResponse.p, response, offset, n);
                lib.qub_doseresponse_peak_add_data(this.obj, n, this.bufDose.p, this.bufResponse.p);
            }
            getResult() {
                const count = lib.qub_doseresponse_peak_result(this.obj, this.maxResults, this.bufDose.p, this.bufResponse.p);
                return {
                    dose: this.bufDose.contents(count),
                    response: this.bufResponse.contents(count)
                };
            }
        };
        
        /* --- qub_filter.h -------------------- */
        const filterIn = lib.fvec();
        const filterOut = lib.fvec();
        lib.filter = function filter(arr, dest, sampling, freqHz) {
            dest = dest || arr;
            const n = arr.length;
            filterIn.set(arr);
            filterOut.size(n);
            lib.qub_filterdataf(filterIn.p, filterOut.p, n, sampling, freqHz);
            return filterOut.contents(n, dest);
        };
        
        /* --- qub_raster_samples.h -------------------- */
        
        lib.raster_samples = function raster_samples(arrayBuf, offset, count, w, spp, res_y) {
            // allocate on emscripten heap:
            const heapData = lib.falloc(count);
            lib.memcpy_in(heapData, arrayBuf, 4*offset, 4*count);
            const heapLo = lib.falloc(w),
                  heapHi = lib.falloc(w),
                  heapGlo = lib.falloc(w),
                  heapGhi = lib.falloc(w),
                  n = lib.qub_raster_samples(w, count, heapData, spp, res_y, heapLo, heapHi, heapGlo, heapGhi),
                  lo = new ArrayBuffer(4*n),
                  hi = new ArrayBuffer(4*n),
                  glo = new ArrayBuffer(4*n),
                  ghi = new ArrayBuffer(4*n);
            lib.memcpy_out(lo, 0, heapLo, 4*n);
            lib.memcpy_out(hi, 0, heapHi, 4*n);
            lib.memcpy_out(glo, 0, heapGlo, 4*n);
            lib.memcpy_out(ghi, 0, heapGhi, 4*n);
            lib.ffree(heapData);
            lib.ffree(heapLo);
            lib.ffree(heapHi);
            lib.ffree(heapGlo);
            lib.ffree(heapGhi);
            return {
                lows: new Float32Array(lo),
                highs: new Float32Array(hi),
                gauss_lows: new Float32Array(glo),
                gauss_highs: new Float32Array(ghi)
            };
        };
        
        /* --- qubx_amp.h -------------------- */
        
        const AMP = {
            data: lib.fvec(),
            classes: lib.ivec(),
            counts: lib.ivec()
        };
        lib.AMP = function(sampling, data, model, autoInit, autoInitUp) {
            const n = data.length;
            AMP.data.set(data);
            AMP.classes.size(n);
            AMP.counts.size(n);
            const pModel = lib.modelFromJson(model),
                  Ndwell = lib.qubx_amp(sampling, n, AMP.data.p, pModel.ptr, autoInit|0, autoInitUp|0, AMP.classes.p, AMP.counts.p),
                  Nclass = model.classes.length,
                  amps = new Float64Array(Nclass),
                  stds = new Float64Array(Nclass);
            let i = 0;
            for (; i<Nclass; ++i) {
                amps[i] = pModel.amp(i);
                stds[i] = Math.sqrt(pModel['var'](i));
            }
            pModel.destroy();
            return {
                amp: amps,
                std: stds,
                classes: AMP.classes.contents(Ndwell),
                counts: AMP.counts.contents(Ndwell)
            };
        };
        
        /* --- qubx_amphist.h ------------------- */
        
        lib.AmpHistMaker = class AmpHistMaker {
            constructor(Nbin) {
                this.Nbin = Nbin || 64;
                this.obj = lib.qubx_amphist_init(this.Nbin);
                this.pData = lib.fvec();
                this.pBins = lib.fvec()
                    .size(this.Nbin);
            }
            destroy() {
                lib.qubx_amphist_free(this.obj);
            }
            reset(countsOnly) {
                lib.qubx_amphist_reset(this.obj, countsOnly|0);
            }
            setNbin(Nbin) {
                this.Nbin = Nbin || 64;
                this.pBins.size(this.Nbin);
                lib.qubx_amphist_setNbin(this.obj, this.Nbin);
            }
            scan(data) {
                this.pData.set(data);
                lib.qubx_amphist_scan(this.obj, this.pData.p, data.length);
            }
            bin(arr) {
                lib.qubx_amphist_bin(this.obj, this.pBins.p);
                return this.pBins.contents(this.Nbin, arr);
            }
            add(data) {
                this.pData.set(data);
                lib.qubx_amphist_add(this.obj, this.pData.p, data.length);
            }
            hist(arr) {
                lib.qubx_amphist_hist(this.obj, this.pBins.p);
                return this.pBins.contents(this.Nbin, arr);
            }
        };
        
        /* --- qubx_durhist.h -------------------- */
        
        lib.DurHistMaker = class DurHistMaker {
            constructor(maxNdwell, maxNbin) {
                maxNdwell = maxNdwell || 1024;
                this.maxNbin = maxNbin || 128;
                this.pDurations = lib.fvec()
                    .size(maxNdwell);
                this.pClasses = lib.ivec()
                    .size(maxNdwell);
                this.pBins = lib.fvec()
                    .size(this.maxNbin);
                this.pHist = [];
                this.pR = lib.dvec()
                    .size(1);
                this.pN = lib.ivec()
                    .size(1);
                this.pQ = lib.dvec();
                this.pClazz = lib.ivec();
                this.pPDF = lib.fvec();
                this.pTau = lib.fvec();
                this.pTauAmp = lib.fvec();
            }
            destroy() {
                this.pTauAmp.destroy();
                this.pTau.destroy();
                this.pPDF.destroy();
                this.pClazz.destroy();
                this.pQ.destroy();
                this.pN.destroy();
                this.pR.destroy();
                this.pHist.forEach(h => h.destroy());
                this.pBins.destroy();
                this.pClasses.destroy();
                this.pDurations.destroy();
            }
            makeBins(nbin, tmin, tmax, tdead, ncls) {
                this.pBins.size(nbin);
                this.pN.set(0, nbin);
                lib.qubx_durhist_binning(0, tmin-tdead, tmax-tdead, this.pN.p, this.pBins.p, tdead||0.0, this.pR.p);
                this.r = this.pR.get(0);
                this.nbin = this.pN.get(0);
                this.pHist.forEach(h => h.size(this.nbin));
                this.clear(ncls);
                this.bins = this.pBins.contents(this.nbin);
                const bb = this.bins;
                let i = 0;
                for (; i<nbin; ++i) {
                    bb[i] = (bb[i] > 0.0) ? Math.log10(bb[i]) : -5;
                }
                return this.bins;
            }
            clear(ncls) {
                ncls = ncls || this.pHist.length;
                this.Nevent = [];
                let c = 0;
                for (; c<ncls; ++c) {
                    this.Nevent[c] = 0;
                    let pHist = this.pHist[c];
                    if ( ! pHist ) {
                        pHist = lib.fvec()
                            .size(this.maxNbin);
                        this.pHist[c] = pHist;
                    }
                    const n = this.nbin;
                    let i = 0;
                    for (; i<n; ++i) {
                        pHist.set(i, 0.0);
                    }
                }
            }
            add(cls, classes, durations, count, tdead, sampling, smooth) {
                this.pClasses.set(classes, count);
                this.pDurations.set(durations, count);
                const addOne = c => {
                    if ( smooth ) {
                        this.Nevent[c] += lib.qubx_durhist_calhist_smooth(c, count, this.pClasses.p, this.pDurations.p,
                                                                          this.nbin, this.pBins.p, this.pHist[c].p, tdead, sampling, this.r);
                    } else {
                        this.Nevent[c] += lib.qubx_durhist_calhist(0.0, c, count, this.pClasses.p, this.pDurations.p,
                                                                   this.nbin, this.pBins.p, this.pHist[c].p, tdead);
                    }
                };
                if ( cls < 0 ) {
                    const ncls = -cls;
                    let i = 0;
                    for (; i<ncls; ++i) {
                        addOne(i);
                    }
                } else {
                    addOne(cls);
                }
                return this;
            }
            read(cls, arr) {
                const hist = this.pHist[cls].contents(this.nbin, arr),
                      scale = this.Nevent[cls] ? (1.0 / this.Nevent[cls]) : 1.0,
                      n = hist.length;
                let i = 0;
                for (; i<n; ++i) {
                    hist[i] *= scale;
                }
                return hist;
            }
            pdf(Nstate, Q, clazz, cls, tdMS) {
                this.pQ.size(Nstate*Nstate);
                let i = 0,
                    p = 0;
                for (; i<Nstate; ++i) {
                    let j = 0;
                    for (; j<Nstate; ++j, ++p) {
                        this.pQ.set(p, Q[i][j]);
                    }
                }
                this.pClazz.set(clazz);
                this.pPDF.size((Nstate+1)*this.nbin);
                this.pTau.size(Nstate);
                this.pTauAmp.size(Nstate);
                const na = lib.qubx_mil_pdf(Nstate, this.pQ.p, this.pClazz.p, cls, tdMS, this.nbin, this.pBins.p, this.pPDF.p, this.pTau.p, this.pTauAmp.p),
                      components = [];
                for (i=0; i<na; ++i) {
                    components[i] = this.pPDF.contents(this.nbin, new Float32Array(this.nbin), (i+1)*this.nbin);
                }
                return {
                    pdf: this.pPDF.contents(this.nbin),
                    components: components,
                    tau: this.pTau.contents(na),
                    amp: this.pTauAmp.contents(na)
                };
            }
        };
        
        /* --- qubx_idealize_mdl.h -------------------- */
        
        lib.MDLIdealizer = class MDLIdealizer {
            constructor(capacity, config) {
                this.config = {
                    Lmin: config.Lmin || 3,
                    split3: (typeof(config.split3)!=='undefined') ? config.split3 : true,
                    combineClose: (typeof(config.combineClose)!=='undefined') ? config.combineClose : false,
                    combineDelta: (typeof(config.combineDelta)!=='undefined') ? config.combineDelta : 1e-3,
                    singlePrecision: (typeof(config.singlePrecision)!=='undefined') ? config.singlePrecision : false,
                };
                this.pData = lib.fvec();
                this.pCounts = lib.ivec();
                this.pAmps = lib.fvec();
                this.pStds = lib.fvec();
                this.capacity = capacity;
                this.pData.size(capacity);
                this.pCounts.size(capacity);
                this.pAmps.size(capacity);
                this.pStds.size(capacity);
                if ( this.config.singlePrecision ) {
                    this.ctx = lib.qubx_idealize_mdl32_context_create();
                    this.idl = lib.qubx_idealizer_mdl32_create(this.ctx);
                } else {
                    this.ctx = lib.qubx_idealize_mdl_context_create();
                    this.idl = lib.qubx_idealizer_mdl_create(this.ctx);
                }
            }
            destroy() {
                if ( this.config.singlePrecision ) {
                    lib.qubx_idealizer_mdl32_free(this.idl);
                    lib.qubx_idealize_mdl32_context_free(this.ctx);
                } else {
                    lib.qubx_idealizer_mdl_free(this.idl);
                    lib.qubx_idealize_mdl_context_free(this.ctx);
                }
                this.pStds.destroy();
                this.pAmps.destroy();
                this.pCounts.destroy();
                this.pData.destroy();
            }
            idealize(arr, sampling) {
                if ( arr.length > this.capacity ) {
                    console.log('MDL data exceeding capacity');
                }
                if ( arr.length === 0 ) {
                    return {
                        classes: [],
                        durations: [],
                        amps: [],
                        stds: []
                    };
                }
                sampling = sampling || 1;
                this.pData.set(arr);
                const Ndwell = this.singlePrecision ?
                    lib.qubx_idealizer_mdl32_idealize(this.idl, this.pData.p, arr.length, this.config.Lmin, this.config.split3, this.config.combineClose, this.config.combineDelta) :
                    lib.qubx_idealizer_mdl_idealize(this.idl, this.pData.p, arr.length, this.config.Lmin, this.config.split3, this.config.combineClose, this.config.combineDelta);
                this.pCounts.size(Ndwell);
                this.pAmps.size(Ndwell);
                this.pStds.size(Ndwell);
                if ( this.singlePrecision ) {
                    lib.qubx_idealizer_mdl32_get_dwells(this.idl, this.pCounts.p, this.pAmps.p, this.pStds.p, Ndwell);
                } else {
                    lib.qubx_idealizer_mdl_get_dwells(this.idl, this.pCounts.p, this.pAmps.p, this.pStds.p, Ndwell);
                }
                const classes = new Int32Array(Ndwell),
                      durations = new Float32Array(Ndwell);
                let i = 0;
                for (; i<Ndwell; ++i) {
                    classes[i] = i;
                    durations[i] = sampling * this.pCounts.get(i);
                }
                return {
                    classes: classes,
                    durations: durations,
                    amps: this.pAmps.contents(),
                    stds: this.pStds.contents()
                };
            }
        };
        
        /* --- qubx_mac_dataprep.h -------------------- */
        
        lib.IdlClassifier = class IdlClassifier {
            constructor(delta) {
                this.obj = lib.qubx_mac_idlcls_create(delta);
                this.durations = lib.fvec();
                this.amps = lib.dvec();
                this.stds = lib.dvec();
                this.classes = lib.ivec();
                this.Ncls = 0;
            }
            destroy() {
                this.classes.destroy();
                this.stds.destroy();
                this.amps.destroy();
                this.durations.destroy();
                lib.qubx_mac_idlcls_free(this.obj);
            }
            add(sampling, classes, durations, amps, stds) {
                const Ndwell = durations.length;
                this.durations.size(Ndwell);
                this.amps.size(Ndwell);
                this.stds.size(Ndwell);
                this.classes.size(Ndwell);
                let i = 0;
                for (; i<Ndwell; ++i) {
                    this.durations.set(i, durations[i]);
                    this.amps.set(i, amps[classes[i]]);
                    this.stds.set(i, stds[classes[i]]);
                }
                this.Ncls = lib.qubx_mac_idlcls_add(this.obj, sampling, Ndwell, this.durations.p, this.amps.p, this.stds.p, this.classes.p);
                const classesOut = [];
                for (i=0; i<Ndwell; ++i) {
                    classesOut[i] = this.classes.get(i);
                }
                return classesOut;
            }
            read() {
                this.amps.size(this.Ncls);
                this.stds.size(this.Ncls);
                lib.qubx_mac_idlcls_read(this.obj, this.Ncls, this.amps.p, this.stds.p);
                const amps = [],
                      stds = [],
                      n = this.Ncls;
                let i = 0;
                for (; i<n; ++i) {
                    amps[i] = this.amps.get(i);
                    stds[i] = this.stds.get(i);
                }
                return {
                    amps: amps,
                    stds: stds
                };
            }
        };
        
        /* --- qubx_model.h -------------------- */

        const setupModelStruct = function setupModelStruct(dest, ptr) {
            lib.addStructAccess(dest, ptr, [
                {name: 'version', typ: 'i32'},
                {name: 'storage', typ: '*'},
                {name: 'Nstate', typ: 'i32'},
                {name: 'Nclass', typ: 'i32'},
                {name: 'clazz', typ: 'i32*'},
                {name: 'usePeq', typ: 'i32'},
                {name: 'IeqFv', typ: 'i32'},
                {name: 'Nchannel', typ: 'i32'},
                {name: 'iVoltage', typ: 'i32'},
                {name: 'r1', typ: 'i32'},
                {name: 'Gleak', typ: 'double'},
                {name: 'Vleak', typ: 'double'},
                {name: 'Vrev', typ: 'double'},
                {name: 'amp0', typ: 'double'},
                {name: 'var0', typ: 'double'},
                {name: 'amp', typ: 'double*'},
                {name: 'var', typ: 'double*'},
                {name: 'P0', typ: 'double*'},
                {name: 'K0', typ: 'double**'},
                {name: 'K1', typ: 'double**'},
                {name: 'L', typ: 'i32**'},
                {name: 'V', typ: 'i32**'},
                {name: 'stopFlag', typ: 'i32'},
                {name: 'progress', typ: 'i32'},//0 to 100
                {name: 'r2', typ: 'i32'},
                {name: 'd_Nchannel', typ: 'double'},
                {name: 'd_amp0', typ: 'double'},
                {name: 'd_var0', typ: 'double'},
                {name: 'd_amp', typ: 'double*'},
                {name: 'd_var', typ: 'double*'},
                {name: 'd_K0', typ: 'double**'},
                {name: 'd_K1', typ: 'double**'},
                {name: 'K2', typ: 'double**'},
                {name: 'd_K2', typ: 'double**'},
                {name: 'P', typ: 'i32**'}
            ]);
            return dest;
        };
        
        lib.Model = class Model {
            constructor(Nstate, Nclass) {
                this.ptr = lib.qubx_model_create(Nstate, Nclass, 0, 0, 0);
                this.signalIndex = {'Current' : 0};
                setupModelStruct(this, this.ptr);
                this.usePeq(1);
            }
            destroy() {
                lib.qubx_model_free(this.ptr);
            }
        };

        lib.MultiModel = class MultiModel{
            constructor(model) {
                this.ptr = lib.qubx_multi_model_create(model.ptr);
                lib.addStructAccess(this, this.ptr,
                                    [{name: 'version', typ: 'i32'},
                                     {name: 'storage', typ: '*'},
                                     {name: 'single_p', typ: '*'},
                                     {name: 'multi_p', typ: '*'},
                                     {name: 'Nstate', typ: 'i32'},
                                     {name: 'Npath', typ: 'i32'},
                                     {name: 'Nclass', typ: 'i32'},
                                     {name: 'NclassOrder', typ: 'i32'},
                                     {name: 'NpathSingle', typ: 'i32'},
                                     {name: 'classOrder', typ: 'i32*'},
                                     {name: 'cmutistate', typ: 'i32*'},
                                     {name: 'path', typ: 'i32*'},
                                     {name: 'mutistate', typ: 'i32*'},
                                     {name: 'mutipath', typ: 'i32*'}
                                    ]);
                this.single = model;
                this.multi = setupModelStruct({}, this.multi_p());
                this.multi.ptr = this.multi_p;
            }
            destroy() {
                lib.qubx_multi_model_free(this.ptr);
            }
            update() {
                lib.qubx_multi_model_update(this.ptr);
            }
        };
        
        lib.Constrained = class Constrained {
            constructor(Npar, ptr, ownPtr) {
                if ( ptr ) {
                    this.ptr = ptr;
                    this.ownPtr = ownPtr;
                } else {
                    this.ptr = lib.qubx_constrained_create(Npar);
                    this.ownPtr = true;
                }
                lib.addStructAccess(this, this.ptr, [
                    {name: 'version', typ: 'i32'},
                    {name: 'storage', typ: '*'},
                    {name: 'Npar', typ: 'i32'},
                    {name: 'NparR', typ: 'i32'},
                    {name: 'NfPar', typ: 'i32'},
                    {name: 'Ncns', typ: 'i32'},
                    {name: 'NcnsR', typ: 'i32'},
                    {name: 'Ain', typ: 'double**'},
                    {name: 'Bin', typ: 'double*'},
                    {name: 'pars', typ: 'double*'},
                    {name: 'fPars', typ: 'double*'},
                    {name: 'Acns', typ: 'double**'},
                    {name: 'Ascal', typ: 'double**'},
                    {name: 'Bcns', typ: 'double*'},
                    {name: 'Bscal', typ: 'double*'}
                ]);
            }
            destroy() {
                if ( this.ownPtr ) {
                    lib.qubx_constrained_free(this.ptr);
                }
            }
            setup() {
                return lib.qubx_constrained_setup(this.ptr);
            }
            freeToPars() {
                lib.qubx_constrained_free_to_pars(this.ptr);
            }
            parsToFree() {
                lib.qubx_constrained_pars_to_free(this.ptr);
            }
            calcStd(pInvHessian) {
                const N = this.Npar(),
                      dPar = lib.falloc(N),
                      result = new Float64Array(N);
                lib.qubx_constrained_calc_std(this.ptr, pInvHessian, dPar);
                lib.memcpy_out_double(result, 0, dPar, N);
                lib.ffree(dPar);
                return result;
            }
        };
        
        lib.StimAmps = class StimAmps {
            constructor(model, Nstim, Nsig, pStimclasses, pScFrac) {
                this.ptr = lib.qubx_stim_amps_create(model.ptr, Nstim, Nsig, pStimclasses, pScFrac);
                lib.addStructAccess(this, this.ptr, [
                    {name: 'version', typ: 'i32'},
                    {name: 'storage', typ: '*'},
                    {name: 'model', typ: '*'},
                    {name: 'Nstim', typ: 'i32'},
                    {name: 'Nsig', typ: 'i32'},
                    {name: 'stimclasses', typ: 'double*'},
                    {name: 'sc_frac', typ: 'double*'},
                    {name: '_cns', typ: '*'},
                    {name: 'sc_amp', typ: 'double**'},
                    {name: 'sc_var', typ: 'double**'},
                    {name: 'sc_std', typ: 'double**'}
                ]);
                this.cns = new lib.Constrained(0, this._cns());
            }
            destroy() {
                this.cns.destroy();
                lib.qubx_stim_amps_free(this.ptr);
            }
            setupScs() {
                lib.qubx_stim_amps_setup_scs(this.ptr);
            }
            parsToModel() {
                lib.qubx_stim_amps_pars_to_model(this.ptr);
            }
            modelToPars() {
                lib.qubx_stim_amps_model_to_pars(this.ptr);
            }
            fixAmp0() {
                lib.qubx_stim_amps_fix_amp0(this.ptr);
            }
            fixVar0() {
                lib.qubx_stim_amps_fix_var0(this.ptr);
            }
            fixAmp(cls) {
                lib.qubx_stim_amps_fix_amp(this.ptr, cls);
            }
            fixVar(cls) {
                lib.qubx_stim_amps_fix_var(this.ptr, cls);
            }
            calcStd(pInvHessian) {
                lib.qubx_stim_amps_calc_std(this.ptr, pInvHessian);
            }
        };
        
        lib.StimRates = class StimRates {
            constructor(model, Nstim, Nsig, pStimclasses, pScFrac, sampling) {
                this.ptr = lib.qubx_stim_rates_create(model.ptr, Nstim, Nsig, pStimclasses, pScFrac, sampling, 0);
                lib.addStructAccess(this, this.ptr, [
                    {name: 'version', typ: 'i32'},
                    {name: 'storage', typ: '*'},
                    {name: 'model', typ: '*'},
                    {name: 'custom_eigen', typ: '*'},
                    {name: 'Nstim', typ: 'i32'},
                    {name: 'Nsig', typ: 'i32'},
                    {name: 'stimclasses', typ: 'double*'},
                    {name: 'sc_frac', typ: 'double*'},
                    {name: '_cns', typ: '*'},
                    {name: 'QQ', typ: 'double***'},
                    {name: 'AA', typ: 'double***'},
                    {name: 'QQ_eig', typ: 'double**'},
                    {name: 'QQ_imag', typ: 'double**'},
                    {name: 'QQ_evec', typ: 'double***'},
                    {name: 'QQ_einv', typ: 'double***'},
                    {name: 'A', typ: 'double**'} // of requested sc, dt
                ]);
                this.cns = new lib.Constrained(0, this._cns());
            }
            destroy() {
                this.cns.destroy();
                lib.qubx_stim_rates_free(this.ptr);
            }
            setupQQ() {
                lib.qubx_stim_rates_setup_QQ(this.ptr);
            }
            setupAA() {
                lib.qubx_stim_rates_setup_AA(this.ptr);
            }
            setupA(sc, dt) {
                lib.qubx_stim_rates_setup_A(this.ptr, sc, dt);
            }
            fixRate(a, b) {
                lib.qubx_stim_rates_fix_rate(this.ptr, a, b);
            }
            fixExp(a, b) {
                lib.qubx_stim_rates_fix_exp(this.ptr, a, b);
            }
            fixPres(a, b) {
                lib.qubx_stim_rates_fix_pres(this.ptr, a, b);
            }
            scaleRate(a, b, c, d) {
                lib.qubx_stim_rates_scale_rate(this.ptr, a, b, c, d);
            }
            scaleExp(a, b, c, d) {
                lib.qubx_stim_rates_scale_exp(this.ptr, a, b, c, d);
            }
            scalePres(a, b, c, d) {
                lib.qubx_stim_rates_scale_pres(this.ptr, a, b, c, d);
            }
            balanceLoop(loopStates) {
                const pLoopstates = lib.ivec();
                pLoopstates.set(loopStates);
                lib.qubx_stim_rates_balance_loop(this.ptr, loopStates.length, pLoopstates.p);
                pLoopstates.destroy();
            }
            balanceLoops() {
                lib.qubx_stim_rates_balance_loops(this.ptr);
            }
            fixAll() {
                lib.qubx_stim_rates_fix_all(this.ptr);
            }
            setupConstraints() {
                return lib.qubx_stim_rates_setup_constraints(this.ptr);
            }
            parsToModel() {
                lib.qubx_stim_rates_pars_to_model(this.ptr);
            }
            modelToPars() {
                lib.qubx_stim_rates_model_to_pars(this.ptr);
            }
            calcStd(pInvHessian) {
                lib.qubx_stim_rates_calc_std(this.ptr, pInvHessian);
            }
        };
        
        lib.modelFromJson = function modelFromJson(obj, signalIndex) {
            const m = new lib.Model(obj.states.length, obj.classes.length),
                  Nstate = obj.states.length;
            m.IeqFv(obj.IeqFv);
            m.Vrev(obj.vRev);
            m.Nchannel(obj.channelCount);
            if ( obj.classes.length ) {
                m.amp0(obj.classes[0].Amp);
                m.var0(obj.classes[0].Std * obj.classes[0].Std);
            }
            let i = 0;
            for (; i<Nstate; ++i) {
                m.clazz(i, obj.states[i].Class);
                m.P0(i, obj.states[i].Pr);
            }
            if ( obj.IeqFv ) {
                const n = obj.classes.length;
                for (i=0; i<n; ++i) {
                    const cls = obj.classes[i];
                    m.amp(i, cls.Cond);
                    m.var(i, cls.CondStd * cls.CondStd);
                }
            } else {
                const n = obj.classes.length;
                for (i=0; i<n; ++i) {
                    const cls = obj.classes[i];
                    m.amp(i, cls.Amp);
                    m.var(i, cls.Std*cls.Std);
                }
            }
            m.signalIndex = signalIndex || {'Current' : 0};
            function si(name) {
                if ( ! (name in m.signalIndex) ) {
                    m.signalIndex[name] = Object.keys(m.signalIndex).length;
                }
                return m.signalIndex[name];
            }
            const n2 = Nstate*Nstate;
            for (i=0; i<n2; ++i) {
                m.K0(i,0).K1(i,0).K2(i,0).L(i,0).V(i,0).P(i,0);
            }
            for (const rate of obj.rates) {
                const j = rate.From * Nstate + rate.To;
                m.K0(j, rate.k0);
                m.K1(j, rate.k1);
                m.K2(j, rate.k2);
                if ( rate.Ligand ) {
                    m.L(j, si(rate.Ligand));
                }
                if ( rate.Voltage ) {
                    m.V(j, si(rate.Voltage));
                }
                if ( rate.Pressure ) {
                    m.P(j, si(rate.Pressure));
                }
            }
            return m;
        };
        
        /* --- qubx_simulate.h -------------------- */
        
        lib.SimConst = class SimConst {
            constructor() {
                this.buf = lib.fvec();
                this.buf.size(1);
            }
            destroy() {
                this.buf.destroy();
            }
            set(x) {
                this.buf.set(0, x);
            }
        };
        
        lib.SimStimulus = class SimStimulus {
            constructor() {
                this.buf = lib.fvec();
                this.unfiltered = lib.fvec();
                this.levels = lib.fvec();
                this.durations = lib.fvec();
                this.n = 0;
            }
            destroy() {
                this.buf.destroy();
                this.unfiltered.destroy();
                this.levels.destroy();
                this.durations.destroy();
            }
            setup(stimType, trace, n, sampling, filter_kHz) {
                this.n = n;
                this.buf.size(n);
                this.unfiltered.size(n);
                const nLevel = trace.levels.length;
                this.levels.set(trace.levels);
                this.durations.set(trace.durations);
                const dest = filter_kHz ? this.unfiltered : this.buf;
                ({
                    'steps': lib.qubx_simulate_stimulus_steps,
                    'ramps': lib.qubx_simulate_stimulus_ramps
                })[stimType](n, dest.p, sampling, nLevel, this.levels.p, this.durations.p);
                if ( filter_kHz ) {
                    lib.qub_filterdataf(this.unfiltered.p, this.buf.p, n, sampling, 1000*filter_kHz);
                }
            }
        };
        
        lib.SimStimuli = class SimStimuli {
            constructor() {
                this.stims = [];
                this.consts = [];
                this.iNextStim = this.iNextConst = 0;
                this.stimCount = lib.ivec();
                this.stimuli = lib.pvec();
            }
            destroy() {
                this.stimCount.destroy();
                this.stimuli.destroy();
                this.stims.forEach(stim => stim.destroy());
                this.consts.forEach(stim => stim.destroy());
            }
            nextStim() {
                var stim;
                if ( this.iNextStim === this.stims.length ) {
                    stim = new lib.SimStimulus();
                    this.stims.push(stim);
                } else {
                    stim = this.stims[this.iNextStim];
                }
                ++this.iNextStim;
                this.assigned.push(stim);
                return stim;
            }
            nextConst(val) {
                var stim;
                if ( this.iNextConst === this.consts.length ) {
                    stim = new lib.SimConst();
                    this.consts.push(stim);
                } else {
                    stim = this.consts[this.iNextConst];
                }
                ++this.iNextConst;
                this.assigned.push(stim);
                stim.set(val);
                return stim;
            }
            releaseAll() {
                this.iNextStim = this.iNextConst = 0;
                this.assigned = [];
                this.offset = 0;
            }
            setup(stimulus, iSeg, n, sampling, signalIndex) {
                // when done, "this" has:
                //   .vSignal (0 === none), .signalNames = ['Current', ...] of sampled signals,
                //   .Nstim (number for sim function, includes "Current" and other constants)
                //   asm buffers: .stimCount, .stimuli
                const names = [];
                Object.keys(signalIndex).forEach(name => {
                    names[signalIndex[name]] = name;
                });
                const Nstim = names.length;
                this.Nstim = Nstim;
                this.stimCount.size(Nstim);
                this.stimuli.size(Nstim);
                this.vSignal = 0;
                this.signalNames = ['Current'];
                
                this.releaseAll();
                let s = this.nextConst(0);
                this.stimCount.set(0, 0);
                this.stimuli.set(0, s.buf.p);
                let i = 1;
                for (; i<Nstim; ++i) { // signalIndex['Current'] === 0; names[0] === 'Current'
                    if ( names[i] === 'Voltage' ) {
                        this.vSignal = i;
                    }
                    const stimRec = stimulus[names[i]] || {};
                    if ( stimRec.stimType === 'steps' || stimRec.stimType === 'ramps' ) {
                        const trace = stimRec.traces[Math.min(iSeg, stimRec.traces.length-1)];
                        s = this.nextStim();
                        s.setup(stimRec.stimType, trace, n, sampling, stimRec.filter_kHz);
                        this.stimCount.set(i, n);
                        this.stimuli.set(i, s.buf.p);
                        this.signalNames.push(names[i]);
                    } else {
                        s = this.nextConst(stimRec.level || 0);
                        this.stimCount.set(i, 0);
                        this.stimuli.set(i, s.buf.p);
                    }
                }
            }
            move(offset) {
                this.offset = offset;
                this.assigned.forEach((stim, i) => stim.n && this.stimuli.set(i, stim.buf.p + 4*offset));
            }
            copyOut(buffers) {
                const nBuf = buffers.length;
                let i = 1;
                for (; i<nBuf; ++i) {
                    if ( i <= this.iNextStim ) {
                        const stimSamples = new Float32Array(buffers[i]);
                        lib.memcpy_out_float(stimSamples, 0, this.stims[i-1].buf.p + 4*this.offset, stimSamples.length);
                    }
                }
            }
        };

        lib.SimulatorState = class SimulatorState {
            constructor() {
                this.startStates = lib.ivec();
            }
            destroy() {
                this.startStates.destroy();
            }
            prep(resetP0, channelCount) {
                resetP0 = resetP0 || (this.startStates.size() !== channelCount);
                this.startStates.size(channelCount);
                if ( resetP0 ) {
                    let i = 0;
                    for (; i<channelCount; ++i) {
                        this.startStates.set(i, -1);
                    }
                }
            }
        };
        
        const SimulatorConfig = {
            addBaseline: true,
            usePeq: true,
            forceQ: false
        };
        lib.Simulator = class Simulator {
            constructor(config) {
                this.config = {
                    addBaseline: (config && "addBaseline" in config) ? config.addBaseline : SimulatorConfig.addBaseline,
                    usePeq : (config && "usePeq" in config) ? config.usePeq : SimulatorConfig.usePeq,
                    forceQ : (config && "forceQ" in config) ? config.forceQ : SimulatorConfig.forceQ
                };
                this.engine = this.config && this.config.engine || Random.engines.mt19937().autoSeed();
                this._amps = lib.dvec();
                this._stds = lib.dvec();
                this._samples = lib.fvec();
                this._states = lib.ivec();
                this._counts = lib.ivec();
                this._startStates = lib.ivec();
            }
            destroy() {
                this._amps.destroy();
                this._stds.destroy();
                this._samples.destroy();
                this._states.destroy();
                this._counts.destroy();
                this._startStates.destroy();
            }
            simulate(model, simState, sampling, Nsample, stimuli, vDef, buffers, zero) {
                const Nstate = model.Nstate();
                this._amps.size(Nstate);
                this._stds.size(Nstate);
                this._samples.size(Nsample);
                this._states.size(Nsample);
                this._counts.size(Nsample);
                
                stimuli.copyOut(buffers);
                
                zero && lib.bzero(this._samples.p, Nsample*4);
                let Nevent = 0;
                if ( model.Nstate() ) {
                    Nevent = lib.qubx_simulate1(model.ptr, simState.startStates.p, this.config.addBaseline, this.config.usePeq,
                                                model.IeqFv() ? stimuli.vSignal : -2, vDef,
                                                sampling,
                                                stimuli.Nstim, stimuli.stimCount.p, stimuli.stimuli.p,
                                                Nsample, this._samples.p, this._states.p, this._counts.p, this._amps.p, this._stds.p,
                                                this.config.forceQ, Random.integer(-(1<<30), 1<<30)(this.engine));
                }
                return {
                    samples: this._samples.contents(Nsample, new Float32Array(buffers[0], 0, Nsample)),
                    Nsample: Nsample,
                    Nevent: Nevent,
                    states: this._states.contents(Nevent),
                    counts: this._counts.contents(Nevent),
                    amps: this._amps.contents(),
                    stds: this._stds.contents(),
                    signalNames: stimuli.signalNames
                };
            }
        };
        
        /* --- qubx_skm.h -------------------- */
        
        lib.SKM = class SKM {
            constructor(sampling, maxN, model) {
                this.ptr = lib.qubx_skm_session_create(sampling, maxN, model.ptr);
                this.pData = lib.fvec();
                this.pLL = lib.dvec();
                this.pLL.size(1);
                this.pFixAmp = lib.ivec();
                this.pFixStd = lib.ivec();
                this.pStateSeq = lib.ivec();
                this.pCounts = lib.ivec();
                this.pA = lib.dvec();
                this.pAmp = lib.dvec();
                this.pStd = lib.dvec();
            }
            destroy() {
                lib.qubx_skm_session_free(this.ptr);
                this.pLL.destroy();
                this.pData.destroy();
                this.pFixAmp.destroy();
                this.pFixStd.destroy();
                this.pStateSeq.destroy();
                this.pCounts.destroy();
                this.pA.destroy();
                this.pAmp.destroy();
                this.pStd.destroy();
            }
            reset(dtMS) {
                lib.qubx_skm_session_reset(this.ptr, dtMS||0.0);
            }
            viterbi(data, stateSeq) {
                this.pData.set(data);
                stateSeq && this.pStateSeq.size(stateSeq.length);
                lib.qubx_skm_viterbi(this.ptr, data.length, this.pData.p, stateSeq ? this.pStateSeq : 0, this.pLL.p);
                stateSeq && this.pStateSeq.contents(stateSeq.length, stateSeq);
                return this.pLL.get(0);
            }
            iterate(data, stateSeq, maxIter, fixA, fixAmp, fixStd) {
                this.pData.set(data);
                stateSeq && this.pStateSeq.size(stateSeq.length);
                fixAmp && this.pFixAmp.set(fixAmp);
                fixStd && this.pFixStd.set(fixStd);
                const iterations = lib.qubx_skm_iterate(this.ptr, data.length, this.pData.p, stateSeq ? this.pStateSeq : 0,
                                                        maxIter, fixA|0, fixAmp ? this.pFixAmp.p : 0, fixStd ? this.pFixStd.p : 0,
                                                        this.pLL.p);
                stateSeq && this.pStateSeq.contents(stateSeq.length, stateSeq);
                return {
                    ll: this.pLL.get(0),
                    iterations
                };
            }
            getA() {
                const nstate = lib.qubx_skm_Nstate(this.ptr);
                this.pA.size(nstate*nstate);
                lib.qubx_skm_read_A(this.ptr, this.pA.p);
                return this.pA.contents();
            }
            getEvents() {
                const nevent = lib.qubx_skm_Nevent(this.ptr);
                this.pStateSeq.size(nevent);
                this.pCounts.size(nevent);
                lib.qubx_skm_read_events(this.ptr, this.pStateSeq.p, this.pCounts.p);
                
                const nclass = lib.qubx_skm_Nclass(this.ptr);
                this.pAmp.size(nclass);
                this.pStd.size(nclass);
                lib.qubx_skm_read_amp(this.ptr, this.pAmp.p, this.pStd.p);
                
                return {
                    classes: this.pStateSeq.contents(),
                    counts: this.pCounts.contents(),
                    amp: this.pAmp.contents(),
                    std: this.pStd.contents()
                };
            }
        };
        
        /* --- qubx_util.h ------------------- */

        lib.QtoPe = function QtoPe(Q) {
            const n = Q.length;
            if ( ! n ) { return []; }
            const mQ = lib.dalloc(n*n),
                  mPe = lib.dalloc(n);
            let i = 0;
            for (; i<n; ++i) {
                let j = 0;
                for (; j<n; ++j) {
                    lib.dset(mQ, i*n+j, Q[i][j]);
                }
            }
            lib.qubx_QtoPe(mQ, mPe, n);
            const Pe = [];
            for (i=0; i<n; ++i) {
                Pe.push(lib.dget(mPe, i));
            }
            lib.dfree(mQ);
            lib.dfree(mPe);
            return Pe;
        };
        
        lib.QtoQe = function QtoQe(Q, clazz, tdead) {
            const n = Q.length,
                  mQ = lib.dalloc(n*n);
            let i = 0;
            for (; i<n; ++i) {
                let j = 0;
                for (; j<n; ++j) {
                    lib.dset(mQ, i*n+j, Q[i][j]);
                }
            }
            const pClazz = lib.ivec().set(clazz),
                  pQe = lib.dvec().size(n*n);
            lib.qubx_QtoQe(mQ, n, pClazz.p, tdead, pQe.p);
            const Qe = [],
                  Qe_flat = pQe.contents();
            let at = 0;
            for (i=0; i<n; ++i) {
                const row = [];
                let j = 0;
                for (; j<n; ++j, ++at) {
                    row.push(Qe_flat[at]);
                }
                Qe.push(row);
            }
            lib.dfree(mQ);
            pQe.destroy();
            pClazz.destroy();
            return Qe;
        };
        
        lib.expm = function expm(Q, dt) {
            const n = Q.length,
                  mQ = lib.dalloc(n*n),
                  mA = lib.dalloc(n*n),
                  A = [];
            let i = 0;
            for (; i<n; ++i) {
                let j = 0;
                for (; j<n; ++j) {
                    lib.dset(mQ, i*n+j, Q[i][j]);
                }
            }
            lib.qubx_expm(n, mA, mQ, dt);
            for (i=0; i<n; ++i) {
                A[i] = [];
                let j = 0;
                for (; j<n; ++j) {
                    A[i][j] = lib.dget(mA, i*n+j);
                }
            }
            lib.dfree(mQ);
            lib.dfree(mA);
            return A;
        };
        
        lib.QtoKeq = function QtoKeq(Q) {
            const n = Q.length,
                  Keq = [];
            let i = 0;
            for (; i<n; ++i) {
                Keq[i] = Q.map(item => 0);
                Keq[i][i] = 1;
            }
            for (i=0; i<n; ++i) {
                let j = i + 1;
                for (; j<n; ++j) {
                    if ( Q[i][j] && Q[j][i] ) {
                        Keq[i][j] = Q[i][j] / Q[j][i];
                        Keq[j][i] = 1.0 / Keq[i][j];
                    }
                }
            }
            return Keq;
        };
        
        lib.KeqToDelG = function KeqToDelG(Keq) {
            const n = Keq.length,
                  delG = [];
            let i = 0;
            for (; i<n; ++i) {
                delG[i] = [];
                let j = 0;
                for (; j<n; ++j) {
                    delG[i][j] = Keq[i][j] ? Math.log10(Keq[i][j]) : 0.0;
                }
            }
            return delG;
        };
        
        lib.makeQflux = function makeQflux(Q) {
            const n = Q.length,
                  mQ = lib.dalloc(n*n),
                  mA = lib.dalloc(n*n),
                  A = [];
            let i = 0;
            for (; i<n; ++i) {
                let j = 0;
                for (; j<n; ++j) {
                    lib.dset(mQ, i*n+j, Q[i][j]);
                }
            }
            lib.qubx_QtoQflux(n, mA, mQ);
            for (i=0; i<n; ++i) {
                A[i] = [];
                let j = 0;
                for (; j<n; ++j) {
                    A[i][j] = lib.dget(mA, i*n+j);
                }
            }
            lib.dfree(mQ);
            lib.dfree(mA);
            return A;
        };
        
        lib.meanFirstPassage = function meanFirstPassage(Q, sampling) {
            const n = Q.length,
                  mQ = lib.dalloc(n*n),
                  mA = lib.dalloc(n*n),
                  A = [];
            let i = 0;
            for (; i<n; ++i) {
                let j = 0;
                for (; j<n; ++j) {
                    lib.dset(mQ, i*n+j, Q[i][j]);
                }
            }
            lib.qubx_MeanFirstPassage(n, mA, mQ, sampling);
            for (i=0; i<n; ++i) {
                A[i] = [];
                let j = 0;
                for (; j<n; ++j) {
                    A[i][j] = lib.dget(mA, i*n+j);
                }
            }
            lib.dfree(mQ);
            lib.dfree(mA);
            return A;
        };
        
        lib.firstLatencies = function firstLatencies(A, pEq, dt) {
            const n = A.length,
                  mA = lib.dalloc(n*n),
                  mPeq = lib.dalloc(n),
                  mFL = lib.dalloc(n*n);
            let i = 0;
            for (; i<n; ++i) {
                let j = 0;
                for (; j<n; ++j) {
                    lib.dset(mA, i*n+j, A[i][j]);
                }
                lib.dset(mPeq, i, pEq[i]);
            }
            lib.qubx_FirstLatencies(mA, mPeq, dt, n, mFL);
            const FL = [];
            for (i=0; i<n; ++i) {
                FL[i] = [];
                let j = 0;
                for (; j<n; ++j) {
                    FL[i][j] = lib.dget(mFL, i*n+j);
                }
            }
            lib.dfree(mFL);
            lib.dfree(mPeq);
            lib.dfree(mA);
            return FL;
        };
        
        /* --- ublas_matrixutil.h ------------------- */

        lib.pinvm = function pinvm(A) {
            const n = A.length,
                  mA = lib.dalloc2(n, n);
            let i = 0;
            for (; i<n; ++i) {
                const pRow = lib.drow2(mA, i);
                let j = 0;
                for (; j<n; ++j) {
                    lib.dset(pRow, j, A[i][j]);
                }
            }
            lib.sv_invm(mA, n, mA);
            const Ainv = [];
            for (i=0; i<n; ++i) {
                const row = [];
                Ainv.push(row);
                const pRow = lib.drow2(mA, i);
                let j = 0;
                for (; j<n; ++j) {
                    row[j] = lib.dget(pRow, j);
                }
            }
            lib.dfree2(mA);
            return Ainv;
        };

        return lib;
    });
});
