/* Copyright 2014-2016 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_scrollvid_main", ["qubx", "qubx_app", "qubx_ui", "qubx_ui_data", "qubx_data", "qubx_data_work", "qubx_lib", "qubx_data_prep", "qubtrace", "RecordRTC"],
    function(qubx, qubx_app, qubx_ui, qubx_ui_data, qubx_data, qubx_data_work, qubx_lib, qubx_data_prep, qubtrace, RecordRTC) {
        "use strict";
        var exports = {};

        var SVMainDefs = {
            fullscreen: false
        };

        var MacSignals = function MacSignals() {
            this.storage = {};
        };
        MacSignals.prototype.get = function(name) {
            var sig;
            if ( name in this.storage ) {
                sig = this.storage[name];
                sig.defval = sig.defval || 0.0;
                sig.delay = sig.delay || 0.0;
                return sig;
            }
            sig = {
                name: name,
                defval: 0.0,
                delta: 0.05,
                delay: 0.0,
                rename: name,
                isLigand: false,
                isVoltage: (name[0] in ['V', 'v']),
                isPressure: (name[0] in ['P', 'p'])
            };
            this.storage[name] = sig;
            return sig;
        };
        MacSignals.prototype.savePrefs = function() {
            localStorage.setItem("qubx_mac_signals", JSON.stringify(this.storage));
        };
        MacSignals.prototype.readPrefs = function() {
            var saved = localStorage.getItem("qubx_mac_signals");
            if ( saved ) {
                this.storage = JSON.parse(saved);
            }
        };
        
        exports.start = function(cfg) {
            var main;
            var prefs = localStorage.getItem("qubx_scrollvid_prefs");
            prefs = prefs ? JSON.parse(prefs) : {};
            var config = qubx.extend({}, SVMainDefs, cfg, prefs, {
                topframe: "#qubx_scrollvid_topframe"
            });
            main = new qubx_app.Main(config);
            main.dataWorker = new qubx_data_work.DataWorkgroup();
            var _dispose = main.dispose.bind(main);
            qubx.extend(main, {
                signals: new MacSignals(),
                sentConst: {},
                dispose: function() {
                    main.savePrefs();
                    main.panels.then(function(panels) {
                        panels.data.dispose();
                        _dispose();
                    });
                    main.dataWorker.destroy();
                },
                savePrefs: function() {
                    var prefs = {};
                    for (var k in SVMainDefs) { // to avoid e.g. config.node
                        prefs[k] = main.config[k];
                    }
                    localStorage.setItem("qubx_scrollvid_prefs", JSON.stringify(prefs));
                    main.signals.savePrefs();
                    qubx_ui.SaveInputs();
                    qubx_ui.resultDlg.savePrefs();
                },
                _makeMenus: function(p) {
                    main.opener = new qubx_ui.Opener();
                    p.mnuQUB = new qubx_ui.Menu({
                        node: p.controls.querySelector("#qubx_mnu_qub"),
                        caption: 'QUB...',
                        groups: [[
                            {
                                caption: 'About...',
                                action: main.showAbout
                            }
                        ], [
                            {
                                caption: 'Simulator',
                                action: function() {
                                    window.location.href = 'index.html#welcome=1';
                                }
                            },
                            {
                                caption: 'Macroscopic solver',
                                action: function() {
                                    window.location.href = 'mac.html#welcome=1';
                                }
                            }
                        ]]
                    });
                    
                    p.mnuLoad = new qubx_ui.Menu({
                        node: p.controls.querySelector("#qubx_mnu_load"),
                        caption: 'Load...',
                        groups: [[
                            {
                                caption: 'Data from disk...',
                                action: function() {
                                    main.opener.run()
                                        .then(main.openData)
                                        .then(function() {
                                            main.opener.reset();
                                        });
                                }
                            }
                        ]]
                    });
                },
                _makeDialogs: function(p, items) {
                    main.aboutDlg = new qubx_ui.Modal({
                        title: "QUB Online Scrolling data videos - About",
                        template: '#qubx_scrollvid_about',
                        canCancel: false,
                        okLabel: 'Close'
                    });
                },
                showAbout: function() {
                    main.aboutDlg.run();
                },
                _setupData: function(p, items) {
                    p.data = main.data = new qubx_ui_data.PlotFrame({
                        node: p.main.querySelector("#qubx_scrollvid_data")
                    });
                    p.data.setData(new qubx_data.Data());
                    /// do something with p.data.plot.chosenControl
                    
                    var ctrls = items.querySelector("#qubx_scrollvid_ctrls");
                    p.controls.appendChild(ctrls);
                    
                    var workarea = p.main.querySelector("#qubx_scrollvid_work");
                    main.canvas = workarea.querySelector("#qubx_scrollvid_canvas");
                    main.canvas.className = 'qubx_scrollvid_heightfill';
                    setTimeout(function() {
                        if ( main.canvas.clientHeight > workarea.clientHeight ) {
                            // some browsers don't calculate 100% height relative to the parent, even though parent has position: relative
                            main.canvas.className = 'qubx_scrollvid_heightfix';
                        }
                    }, 700);

                    main.txtWidth = new qubx_ui.AcceptedInput(ctrls.querySelector("#scrollvid_width"),
                                                                  800,
                                                                  {
                                                                      accept: qubx.acceptIntAtLeast(100),
                                                                      format: qubx.formatStr,
                                                                      widthChars: 4
                                                                  });
                    main.canvas.width = main.txtWidth.value;
                    main.txtWidth.ev.change.sub(function(w) {
                        main.canvas.width = w;
                    });
                    main.txtHeight = new qubx_ui.AcceptedInput(ctrls.querySelector("#scrollvid_height"),
                                                                  450,
                                                                  {
                                                                      accept: qubx.acceptIntAtLeast(100),
                                                                      format: qubx.formatStr,
                                                                      widthChars: 4
                                                                  });
                    main.canvas.height = main.txtHeight.value;
                    main.txtHeight.ev.change.sub(function(h) {
                        main.canvas.height = h;
                    });
                    main.txtDur = ctrls.querySelector("#scrollvid_dur");
                    main.txtFPS = ctrls.querySelector("#scrollvid_fps");
                    main.txtRate = ctrls.querySelector("#scrollvid_rate");
                    main.btnRun = ctrls.querySelector("#scrollvid_run");
                    main.btnRun.addEventListener('click', main.run);
                    main.dl = ctrls.querySelector("#scrollvid_dl");
                    main.msg = ctrls.querySelector("#scrollvid_msg");
                },

                progressDlg: new qubx_ui.ModalProgress(),

                openData: function(files) {
                    var f = files[0];
                    function readAll(rdr, index, name, delta, idealized, getDeltaForN) {
                        var data = main.data.data;
                        data.segmentCount(0);
                        data.sampling(rdr.sampling);
                        data.signalCount(rdr.signalNames.length);
                        rdr.signalNames.forEach(function(nm, i) {
                            data.signals[i].name = nm;
                        });
                        data.segmentCount(rdr.counts.length);
                        rdr.counts.forEach(function(count, iSeg) {
                            data.segmentSize(iSeg, count);
                        });
                        data.doneSegmentSize();
                        rdr.readSegments(function(rdr, seg, iSeg) {
                            rdr.signalNames.forEach(function(nm, iSig) {
                                data.setSampled(iSeg, iSig, 0, seg.buffers[iSig]);
                                var idlSegs = idealized[iSig];
                                if ( idlSegs && idlSegs.length ) {
                                    console.log(idlSegs);
                                    /// data.setIdeal(iSeg, iSig, counts, classes); ///
                                }
                            });
                        });
                        main.getDeltaForN = getDeltaForN;
                        main.prepData();
                    }
                    var rdrType;
                    var reader = new FileReader();
                    reader.onload = function(e) {
                        var rdr = new rdrType(e.target.result);
                        while ( isNaN(1/rdr.sampling) ) {
                            rdr.sampling = rdr.sampling || 1e-3/(+prompt("Sampling rate [kHz]:", "10.0"));
                        }
                        qubx_data_prep.openDlg.runSignals(main.dataWorker, f.name, rdr, main.signals)
                            .then(function(info) {
                                main.progressDlg.show('Preparing data...');
                                return qubx_data_prep.prepData(main.dataWorker, rdr, info.depVar, main.signals, function(prog) {
                                    main.progressDlg.progress(prog);
                                    return ! main.progressDlg.canceled;
                                });
                            })
                            .then(function(info) {
                                main.progressDlg.hide();
                                readAll(rdr, info.index, info.name, info.delta, info.idealized, info.getDeltaForN);
                            })
                            .catch(qubx.reportErr)
                            .catch(function(err) {
                                main.progressDlg.hide();
                            });
                    };
                    var ext = (f.name.length > 4) ? f.name.slice(f.name.length-4).toLowerCase() : '.txt';
                    if ( ext === '.qdf' ) { // really ought to open/multi with .qsf if available
                        rdrType = qubx_data.QDFReader;
                        reader.readAsArrayBuffer(f);
                    } else/*if ( ext === '.txt' )*/{
                        rdrType = qubx_data.TextReader;
                        reader.readAsText(f);
                    }
                },
                prepData: function() {
                    qubx_lib.then(function(lib) {
                        main.raster_samples = lib.raster_samples;
                    });
                    main.data.plot.drawHiresSoon();
                },

                running: false,
                interrupted: false,
                run: function() {
                    if ( main.running ) {
                        main.interrupted = true;
                        return;
                    }
                    main.running = true;
                    main.interrupted = false;
                    main.btnRun.innerHTML = 'Interrupt';
                    main.dl.download = '';
                    main.dl.href = '';
                    main.dl.innerHTML = '';
                    main.msg.innerHTML = 'Please wait...';
                    var fps = +main.txtFPS.value;
                    var rate = +main.txtRate.value;
                    var dur = +main.txtDur.value;
                    var sampling = main.data.data.sampling();
                    var recorder = new RecordRTC.CanvasRecorder(main.canvas, {
                        disableLogs: false
                    });
                    recorder.record();
                    //var encoder = new whammy.Video(fps);
                    var empty = true;
                    function frame() {
                        empty = false;
                        //encoder.add(main.canvas);
                    }
                    function renable() {
                        main.btnRun.disabled = false;
                        main.btnRun.innerHTML = 'Make video';
                        main.running = false;
                    }
                    function finish() {
                        if ( main.interrupted ) {
                            main.msg.innerHTML = 'Ready';
                            renable();
                        } else if ( empty ) {
                            alert('Please load and highlight some data');
                            main.msg.innerHTML = 'Ready';
                            renable();
                        } else {
                            main.btnRun.disabled = true;
                            main.btnRun.innerHTML = 'Finishing video';
                            recorder.stop(function(blob) {
                            //encoder.compile(false, function(blob) {
                                main.dl.download = 'data.webm';
                                main.dl.href = URL.createObjectURL(blob);
                                main.dl.innerHTML = 'Download video...';
                                main.msg.innerHTML = '';
                                renable();
                            });
                        }
                    }
                    var w = main.canvas.width;
                    var h = main.canvas.height;
                    var spf = rate/(fps*sampling); // samples per frame
                    var samples_screen = Math.round(dur / sampling)|0;
                    var spp = samples_screen / w; // samples per pixel
                    var ctx = main.canvas.getContext('2d');
                    ctx.translate(0.5, 0); // for even raster pixels
                    var raster = main.raster_samples || qubtrace.raster_samples;
                    function draw(px0, pxw, points, offset, n, yLo, ypp) {
                        //console.log(px0, pxw, points, offset, n, yLo, ypp, spp);
                        if ( spp >= main.data.plot.config.minSppRaster ) {
                            var rastered = raster(points.buffer, offset, n, pxw, spp, ypp);
                            qubtrace.drawRastered(ctx, px0, pxw, h, rastered, "rgb(100,100,100)", "rgb(0,0,0)", spp, yLo, ypp);
                        } else {
                            qubtrace.drawLines(ctx, px0, pxw, h, points, offset, n, "rgb(0,0,0)", spp, yLo, ypp);
                        }
                    }
                    main.data.plot.reSelect(function(offset, count, segments) {
                        if ( ! (count && segments && segments.length) ) {
                            return;
                        }
                        var chosen = main.data.plot.chosen();
                        if ( chosen >= 0 ) {
                            segments = segments.slice(chosen, chosen+1);
                        }
                        var yLo, yHi;
                        segments.forEach(function(seg) {
                            var data = seg.sources[0];
                            if ( typeof(yLo) === 'undefined' ) {
                                yLo = data[seg.selFrom];
                                yHi = yLo;
                            }
                            for (var i=seg.selFrom; i<=seg.selTo; ++i) {
                                var d = data[i];
                                yLo = Math.min(yLo, d);
                                yHi = Math.max(yHi, d);
                            }
                        });
                        var dy = yHi - yLo;
                        var cy = (yHi + yLo) / 2;
                        yLo = cy - .6*dy;
                        yHi = cy + .6*dy;
                        var ypp = (yHi - yLo) / h;
                        var iseg = -1;
                        var i0, n, data, i;
                        requestAnimationFrame(function animate() {
                            if ( main.interrupted ) {
                                return finish();
                            }
                            if ( (iseg < 0) || (i >= n) ) {
                                ++iseg;
                                if ( iseg === segments.length ) {
                                    return finish();
                                }
                                i0 = segments[iseg].selFrom;
                                n = segments[iseg].selTo - i0 + 1;
                                data = segments[iseg].sources[0];
                                i = iseg ? (-samples_screen) : 0;
                            } else {
                                i += spf;
                            }
                            var px0, pxw, ndraw, offset;
                            ctx.fillStyle = "rgb(255,255,255)";
                            ctx.fillRect(0, 0, w, h);
                            if ( i < 0 ) {
                                ndraw = Math.ceil(Math.min(i+samples_screen, n))|0;
                                px0 = Math.round(-i/spp)|0;
                                offset = i0;
                            }
                            else {
                                px0 = 0;
                                offset = Math.max(0,Math.round(i)|0);
                                ndraw = Math.max(0, Math.ceil(Math.min(samples_screen, n-offset))|0);
                                offset += i0;
                            }
                            if ( ndraw ) {
                                pxw = Math.ceil(ndraw/spp)|0;
                                draw(px0, pxw, data, offset, ndraw, yLo, ypp);
                            }
                            frame();
                            requestAnimationFrame(animate);
                        });
                    });
                }
            });

            qubx_ui.loadTemplate("#qubx_scrollvid_components").then(function(items) {
                main.panels.then(function(p) {
                    window.addEventListener("pagehide", main.savePrefs.bind(main));

                    main.signals.readPrefs();
                    main._setupData(p, items);
                    main._makeMenus(p);
                    main._makeDialogs(p);

                    if ( (! navigator.vendor) || (0 > navigator.vendor.indexOf('Google')) ) {
                        alert('This page requires the Google Chrome browser, because it is the only one with built-in WebM facilities. Sorry for the inconvenience.');
                    }

                    return p;
                });
            });

            return main;
        };
        
        return exports;
    });
