/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

"use strict";

let config = {},
    models = [],
    serial = null,
    state = 0,
    simModels = null,
    simState = [],
    sim = null,
    stimuli = null,
    signalIndex = null;

const QUBX_LIB_MEMORY = 2*16*1024*1024;

importScripts('qubx_worker_require.js');

let earlyOnes = [];
const earlyListener = event => earlyOnes.push(event.data);
self.addEventListener('message', earlyListener);

qubx_require(['qubx_lib', 'random']).then(libs => {
    const [lib, Random] = libs;
    /*
    var engine = Random.engines.mt19937().autoSeed();
    var uniDist = Random.real(0, 1, false);
    function expo(Qdiag) {
        return ( Math.log(1 - uniDist(engine)) / Qdiag );
    }
    var last_norm = null;
    function norm() {
        var result;
        if ( last_norm === null ) {
            a = Math.sqrt(-2 * Math.log(uniDist(engine)));
            b = 2*Math.PI*uniDist(engine);
            last_norm = a*Math.cos(b);
            result = a*Math.sin(b);
        } else {
            result = last_norm;
            last_norm = null;
        }
        return result;
    }
    
    function makeQ() {
        var n = model.states.length;
        var Q = model.states.map(function(state) { return model.states.map(function(state) { return 0; }); });
        for (var r=0, nr=model.rates.length; r<nr; ++r) {
            var rate = model.rates[r];
            Q[rate.From][rate.To] = rate.k0;
        }
        for (var i=0; i<n; ++i) {
            var sum = 0.0;
            for (var j=0; j<n; ++j) {
                sum += Q[i][j];
            }
            Q[i][i] = - sum;
        }
        return Q;
    }
    
    function pick(Q, qdiag, state) {
        var x = -qdiag * uniDist(engine);
        var tot = 0.0;
        var qrow = Q[state];
        for (var i=0, n=Q.length; i<n; ++i) {
            if ( i != state ) {
                tot += qrow[i];
                if ( tot >= x ) {
                    return i;
                }
            }
        }
        console.log('!');
        return state;
    }
    */
    function statesToClasses(simResult, out, states) {
        const clazz = states.map(s => s.Class),
              ss = simResult.states,
              n = ss && ss.length || 0,
              cc = out.classes = new Int32Array(n);
        out.counts = simResult.counts;
        out.amps = [];
        out.stds = [];
        let i = 0;
        for (; i<n; ++i) {
            const c = clazz[ss[i]];
            cc[i] = c;
            out.amps[c] = simResult.amps[ss[i]];
            out.stds[c] = simResult.stds[ss[i]];
        }
    }
    
    function ensureModel(i) {
        if ( ! simModels ) {
            simModels = [];
        }
        if ( ! signalIndex ) {
            signalIndex = {'Current': 0};
        }
        if ( models[i] ) {
            if ( ! simModels[i] ) {
                simModels[i] = lib.modelFromJson(models[i], signalIndex);
            }
            if ( ! simState[i] ) {
                simState[i] = new lib.SimulatorState();
            }
            return true;
        }
        return false;
    }
    
    sim = new lib.Simulator();
    stimuli = new lib.SimStimuli();

    self.removeEventListener('message', earlyListener);
    const mainListener = event => {
        if ( event.data.t === 'STOP' ) {
            simModels && simModels.forEach((simModel, i) =>simModel && simModel.destroy());
            simState.forEach(simState => simState & simState.destroy());
            stimuli && stimuli.destroy();
            sim.destroy();
            close();
        } else if ( event.data.t === 'CONFIG' ) {
            config = event.data.config;
            serial = null;
            // transcribe constants into stimulus constant notation
            const stimulus = config.stimulus || {};
            Object.keys(config.constants).forEach(stimName => {
                if ( ! (stimName in stimulus) ) {
                    stimulus[stimName] = {
                        stimType: 'constant',
                        level: config.constants[stimName]
                    };
                }
            });
        } else if ( event.data.t === 'MODELS' ) {
            models = event.data.models;
            serial = null;
            if ( simModels ) {
                simModels.forEach(simModel => simModel.destroy());
                simModels = signalIndex = null;
            }
        } else if ( event.data.t === 'MODEL' ) {
            const i = event.data.i
            models[i] = event.data.model;
            serial = null;
            if ( simModels && simModels[i] ) {
                simModels[i].destroy();
                simModels[i] = signalIndex = null;
            }
        } else if ( event.data.t === 'RATE' ) {
            const i = event.data.i
            if ( ensureModel(i) ) {
                const r = models[i].rates[event.data.info.row],
                      column = event.data.info.column;
                r[column] = event.data.info.value;
                if ( column === 'k0' ) {
                    simModels[i].K0(r.From*models[i].states.length + r.To, event.data.info.value);
                } else if ( column === 'k1' ) {
                    simModels[i].K1(r.From*models[i].states.length + r.To, event.data.info.value);
                } else if ( column === 'k2' ) {
                    simModels[i].K2(r.From*models[i].states.length + r.To, event.data.info.value);
                }
            }
            serial = null;
        } else if ( event.data.t === 'CC' ) {
            const i = event.data.i
            if ( ensureModel(i) ) {
                models[i].channelCount = event.data.n;
                simModels[i].Nchannel(event.data.n);
            }
            serial = null;
        } else if ( event.data.t === 'FILL' ) { // data.buf, data.n
            event.data.signalNames = [];
            event.data.classAmp = [];
            const use = models.map((m,i) => ({
                model: m,
                index: i
            })).filter(u => u.model && u.model.states.length && u.model.rates.length);
            if ( use.length === 1 ) {
                const model = use[0].model,
                      nc = model.classes.length;
                let i = 0;
                for (; i<nc; ++i) {
                    event.data.classAmp.push(model.IeqFv ? model.classes[i].Cond : model.classes[i].Amp); /// could get it right if Voltage is constant
                }
            }
            if ( (serial === null) && (event.data.first === 0) && (event.data.iSeg === 0) ) {
                serial = event.data.serial;
            }
            if ( (serial !== event.data.serial) || (use.length === 0) ) {
                postMessage(event.data, event.data.buffers);
                return;
            }
            sim.config.usePeq = config.usePeq;
            use.forEach(u => {
                ensureModel(u.index);
                u.state = simState[u.index];
                u.state.prep(event.data.first === 0, u.model.channelCount);
                u.simModel = simModels[u.index];
            });

            if ( event.data.first === 0 ) {
                stimuli.setup(config.stimulus, event.data.iSeg, event.data.total, event.data.sampling, signalIndex);
            } else {
                stimuli.move(event.data.first);
            }
            
            const vDef = config.constants['Voltage'] || 0.0;
            use.forEach((u, ui) => {
                const result = sim.simulate(u.simModel, u.state, event.data.sampling, event.data.n, stimuli, vDef, event.data.buffers, ui===0);
                event.data.signalNames = result.signalNames;
                (use.length === 1) && statesToClasses(result, event.data, u.model.states);
            });
            postMessage(event.data, event.data.buffers);
        }
    };
    for (const data of earlyOnes) {
        mainListener({data});
    }
    earlyOnes = [];
    self.addEventListener('message', mainListener);
});
