/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

'use strict';

importScripts('require.js');

const rqPaths = {};
// Import this script minified, and it will configure require to load the following modules minified as well:
if ( location.pathname.endsWith('.min.js') ) {
    rqPaths = {
        "asmffi": "asmffi.min",
        "hjcfit": "hjcfit.min",
        "qubx_fit": "qubx_fit.min",
        "qubx_lib": "qubx_lib.min",
        "qubx_worker_util": "qubx_worker_util.min",
        "qubx_xmeans_math": "qubx_xmeans_math.min",
        "random": "random.min"
    };
}
require.config({
    waitSeconds: 60,
    baseUrl: '.',
    paths: rqPaths,
    shim: {
        "random" : {
            deps: [],
            exports: "Random"
        }
    }
});

function qubx_require(names) {
    return new Promise((resolve, reject) => require(names, (...args) => {
        const libs = Array.prototype.slice.apply(args);
        let init = Promise.resolve(true);
        libs.forEach((lib, i) => {
            if ( ! lib ) {
                init = init.then(() => Promise.reject('Failed to load '+names[i]));
            } else if ( lib.then ) {
                init = init.then(() => lib.then(libResolved => {
                    libs[i] = libResolved;
                }));
            }
        });
        init.then(() => resolve(libs))
            .catch(err => reject(err));
    }));
}
