/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_workers", ["qubx"], qubx => {
    'use strict';

    const exports = [];

    exports.Worker = class _Worker {
        constructor(config) {
            this.config = Object.assign({}, config);
            this.workerReady = new Promise((resolve, reject) => {
                this.resolveReady = resolve;
                this.rejectReady = reject;
            }).catch(qubx.reportErr);
            const wantMinified = (location.pathname.indexOf('test') < 0); //  && (location.pathname.indexOf('xmeans') < 0);
            this.worker = new Worker(wantMinified && this.config.hrefMin || this.config.href);
            this.worker.addEventListener('message', this._onMessage.bind(this), false);
            this.worker.addEventListener('error', this._onError.bind(this), false);
            this.sched = new qubx.PromiseTickets();
        }
        destroy() {
            return this.request({t: 'STOP'});
        }
        ask(verb, data, transfers) {
            data = data || {};
            data.verb = verb;
            return this.request(data, transfers);
        }
        send(verb, data, transfers) {
            data.verb = verb;
            return this.post(data, transfers);
        }
        method(verb, noreply) {
            this[verb] = (data, transfers) => {
                if ( noreply ) {
                    return this.send(verb, data, transfers);
                } else {
                    return this.ask(verb, data, transfers);
                }
            };
            return this;
        }
        request(data, transfers, onprogress) {
            return new Promise((resolve, reject) => {
                data.tkt = this.sched.register(resolve, reject, onprogress);
                this.post(data, transfers);
            }).catch(qubx.reportErr);
        }
        post(data, transfers) {
            return this.workerReady.then(worker => worker.postMessage(data, transfers));
        }
        _onMessage(event) {
            if ( event.data.tkt ) {
                if ( event.data.t === 'PROGRESS' ) {
                    this.sched.progress(event.data.tkt, event.data.progress);
                } else if ( event.data.tktErr ) {
                    this.sched.reject(event.data.tkt, event.data);
                } else {
                    this.sched.resolve(event.data.tkt, event.data);
                }
            } else if ( this.resolveReady && (event.data.t === 'READY') ) {
                this.resolveReady(this.worker);
                this.resolveReady = this.rejectReady = undefined;
            } else if ( this.rejectReady && (event.data.t === 'LOADERR') ) {
                this.rejectReady(event.data.err);
                this.resolveReady = this.rejectReady = undefined;
            } else if ( this.config.onmessage ) {
                this.config.onmessage(event);
            }
        }
        _onError(e) {
            console.log(e);
            qubx.errOut.send(e);
            if ( this.rejectReady ) {
                this.rejectReady(e);
                this.resolveReady = this.rejectReady = undefined;
            }
            if ( this.config.onerror ) {
                this.config.onerror(e);
            }
        }
    };

    return exports;
});

