/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_xmeans", ["qubx", "qubx_workers"], (qubx, qubx_workers) => {
    "use strict";
    
    const exports = {};

    exports.XMeans = class XMeans {
        constructor() {
            this.worker = new qubx_workers.Worker({
                href: 'js/qubx_xmeans_worker.js'
                // hrefMin: 'js/qubx_xmeans_worker.min.js'
            })
                .method('stop')
                .method('setData')
                .method('getClusters')
                .method('iterate');
        }
        destroy() {
            return this.worker.stop();
        }
        // if columns[i] and weight are typed arrays, they are transfered into the worker (locally becoming empty)
        setData(data) {
            const transfer = (data.weight && data.weight.buffer) ? [data.weight.buffer] : [];
            data.columns.forEach(col => col.buffer && transfer.push(col.buffer));
            return this.worker.setData({
                columns: data.columns,
                weight: data.weight
            }, transfer);
        }
        getClusters(k, trials, iterations) {
            return this.worker.getClusters({
                k: k,
                trials: trials,
                iterations: iterations
            });
        }
        iterate(k, color, iterations) {
            return this.worker.iterate({
                k: k,
                color: color,
                iterations: iterations
            });
        }
    };

    return exports;
});
