/* Copyright 2014-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

define("qubx_xmeans_main", ["qubx", "qubx_app", "qubx_ui", "qubx_coldata", "qubx_data", "qubx_xmeans", "d3", "c3", "nv.d3"],
    (qubx, qubx_app, qubx_ui, qubx_coldata, qubx_data, qubx_xmeans, d3, c3, nv) => {
        "use strict";
        const exports = {};

        const COLORS = ['#000', '#f00', '#00f', '#0f0', '#00ffff', '#ff0', '#f0f', '#800', '#865203', '#728c07', '#26930b', '#0f9946', '#149f9f', '#1951a6', '#3b1eac', '#9624b3', '#b92a80', '#bf3030', '#c68c36', '#afcc3d'],
              COLOR_BEYOND = '#666',
              COLORS_HIST = function(k) {
                  const cc = COLORS.slice(1, k+1);
                  cc.push(COLORS[0]);
                  return cc;
              };

        function readColumns(rows, colDefs) {
            const columns = colDefs.map(() => new Float32Array(rows.length)),
                  weight = new Float32Array(rows.length),
                  n = rows.length,
                  m = colDefs.length;
            let i = 0;
            for (; i<n; ++i) {
                const row = rows[i];
                let j = 0;
                for (; j<m; ++j) {
                    columns[j][i] = row[colDefs[j].name];
                }
                weight[i] = row.Weight;
            }
            return {
                columns: columns,
                weight: weight
            };
        }

        function stackedHistogram(rows, colName, Nbin, K) {
            const countsInColor = [];
            let c = 0;
            for (; c<=K; ++c) {
                countsInColor[c] = new Float64Array(Nbin);
            }
            const bins = new Float64Array(Nbin);
            if ( rows.length) {
                let lo = rows[0][colName],
                    hi = lo;
                rows.forEach(row => {
                    const d = row[colName];
                    if ( d < lo ) {
                        lo = d;
                    }
                    if ( d > hi ) {
                        hi = d;
                    }
                });
                const rng = hi - lo,
                      dy = rng / Nbin,
                      mid0 = lo + dy/2;
                let b = 0;
                for (; b<Nbin; ++b) {
                    bins[b] = mid0 + dy*b;
                }
                rows.forEach(row => {
                    const c = row.Color,
                          d = row[colName].
                          b = Math.min(Nbin-1, ((d - lo) / dy) | 0);
                    countsInColor[c][b] += 1;
                });
                const toFrac = 1.0 / rows.length;
                countsInColor.forEach(counts => {
                    let b = 0;
                    for (; b<Nbin; ++b) {
                        counts[b] *= toFrac;
                    }
                });
            }
            const datum = countsInColor.map((counts, c) => {
                const values = [];
                let b = 0;
                for (; b<Nbin; ++b) {
                    values[b] = {
                        x: bins[b],
                        y: counts[b]
                    };
                }
                return {
                    key: ''+c,
                    values: values
                };
            });
            datum.push(datum[0]);
            return datum.slice(1);
        }

        const DemoMainDefs = {
            fullscreen: false
        };
        exports.start = function start(config) {
            config = Object.assign({}, DemoMainDefs, config);
            const main = new qubx_app.Main(config),
                  _dispose = main.dispose.bind(main);
            Object.assign(main, {
                opener: new qubx_ui.Opener(),
                reader: new qubx_coldata.TableReader({
                    msg: "Please verify the data have been read correctly, and select columns.  For each row, the N selected columns make up a point in N-dimensional space. We will find the clusters that minimize sum-square-distance from the points' respective centroids."
                }),
                saver: new qubx_ui.Saver(),
                progressDlg: new qubx_ui.ModalProgress(),
                xmeans: new qubx_xmeans.XMeans(),

                rows: [],
                columns: [],
                results: [],
                resultsForK: [],
                K: 1,
                noData: true,

                dispose: function() {
                    main.savePrefs();
                    _dispose();
                },
                savePrefs: function() {
                    qubx_ui.SaveInputs();
                },
                _makeMenus: function(p) {
                    const dotdot = qubx_ui.querydotdot(p.controls);
                    dotdot('#qubx_mnu_file', node => new qubx_ui.Menu({
                        node: node,
                        caption: 'File',
                        groups: [[
                            {
                                caption: 'Load data file...',
                                action: () => main.opener.run()
                                    .then(main.reader.open.bind(main.reader))
                                    .then(main.setData)
                                    .then(() => main.opener.reset())
                                    .catch(qubx.reportErr)
                            },
                            {
                                caption: 'Enter/edit data...',
                                action: () => main.reader.edit(main.rows, main.columns)
                                    .then(main.setData)
                                    .catch(qubx.reportErr)
                            },
                            {
                                caption: 'Save data...',
                                action: main.saveData
                            },
                            {
                                caption: 'Save color stats...',
                                action: main.saveColorStats
                            }
                        ],[
                            {
                                caption: 'About QUB X-Means...',
                                action: () => main.aboutDlg.run()
                            }
                        ]]
                    }));
                    dotdot.listen('#qubx_xmeans_showOptions', 'click', () => main.optionsDlg.run());
                    main.btnIterate = dotdot.listen("#qubx_xmeans_iterate", 'click', main.iterate);
                    dotdot.listen("#qubx_xmeans_trim", 'click', main.promptTrim);
                    dotdot.listen("#qubx_xmeans_showStats", 'click', main.showStats);
                    main.divStatus = dotdot("#qubx_xmeans_status"); ///?
                },
                _makeDialogs: function(p) {
                    main.aboutDlg = new qubx_ui.Modal({
                        title: 'QUB X-Means - About',
                        template: '#qubx_xmeans_about',
                        canCancel: false,
                        okLabel: 'Close'
                    });

                    main.optionsDlg = new qubx_ui.Modal({
                        title: 'QUB X-Means - Options',
                        template: "#qubx_xmeans_options",
                        canCancel: false,
                        okLabel: 'Close'
                    });
                    main.optionsDlg.dom.then(dom => {
                        const dotdot = qubx_ui.querydotdot(dom.content);
                        main.txtTrialsPerK = dotdot('#qubx_xmeans_trialsPerK', node => new qubx_ui.AcceptedInput(node, 100, {
                            accept: qubx.acceptIntAtLeast(1),
                            format: qubx.formatStr,
                            localKey: "xmeans.trialsPerK",
                            widthChars: 5
                        }));
                        main.txtIterPerTrial = dotdot('#qubx_xmeans_iterPerTrial', node => new qubx_ui.AcceptedInput(node, 100, {
                            accept: qubx.acceptIntAtLeast(1),
                            format: qubx.formatStr,
                            localKey: "xmeans.iterPerTrial",
                            widthChars: 5
                        }));
                        main.txtIterPerButton = dotdot('#qubx_xmeans_iterPerButton', node => new qubx_ui.AcceptedInput(node, 10, {
                            accept: qubx.acceptIntAtLeast(1),
                            format: qubx.formatStr,
                            localKey: "xmeans.iterPerButton",
                            widthChars: 5
                        }));
                    }).catch(qubx.reportErr);

                    main.trimDlg = new qubx_ui.Modal({
                        title: 'QUB X-Means - Trim',
                        template: "#qubx_xmeans_trim_options",
                        okLabel: 'Trim'
                    });
                    main.trimDlg.dom.then(dom =>  {
                        const dotdot = qubx_ui.querydotdot(dom.content);
                        main.txtTrimStd = dotdot('#qubx_xmeans_trimStd', node => new qubx_ui.AcceptedInput(node, 1.0, {
                            accept: qubx.acceptFloatGreaterThan(0.0),
                            format: qubx.formatK0,
                            localKey: "xmeans.trimStd",
                            widthChars: 6
                        }));
                        const run = main.trimDlg.run.bind(main.trimDlg);
                        main.trimDlg.run = () => {
                            setTimeout(() => main.txtTrimStd.node.focus(), 10);
                            return run();
                        };
                    }).catch(qubx.reportErr);

                    main.statsDlg = new qubx_ui.Modal({
                        title: 'QUB X-Means - Color stats',
                        template: "#qubx_xmeans_colorStats",
                        canCancel: false,
                        okLabel: 'Close'
                    });
                    main.statsDlg.dom.then(dom => {
                        main.statsTable = dom.content.querySelector("#qubx_xmcs_stats > table");
                    }).catch(qubx.reportErr);
                },
                showAbout: function() {
                    return main.aboutDlg.run();
                },
                _setupData: function(p, components) {
                    const ctrls = components.querySelector("#qubx_xmeans_ctrls");
                    p.controls.appendChild(ctrls);
                    main.sub = components.querySelector("#qubx_xmeans_submain");
                    p.main.appendChild(main.sub);
                    main.divDplots = main.sub.querySelector("#qubx_xmeans_dataplots .qubx_xmeans_plots");
                    main.gridDplots = new qubx_ui.AspectGrid3({node: main.divDplots});
                    main.divKplots = main.sub.querySelector("#qubx_xmeans_kplots .qubx_xmeans_plots");
                    main.gridKplots = new qubx_ui.AspectGrid({node: main.divKplots});
                    main.kTable = main.sub.querySelector("#qubx_xmeans_ktable > table");

                    const tbl = d3.select(main.kTable),
                          tbody = tbl.append('tbody');
                    main.kRowK = tbody.append('tr');
                    main.kRowK.append('td').html('<b>K</b>').classed('qubx_xmeans_ktable_rowhead', true);
                    main.kRowSSR = tbody.append('tr');
                    main.kRowSSR.append('td').html('<b>SSR</b>').classed('qubx_xmeans_ktable_rowhead', true);
                    main.kRowAICc = tbody.append('tr');
                    main.kRowAICc.append('td').html('<b>AICc</b>').classed('qubx_xmeans_ktable_rowhead', true);

                    main.chartSSR = nv.models.scatterChart();
                    main.chartSSR.xAxis.tickFormat(d3.format('.02f'));
                    main.chartSSR.yAxis.tickFormat(d3.format('.02f'));
                    main.plotSSR = d3.select(main.gridKplots.insert(undefined, undefined, this.__onLayoutSSR.bind(this)))
                        .append('svg');
                    main.plotSSR.node().parentNode.className = 'qubx_xmeans_plot';
                    main.chartAICc = nv.models.scatterChart();
                    main.chartAICc.xAxis.tickFormat(d3.format('.02f'));
                    main.chartAICc.yAxis.tickFormat(d3.format('.02f'));
                    main.plotAICc = d3.select(main.gridKplots.insert(undefined, undefined, this.__onLayoutAICc.bind(this)))
                        .append('svg');
                    main.plotAICc.node().parentNode.className = 'qubx_xmeans_plot';
                },

                columnStore: {},
                getColumn: function(name) {
                    if ( ! main.columnStore[name] ) {
                        main.columnStore[name] = {name: name};
                    }
                    return main.columnStore[name];
                },
                setData: function(data) {
                    const rows = data.rows,
                          columnNames = data.columns.map(qubx.m('name'));
                    if ( main.noData ) {
                        main.noData = false;
                        d3.select(main.divDplots).select("instructions").remove();
                    }
                    rows.forEach((row, r) => {
                        row.Color = row.Color || 0;
                        if ( typeof(row.Weight) === 'undefined' ) {
                            row.Weight = 1.0;
                        }
                    });
                    main.rows = rows;
                    main.columns = columnNames.map(main.getColumn);
                    main.displayData();
                    main.xmeans.setData(readColumns(main.rows, main.columns));
                    main.ssdOfK = [0.0];
                    main.displayK();
                    main.scan();
                },
                displayData: function() {
                    const setupPlot = function(d,i) {
                        this.classList.add('qubx_xmeans_plot');
                        d.svg = d3.select(this).html('').append('svg');

                        d.controls = d3.select(this).append('div');
                        d.controls.classed('qubx_xmeans_plot_controls', true);
                        d.txtK = d.controls.append('span');
                        d.imageLink = d.controls.append('a');
                        d.imageLink.append('img')
                            .attr({
                                src: 'images/dl.svg',
                                alt: '[Download]',
                                title: 'Download'
                            }).style({
                                width: '1em',
                                height: '1em'
                            });
                        const lbl = d.controls.append('label');
                        d.chkHist = lbl.append('input');
                        d.chkHist.property({type: 'checkbox'});
                        d.chkHist.on('change', () => {
                            d.hist = d.chkHist.node().checked;
                            d3.select(this).select('svg').html('');
                            [d.lblBins, d.txtBins, d.slBins].forEach(ctrl => ctrl.style('visibility', d.hist ? 'visible' : 'hidden'));
                            renderPlot.bind(this)(d,i);
                        });
                        lbl.append('span').text('Histogram');
                        d.lblBins = d.controls.append('span');
                        d.lblBins.text('; bins: ');
                        d.lblBins.style('visibility', 'hidden');
                        d.bins = d.bins || 16;
                        d.txtBins = d.controls.append('span');
                        d.txtBins.text(''+d.bins);
                        d.txtBins.style('visibility', 'hidden');
                        d.slBins = d.controls.append('input');
                        d.slBins.property({
                            type: 'range',
                            min: 4,
                            max: 256,
                            step: 4,
                            value: d.bins
                        }).style('visibility', 'hidden');
                        d.slBins.on('input', function() {
                            d.bins = d.slBins.node().value|0;
                            d.txtBins.text(''+d.bins);
                            d3.select(this).select('svg').html('');
                            renderPlot.bind(this)(d,i);
                        }.bind(this));
                        
                        d.scatterChart = nv.models.scatterChart();
                        d.scatterChart.xAxis.tickFormat(d3.format('.02f'));
                        d.scatterChart.yAxis.tickFormat(d3.format('.02f'));

                        d.histChart = nv.models.multiBarChart()
                            .stacked(true)
                            .reduceXTicks(true)
                            .rotateLabels(0)
                            .showControls(false)
                            .showLegend(false)
                            .groupSpacing(0.1);
                        d.histChart.xAxis.tickFormat(d3.format(',f'));
                        d.histChart.yAxis.tickFormat(d3.format(',.2f'));
                    };
                    const onLayoutPlot = d => d.activeChart && d.activeChart.update();
                    const teardownPlot = d => {
                        d.scatterChart = d.histChart = d.activeChart = null;
                    };
                    const renderPlot = function(d,i) {
                        const qubplot = this;
                        nv.addGraph(() => {
                            d.txtK.text('K = '+main.K+' ');
                            if ( d.hist ) {
                                d.histChart.width(this.clientWidth).height(this.clientHeight)
                                    .color(COLORS_HIST(main.K))
                                d3.select(qubplot).select('svg')
                                    .datum(stackedHistogram(main.rows, d.name, d.bins, main.K))
                                    .transition().duration(10)
                                    .call(d.histChart);
                                d.activeChart = d.histChart
                                return d.histChart;
                            } else {
                                d.scatterChart.width(this.clientWidth).height(this.clientHeight);
                                d3.select(qubplot).select('svg')
                                    .datum([{
                                        key: d.name,
                                        values: main.rows.map((row, r) => ({
                                            x: r,
                                            y: row[d.name],
                                            color: COLORS[row.Color] || COLOR_BEYOND
                                        }))
                                    }])
                                    .transition().duration(100)
                                    .call(d.scatterChart);
                                d.activeChart = d.scatterChart
                                return d.scatterChart;
                            }
                        }, () => {
                            if ( ! d.hist ) {
                                d.scatterChart.scatter.dispatch.on('elementClick', e => {
                                    // e: point, pointIndex, pos[2], series {key, values}, seriesIndex
                                    const pt = main.rows[e.pointIndex];
                                    pt.Color = (pt.Color + 1) % (main.K + 1);
                                    e.point.color = COLORS[pt.Color] || COLOR_BEYOND;
                                    d.scatterChart.update();
                                    //main.columns.forEach(function(col) {
                                    //    col.plot.select(iPt);
                                    //});
                                });
                            }
                            // capture svg in d.imageLink
                            if ( main.chartStyles ) {
                                const svg = d.svg.node().cloneNode(true);
                                svg.setAttribute("xmlns", "http://www.w3.org/2000/svg");
                                svg.setAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink");
                                svg.setAttribute("version", "1.1");
                                const style = document.createElement('style');
                                style.innerHTML = main.chartStyles;
                                svg.insertBefore(style, svg.firstChild);
                                const uri = 'data:image/svg+xml;utf8,'+svg.outerHTML;
                                d.imageLink.attr({
                                    href: uri,
                                    download: d.name+' (X-Means).svg'
                                });
                            }
                        });
                    };
                    main.gridDplots.data(main.columns, qubx.m('name'), setupPlot, onLayoutPlot, teardownPlot)
                        .each(renderPlot);
                },
                // click: highlight in all data plots
                displayK: function() {
                    // main.results: [ {k, ssd, aicc, colors, mean, std} ]
                    const onclick = d => main.displayClusters(d.k);
                    let sel = main.kRowK.selectAll('.ktable_dcell')
                        .data(main.results, qubx.m('k'));
                    sel.enter().insert('td')
                        .text(d => ''+d.k)
                        .on('click', onclick)
                        .classed('ktable_dcell', true);
                    sel.exit().remove();

                    sel = main.kRowSSR.selectAll('.ktable_dcell')
                        .data(main.results, qubx.m('k'))
                        .text(d => qubx.formatK0(d.ssd));
                    sel.enter().insert('td')
                        .text(d => qubx.formatK0(d.ssd))
                        .on('click', onclick)
                        .classed('ktable_dcell', true);
                    sel.exit().remove();

                    sel = main.kRowAICc.selectAll('.ktable_dcell')
                        .data(main.results, qubx.m('k'))
                        .text(d => qubx.formatK0(d.aicc));
                    sel.enter().insert('td')
                        .text(d => qubx.formatK0(d.aicc))
                        .on('click', onclick)
                        .classed('ktable_dcell', true);
                    sel.exit().remove();

                    [{y: 'ssd',  lbl: 'SSR',  chart: main.chartSSR,  plot: main.plotSSR},
                     {y: 'aicc', lbl: 'AICc', chart: main.chartAICc, plot: main.plotAICc}]
                        .forEach(kplot => {
                            nv.addGraph(function() {
                                kplot.chart.width(this.clientWidth, this.clientHeight);
                                kplot.plot.datum([{
                                    key: kplot.lbl,
                                    values: main.results.map(result => ({
                                        x: result.k,
                                        y: result[kplot.y]
                                    }))
                                }])
                                    .transition().duration(100)
                                    .call(kplot.chart);
                                return kplot.chart;
                            }, () => kplot.chart.scatter.dispatch.on('elementClick', e => {
                                    main.displayClusters(main.results[e.pointIndex].k);
                            }));
                        });
                },
                __onLayoutSSR: function() {
                    main.chartSSR.update && main.chartSSR.update();
                },
                __onLayoutAICc: function() {
                    main.chartAICc.update && main.chartAICc.update();
                },
                
                // click: displayClusters(k)
                displayClusters: function(k) {
                    main.K = k;
                    const color = main.resultsForK[k].color;
                    main.rows.forEach((row, r) => {
                        row.Color = color[r];
                    });
                    main.displayData();
                    /// preserve data plots highlighted point?
                },
                scan: function() {
                    main.results = [];
                    main.resultsForK = [];
                    main.progressDlg.show('Scanning all K...');
                    //   main.progressDlg.hide();
                    //   main.displayData();
                    //   main.displayK();
                    let process = Promise.resolve(true),
                        i = 1;
                    for (let k=1; k<=10; ++k) {
                        process = process.then(() => {
                            if ( main.progressDlg.canceled ) {
                                return false;
                            }
                            return main.xmeans.getClusters(k, (k===1) ? 1 : (k*main.txtTrialsPerK.value), (k===1) ? 1 : main.txtIterPerTrial.value)
                                .then(result => {
                                    // result: {clusters: {color, ssd, mean, std}}
                                    main.setResult(result.clusters);
                                    main.progressDlg.progress(k/10);
                                });
                        });
                    }
                    return process.then(() => {
                        main.progressDlg.hide();
                        let lastAICc = main.resultsForK[1] ? main.resultsForK[1].aicc : 0,
                            lastRise = 0.0,
                            bestK = 1,
                            i = 2;
                        for (; i<10; ++i) {
                            const aicc = main.resultsForK[i] ? main.resultsForK[i].aicc : 0,
                                  rise = Math.max(aicc - lastAICc, 0.0);
                            if ( rise ) {
                                if ( rise > lastRise ) {
                                    bestK = i - 1;
                                    lastRise = rise;
                                }
                            } else {
                                lastRise = 0.0;
                            }
                            lastAICc = aicc;
                        }
                        main.displayClusters(bestK);
                        return (! main.progressDlg.canceled);
                    }).catch(qubx.reportErr);
                },
                setResult: function(result) {
                    const rPrev = main.resultsForK[result.k];
                    if ( rPrev ) {
                        main.results.splice(main.results.indexOf(rPrev), 1);
                    }
                    main.results.push(result);
                    main.resultsForK[result.k] = result;
                    main.displayK();
                    main.displayClusters(result.k);
                },

                iterate: function() {
                    main.btnIterate.disabled = true;
                    main.xmeans.iterate(main.K, main.rows.map(qubx.m('Color')), main.txtIterPerButton.value)
                        .then(data => {
                            main.setResult(data.clusters);
                            main.btnIterate.disabled = false;
                        }).catch(qubx.reportErr);
                },
                
                promptTrim: function() {
                    main.trimDlg.run()
                        .then(() => main.trim(main.txtTrimStd.value))
                        .catch(qubx.reportErr);
                },
                trim: function(cutoffSD) {
                    const K = main.K,
                          result = main.resultsForK[K],
                          lo = [],
                          hi = [],
                          trimmed = main.rows.map(qubx.m('Color'));
                    main.columns.forEach((col, c) => {
                        const mm = result.mean[c],
                              ss = result.std[c];
                        let i = 1;
                        for (; i<=K; ++i) {
                            lo[i] = mm[i] - cutoffSD * ss[i];
                            hi[i] = mm[i] + cutoffSD * ss[i];
                        }
                        main.rows.forEach((row, r) => {
                            const d = row[col.name],
                                  k = row.Color;
                            if ( k && ((d < lo[k]) || (d > hi[k])) ) {
                                row.Color = trimmed[r] = 0;
                            }
                        });
                    });
                    main.xmeans.iterate(K, trimmed, 0) // 0 iterations: measure only
                        .then(data => main.setResult(data.clusters))
                        .catch(qubx.reportErr);
                },
                
                saveData: function() {
                    let txt = main.columns.map(qubx.m('name')).join('\t') + '\n';
                    main.rows.forEach(row => {
                        const r = [];
                        main.columns.forEach(col => r.push(row[col.name].toPrecision(8)));
                        txt += r.join('\t') + '\n';
                    });
                    main.saver.prompt({name: "X-Means Data.txt"})
                        .then(fi => main.saver.saveText(fi.fullname, txt, "application/octet-stream;charset=UTF-8"));
                },
                
                showStats: function() {
                    const columns = ['Color', 'Count'];
                    main.columns.forEach(col => {
                        columns.push(col.name+' mean');
                        columns.push(col.name+' std');
                    });
                    const clusters = main.resultsForK[main.K],
                          rows = clusters.mean[0].map((x,i) => {
                              const row = {
                                  Count: 0,
                                  Color: i
                              };
                              if ( i ) {
                                  clusters.mean.forEach((mm, j) => {
                                      row[columns[2+2*j]] = mm[i];
                                      row[columns[2+2*j+1]] = clusters.std[j][i];
                                  });
                              }
                              return row;
                          });
                    clusters.color.forEach(c => {
                        rows[c].Count += 1
                    });
                    main.statsTable.innerHTML = '';
                    main.statsView = new qubx_ui.ColumnTable({
                        node: main.statsTable,
                        columns: columns.map(name => qubx_ui.ColumnTable.txtCol(name, name, qubx.formatK0))
                    });
                    main.statsView.render(rows.slice(1));
                    return main.statsDlg.run();
                },
                saveColorStats: function() {
                    const clusters = main.resultsForK[main.K];
                    if ( ! clusters ) {
                        return;
                    }
                    const count = clusters.mean[0].map(() => 0);
                    clusters.color.forEach(c => {
                        count[c] += 1;
                    });
                    let txt = "Color\tCount";
                    main.columns.forEach(col => {
                        txt += '\t' + col.name + ' mean\t' + col.name + ' std';
                    });
                    txt += '\n';
                    count.forEach((count, i) => {
                        if ( i ) {
                            txt += ''+i+'\t'+count;
                            clusters.mean.forEach((mm, j) => {
                                txt += '\t'+mm[i]+'\t'+clusters.std[j][i];
                            });
                            txt += '\n';
                        }
                    });
                    main.saver.prompt({name: "X-Means "+main.K+"-Coloring.txt"})
                        .then(fi => main.saver.saveText(fi.fullname, txt, "application/octet-stream;charset=UTF-8"));
                },

                _onDrop: function(event) {
                    event.stopPropagation();
                    event.preventDefault();
                    main.openData(event.dataTransfer.files);
                    main.panels.then(p => p.main.classList.remove('qubx_dragtarget'));
                }
            });

            qubx.get("css/nv.d3.css").then(styles => {
                main.chartStyles = styles;
            }).catch(qubx.reportErr);
            
            qubx_ui.loadTemplate("#qubx_xmeans_components").then(components => main.panels.then(p => {
                window.addEventListener("pagehide", main.savePrefs.bind(main));
                
                p.main.removeChild(p.main.querySelector("#qubx_topbot"));
                main._setupData(p, components);
                main._makeMenus(p);
                main._makeDialogs(p);
                
                p.main.addEventListener('dragenter', qubx_ui.dragEnter(p.main));
                p.main.addEventListener('dragleave', qubx_ui.dragLeave(p.main));
                p.main.addEventListener('dragover', qubx.stopAndPrevent);
                p.main.addEventListener('drop', main._onDrop);
                
                return p;
            })).catch(qubx.reportErr);
        
            return main;
        };
        
        return exports;
    });

// download all charts: maybe impossible with nv.d3's clip-path: url(...) nonsense

// better multi-var chart layout?

