    var QTR_MAGIC = "QUB_(;-)_QFS";
    var PRELOAD = 2;
    var DATA_IN_NODE = 4;
    var QTR_NODE_BYTES = 32;
    var QTR_TYPE_STRING = 3;
    var QTR_VIEWER = [null, null, null, null, 'getUint8', 'getInt8', 'getUint16', 'getInt16', 'getUint32', 'getInt32', 'getUint64', 'getInt64', 'getFloat32', 'getFloat64', null];
    var QTR_BUFTYPE = [null, null, null, null, Uint8Array, Int8Array, Uint16Array, Int16Array, Uint32Array, Int32Array, null, null, Float32Array, Float64Array, null];
    var QTR_UNIT = [0, 0, 4, 1, 1, 1, 2, 2, 4, 4, 8, 8, 4, 8, 10];

    exports.readTree = function readTree(arraybuf, view, offset, parent) {
        if ( ! view ) {
            view = new DataView(arraybuf);
        }
        if ( ! offset ) {
            if ( arraybuf.byteLength < QTR_MAGIC.length ) {
                return null;
            }
            var magic = '';
            for (var i=0; i<QTR_MAGIC.length; ++i) {
                magic += String.fromCharCode(view.getUint8(i));
            }
            if ( magic !== QTR_MAGIC ) {
                return null;
            }
            offset = QTR_MAGIC.length;
        }
        if ( arraybuf.byteLength < (offset + QTR_NODE_BYTES) ) {
            return null;
        }
        var node = {name: "", dataType: 0, dataSize: 0, dataCount: 0, rows: 0, columns: 0, children: [], ch: {}};
        node.flags = view.getUint8(offset);
        node.dataType = view.getUint8(offset+3, true);
        node.dataSize = view.getUint32(offset+4, true);
        node.dataCount = view.getUint32(offset+8, true);
        readTreeData(node, arraybuf, view, offset);
        var pos = view.getUint32(offset+16, true);
        if ( pos ) {
            node.children.push(readTree(arraybuf, view, pos, node));
            node.children.reverse();
            node.ch[node.children[0].name] = node.children[0];
        }
        pos = view.getUint32(offset+20, true);
        if ( pos && parent ) {
            var sib = readTree(arraybuf, view, pos, parent);
            parent.children.push(sib);
            parent.ch[sib.name] = sib;
        }
        var namelen = view.getUint8(offset+31);
        if ( namelen ) {
            var name = "";
            for (var i=0; i<namelen; ++i) {
                name += String.fromCharCode(view.getUint8(offset+32+i));
            }
            node.name = name;
        }
        return node;
    };

    function readTreeData(node, arraybuf, view, offset) {
        if ( ! node.dataType ) { return; }
        var unit = QTR_UNIT[node.dataType];
        node.rows = node.dataCount;
        node.columns = node.dataSize / unit;
        var pos = (node.flags & DATA_IN_NODE) ? (offset+12) : view.getUint32(offset+12, true);
        if ( node.dataType === QTR_TYPE_STRING ) {
            var d = "";
            for (var i=pos, n=pos+node.dataCount; i<n; ++i) {
                var c = view.getUint8(i);
                if ( c ) {
                    d += String.fromCharCode(c);
                }
            }
            node.data = d;
        } else {
            var method = QTR_VIEWER[node.dataType];
            if ( method ) {
                if ( ! (node.flags & PRELOAD) ) { // potentially large e.g. sampled data
                    node.data = new QTR_BUFTYPE[node.dataType](node.rows*node.columns);
                    if ( pos % unit ) { // have to read them one-by-one if the array is not aligned within arraybuf
                        for (var p=pos, i=0, n=node.rows*node.columns; i<n; ++i, p += unit) {
                            node.data[i] = view[method](p, true);
                        }
                    } else {
                        node.data.set(new QTR_BUFTYPE[node.dataType](arraybuf, pos, node.columns*node.rows));
                    }
                } else {
                    node.data = [];
                    for (var p=pos, i=0, n=node.rows; i<n; ++i) {
                        if ( node.columns === 1 ) {
                            node.data[i] = view[method](p, true);
                            p += unit;
                        } else {
                            var row = [];
                            node.data[i] = row;
                            for (var j=0, m=node.columns; j<m; ++j, p += unit) {
                                row[j] = view[method](p, true);
                            }
                        }
                    }
                }
            }
        }
    }

