'use strict';

const CACHE_VERSION = "v17";
const CACHE_PREFIX = 'online_';
const staticPaths = [ // make the first one small, and if any files are new, put one of them first
    '/online/images/curvefit.svg',
    '/online/qub.ico',
    '/online/qub.png',
    '/online/css/c3.css',
    '/online/css/c3-embed.css',
    '/online/css/c3.min.css',
    '/online/css/nv.d3.css',
    '/online/images/appbar.clipboard.paste.svg',
    '/online/images/appbar.edit.svg',
    '/online/images/appbar.folder.open.svg',
    '/online/images/appbar.graph.bar.svg',
    '/online/images/appbar.graph.line.up.svg',
    '/online/images/appbar.group.svg',
    '/online/images/appbar.page.corner.cells.svg',
    '/online/images/appbar.settings.svg',
    '/online/images/appbar.stock.down.svg',
    '/online/images/baseline.png',
    '/online/images/blank.png',
    '/online/images/dl.svg',
    '/online/images/landscape.svg',
    '/online/images/lists.svg',
    '/online/images/scaling.svg',
    '/online/images/stimbuild.svg',
    '/online/js/c3.js',
    '/online/js/d3.js',
    '/online/js/domReady.js',
    '/online/js/element-resize-detector.js',
    '/online/js/erf.js',
    '/online/js/hammer.js',
    '/online/js/hjcfit.js',
    '/online/js/nv.d3.js',
    '/online/js/random.js',
    '/online/js/require.js',
    '/online/js/RecordRTC.js',
    '/online/js/screenfull.js',
    '/online/js/snap.svg.js',
    '/online/js/stringview.js',
    '/online/js/whammy.js'
];
const staticOptions = [
];
const dynPaths = [
    '/online/',
    '/online/index.html',
    '/online/xmeans.html',
    '/online/commitor.html',
    '/online/fitness.html',
    '/online/fitframe.html',
    '/online/js/lmfit.js',
    '/online/qubx_lib_asm.js.mem',
    '/online/service-worker.js',
    '/online/manifest.json',
    '/online/qubx_templates.html',
    '/online/signin.html',
    '/online/css/fitness.css',
    '/online/css/qubtrace.css',
    '/online/css/qubx_model.css',
    '/online/css/qubx_panels.css',
    '/online/css/qubx_resultDoc.css',
    '/online/js/asmffi.js',
    '/online/js/fitframe.js',
    '/online/js/fitness_core.js',
    '/online/js/fitness_ui.js',
    '/online/js/jsotan.js',
    '/online/js/cloud.js',
    '/online/js/models.js',
    '/online/js/qubtrace.js',
    '/online/js/qubx_app.js',
    '/online/js/qubx_chart.js',
    '/online/js/qubx_cloud_ui.js',
    '/online/js/qubx_coldata.js',
    '/online/js/qubx_commitor.js',
    '/online/js/qubx_data.js',
    '/online/js/qubx_data_axg.js',
    '/online/js/qubx_data_prep.js',
    '/online/js/qubx_data_worker.js',
    '/online/js/qubx_data_work.js',
    '/online/js/qubx_demo_dataentry.js',
    '/online/js/qubx_demo_mac.js',
    '/online/js/qubx_demo_main.js',
    '/online/js/qubx_demo_simu.js',
    '/online/js/qubx_demo_single.js',
    '/online/js/qubx_demo_vid.js',
    '/online/js/qubx_fit.js',
    '/online/js/qubx_fitness_main.js',
    '/online/js/qubx_hist.js',
    '/online/js/qubxhjc.js',
    '/online/js/qubx_idl.js',
    '/online/js/qubx.js',
    '/online/js/qubx_lib_asm.js',
    '/online/js/qubx_lib_asm.js.mem',
    '/online/js/qubx_lib.js',
    '/online/js/qubx_mac.js',
    '/online/js/qubx_mac_worker.js',
    '/online/js/qubx_model.js',
    '/online/js/qubx_model_math.js',
    '/online/js/qubx_optimize.js',
    '/online/js/qubx_simulate.js',
    '/online/js/qubx_simulate_worker.js',
    '/online/js/qubx_simu_results.js',
    '/online/js/qubx_single.js',
    '/online/js/qubx_single_worker.js',
    '/online/js/qubx_ui_data.js',
    '/online/js/qubx_ui.js',
    '/online/js/qubx_ui_model.js',
    '/online/js/qubx_worker_require.js',
    '/online/js/qubx_workers.js',
    '/online/js/qubx_worker_util.js',
    '/online/js/qubx_xmeans.js',
    '/online/js/qubx_xmeans_main.js',
    '/online/js/qubx_xmeans_math.js',
    '/online/js/qubx_xmeans_worker.js',
    '/online/version'
];

self.addEventListener('install', function(event) {
    event.waitUntil(
        caches.open(CACHE_PREFIX+'static').then(function(cache) {
            return cache.match(staticPaths[0])
                .then(function(response) {
                    return response ? Promise.resolve(true) : cache.addAll(staticPaths);
                });
        }).then(function() {
            return caches.open(CACHE_PREFIX+CACHE_VERSION).then(function(cache) {
                return cache.addAll(dynPaths);
            });
        })
    );
});

self.addEventListener('activate', function(event) {
    var cacheWhitelist = [CACHE_PREFIX+'static', CACHE_PREFIX+CACHE_VERSION];
    
    event.waitUntil(
        caches.keys().then(function(keyList) {
            return Promise.all(keyList.map(function(key, i) {
                if ( key.startsWith(CACHE_PREFIX) && (cacheWhitelist.indexOf(key) === -1) ) {
                    return caches.delete(keyList[i]);
                }
            }));
        })
    );
});

self.addEventListener('message', function(event) {
    if ( event.data === 'update' ) {
        event.waitUntil(
            caches.open(CACHE_PREFIX+CACHE_VERSION).then(function(cache) {
                return cache.addAll(dynPaths);
            })
        );
    }
});


function rqPath(request) {
    return /^.*:\/\/[^\/]+(\/.*)/.exec(request.url)[1];
}

self.addEventListener('fetch', function(event) {
    //console.log('seeking',event.request);
    const path = rqPath(event.request);
    event.respondWith(
        // respond from the cache if available
        caches.match(event.request)
            .then(function(r) {
                if ( r && (dynPaths.indexOf(path) >= 0) ) {
                    // update dynamic cached content for next time
                    var newRq = (event.request.mode === 'navigate') ? event.request
                        : new Request(event.request, {
                            cache: 'no-cache'
                        });
                    fetch(newRq).then(function(r2) {
                        if ( r2 && (r2.status !== 304) ) {
                            return caches.open(CACHE_PREFIX+CACHE_VERSION).then(function(cache) {
                                return cache.put(event.request, r2);
                            });
                        }
                    });
                }
                if ( ! r ) {
                    console.log('not in SW cache:',event.request.url);
                    // load missing content
                    r = fetch(event.request)
                        .then(function(r) {
                            let step1 = Promise.resolve(true);
                            if ( staticOptions.indexOf(path) >= 0 ) {
                                // cache optional static content e.g. font variants
                                step1 = caches.open(CACHE_PREFIX+'static').then(function(cache) {
                                    return cache.put(event.request, r.clone());
                                });
                            }
                            return step1.then(function() { return r; });
                        });
                }
                return r;
            })
    );
});

