var TwoPaneDefs = {
    vertical: false,
    skews: [.3, .5, .7],
    skewType: 1,
    snap: 0.05
    // node, first, second
};
let TwoPane = function(config) {
    this.config = {
        vertical: config.vertical || TwoPaneDefs.vertical,
        skews: config.skews || TwoPaneDefs.skews,
        skewType: (typeof(config.skewType) !== 'undefined') ? config.skewType : TwoPaneDefs.skewType,
        snap: (typeof(config.snap) !== 'undefined') ? config.snap : TwoPaneDefs.snap
    };
    this.node = config.node || document.createElement('div');
    this.node.classList.add('qubx_twopane');
    this.node.classList.add(this.config.vertical ? 'qubx_twopane_v' : 'qubx_twopane_h');
    this.first = config.first || document.createElement('div');
    this.first.classList.add('qubx_twopane_item');
    this.first.classList.add('qubx_twopane_first');
    this.second = config.second || document.createElement('div');
    this.second.classList.add('qubx_twopane_item');
    this.second.classList.add('qubx_twopane_second');
    this.sep = document.createElement('div');
    this.sep.className = 'qubx_twopane_sep qubx_noselect';
    this.sepHandle = document.createElement('div');
    this.sepHandle.className = 'qubx_twoPane_sepHandle';
    this.sep.appendChild(this.sepHandle);
    this.btnMoreB = document.createElement('button');
    this.btnMoreB.className = 'qubx_twopane_moreA';
    this.btnMoreB.addEventListener('mousedown', function(event) {
        event.preventDefault();
        event.stopPropagation();
    }.bind(this));
    this.btnMoreB.addEventListener('mouseup', function(event) {
        this.moreB();
    }.bind(this));
    this.btnMoreB.textContent = this.config.vertical ? '^' : '<';
    this.sepHandle.appendChild(this.btnMoreB);
    this.btnMoreA = document.createElement('button');
    this.btnMoreA.className = 'qubx_twopane_moreA';
    this.btnMoreA.addEventListener('mousedown', function(event) {
        event.preventDefault();
        event.stopPropagation();
    }.bind(this));
    this.btnMoreA.addEventListener('mouseup', function(event) {
        this.moreA();
    }.bind(this));
    this.btnMoreA.textContent = this.config.vertical ? 'v' : '>';
    this.sepHandle.appendChild(this.btnMoreA);
    this.sep.addEventListener('mousedown', this.__onMouseDown.bind(this));
    this.node.appendChild(this.first);
    this.node.appendChild(this.sep);
    this.node.appendChild(this.second);
    setTimeout(this.onResize.bind(this), 10);
};
TwoPane.prototype.addControl = function(node) {
    node.classList.add('qubx_twopane_sepControl');
    this.sepHandle.appendChild(node);
    return this;
};
TwoPane.prototype.onResize = function(setOnly, pos) {
    var limit = this.config.vertical ? this.node.clientHeight : this.node.clientWidth;
    if ( ! setOnly ) {
        this.skewsPix = this.config.skews.map(function(skew) {
            return (skew * limit) | 0;
        }.bind(this));
    }
    if ( typeof(pos) !== 'undefined' ) {
        pos = pos + 'px';
    } else if ( this.config.skewType < 0 ) {
        pos = ((this.config.snap*limit)|0) + 'px';
    } else if ( this.config.skewType < this.skewsPix.length ) {
        pos = this.skewsPix[this.config.skewType] + 'px';
    } else {
        pos = (((1-this.config.snap)*limit)|0) + 'px';
    }
    if ( this.config.vertical ) {
        this.first.style.height = pos;
    } else {
        this.first.style.width = pos;
    }
};
TwoPane.prototype.moreA = function() {
    if ( this.config.skewType < this.skewsPix.length ) {
        ++this.config.skewType;
        this.onResize(true);
    }
};
TwoPane.prototype.moreB = function() {
    if ( this.config.skewType >= 0 ) {
        --this.config.skewType;
        this.onResize(true);
    }
};
TwoPane.prototype.__onMouseDown = function(event) {
    var p0 = this.config.vertical ? event.clientY : event.clientX;
    var limit = this.config.vertical ? this.node.clientHeight : this.node.clientWidth;
    var skewPix = 0;
    if ( 0 <= this.config.skewType ) {
        if ( this.config.skewType < this.skewsPix.length ) {
            skewPix = this.skewsPix[this.config.skewType];
        } else {
            skewPix = limit;
        }
    }
    var onmousemove = function(evt) {
        var p = (this.config.vertical ? evt.clientY : evt.clientX)
        skewPix = Math.max(0, Math.min(limit, skewPix + (p - p0)));
        p0 = p;
        this.onResize(true, skewPix);
    }.bind(this);
    var onmouseup = function(evt) {
        onmousemove(evt);
        var Nskew = this.skewsPix.length;
        var skew = skewPix / limit;
        if ( skew < this.config.snap ) {
            this.config.skewType = -1;
        } else if ( skew > (1 - this.config.snap) ) {
            this.config.skewType = Nskew;
        } else {
            var skewFrac = 1/Nskew;
            for (var s=0; s<Nskew; ++s) {
                if ( skew < (s+1)*skewFrac ) {
                    this.config.skewType = s;
                    this.config.skews[s] = skew;
                    this.skewsPix[s] = skewPix;
                    break;
                }
            }
        }
        this.onResize(true);
        this.node.removeEventListener('mousemove', onmousemove);
        document.body.removeEventListener('mouseup', onmouseup);
    }.bind(this);
    this.node.addEventListener('mousemove', onmousemove);
    document.body.addEventListener('mouseup', onmouseup);
};

window.addEventListener('load', function() {
    let tp = new TwoPane({
        node: document.querySelector("box"),
        //vertical: true
    });
    tp.first.textContent = 'A';
    tp.second.textContent = 'B';

    window.addEventListener('resize', function() {
        tp.onResize();
    });
});


