define("test_qubx", ["qubx"], function(qubx) {
    describe("Core utilities", function() {
	//it("does something nice", function() {
	//    expect( expression ).toBe( result ); // .not.toBe(...); .toEqual(...); .not.toMatch(/foo/); .toBeDefined(); .toBeCloseTo(x, d)
        //});

	it("validates string input with accept functions", function() {
	    expect( qubx.acceptInt("5") ).toBe(5);
	    expect( qubx.acceptInt("-5") ).toBe(-5);
	    expect( function() {qubx.acceptInt("")} ).toThrow();
	    expect( function() {qubx.acceptInt("1e5")} ).toThrow();
	    expect( function() {qubx.acceptInt("1.1")} ).toThrow();
	    expect( function() {qubx.acceptInt("foo")} ).toThrow();
	    
	    expect( qubx.acceptFloat("5") ).toBe(5.0);
	    expect( qubx.acceptFloat("-5.0") ).toBe(-5.0);
	    expect( qubx.acceptFloat("1e5") ).toBe(1e5);
	    // expect( function() {qubx.acceptFloat("5foo")} ).toThrow(); // returns 5;
	    expect( function() {qubx.acceptFloat("")} ).toThrow();
	    expect( function() {qubx.acceptFloat("foo")} ).toThrow();
	    
	    expect( function() {qubx.acceptNone(null)} ).toThrow();
        });
	
	it("formats field data with format functions", function() {
	    expect( qubx.formatStr(5) ).toBe("5");
	});
    });
    
    describe("Table class", function() {
	var t = new qubx.Table({label: 'lb'});
	var ev = {addField: function(spec) {},
		  inserting: function(row) {},
		  inserted: function(row) {},
		  removing: function(row) {},
		  removed: function(row) {},
		  set: function(info) {}
		 };
	it("is initialized", function() {
	    expect( t.label ).toBe('lb');
	    expect( t.size ).toBe(0);
	    expect( t.fieldIndex['Index'] ).toBe(0);
	    expect( t.fieldIndex['A'] ).toBeUndefined();
	});
	
	it("adds a field (empty)", function() {
	    for (var k in ev) {
		spyOn(ev, k).and.callThrough();
                t.ev[k].sub(ev[k]);
	    }
	    t.addField({name: 'A'});
	    expect(t.fields[1].name).toBe('A');
	    expect(t.fields[1].defval).toBe(0.0);
	    expect(t.fields[1].accept).toBe(qubx.acceptFloat);
	    expect(t.fields[1].format(0.0)).toBe("0.00");
	    expect(t.row_default['A']).toBe(0.0);
	    expect(t.fieldIndex['A']).toBe(1);
	    expect(ev.onAddField).toHaveBeenCalled();
	    expect(ev.onAddField.calls.argsFor(0)[0].defval).toBe(0.0);
	    for (var k in ev) {
                t.ev[k].unsub(ev[k]);
	    }
	});

	it("appends a row", function() {
	    for (var k in ev) {
		spyOn(ev, k).and.callThrough();
                t.ev[k].sub(ev[k]);
	    }
	    t.append({});
	    expect(t.size).toBe(1);
	    expect(t.entries.length).toBe(1);
	    expect(t.entries[0].Index).toBe(0);
	    expect(t.entries[0].A).toBe(0.0);
	    expect(t.get(0, 'Index')).toBe(0);
	    expect(t.get(0, 'A')).toBe(0.0);
	    expect(ev.onInserting).toHaveBeenCalled();
	    expect(ev.onInserting.calls.argsFor(0)[1].Index).toBe(0);
	    expect(ev.onInserted).toHaveBeenCalled();
	    expect(ev.onInserted.calls.argsFor(0)[1].A).toBe(0.0);
	    for (var k in ev) {
                t.ev[k].unsub(ev[k]);
	    }
	});
	
	it("inserts a row", function() {
	    for (var k in ev) {
		spyOn(ev, k).and.callThrough();
                t.ev[k].sub(ev[k]);
	    }
	    t.insert(0, {A : 1.0});
	    expect(t.size).toBe(2);
	    expect(t.entries.length).toBe(2);
	    expect(t.get(0, 'A')).toBe(1.0);
	    expect(t.get(0, 'Index')).toBe(0);
	    expect(t.get(1, 'Index')).toBe(1);
	    expect(t.get(1, 'A')).toBe(0.0);
	    for (var k in ev) {
                t.ev[k].unsub(ev[k]);
	    }
	});
	
	it("removes a row", function() {
	    for (var k in ev) {
		spyOn(ev, k).and.callThrough();
                t.ev[k].sub(ev[k]);
	    }
	    t.remove(0);
	    expect(t.size).toBe(1);
	    expect(t.entries.length).toBe(1);
	    expect(t.get(0, 'A')).toBe(0.0);
	    expect(t.get(0, 'Index')).toBe(0);
	    expect(ev.onRemoving).toHaveBeenCalled();
	    expect(ev.onRemoving.calls.argsFor(0)[1].A).toBe(1.0);
	    expect(ev.onRemoved).toHaveBeenCalled();
	    expect(ev.onRemoved.calls.argsFor(0)[1].Index).toBe(0);
	    for (var k in ev) {
                t.ev[k].unsub(ev[k]);
	    }
	});
	
	it("inserts and removes several rows", function() {
	    for (var k in ev) {
		spyOn(ev, k).and.callThrough();
                t.ev[k].sub(ev[k]);
	    }
	    for (var i=0; i<5; ++i) {
		t.insert(1, {A:i+1});
	    }
	    expect(t.size).toBe(6);
	    expect(t.get(0, 'A')).toBe(0.0);
	    expect(t.get(1, 'A')).toBe(5);
	    expect(t.get(2, 'A')).toBe(4);
	    for (var i=0; i<5; ++i) {
		t.remove(0);
	    }
	    expect(t.size).toBe(1);
	    expect(t.entries.length).toBe(1);
	    expect(t.get(0, 'A')).toBe(1);
	    expect(t.get(0, 'Index')).toBe(0);
	    for (var k in ev) {
                t.ev[k].unsub(ev[k]);
	    }
	});
	
	it("sets and retrieves a cell value", function() {
	    for (var k in ev) {
		spyOn(ev, k).and.callThrough();
                t.ev[k].sub(ev[k]);
	    }
	    t.set(0, 'A', 5.0);
	    expect(t.get(0, 'A')).toBe(5.0);
	    expect(ev.onSet).toHaveBeenCalled();
	    expect(ev.onSet.calls.argsFor(0)[1].row).toBe(0);
	    expect(ev.onSet.calls.argsFor(0)[1].column).toBe('A');
	    expect(ev.onSet.calls.argsFor(0)[1].value).toBe(5.0);
	    expect(ev.onSet.calls.argsFor(0)[1].prev).toBe(1.0);
	    for (var k in ev) {
                t.ev[k].unsub(ev[k]);
	    }
	});
    });

    describe("Multi-level Undo", function() {
	var evt;
	var undo = new qubx.Undo();
	
	beforeEach(function() {
	    evt = {onChange: function() {}, undo: function() {}, redo: function() {}, undo2: function() {}, redo2: function() {}};
	    for (var k in evt) {
		spyOn(evt, k);
            }
            undo.ev.change.sub(evt.onChange);
	});
	afterEach(function() {
            undo.ev.change.unsub(evt.onChange);
	});
	
	it("starts empty", function() {
	    expect( undo.canUndo() ).toBe( false );
        });
	it("receives an action", function() {
	    undo.pushUndo(evt.undo, evt.redo);
	    undo.sealUndo("1");
	    expect( undo.canUndo() ).toBe( true );
	    expect( undo.undoLabel() ).toBe( "1" );
	    expect( undo.canRedo() ).toBe( false );
	});
	it("clears", function() {
	    undo.clear();
	    expect( undo.canUndo() ).toBe( false );
	    expect( undo.canRedo() ).toBe( false );

	});
	it("receives and undoes an action", function() {
	    undo.pushUndo(evt.undo, evt.redo);
	    undo.sealUndo("1");
	    undo.undo();
	    expect(evt.undo).toHaveBeenCalled();
	    expect(evt.redo).not.toHaveBeenCalled();
	    expect(evt.onChange).toHaveBeenCalled();
	    expect(undo.canUndo()).toBe(false);
	    expect(undo.canRedo()).toBe(true);
	});
	it("receives and undoes and redoes an action", function() {
	    undo.pushUndo(evt.undo, evt.redo);
	    undo.sealUndo("1");
	    undo.undo();
	    undo.redo();
	    expect(evt.redo).toHaveBeenCalled();
	    expect(evt.onChange).toHaveBeenCalled();
	    expect(undo.canUndo()).toBe(true);
	    expect(undo.canRedo()).toBe(false);
	});
	it("receives and undoes two actions", function() {
	    undo.clear();
	    undo.pushUndo(evt.undo, evt.redo);
	    undo.pushUndo(evt.undo2, evt.redo2);
	    undo.sealUndo("1");
	    undo.undo();
	    expect(evt.undo).toHaveBeenCalled();
	    expect(evt.undo2).toHaveBeenCalled();
	    expect(undo.canUndo()).toBe(false);
	    expect(undo.canRedo()).toBe(true);
	});
	it("seals 2 and undoes one", function() {
	    undo.clear();
	    undo.pushUndo(evt.undo, evt.redo);
	    undo.sealUndo("1");
	    undo.pushUndo(evt.undo2, evt.redo2);
	    undo.sealUndo("2");
	    expect(undo.undoLabel()).toBe("2");
	    undo.undo();
	    expect(undo.canUndo()).toBe(true);
	    expect(undo.undoLabel()).toBe("1");
	    expect(undo.canRedo()).toBe(true);
	    expect(undo.redoLabel()).toBe("2");
	    expect(evt.undo).not.toHaveBeenCalled();
	    expect(evt.undo2).toHaveBeenCalled();
	});
	it("seals 2 and undoes and redoes one", function() {
	    undo.clear();
	    undo.pushUndo(evt.undo, evt.redo);
	    undo.sealUndo("1");
	    undo.pushUndo(evt.undo2, evt.redo2);
	    undo.sealUndo("2");
	    undo.undo();
	    undo.redo();
	    expect(undo.canUndo()).toBe(true);
	    expect(undo.undoLabel()).toBe("2");
	    expect(undo.canRedo()).toBe(false);
	    expect(evt.redo2).toHaveBeenCalled();
	    expect(evt.redo).not.toHaveBeenCalled();
	});
	it("honors config.maxDepth", function() {
	    undo = new qubx.Undo({maxDepth: 2});
	    undo.pushUndo(evt.undo, evt.redo);
	    undo.pushUndo(evt.undo2, evt.redo2);
	    undo.sealUndo("1");
	    undo.pushUndo(evt.undo, evt.redo);
	    undo.sealUndo("2");
	    undo.pushUndo(evt.undo, evt.redo);
	    undo.sealUndo("3");
	    undo.undo();
	    undo.undo();
	    expect(undo.canUndo()).toBe(false);
	    expect(undo.canRedo()).toBe(true);
	    expect(undo.redoLabel()).toBe("2");
	});
    });

    describe("Table Undo", function() {
	var t, undo, tUndo, evt;
	
	beforeEach(function() {
	    t = new qubx.Table();
	    t.addField({name: 'A'});
	    t.addField({name: 'B'});
	    undo = new qubx.Undo();
	    tUndo = new qubx.TableUndo(t, undo);
	});
	it("remains consistent", function() {
	    t.append({A:1, B:2});
	    undo.sealUndo();
	    t.append({A:5, B:6});
	    undo.sealUndo();
	    t.insert(1, {A:3, B:4});
	    undo.sealUndo();
	    undo.undo();
	    expect(t.size).toBe(2);
	    expect(t.get(1, 'A')).toBe(5);
	    undo.redo();
	    expect(t.size).toBe(3);
	    expect(t.get(1, 'A')).toBe(3);
	    expect(t.get(0, 'A')).toBe(1);
	    t.set(0, 'B', 10);
	    undo.sealUndo();
	    expect(t.get(0, 'B')).toBe(10);
	    undo.undo();
	    expect(t.get(0, 'B')).toBe(2);
	    undo.redo();
	    expect(t.get(0, 'B')).toBe(10);
	});
    });
    
    return qubx;
});
