define("vcheck", ["express", "fs", "multiparty", "mysql"], function(express, fs, multiparty, mysql) {
    'use strict';

    const MAXV = 5;
    function parseVersion(s) {
        const v = s.split('.').map(token => (token|0));
        while ( v.length < MAXV ) {
            v.push(0);
        }
        return v;
    }
    function vcmp(a, b) {
        {let i=0; for (i=0; i<MAXV; ++i) {
            if ( a[i] < b[i] ) {
                return -1;
            } else if ( a[i] > b[i] ) {
                return 1;
            }
        }}
        return 0;
    }

    function merged(a, b, k, cmp) {
        const m = [];
        const sgn = -(((a.length > 1) && cmp(a[0][k], a[1][k])) ||
                    ((b.length > 1) && cmp(b[0][k], b[1][k])) || 1);
        let i=0, j=0;
        while ( (i < a.length) || (j < b.length) ) {
            if ( (i < a.length) && (j < b.length) ) {
                let d = cmp(a[i], b[j]);
                if ( d === 0 ) {
                    m.push(Object.assign({}, a[i], b[j]));
                    ++i;
                    ++j;
                } else if ( d === sgn ) {
                    m.push(a[i]);
                    ++i;
                } else {
                    m.push(b[j]);
                    ++j;
                }
            } else if ( i < a.length ) {
                m.push(a[i]);
                ++i;
            } else {
                m.push(b[j]);
                ++j;
            }
        }
        return m;
    }

    const pool = mysql.createPool({
        connectionLimit: 100,
        host: 'localhost',
        user: 'dlposter',
        password: 'qubdl',
        database: 'downloads',
        charset: "utf8mb4_unicode_ci"
    });
    const query = (q, inserts) => {
        return new Promise((resolve, reject) => {
            pool.getConnection((err,connection) => {
                if (err) {
                    reject("No database connection: "+err);
                    return;
                }
                connection.query(q, inserts||[], (err, rows) => {
                    connection.release();
                    if ( err ) {
                        reject("Database error: "+err);
                    } else {
                        resolve(rows);
                    }
                }).on('error', err => {
                    reject("Database error: "+err);
                    return;
                });
            });
        });
    };

    const promisify = fsCall => path => new Promise((resolve, reject) => {
        fsCall(path, (err, result) => {
            if ( err ) { reject(err); }
            else { resolve(result); }
        });
    });
    const promisify2 = fsCall => (path, arg) => new Promise((resolve, reject) => {
        fsCall(path, arg, (err, result) => {
            if ( err ) { reject(err); }
            else { resolve(result); }
        });
    });
    const promis = {
        stat: promisify(fs.stat),
        readFile: promisify2(fs.readFile)
    };

    const vdetect = /^\d\d\d\d-\d\d-\d\d +version +([0-9.]+)/;
    const parseChanges = text => {
        let entries = [];
        var entry;
        text.split(/\r\n|\r|\n/).forEach(line => {
            let match = vdetect.exec(line);
            if ( match ) {
                if ( entry ) {
                    entry.text += '\n';
                    entries.push(entry);
                }
                entry = {version: parseVersion(match[1]), text: line}
            } else {
                entry.text += '\n'+line;
            }
        });
        if ( entry ) {
            entry.text += '\n';
            entries.push(entry);
        }
        return entries;
    };

    const CHANGEBASE = '/var/www/';
    
    const apps = {
        items: {},
        get: function(key, os) {
            key = (''+key) || 'qubx';
            os = (''+os) || 'win';
            const x = this.items[key+os];
            if ( x ) {
                return Promise.resolve(x);
            }
            return query(`SELECT Id, Changelog FROM Apps WHERE Code = ${mysql.escape(key)};`)
                .then(rows => this._newApp(key, rows[0].Id, rows[0].Changelog, os));
        },
        _newApp: function(key, Id, Changelog, OS) {
            const item = {
                Id: Id,
                Changelog: Changelog,
                changesModified: null,
                _changes: [],
                _changesReady: function() {
                    return promis.stat(CHANGEBASE + Changelog).then(stats => {
                        const mod = (stats && stats.mtime) || null;
                        if ( mod === this.changesModified ) {
                            return this._changes;
                        }
                        return promis.readFile(CHANGEBASE + Changelog, 'utf8').then(data => {
                            const entries = parseChanges(data);
                            return query(`SELECT Filename, Upgrade, UpgradeSince, Number, Date FROM Files INNER JOIN Versions ON Version = Versions.Id WHERE OS = ${mysql.escape(OS)} AND App = ${Id} ORDER BY Number DESC LIMIT 1;`).then(rows => {
                                rows = rows.map(row => Object.assign({}, row, {version: parseVersion(row.Number)}));
                                this._changes = merged(entries, rows, 'version', vcmp);
                                return this._changes;
                            });
                        });
                    });
                },
                changes: function(version, q) {
                    return this._changesReady().then(changes => {
                        const newStuff = [];
                        for (var i=0; i<changes.length; ++i) {
                            if ( vcmp(version, changes[i].version) < 0 ) {
                                newStuff.push(changes[i]);
                            } else {
                                break;
                            }
                        }
                        return newStuff;
                    });
                }
            };
            this.items[key+OS] = item;
            return item;
        }
    };

    const router = express.Router();

    router.use((req, res) => {
        if ( req.method === 'POST' ) {
            const form = new multiparty.Form();
            form.parse(req, (err, fields, files) => {
                const q = {};
                for (var name in fields) {
                    q[name] = fields[name][0];
                }
                checkVersion(q, req, res);
            });
        } else {
            checkVersion(req.query, req, res);
        }
    });

    const URLBASE = 'https://www.qub.buffalo.edu/download/#f=';
    const HEAD = '<!DOCTYPE html><html lang="en"><meta http-equiv="Content-Type" content="text/html; charset=utf-8" /><meta name="viewport" content="initial-scale=1, width=device-width"><title>QUB - Version check</title><link rel="stylesheet" href="qub.css", type="text/css" charset="utf-8" /></head><body><a href="index.html"><img src="qublogo.png" style="float:left; margin: 1em" /></a>';
    const FOOT = '</body></html>';

    function checkVersion(q, req, res) {
        query(`INSERT INTO Vcheck (Date, Time, Code, Version, OS, IP, GUID) VALUES (CURDATE(), NOW(), ${mysql.escape(q.key)}, ${mysql.escape(q.version)}, ${mysql.escape(q.os)}, ${mysql.escape(req.ip)}, ${mysql.escape(q.guid)});`);
        const version = parseVersion(q.version);
        return apps.get(q.key, q.os).then(app => {
            return app.changes(version, q).then(changes => {
                var replyText;
                if ( changes.length ) {
                    let url = URLBASE + changes[0].Filename;
                    if ( changes[0].Upgrade && (vcmp(version, parseVersion(changes[0].UpgradeSince)) >= 0) ) {
                        url = URLBASE + changes[0].Upgrade;
                    }
                    replyText = `<version>${changes[0].version.slice(0,3).join('.')}</version>\n<url>${url}</url><pre><changes>` +
                        changes.map(ch => ch.text).join('') + '</changes></pre>';
                } else {
                    replyText = 'Up to date.';
                }
                res.setHeader('Content-Type', 'text/html');
                res.writeHead(200); // ? res.statusCode);
                res.write(HEAD + replyText + FOOT);
                res.end();
            });
        }).catch(err => {
            console.error(err);
            console.error(err.stack);
            res.setHeader('Content-Type', 'text/html');
            res.writeHead(404);
            res.end();
        });
    }

    return router;
});


    
